/* modified NKC 12/30/09
 * Authenticate.java
 * Transport Layer Multicaster
 * Created on June 18, 2001
 * @author  mpullen & students
 * @version 5.1.0
 * @modified JMP 1/9/10; 5/20/10
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2010 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
// The Authenticate class contains logic to authenticate users.
/*********************SYSTEM LOGIC**************************
 *
 * Auth.java class logic:
 * ->contructor:
 * ->if using file Read {username,password} from authenticate file (default: Auth.conf)
 *
 * ->method authenticate(hostIndex,username,password,courseID)
 * ->  if using file,
 * ->    check values against stored table
 * ->    if present, return fill in values at hostIndex postition and true
 * ->    if not present, re-read the file and try again
 * ->    if present, return true
 * ->    if still not present, return false
 * ->  otherwise authenticate in database
 *
 * -> method nameAvailable(username)
 * -> if notDatabase, check username table for existence of name;
 *    return true if not there
 * -> otherwise look in database for username, return true is not there
 *
 * -> method validateSessionVariable(sessionVariable)
 *
 * -> method sendBreakoutData()
 *    get all possible users from database
 *    get persistent group assignments from database
 *    resolve any differences between the two
 *    load host table from it
 *    format and send TC 31 message to all FC
 *
 * -> method sendFCBreakoutMessage(hostIndex)
 *    formats breakoutDataMessage and sends it
 *    either to all connected hosts or to new joiner
 *
 * -> method setBreakoutGroup(group,username)
 *    update host table and database persistent group
 *    send updated group to all FC 
 *
 ***********************************************************
 Definition of authentication file for TLM:
  One line per user, fields separated by /
  Comments have # in first position
  1. username - up to 8 characters
  2. password - up to 8 characters
  3. Unicast Address of user, may be DNS or IP address; if present, user must
     be at this IP address; if missing, user may be at any IP address
  4. drawGrid - 0 or 1 (draw a floor control grid - defaults to 0)
  5. controlPrivilege (user is floor controller - defaults to 0)
  Items 2 to 5 are optional, as are slashes unless needed, however at least one
  character must appear between any two slashes.
 ***********************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.Date.*;
import java.util.TimeZone.*;
import java.sql.*;

public class Auth extends TLM
{
  private static String[] authUsername;         // 8-char username of authenticated user
  private static String[] authFullname;         // 40-char name of authenticated user
  private static String[] authPassWord;         // 8-char key for user login
  private static InetAddress[] authHostAddr;    // Internet address user must use
                                                // 1 = allow any address
  private static char[] authControlPrivilege;   // 0,1 or 2 indicating whether user
                                                // floor control privilege
  private static boolean[] authDrawGrid;        // 0 or 1 indicating whether client on this
                                                // host displays a floor control grid
  private static BufferedReader authFile;       // file contains above authentication values
  private static int authTableSize;
  private static boolean authFileHasBeenRead = false;
  private static boolean tableHasBeenPrinted = false;
  //private static Connection databaseConnection = null; // for MySQL
  private static DBQuery dbq = null; // nick: DBQuery modifications

  // constructor - sets up tables and loads them first time
  public Auth()
  {
    // set up storage for file contents
    authUsername = new String[MAX_USERS];
    authFullname = new String[MAX_USERS];
    authPassWord = new String[MAX_USERS];
    authHostAddr = new InetAddress[MAX_USERS];
    authControlPrivilege = new char[MAX_USERS];
    authDrawGrid = new boolean[MAX_USERS];

    // obtain semester code from the database
    // unless it was set by command line
    if(thisSemester.trim().length() > 0)return;

    // data object for MySQL
    String semesterQuery = null;

    // connect to database
    if(connectToDatabase())return;
	
    // build query
    if(noDatabase || (thisSemester.trim().length() > 0))
    {
      System.out.print("database connection failed - ");
      if (thisSemester.trim().length() > 0)
        System.out.println("using command line/conf file semester code " +
                           thisSemester);
      else
        System.out.println("operating with no required semester code");
      return;
    }

    //old queries:
    //     result = statement.executeQuery(
    //    "SELECT semester from semesterinfo where conference='current';");
    // result = statement.executeQuery(
    //          "SELECT semester from semesterinfo where conference='currdisted';");

    // Retrieve current seemster from the database
    String tSem = dbq.getCurrentSemester();
    if (tSem == null) {
       if (thisSemester.trim().length() > 0)
         System.out.println("using command line semester code " + thisSemester);
       else
         System.out.println("operating with no required semester code");
    } else {
        thisSemester=tSem;
    }
    
    thisSemesterBytes = thisSemester.getBytes();

    // close database connection
    dbq.close();
  }

  // method to read the file and construct table from it
  // also prints the table the first time it is read, if argument is true
  // returns true if file read; false if not found

  public static boolean readAuthFile(boolean printTable)
  {
    boolean printThisTable = printTable && !tableHasBeenPrinted;

    if(authFileName == null)
    {
      System.out.println("Authentication file name not found");
      return false;
    }
    if(authFileName == "")
    {
      System.out.println("Authentication file name not found");
      return false;
    }
    // open the file
    nextCourseID[0] = roomID;
    try
    {
      System.out.print( "Loading authentication table from file "
        + authFileName + "\n" );
      authFile = new BufferedReader( new FileReader( authFileName ));
    }
    catch( FileNotFoundException e )
    {
      System.out.println("Authentication file not found");
      return false;
    }
    synchronized( authUsername )
    {
      // read the file and parse values into memory
      try
      {
        int userIndex = 0;
        while( userIndex < MAX_USERS )
        {
          String line = null;
          while(line == null)line = authFile.readLine();
          line = line.trim();
          if( line.equals(endTag) )break;
          if( line.startsWith( commentTag ) )
          {
            if(printThisTable)System.out.print( line );
          }
          else
          {
            // at this point we have a user entry line - parse it
            // "/" separates authenticators
            StringTokenizer parseLine = new StringTokenizer( line,"/" );

            // parse userName
            if( parseLine.hasMoreTokens() )
            {
              authUsername[ userIndex ] = parseLine.nextToken();
              if(printThisTable)System.out.print( userIndex + ":"
                + authUsername[ userIndex ] );
              if( authUsername[ userIndex ].length() > 8 )
              {
                System.out.print( " userName has more than 8 characters- truncated\n" );
                authUsername[ userIndex ] = authUsername[ userIndex ].substring(0,7);
              }
            }
            else
            {
              System.out.println("end; not found in authentication file - execution aborted.\n");
              stopTLM = true;
              System.exit(-1);
            }

            // pad out username with blanks
            while( authUsername[ userIndex ].length() < 8 )
              authUsername[ userIndex ] = authUsername[ userIndex ].concat(" ");

            // check that user name is unique
            for( int userCheck=0; userCheck < userIndex-1; ++userCheck )
            {
              if( authUsername[ userIndex ].equals( authUsername[ userCheck ] ) )
              {
                System.out.print( "\nUsername:" + authUsername[ userCheck ]
                  + " is not unique- execution aborted.\n" );
                stopTLM = true;
                System.exit(-1);
              } // end if( userName...
            } // end for(

            // parse password
            if( parseLine.hasMoreTokens() )
            {
              authPassWord[ userIndex ] = parseLine.nextToken();
              if( authPassWord[ userIndex ].length() > 8 )
              {
                System.out.print( "password has more than 8 characters- truncated\n" );
                authPassWord[ userIndex ] = authPassWord[ userIndex ].substring(0,7);
              } // end if(authuserName...
            } // end if( parseLine...
            else authPassWord[ userIndex ] = "        ";

            // pad out password with blanks
            while( authPassWord[ userIndex ].length() < 8 )
              authPassWord[ userIndex ] = authPassWord[ userIndex ].concat(" ");
            if(printThisTable)System.out.print( "/" + authPassWord[ userIndex] );

            // parse IP Address
            if( parseLine.hasMoreTokens() )
            {
              String hostAddrText = parseLine.nextToken();
              if( hostAddrText.equals(" ") )
              {
                authHostAddr[ userIndex ] = null;
              }
              else
              {
                authHostAddr[ userIndex ] = InetAddress.getByName( hostAddrText );
              } // end else/if( hostAddrText...
            } // end if( parseLine.hasMoreTokens() )
            else authHostAddr[ userIndex ] = null;
            if(printThisTable)System.out.print( "/" + authHostAddr[ userIndex ]);

            // parse drawGrid
            if( parseLine.hasMoreTokens() )
            {
              String drawText = parseLine.nextToken();
              authDrawGrid[ userIndex ] = drawText.equals("1");
            }
            else authDrawGrid[ userIndex ] = false;
            if(printThisTable)System.out.print( "/" + authDrawGrid[ userIndex ] );

            // parse controlPrivilege
            if( parseLine.hasMoreTokens() )
            {
              String controlText = parseLine.nextToken();
              authControlPrivilege[ userIndex ] = (controlText.toCharArray())[0];
            }
            else authControlPrivilege[ userIndex ] = '0';
            if(printThisTable)System.out.print( "/" + authControlPrivilege[ userIndex ] );

            // parse fullname
            if( parseLine.hasMoreTokens() )
              authFullname[ userIndex ] = parseLine.nextToken();
            else authFullname[ userIndex ] = "";
            if(printThisTable)System.out.print( "/" + authFullname[ userIndex ]);

            userIndex++;

          } // end else/if(line.startsWith...
          if(printThisTable)System.out.print( "\n" );
        } // end while( userIndex < MAX_USERS )
        if( userIndex >= MAX_USERS )
        {
          System.out.println( "Authentication file has too many entries (max "
            + MAX_USERS + ") - truncated table (last valid username:"
            + authUsername[ MAX_USERS-1 ] + ")" );
        }   // end if( userIndex...
        authTableSize = userIndex;
        authFile.close();
      } // end try for read file line
      catch( IOException ioe )
      {
        System.out.println("Error reading authentication file " + authFileName
          + " - shutting down");
        System.exit(-1);
      } // end catch
    } // end synchronized( authUsername )
    if(printTable)tableHasBeenPrinted = true;
    return true;
  } // end readAuthfile method

  // perform authentication test and, if passed, move data to hosts table
  // returns valid loginData{username,password} if passed; otherwise, null
  public loginData authenticate( int hostIndex,
                                 String testUsername,
                                 String testPassword,
                                 String testHostAddr,
                                 int testSourcePort,
                                 String testCourseID,
                                 String testSessionVariable)
  {
    // file authentication does not use database
    if(authFileName != null)
      return fileAuthentication(hostIndex, testUsername,
                                testPassword, testHostAddr, testCourseID);
    // data objects for MySQL
    if(noDatabase)return null;
     
    // database authentication
    String readPassword = null, readFullname = null, readIPAddress = null;
    String readController = null, readShowGrid = null, readInstructor = null;
    String authenticateQuery = null;
    loginData login = new loginData();

    // connect to database
    if(connectToDatabase())return null;


    // login through newfcClient with no login panel; this happens when
    // using webserver authentication; we find user information from the
    // database by session variable and date-time the webserver places there
    if(testUsername.equals("        ") && testPassword.equals("        "))
    {
      // query below was moved into DBQuery class, below is for reference
      // lookup this sessionVariable in the database;
      // consider only posting in last 60 seconds
      // where courseID matches requiredCourseID
      //authenticateQuery =
      //  "SELECT semester,courseID,username,fullname,instructor," +
      //  "controller,showGrid,lastLogon from userinfo where NEWSessID ='" +
      //   testSessionVariable + "' and userDeleted='0';";

      // access database
      String readSemester=null, readUsername=null;
      String readCourseID=null, readLastLogon=null;
      boolean authenticatedOK = false;
      try
      {
        // Query DB to authenticate based on session variable:
        //    if authUserWithToken returns true, then the user
        //       details will be set in the dbq.user object
        //    note: authUserWithToken also valiates the lastLogon
        authenticatedOK = dbq.authUserWithToken(testSessionVariable, roomID);
        if (authenticatedOK) {
          readSemester = dbq.user.semester;
          readCourseID = dbq.user.courseID;
          readUsername = dbq.user.username;
          readFullname = dbq.user.fullname;
          readInstructor = String.valueOf(dbq.user.instructor);
          readController = String.valueOf(dbq.user.controller);
          readShowGrid = String.valueOf(dbq.user.showGrid);
          readLastLogon = dbq.user.lastLogon;
        } else {
          System.out.println("no match found for user at IP address " +
                             testHostAddr + "/" + testSourcePort);
          return null;
        }

      } catch (SQLException ex) {

        System.out.println("Exception in reading sessionVariable from database:"+ex);
		    ex.printStackTrace(System.out);
		    return null;
      }

      dbq.close();  //close database
      
      // self protection
      if(readUsername.length() > 8)readUsername = readUsername.substring(0,8);
      if(readCourseID.length() > 11)readCourseID = readCourseID.substring(0,11);

      // if valid username was found, pad it out to 8 bytes;
      // then set host table values from database entries found
      if(!authenticatedOK)return null;
      while(readUsername.length() < 8)readUsername = readUsername+" ";
      while(readCourseID.length() < 11)readCourseID = readCourseID+" ";
      checkForDeadHostTableEntry(readUsername);
      username[hostIndex] = readUsername;
      fullname[hostIndex] = readFullname;
      controlPrivilege[hostIndex] = (readController.toCharArray())[0];
      instructorPrivilege[hostIndex] = (new Integer(readInstructor)).intValue();
      drawGrid[ hostIndex ] = readShowGrid.equals("1");
      login.username = readUsername;
      login.courseID = readCourseID.toUpperCase();
      return login;

    } // end if(testUsername.equals("        ")

    // login through newfcClient or newfcApplet login panel or passed parameters
    else
    {
      // old db query, moved into DBQuery:
      // access database
      // if(thisSemester.trim().length()>0)
      //  authenticateQuery =
      //    "SELECT password,fullname,IPAddress,instructor,controller," +
      //    "showGrid from userinfo where username ='" +
      //    testUsername.trim() + "' and courseID = '" +
      //    testCourseID.trim() + "' and semester ='" +
      //    thisSemester.trim() + "' and userDeleted='0';";
      //else
      //  authenticateQuery =
      //    "SELECT password,fullname,IPAddress,instructor,controller," +
      //    "showGrid from userinfo where username ='" +
      //    testUsername.trim() + "' and courseID = '" +
      //    testCourseID.trim() + "' and userDeleted='0';";
      try
      {
        // query the database to authenticate, if query is successful
        //      such that the record is found, return true and set
        //      dbq.user object with required fields
        if (!dbq.authUserWithPassword( testUsername.trim(),
                                       testCourseID.trim(),
                                       thisSemester.trim() ) ) {
            return null;
        }
        readPassword = dbq.user.password;
        readFullname = dbq.user.fullname;
        readIPAddress = dbq.user.IPAddress;
        readInstructor = String.valueOf(dbq.user.instructor);
        readController = String.valueOf(dbq.user.controller);
        readShowGrid = String.valueOf(dbq.user.showGrid);
      }
      catch (SQLException ex)
      {
        return null;
      }
    } // end else/if(testUsername...

    dbq.close(); //close database connection
    
    // check password against the user's input
    // if readPassword is null, the user was not in database
    if(readPassword == null)return null;
    // next 4 commented out lines would perform IP address checking
    // if database held a required IP address; IPAddress field was
    // hijacked for Web authentication use
    //if(readIPAddress != null)
    //  if(readIPAddress.trim().equals(""))readIPAddress = null;
    if(checkPassword)
    {
      if(!readPassword.equals(testPassword.trim()))return null;
      //if(readIPAddress != null)
      //  if(!readIPAddress.equals(testUAddr.getHostAddress()))return null;
    }
     
    // authentication passed; now set associated table data
    checkForDeadHostTableEntry(testUsername);
    username[hostIndex] = padto8(testUsername);
    fullname[hostIndex] = readFullname;
    controlPrivilege[hostIndex] = (readController.toCharArray())[0];
    instructorPrivilege[hostIndex] = (new Integer(readInstructor)).intValue();
    drawGrid[hostIndex] = readShowGrid.equals("1");
    login.username = testUsername;
     
    return login;
    
  } // end authenticate method

  // file-based authentication; returns valid login(username only)
  // if matched in file, otherwise null
  public loginData fileAuthentication( int hostIndex, String testUsername,
    String testPassword, String testHostAddr, String testCourseID )
  {
    loginData login = new loginData();
    int userIndex = 0;
    boolean foundMatch = false;
    if(!authFileHasBeenRead)
      if(!readAuthFile(true))return null;
    
    // username blank is allowed for database with webserver but
    // not for file-based authentication
    if(testUsername.equals("        "))
      return null;

    // check for courseID match
    if( thisCourseID.trim().length() > 0 )
      if( !thisCourseID.trim().equals( testCourseID.trim() ) )return null;

    // assign an empty courseNum
    thisCourseNum = "00000";
    thisCourseNumBytes = thisCourseNum.getBytes();

    // search authentication table for match
    for( ; userIndex < authTableSize; ++userIndex )
      if( testUsername.equals(authUsername[ userIndex ] )
          && testPassword.equals(authPassWord[ userIndex ]) )break;
    if( userIndex < authTableSize )
    {
      if( (testHostAddr == null) || (authHostAddr[ userIndex ] == null) )foundMatch = true;
      else if( testHostAddr.equals(authHostAddr[ userIndex ]) )foundMatch = true;
    }
    if( !foundMatch)
    {
      // entry was not in table, but the table may have changed; read it again
      if(!readAuthFile(false))return null;

      // search authentication table again for match
      for( userIndex = 0; userIndex < authTableSize; ++userIndex)
        if( testUsername.equals(authUsername[ userIndex ])
          && testPassword.equals(authPassWord[ userIndex ]) )break;
      if( userIndex < authTableSize )
      {
        if( (testHostAddr == null) || (authHostAddr[ userIndex ] == null) )
          foundMatch = true;
        else if( testHostAddr.equals(authHostAddr[ userIndex ]) )
          foundMatch = true;
      }
    }
    if( !foundMatch )return null;

    // userIndex table entry matches authentication data;
    // move details from user table to host table
	checkForDeadHostTableEntry(testUsername);
    username[hostIndex] = padto8(testUsername);
    fullname[hostIndex] = padto40(authFullname[userIndex]);
    controlPrivilege[hostIndex] = authControlPrivilege[userIndex];
    drawGrid[hostIndex] = authDrawGrid[userIndex];
    login.username = testUsername;
    login.courseID = testCourseID;
    return login;
  } // end fileAuthentication method

  // confirm that a username is available (because it is not in authUserName[])
  // returns true if the name is not found
  public static boolean nameAvailable( String testUsername, String testCourseID )
  {
    // file-based authentication
    // search authentication table for match
    if(authFileName != null)
    {
      for (int userIndex = 0; userIndex < authTableSize; ++userIndex)
        if (testUsername.equals(authUsername[userIndex]))
          return false;
      return true;
    }
    // data objects for MySQL
    if(noDatabase)return true;
    
    // connect to database
    if(connectToDatabase())return true;

    // database authentication
    String readPassword = null;

    // old query, moved into DBQuery:
    // access database
    //    String authenticateQuery =
    //  "SELECT password from userinfo where username ='" +
    //  testUsername.trim() + "' and courseID = '" +
    //  testCourseID.trim() + "' and UserDeleted='0';";
    try
    {
      // query DB, return true if record is found and sets dbq.user
      if (dbq.nameAvailable(testUsername,testCourseID)) {
        readPassword = dbq.user.password;
      }
    }
    catch(SQLException ex)
    {
      // could not find username/courseID
      readPassword = null;
    }
    // close database connection
    dbq.close();
    // check password against the user's input
    return (readPassword != null);
  } // end nameAvailable method


  // method to read database and confirm a sessionVariable
  // is present and current
  // returns true for valid, false for invalid
  boolean validateSessionVariable(String sessionVariable,String testHostAddr)
  {
	// database authentication
    String readSessionVariable = null, readIPAddress = null;
    String authenticateQuery = null;

    // connect to database
    if(connectToDatabase())return false;


    // old query, moved into DBQuery:
    // lookup this sessionVariable in the database;
    // consider only posting in last 60 seconds
    // where courseID matches requiredCourseID
    //authenticateQuery =
    //    "SELECT courseID from userinfo,lastlogon " + // need IPAddress here...
    //    "where NEWSessID ='" + sessionVariable + "' and userDeleted='0';";
    // note from nick during DB SQL conversion for Moodle:
    //              the query above would never have worked but I put a fixed
    //              version in the new DBQuery class anyway

    boolean authenticatedOK = false;
    try
    {
        // Query database for session variable and check lastLogon
        if (dbq.validateSessionVariable(sessionVariable)) {
            return true;
        } else {
            System.out.println("no match found for session variable at IP address " +
                           testHostAddr);

        }

    } catch (SQLException ex) {
        System.out.println("MySQL exception querying database for session variable" + ex.getMessage());
        return false;
    }
    // close database connection
    dbq.close();
    return false;
  }

  // method to read the certifinfo table
  // parameter: certificate number
  // returns all info in table or null if not available
  public static synchronized certificateInfo getCertificateInfo(long certificateNumber)
  {
    certificateInfo returnData = new certificateInfo();

    // oldquery below was moved into DBQuery:
    // lookup this certificateNumber in the database;
    //String certificateQuery =
    //    "SELECT semester,courseID,username,downloaded," +
    //    "commonName,orgName,orgUnit,orgLoc,orgState,orgCountry," +
    //    "emailAddr,createDt,downloadDt,expiryDt " +
    //    "from certifinfo where serialNum = " +
    //    certificateNumber + ";" ;

    // connect to database
    if(connectToDatabase())return null;
     // access database
     try
     {
         //dbq.getCertificateInfo returns a TLM.certificateInfo object if found
         returnData = dbq.getCertificateInfo(certificateNumber);
     } catch (SQLException ex) {
         System.out.println("database exception getting certificate info " + ex.getMessage());
     }
    dbq.close();

    return returnData;

  } // end getCertificateInfo()


  // method to locate the next course start time in the database classlist table
  // returns start time for class in format of System.currentTimeMillis()
  // or zero if no class is scheduled
  //
  // assumption: no class session spans more than one calendar day
  //
  // side effects:
  // - sets todaysDatabaseDayCode to database code for today (see databaseCodeIsCloser below)
  // - sets printCourseDayCode to Mon, Tue, Wed, Thu, Fri, Sat, Sun matching scheduled course
  // - sets printThisDayOfWeek to Mon, Tue, Wed, Thu, Fri, Sat, Sun matching thisDayOfWeek
  // - sets nextCourseID[] to database value of scheduled courseIDs
  // - sets courseStartTime to database start in format of System.currentTimeMillis()
  // - sets printStartTime to database start in format hh:mm:ss
  // - sets courseEndTime to database end in format of System.currentTimeMillis()
  // - sets printEndTime to database end in format hh:mm:ss
  static boolean schedulesOverlap = false;
  static int todaysDatabaseDayCode;
  static long closestCourseDatabaseDayCode, candidateCourseDatabaseDayCode;
  static String timeOfDayNow;
  static synchronized long getNextClassTime()
  {
    // this method needs to access the database
	if(connectToDatabase())return 0;

    // code today's day of week so it will be available to compare
    todaysDatabaseDayCode = getTodaysDatabaseCode();

    // find time now and the number of seconds passed today
    GregorianCalendar gcalendar = new GregorianCalendar();
    int hour = gcalendar.get(Calendar.HOUR_OF_DAY);
    int minute = gcalendar.get(Calendar.MINUTE);
    int second = gcalendar.get(Calendar.SECOND);
    long secondsToday = 3600*hour + 60*minute + second;
    String strHour = "" + ( hour < 10 ? "0" + hour : "" + hour );
    String strMinute = "" + ( minute < 10 ? "0" + minute : "" + minute );
    String strSecond = "" + ( second < 10 ? "0" + second : "" + second );
    timeOfDayNow = strHour + ":" + strMinute + ":" + strSecond;
    
    // get the next class start time: classDayID+classStartTime
    // where the day code is 10000000L times the database day ID
    // code, which in turn is the decimal sum of day codes:
     //  Monday    = 1
     //  Tuesday   = 10
     //  Wednesday = 100
     //  Thursday  = 1000
     //  Friday    = 10000
     //  Saturday  = 1000000
     //  Sunday    = 10000000
    String closestCourseID = null;
    String closestCourseStartTime = "99:99:99";
    String closestCourseEndTime = "99:99:99";
    closestCourseDatabaseDayCode = 1000000000000000L;

    ResultSet result = null;

    // old query, moved to DBQuery:
    // access database to find the closest course start time
    //String courseTimeQuery =
    //       "SELECT courseID,courseNum,classDayID,classStartTime,hideName,classEndTime" +
    //       " from courselist where semester='" + thisSemester +
    //       "' and CID_live='" + roomID + "';";

    // access database to find the closest course start time
    String readCourseID, readCourseNum="", readClassStartTime, readClassEndTime;
    int readClassDayID;
    try
    {
        //Note: using DBQuery for this, but not elegant since we are passing
        //       the result set back, it may make it harder to modify columns names
        //       when using a different database
      result = dbq.getCourseInfo(thisSemester, roomID);

      // scan through the result table to find a closest courseID/day/end time
      // (we look for end time because the server might start in the middle
      // of a scheduled course time)
      while (result.next())
      {
        readCourseID = result.getString(1);
        readCourseNum = result.getString(2);
        String readClassDayString = result.getString(3);
        readClassStartTime = result.getString(4);
        hideFullname = !(result.getString(5).equals("0"));
        readClassEndTime = result.getString(6);

        // self-protection
        if((readCourseID == null) || (readCourseNum == null) ||
           (readClassDayString == null) ||
           (readClassStartTime == null) || (readClassEndTime == null))
        {
          System.out.println(
            "*** warning *** null entry in database courselist for course " +
            readCourseID);
          continue;
        }
        try
        {
          readClassDayID = (new Integer(readClassDayString)).intValue();
        }
        catch(NumberFormatException nfe)
        {
          System.out.println(
            "*** warning *** invalid entry in courselist classDayID for course " +
            readCourseID);
          continue;
        }

        // more self-protection
        if(readCourseID.length() > 11)readCourseID = readCourseID.substring(0,11);
        if(readCourseNum.length() > 5)readCourseNum = readCourseNum.substring(0,5);

        // check class time
        if(readClassStartTime.compareTo(readClassEndTime) >= 0)
        {
          System.out.println(
            "*** warning *** skipped entry " + readCourseID +
            " in course database because start time not less than end time");
          continue;
        }
        // compare first the day (modulo 7 days)
        if(!databaseCodeIsFarther(readClassDayID,
                                  readClassEndTime,
                                  closestCourseDatabaseDayCode,
                                  closestCourseEndTime))
        {
          // if the days are the same, compare the time
          if(nearestFutureDatabaseCode(readClassDayID, readClassEndTime) ==
             closestCourseDatabaseDayCode)
          {
            if(readClassEndTime.compareTo(closestCourseEndTime) < 0)
            {
              // check for overlapping (not same slot) courses case
              // where the previous closestCourse overlaps the new one
              if(readClassEndTime.compareTo(closestCourseStartTime) > 0)
              {
                System.out.print("*** warning *** overlapping courses in schedule:");
                System.out.println("  " + readCourseID + " and " + closestCourseID);
                schedulesOverlap = true;
              }

              // course just read ends earlier than current
              // candidate in the day - schedule it first
              numberOfCourseIDAliases = 0;
              closestCourseID = readCourseID;
              closestCourseDatabaseDayCode = candidateCourseDatabaseDayCode;
              closestCourseStartTime = readClassStartTime;
              closestCourseEndTime = readClassEndTime;
            }
            else
            {
              // check for overlapping (not same slot) courses case
              // where the course to be skipped overlaps the closestCourse
              if((readClassEndTime.compareTo(closestCourseEndTime) > 0) &&
                 (readClassStartTime.compareTo(closestCourseEndTime) < 0))
              {
                System.out.print("*** warning *** overlapping courses in schedule:");
                System.out.println("  " + readCourseID + " and " + closestCourseID);
                schedulesOverlap = true;
              }
            }
          } // end if(nearestFutureDatabaseCode...
          else
          // the days are not the same; therefore
          // nearestFutureDatabaseCode(readClassDayID) is smaller
          {
            numberOfCourseIDAliases = 0;
            closestCourseID = readCourseID;
            closestCourseDatabaseDayCode =
              nearestFutureDatabaseCode(readClassDayID, readClassEndTime);
            closestCourseStartTime = readClassStartTime;
            closestCourseEndTime = readClassEndTime;
          }

          // check for multiple course sections at the same time
          if(readClassStartTime.equals(closestCourseStartTime))
          {
            // section at same time- add to table if there is space
            if(numberOfCourseIDAliases >= MAX_COURSEID_ALIASES)
            {
              System.out.println(
                "Panic shutdown: too many courses in same time slot");
              System.out.println(
                "accepted courseIDs:" + CourseTimer.printCourseID());
              System.out.println(
                "acceptable number of simultaneous courses:" +
                MAX_COURSEID_ALIASES);
              System.out.println("first course over limit:" + readCourseID);
              System.exit(-1);
            }
            nextCourseID[numberOfCourseIDAliases] = readCourseID;
            nextCourseNum[numberOfCourseIDAliases++] = readCourseNum;
            if(!readClassEndTime.equals(closestCourseEndTime))
              System.out.println("*** warning *** course/conferenceID " +
                                 readCourseID + " shares time slot with " +
                                 nextCourseID[0] +
                                 " but does not have same end time");
          } // end if(readClassStartTime...
        } // end if(!database...
      } // end while(result.next())
    }
    catch (SQLException ex)
    {
      dbq.close();
      return 0;
    }
    dbq.close();
    // if overlapping schedules were found, bail out
    if(schedulesOverlap)
    {
      System.out.println(
        "*** autoscheduler shut down due to overlapping schedules ***");
      return 0;
    }

    // convert class start and end times to System.currentTimeMillis() format
    //
    // find the number of seconds to the start and end in their days
    long startSeconds =
      3600L *(new Integer(closestCourseStartTime.substring(0,2))).intValue() +
      60L *(new Integer(closestCourseStartTime.substring(3,5))).intValue() +
      (new Integer(closestCourseStartTime.substring(6,8))).intValue();

    long endSeconds =
        3600L *(new Integer(closestCourseEndTime.substring(0,2))).intValue() +
        60L *(new Integer(closestCourseEndTime.substring(3,5))).intValue() +
        (new Integer(closestCourseEndTime.substring(6,8))).intValue();

    // find the number of days from today the session will start
    int daysAway = 0;
    long checkCourseDatabaseDayCode = closestCourseDatabaseDayCode;

    // count days the course start is away from today, considering
    // it will be a week longer if the endTime has past today already
    if((checkCourseDatabaseDayCode == todaysDatabaseDayCode) &&
       (endSeconds <= secondsToday) &&
       (checkCourseDatabaseDayCode < 10000000L))
      checkCourseDatabaseDayCode *= 10000000L;
    while((checkCourseDatabaseDayCode != todaysDatabaseDayCode) &&
          (daysAway < 8))//bug-proofing
    {
      daysAway++;
      checkCourseDatabaseDayCode = checkCourseDatabaseDayCode / 10L;
    } // while(checkCourseDatabaseDayCode...

    // figure out daylight savings time offset
    long msNow = System.currentTimeMillis();
    java.util.Date thisDate = new java.util.Date(msNow);
	java.util.Date nextDate = new java.util.Date(msNow + 86400000L*daysAway);
	boolean thisDST = TimeZone.getDefault().inDaylightTime(thisDate);
    boolean nextDST = TimeZone.getDefault().inDaylightTime(nextDate);

    // Spring forward, Fall back by number of ms in an hour
    long daylightDelta = 0;
    if(!thisDST && nextDST)daylightDelta = -3600000L;
    if(thisDST && !nextDST)daylightDelta = 3600000L;

    // save values in Strings and ms
    printCourseDayCode = printDay(closestCourseDatabaseDayCode);
    nextCourseStart = (86400L*daysAway + startSeconds-secondsToday)*1000L +
                      msNow + daylightDelta;
    printStartTime = closestCourseStartTime;
    nextCourseEnd = (86400L*daysAway + endSeconds-secondsToday)*1000L +
                    msNow + daylightDelta;
    printEndTime = closestCourseEndTime;
    return nextCourseStart;
  } // end getNextClassTime()

  // method to compare two database class day codes in the future;
  // returns true if first is farther away than second from current day of
  // the week (e.g. 10000=F is farther than 10=T on 1=M, but not on 100=W)
  // where database day ID code is the decimal sum of day codes:
  //  Monday    = 1
  //  Tuesday   = 10
  //  Wednesday = 100
  //  Thursday  = 1000
  //  Friday    = 10000
  //  Saturday  = 1000000
  //  Sunday    = 10000000
  //  thus MW  = 101, TR  = 1010, MWF = 10101, etc.
  static boolean databaseCodeIsFarther(long dayOne,
                                       String timeOfDayOne,
                                       long dayTwo,
                                       String timeOfDayTwo)
  {
    // find nearest single day, ignoring days of week after the nearest
    // and considering whether the course time of day has already past today
    long nearestDayOne = nearestFutureDatabaseCode(dayOne, timeOfDayOne);
    long nearestDayTwo = nearestFutureDatabaseCode(dayTwo, timeOfDayTwo);

    // the smaller of the two is the one we want
    if(nearestDayOne > nearestDayTwo)
    {
      candidateCourseDatabaseDayCode = nearestDayTwo;
      return true;
    }
    candidateCourseDatabaseDayCode = nearestDayOne;
    return false;
  } // end databaseCodeIsFurther


  // method to find the nearest future day code of the seven that a
  // databaseCode can hold, always larger than the current day value
  static long nearestFutureDatabaseCode(long databaseCode, String courseTimeOfDay)
  {
    long result = 1000000000000000L;// larger than any possible day code
    long singleDayCode = 1L;
    long unprocessedValue = databaseCode;
    while(unprocessedValue > 0L)
    {
      // consider the low-order digit of the unprocessedValue
      if(unprocessedValue%2L > 0L)
      {
        // this means singleDayCode is present in databaseCode
        // if it has already occurred this week (including the case
        // for where it is today and the time of day has occurred)
        // shift it to next week
        long shiftDayCode = singleDayCode;
        if((shiftDayCode < todaysDatabaseDayCode) ||
           ((shiftDayCode == todaysDatabaseDayCode) &&
            courseTimeOfDay.compareTo(timeOfDayNow) <= 0))
          shiftDayCode *= 10000000L;

        //after shifting, is it the lowest considered to date?
        if(shiftDayCode < result)result = shiftDayCode;
      }
      // prepare for the next digit
      unprocessedValue = unprocessedValue/10L;
      singleDayCode *= 10L;
    } // end while(unprocessedValue...

    return result;

  } // end nearestFutureDatabaseCode()


  // method to return database class day code (see above) for today
  static int getTodaysDatabaseCode()
  {
    switch((new GregorianCalendar()).get(Calendar.DAY_OF_WEEK))
    {
      case Calendar.MONDAY:
        return 1;
      case Calendar.TUESDAY:
        return 10;
      case Calendar.WEDNESDAY:
        return 100;
      case Calendar.THURSDAY:
        return 1000;
      case Calendar.FRIDAY:
        return 10000;
      case Calendar.SATURDAY:
        return 100000;
      case Calendar.SUNDAY:
        return 1000000;
      default:
        return 0;
    } // end switch...

  } // end getTodaysDatabaseCode()


  // method to return String Mon, Tue, Wed, Thu, Fri, Sat, Sun
  // corresponding to high-order digit of database day code
  static String[] daysOfWeek = {"Mon","Tue","Wed","Thu","Fri","Sat","Sun"};
  static String printDay(long databaseDay)
  {
    int day = -1;
    long testDatabaseDay = databaseDay;
    while(testDatabaseDay > 1000000)
      testDatabaseDay = testDatabaseDay/10000000L;
    while (testDatabaseDay > 0L)
    {
      ++day;
      testDatabaseDay = testDatabaseDay/10L;
    }
    if(day < 0)return "";
    return daysOfWeek[day];
  } // end printDay()
  
  // common procedure for database access
  // returns true if unsuccesful, false otherwise
  static boolean connectToDatabase()
  {
    // test whether this failed before
    if(noDatabase)return true;
	
    // connect to database
	//databaseConnection = null;
        dbq = null; //nick
    try {
      String databaseURL =
          "jdbc:mysql://" + databaseHost + "/" + databaseName + "?user=" +
          databaseUsername;

      dbq = new DBQuery("jdbc:mysql://" + databaseHost + "/" + databaseName,
                        databaseUsername, databasePassword, databasePrefix); //nick

      //Class.forName("com.mysql.jdbc.Driver");
      //databaseConnection = DriverManager.getConnection(databaseURL,
      //                                                 databaseAccount,
      //                                                 databasePassword);
    }
    catch (java.lang.ClassNotFoundException ee) {
      warnNoDatabase("MySQL access ClassNotFound:" + ee);
	  return true;
    }
    catch (SQLException ex) {
      warnNoDatabase("MySQL exception:" + ex);
	  return true;
    }
    return false;
	
  }// end connectToDatabase()
  
  //==== below here are classes & methods for breakout groups ====
   
  // classes to group data for one user
  private class UserData
  {
    String username;
	String fullname;
	String courseID;
	byte groupNumber;
  }
  private class GroupData
  {
    String username;
	String courseID;
	byte groupNumber;
  }
  UserData[] userData = new UserData[MAX_USERS];
  GroupData[] groupData = new GroupData[MAX_USERS];
  int numberOfConfiguredUsers = 0;
  int numberOfUsersInGroupData = 0;
  
  
  // copies a String and pads it with blanks, or clips if too long
  String padString(String oldString, int padLength)
  {
    String result;
	if(oldString == null)result = "";
	else 
	  if(oldString.length() <= padLength)
	    result = new String(oldString);
	  else result = new String(oldString.substring(padLength));
	while(result.length() < padLength)result += ' ';
	return result;
  }
	
  
  // extract from database and send to all FC the
  // current breakout group assignments 
  // returns true if problem in making/sending; false otherwise
  boolean sendBreakoutData()
  {
    String readUsername, readFullname;
    byte readGroupNumber;
    Statement statement = null;
    ResultSet result = null;
    boolean foundDummyUsername = false;
    int userIndex = 0;
    int groupDataIndex = 0;
    int courseIDIndex = 0;
    numberOfBreakoutGroups = 0;
    	
    // connect to database
    if(connectToDatabase())return true;
    
    // read all valid usernames with courseID and fullname from database USERINFO
    userIndex = -1;
    for(courseIDIndex = 0; 
        courseIDIndex <= numberOfCourseIDAliases;
        courseIDIndex++)
    {
      // old query,  moved to DBQuery
      // lookup all users for this courseID and semster in the database
      //String userDataQuery =
      //  "SELECT username,fullname from userinfo" +
      //  " where semester= '" + thisSemester + 
      //	"' and courseID= '" + nextCourseID[courseIDIndex] +
      //	"' and userDeleted='0' ORDER BY username;";

      // access database
      try
      {
        // lookup all users for this courseID and semster in the database
        //    returns the ResultSet object from the query
        //result = dbq.getUsersInCourse(thisSemester, nextCourseID[courseIDIndex]);
        result = dbq.getUsersInCourse(roomID, nextCourseID[courseIDIndex]);
        
        // copy the result table to userData
        while(result.next())
        {
          readUsername = result.getString(1);
          readFullname = result.getString(2);
          if(++userIndex < userData.length)
          {
            userData[userIndex] = new UserData();
            userData[userIndex].username = padString(readUsername,8);
            userData[userIndex].fullname = padString(readFullname,40);
            userData[userIndex].courseID = padString(nextCourseID[courseIDIndex],11);

            // default group number is 1
            userData[userIndex].groupNumber = 1;
             
            // breakoutController group is 0
            if(userData[userIndex].username.equals(breakoutControllerUsername))
              userData[userIndex].groupNumber = 0;
          }
          else // table overflow
          {
            System.out.println("*** userData table overflow in Auth.sendBreakoutData");
            return true;
          }
        }// end while(result.next())
      }
      catch(SQLException s)
      {
        System.out.println("*** SQL exception in Auth.sendBreakoutData Q1:" + s.getMessage());
        return true;
       }
    }// end for(courseIDIndex = 0
    numberOfConfiguredUsers = userIndex+1;
    
    if(userIndex < 0)
    {
      System.out.println("*** no userData found in Auth.sendBreakoutData");
      return true;
    }
    
    // read persistent group assignments from database table usergroups
    groupDataIndex = -1;
    for(courseIDIndex = 0; 
        courseIDIndex <= numberOfCourseIDAliases;
        courseIDIndex++)
    {
      // old query, moved to DBQuery
      // String groupDataQuery =
      //  "SELECT username,groupnum from usergroups" +
      //  " where semester= '" + thisSemester +
      //  "' and courseID= '" + nextCourseID[courseIDIndex] +
      //  "' ORDER BY username;";
      // access database
      try
      {

        // get ResultSet object from database query
        result = dbq.getUserGroupAssignments(roomID, nextCourseID[courseIDIndex]);
        
        // copy the result table to groupData
        while(result.next())
        {
          readUsername = result.getString(1);
          readGroupNumber = result.getByte(2);
          if(++groupDataIndex < MAX_USERS)
          {
            if(readUsername.equals("--------"))// dummy holding number of groups
            {
              foundDummyUsername = true;
              --groupDataIndex;
	      numberOfBreakoutGroups = readGroupNumber;	
              continue;
            }
            groupData[groupDataIndex] = new GroupData();
            groupData[groupDataIndex].username = padString(readUsername,8);
            groupData[groupDataIndex].courseID = nextCourseID[courseIDIndex];
            if(readGroupNumber > 0)
              groupData[groupDataIndex].groupNumber = readGroupNumber;
            else
              groupData[groupDataIndex].groupNumber = 1;
          }
          else // table overflow
          {
            System.out.println("*** groupData table overflow in Auth.sendBreakoutData");
            return true;
          }
        }// end while(result.next())
      }
      catch(SQLException s)
      {
        System.out.println("*** SQL exception in Auth.sendBreakoutData Q2:" + s.getMessage());
        return true;
      }
    }//end for(courseIdIndex...
    numberOfUsersInGroupData = groupDataIndex+1;
     
    // sort the userData on username using bubble sort
    for(int highEnd = numberOfConfiguredUsers;
        highEnd > 0;
        --highEnd)
    {
      for(int bubbleUp = 1;
          bubbleUp < highEnd;
          ++bubbleUp)
        if(userData[bubbleUp-1].username.
           compareTo(userData[bubbleUp].username) > 0)
        {
          UserData swap = userData[bubbleUp-1];
          userData[bubbleUp-1] = userData[bubbleUp];
          userData[bubbleUp] = swap;
        }// end if(userData...
      }// end for(int highEnd
      
      // eliminate any multiples of username in userData
      int checkDup = 1;
      while(checkDup < numberOfConfiguredUsers)
      {
        if(userData[checkDup].username.equals(userData[checkDup-1].username))
        {
          // ripple down entries in userData to cover the duplicate
          for(int rippleDown = checkDup;
              rippleDown < numberOfConfiguredUsers-1;
              ++rippleDown)
              userData[rippleDown] = userData[rippleDown+1];
          numberOfConfiguredUsers--;
        }
        else checkDup++;
      }// end while(checkDup...
      
      // sort the groupData on username using bubble sort
      for(int highEnd = numberOfUsersInGroupData;
          highEnd > 0;
          --highEnd)
      {
        for(int bubbleUp = 1;
            bubbleUp < highEnd;
            ++bubbleUp)
        if(groupData[bubbleUp-1].username.
           compareTo(groupData[bubbleUp].username) > 0)
        {
          GroupData swap = groupData[bubbleUp-1];
          groupData[bubbleUp-1] = groupData[bubbleUp];
          groupData[bubbleUp] = swap;
        }// end if(groupData...
      }// end for(int highEnd
      
      // eliminate any multiples of username in groupData
      checkDup = 1;
      while(checkDup < numberOfUsersInGroupData)
      {
        if(groupData[checkDup].username.equals(groupData[checkDup-1].username))
        {
          // ripple down entries in userData to cover the duplicate
          for(int rippleDown = checkDup;
              rippleDown < numberOfUsersInGroupData-1;
              ++rippleDown)
           groupData[rippleDown] = groupData[rippleDown+1];
          numberOfUsersInGroupData--;
        }
	      else checkDup++;
      }// end while(checkDup...
      
      // resolve any differences by updating userGroup
      
      // if we didn't find a dummy username "--------" in usergroups, add one with group 0
      String addUserQuery;
      if(!foundDummyUsername)
      {
        // old query, moved to DBQuery
        //addUserQuery = "INSERT INTO usergroups VALUES('" +
	//			 thisSemester.trim() +
	//			 "','" +
        //                       nextCourseID[0] +
        //                       "','--------','0');";
	try
	{
          dbq.insertToUserGroups(roomID, nextCourseID[0], "--------", 0);
	}
	catch(SQLException s)
	{
	  System.out.println("MySQL exception while adding dummy user to database:" +
                             s.getMessage());
          return true;
	}
      }// end if(!foundDummyUsername)
	
      // step through all usernanmes from userData, matching with groupData
      groupDataIndex = 0;
      for(userIndex = 0; 
          userIndex < numberOfConfiguredUsers;
          ++userIndex)
      {
        // if the username in userData is greater than the one in groupData, 
        // that groupData has been deleted from userinfo and we can step over it
        while(groupData[groupDataIndex] != null)
          if(userData[userIndex].username.compareTo(groupData[groupDataIndex].username) > 0)
   	    groupDataIndex++;
          else
            break;
        
        // if the usernames in userData and groupData are equal, copy
	      // the groupNumber from groupdata to userData and add to host table
        if(groupData[groupDataIndex] != null)
        {
          if(userData[userIndex].username.equals(groupData[groupDataIndex].username))
          {
            userData[userIndex].groupNumber = groupData[groupDataIndex].groupNumber;
          
            // also update the group number in host table, if the user is logged in
	          addGroupNumberToHostTable(userData[userIndex].username,
            groupData[groupDataIndex].groupNumber);
	        }
          // at this point the username exists in userData but is less than the one in
          // groupData, indicating it has been inserted since last run of this method; 
          // so insert it in the groupData (bump all other entries up by one position 
          // to do this)
          else
          {
            // move all groupData entries up one position to make room for this entry
            if(numberOfUsersInGroupData >= MAX_USERS)
            {
              System.out.println("user count exceeds table space " + MAX_USERS +
                                 " while adding group users 1 to database");
            return true;
          }
          for(int ripplePosition = numberOfUsersInGroupData;
              ripplePosition >= groupDataIndex;
              --ripplePosition)
	          groupData[ripplePosition+1] = groupData[ripplePosition];
          numberOfUsersInGroupData++;
	  	  
          // insert the groupData
          groupData[groupDataIndex].username = userData[userIndex].username;
          groupData[groupDataIndex].courseID = userData[userIndex].courseID;
          groupData[groupDataIndex].groupNumber = userData[userIndex].groupNumber;

          // old query, moved to DBQuery:
          // add the new groupData entry to the database
          //addUserQuery = "INSERT INTO usergroups VALUES('" +
          //				 thisSemester.trim() +
          //				 "','" +
          //				 userData[userIndex].courseID.trim() +
          //				 "','" +
          //				 userData[userIndex].username.trim() +
          //				   "','0');";
          try
          {
            dbq.insertToUserGroups(roomID,
                                   userData[userIndex].courseID.trim(),
                                   userData[userIndex].username.trim(), 0);
          }
          catch(SQLException s)
          {
            System.out.println("MySQL exception while adding group users 3 to database:" +
                               s.getMessage());
            return true;
          }
		  
          // also update the group number in host table, if the user is logged in
	        addGroupNumberToHostTable(userData[userIndex].username,
                                   groupData[groupDataIndex].groupNumber); 
	      }
      }// end if(groupData
      else
      {
        // the group data table is exhausted; just move the data from userData
        groupData[groupDataIndex] = new GroupData();
        groupData[groupDataIndex].username = userData[userIndex].username;
        groupData[groupDataIndex].courseID = userData[userIndex].courseID;
        groupData[groupDataIndex].groupNumber = userData[userIndex].groupNumber;
        numberOfUsersInGroupData++;
	
        // old query, moved to DBQuery:
        // and add it to the groupData in database
        //addUserQuery = "INSERT INTO usergroups VALUES('" +
        //                           thisSemester.trim() +
        //                           "','" +
        //                           userData[userIndex].courseID.trim() +
        //                           "','" +
        //                           userData[userIndex].username.trim() +
        //                           "','0');";
	try
	{
          dbq.insertToUserGroups(roomID,
                                 userData[userIndex].courseID.trim(),
                                 userData[userIndex].username.trim(), 0);
	}
	catch(SQLException s)
	{
          System.out.println("MySQL exception while adding group users 2 to database:" +
		              s.getMessage());
          return true;
	}
      }// end else/if(groupData
      groupDataIndex++;
      
    }// end for(userIndex...
    
    // resort the userData on fullname using bubble sort
    for(int highEnd = numberOfConfiguredUsers;
        highEnd > 0;
        --highEnd)
    {
      for(int bubbleUp = 1;
          bubbleUp < highEnd;
          ++bubbleUp)
        if(userData[bubbleUp-1].fullname.
           compareTo(userData[bubbleUp].fullname) > 0)
        {
          UserData swap = userData[bubbleUp-1];
          userData[bubbleUp-1] = userData[bubbleUp];
          userData[bubbleUp] = swap;
        }// end if(userData...
    }// end for(int highEnd
     
    // format and send a breakout group data message to all FC
    // if using multicast for FC, the UDP message will only support 29 users
    if(usingMcFC && (numberOfConfiguredUsers > mcFCUserLimit ))
    {
       System.out.println("*** multicast breakout FC usercount exceeded -- limit " +
                          mcFCUserLimit + " found " + numberOfConfiguredUsers);
       return true;
    }
	
    // build and send the breakout initialization message
    if(sendFCBreakoutMessage(-1))
    {
      dbq.close();
      return true;
    }
    
    // close database and quit    
    dbq.close();
    return false;
	
  }// end sendBreakoutData
    
  
  // build and send a message with FC breakout data
  // if parameter is -1, send to all connected hosts
  // returns true if failure
  boolean sendFCBreakoutMessage(int hostIndex)
  {
    // if using multicast for FC, the UDP message will only support 24 users
    if(usingMcFC && (numberOfConfiguredUsers > mcFCUserLimit ))
    {
       System.out.println("*** multicast breakout FC usercount exceeded -- limit " +
                          mcFCUserLimit + " found " + numberOfConfiguredUsers);
       return true;
    }
     
    // build the message
    int breakoutDataMessageLength = 3 + 49*numberOfConfiguredUsers;
    breakoutDataMessage = new byte[breakoutDataMessageLength];
    breakoutDataMessage[0] = (byte)numberOfBreakoutGroups;
    breakoutDataMessage[1] = intToByte1(numberOfConfiguredUsers);
    breakoutDataMessage[2] = intToByte2(numberOfConfiguredUsers);
    int basePosition = 3;
    for(int userNumber = 0; 
        userNumber<numberOfConfiguredUsers; 
        ++userNumber)
    { 
      // make an entry for the next user, padding to fill fields
      System.arraycopy(userData[userNumber].username.getBytes(),0,
                       breakoutDataMessage,basePosition,
                       8);
      System.arraycopy(userData[userNumber].fullname.getBytes(),0,
                       breakoutDataMessage,basePosition+8,
                       40);
      if(userData[userNumber].username.equals(breakoutControllerUsername))
        breakoutDataMessage[basePosition+48] = 0;
      else
        breakoutDataMessage[basePosition+48] = 
          userData[userNumber].groupNumber;
      basePosition += 49;
    }
    
    // send the message
    if(hostIndex < 0)
      fcConnect.writeMessageToAll(31,
                                  breakoutDataMessageLength,
                                  breakoutDataMessage,
                                  false);
    else
      fcConnect.sendToFC(31,breakoutDataMessage.length,breakoutDataMessage,hostIndex);
    return false;
    
  }// end sendFCBreakoutMessage()  
    

  // adds groupNumber to position in host table matching that of username
  // returns true if position can't be found, false otherwise 
  boolean addGroupNumberToHostTable(String findUsername,byte useGroupNumber)
  {
    // lookup matching position (this could benefit from some sort
    // of ordered table for binary search, or from hashing)
    int searchTable = 0;
    for(searchTable = 0; searchTable < maxConnections; ++searchTable)
      if(findUsername.equals(username[searchTable]))break;
    if(searchTable >= maxConnections)return true;
    if(searchTable == breakoutController)return false;
    breakoutGroup[searchTable] = useGroupNumber;
    return false;
  }
  
  
  // updates breakout group assignment in host table, database, and all FC
  // returns true if problem in making/sending; false otherwise
  boolean setBreakoutGroup(byte setGroupNumber,String setUsername)
  {
    if(connectToDatabase())return true;
	
    // find user's position in host table
    int tablePosition = fcConnect.getUserPosition(setUsername);
      
    // update the users group in host table
    if(tablePosition >= 0)
      breakoutGroup[tablePosition] = setGroupNumber;
	  
    // local user in userData
    int userIndex = 0;
    for(; userIndex<numberOfConfiguredUsers; ++userIndex)
      if(userData[userIndex].username.equals(setUsername))break;
    if(userIndex >= numberOfConfiguredUsers)
    {
      System.out.println("*** unmatched username |"+setUsername+
			 "| in setBreakoutGroup");
      return true;
    }
    
    // update the user's group in database
    if(connectToDatabase())return true;
    String setGroupNumberString = (new Integer(setGroupNumber)).toString();

    // old query, moved to DBQuery:
    //String updateUserQuery = "UPDATE usergroups SET groupnum='" +
    //                                                 setGroupNumberString.trim() +
    //                                                 "' WHERE semester='" +
    //                                                     thisSemester.trim() +
    //                                                     "' and username= '" +
    //                                                 setUsername.trim() +
    //                                                 "' and courseID='" +
    //                                                 userData[userIndex].courseID.trim() +
    //                                                 "';";
				 
    try
    {
        dbq.updateUserGroups(roomID.trim(), setUsername.trim(), setGroupNumber);
    }
    catch(SQLException s)
    {
      System.out.println("MySQL exception while updating breakout groups in database:" +
                         s.getMessage());
      return true;
    }
	
    // build message to update all FC with new group assignment
    byte[] changeBreakoutGroupMessage = new byte[9];
    changeBreakoutGroupMessage[0] = setGroupNumber;
    System.arraycopy(padString(setUsername,8).getBytes(),0,changeBreakoutGroupMessage,1,8);
    fcConnect.writeMessageToAll(26,9,changeBreakoutGroupMessage,false);
    return false;

  }// end setBreakoutGroup
  
  
  // updates number of groups for current breakout
  // does not change breakout status
  boolean setNumberOfGroups(byte setNumber)
  {
      // persist the number in database table usergroups
    if(connectToDatabase())return true;

    // old query, moved to DBQuery:
    //String updateUserQuery = "UPDATE usergroups SET groupnum='" +
    //                     setNumber +
    //                     "' WHERE semester='" +
    //                         thisSemester.trim() +
    //                         "' and username='--------' and courseID='" +
    //                     nextCourseID[0].trim() +
    //                     "';";

    try
    {
      dbq.updateUserGroups(roomID.trim(), "--------", setNumber);
    }
    catch(SQLException s)
    {
      System.out.println("MySQL exception while updating number of groups in database:" +
                         s.getMessage());
      return true;
    }
	
	// if number of groups got smaller, reset database entry for any user with larger value
	// do this both for userData and host table; for userData also update database
	if(numberOfBreakoutGroups > setNumber)
	{
	  for(int userNumber=0; userNumber<numberOfConfiguredUsers; ++userNumber)
	  {
	    if(userData[userNumber].groupNumber > setNumber)
            {
              userData[userNumber].groupNumber = 1;	  
              try
              {
                dbq.updateUserGroups(thisSemester.trim(),
                      userData[userNumber].username.trim(), 0);
              }
              catch(SQLException s)
              {
                System.out.println("MySQL exception while updating number of groups in database:" +
                s.getMessage());
		return true;
              }		  
            }// end if(userData...
	  }// end for(int userNumber...
	  
	  // reset users in host table to group 0
	  for(int hostNumber = 0; hostNumber<maxConnections; hostNumber++)
	    if(breakoutGroup[hostNumber] > setNumber)
		  breakoutGroup[hostNumber] = 0;
	  
	}// end if(numberofBreakoutGroups > setNumber)
	 
	// reset the server number of groups and return
	numberOfBreakoutGroups = setNumber;
	return false;
    
  }// end setNumberOfGroups
  
  
  // checks groupData[] for matching username;
  // returns user's group number if found, otherwise -1
  int getGroup(String findUsername)
  {
    if(groupData == null)return -1;
	for (int userScan=0; userScan < numberOfConfiguredUsers; ++userScan)
	  if(findUsername.equals(groupData[userScan].username))
	    return groupData[userScan].groupNumber;
	return -1;
  }// end getGroup()
  
} // end class Auth
