/*----------------------------------------------------------------*
|    Copyright 2001-2008 Networking and Simulation Laboratory     |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

/**
 * Title:        CoverSlide
 * Description:  generates a cover slide for a NEW .nwr file
 * Copyright:    Copyright (c) 2003-2004
 * Company:      GMU NETLAB
 * @author       Mark Pullen (JMP)
 * @version 1.0
 */


/** invocation parameters:
 *
 *  outputFile        : a random access NEW nwr file, opened for read/write
 *  outFilePathName   : text string, path + filename for the output file
 *  coverSlidePathName: text string, path + filename for the cover slide file
 *                      (if null or "0", no cover slide will be generated)
 *  coverAudioPathName: text string, path + filename for the cover slide file
 *                      (if null or "0", cover slide will be silent)
 *  messageHeader     : byte array containing a header from input file;
 *                      needed for timestamp continuity
 *  pageBlock         : byte containing sequence number for block of WBD pages
 *
 *  format of file at coverSlidePathName:
 *      line 1: URL for cover slide
 *      line 2: subtitle text1 for cover slide (semester etc)
 *      line 3: subtitle text2 for cover slide (semester etc)
 *      line 4: copyright line for cover slide
 *      line 5: copyright conditions (max 500 bytes; no hard returns)
 *    The lines are separated with carriage returns.
 *
 *    coverAudio file is a random access NEW nwr file, containing sound
 *      track 10 seconds or less to be inserted after cover slide
 *
 *  returns: 0 if cover slide successfully built, non-zero otherwise
 *
 *  Produces a cover slide:
 *   black::Network EducationWare Recording (centered)
 *   blue::<URL> (centered)
 *   blue::filename (centered)
 *   blue::subtitle text1 (centered)
 *   blue::subtitle text2 (centered)
 *   red::copyright line (centered)
 *   red::copyright conditions (left justified, autowrap)
 */

import java.io.*;
import java.util.*;

public class CoverSlide
{
  protected static final int MAX_MESSAGE_LENGTH = 1455;
  protected static final byte audio = 0, whiteboard = 1;
  static int bytesWritten, commandNumber = 0;
  static RandomAccessFile coverSlideFile;

  static byte[] carriageReturn = ("\r").getBytes();
  static byte[] lineFeed = ("\n").getBytes();
  static byte[] blank = (" ").getBytes();

  static byte[] coverSender = {127,0,0,1};

  static byte[] drawNewPageMessage =
  {127,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  1,127,  0,  0,  1,  0,  0,  0,  1,
     0,  0,  0,  1,  0,  0,  0,  1,  0,  0,  0,  0,  0, 23,  0, 16,117,110,107,110,
   111,119,110, 65, 78, 83, 79, 78, 46,103,109,117, 46,101,100,117, 58, 49,  0};

  static byte[] drawNewLabelMessage =
  {127,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  1,127,  0,  0,  1,  0,  0,  0,  1,
     0,  0,  0,  1,  0,  0,  0,  1,  0,  0,  0,  0,  0, 32,  0, 17, 67,111,112,121,
   114,105,103,104,116,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0};

  static byte[] drawTextCore =
  {127,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  1,127,  0,  0,  1,  0,  0,  0,  1,
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 24,  0, 19,  0,  0,  0, 90,
     0,  0,  0,  0,  0,  0,  0,  0};

  static byte[] colorRed   = { -2,  0,  0,  0};

  static byte[] colorGreen = {  0,127,  0,  0};

  static byte[] colorBlue  = {  0,  0, -2,  0};

  static byte[] colorBlack = {  0,  0,  0,  0};

  static byte[] helvetiaFontSize36 = {0,  6,  0,  36};

  static byte[] helvetiaFontSize32 = {0,  6,  0,  32};

  static byte[] helvetiaFontSize24 = {0,  6,  0,  24};

  static byte[] helvetiaFontSize18 = {0,  6,  0,  18};

  static byte[] newRecording = ("Network EducationWare Recording").getBytes();

  static float[] fontWidth = // offset by 32 because first 32 are not used
  {11,12,17,20,20,28,25,10,13,13,14,21,11,13,11,11,20,20,20,20,20,20,20,20,
   20,20,12,12,21,21,21,22,33,25,24,25,25,23,22,27,24,10,19,25,22,29,24,26,
   22,26,24,23,22,24,23,33,23,22,22,13,11,13,21,20,13,20,21,20,21,20,13,21,
   21,12,12,20,12,30,21,21,21,21,15,20,13,21,19,27,20,19,18,14,11,14,21,26};

  static float leftXPosition = 50;
  static float centerXPosition = 396;
  static float lineOneYPosition = 100;
  static float lineTwoYPosition = 150;
  static float lineThreeYPosition = 200;
  static float lineFourYPosition = 250;
  static float lineFiveYPosition = 300;
  static float lineSixYPosition = 350;
  static float lineSevenYPosition = 400;
  static float screenXMax = 792;

 /**
  *  creates a new cover slide in the output file
  *
  *  returns number of bytes written, or a negative error code
  */
  public static int writeCoverSlide(RandomAccessFile outFile,
                                   String outFilePathName,
                                   String coverSlidePathName,
                                   String coverAudioPathName,
                                   byte[] messageHeader,
                                   byte pageBlock)
  {
    // open the file with information for cover slide
    if(coverSlidePathName == null)return -1;
    if(coverSlidePathName == "0")return -1;
    try{coverSlideFile = new RandomAccessFile(coverSlidePathName,"r");}
    catch(IOException ioe)
    {
      System.out.println("file " + coverSlidePathName +
        " not found - no cover slides will be generated");
      return -1;
    }

    // initialize output counter
    bytesWritten = 0;

    // write an inital SessionMessage to the output file 10000 ms before
    // the template message
    byte[] newSessionPage = {0,0,0,0};
    newSessionPage[1] = pageBlock;
    RESEQ.addPageNumber(newSessionPage,0);
    byte[] newSessionMessage = makeSessionMessage(coverSender);
    if(!writeMessage(outFile,messageHeader,10000,newSessionMessage,
                     newSessionMessage.length,false,whiteboard))return -1;

    // write a drawNewLabelMessage to the output file
    System.arraycopy(intToBytes(commandNumber),0,drawNewLabelMessage,24,4);
    System.arraycopy(intToBytes(commandNumber),0,drawNewLabelMessage,28,4);
    drawNewLabelMessage[17] = pageBlock;
    drawNewLabelMessage[19] = 0;
    RESEQ.updatePageNumber(drawNewLabelMessage,16);
    if(!writeMessage(outFile,messageHeader,10000,drawNewLabelMessage,
                     drawNewLabelMessage.length,false,whiteboard))return -1;

    // insert black::Network EducationWare Recording (centered)
    drawTextCore[17] = pageBlock;
    drawTextCore[19] = 0;
    if(!emitText(outFile,newRecording,messageHeader,
                 centerXPosition,lineOneYPosition,
                 colorBlack,helvetiaFontSize36,true))
      return -1;

    // insert blue::<URL> (centered)
    if(!emitText(outFile,readCoverSlideFile(),messageHeader,
                 centerXPosition,lineTwoYPosition,
                 colorBlue,helvetiaFontSize32,true))
      return -1;

    // insert blue::filename (centered)
    int backSlashPosition = outFilePathName.lastIndexOf('\\');
    int slashPosition = outFilePathName.lastIndexOf('/');
    if(backSlashPosition > slashPosition)slashPosition = backSlashPosition;
    byte[] fileName =
      outFilePathName.substring(slashPosition+1,outFilePathName.length()).getBytes();
    if(!emitText(outFile,fileName,messageHeader,
                 centerXPosition,lineThreeYPosition,
                 colorBlue,helvetiaFontSize32,true))
      return -1;

    // insert blue::subtitle text1 (centered)
    if(!emitText(outFile,readCoverSlideFile(),messageHeader,
                 centerXPosition,lineFourYPosition,
                 colorBlue,helvetiaFontSize24,true))
      return -1;

    // insert blue::subtitle text2 (centered)
    if(!emitText(outFile,readCoverSlideFile(),messageHeader,
                 centerXPosition,lineFiveYPosition,
                 colorBlue,helvetiaFontSize24,true))
      return -1;

    // insert red::copyright line (centered)
    if(!emitText(outFile,readCoverSlideFile(),messageHeader,
                 centerXPosition,lineSixYPosition,
                 colorRed,helvetiaFontSize32,true))
      return -1;

    // insert red::copyright conditions (left justified, autowrap)
    if(!emitText(outFile, readCoverSlideFile(),messageHeader,
                 leftXPosition,lineSevenYPosition,
                 colorRed,helvetiaFontSize24,false))
      return -1;

    // open Welcome.nwr file; if missing, return
    RandomAccessFile welcomeFile = null;
    try
    {
      welcomeFile = new RandomAccessFile(coverAudioPathName,"r");
    }
    catch(FileNotFoundException f){return bytesWritten;}
    if(welcomeFile == null || coverAudioPathName == "0")return bytesWritten;
    if(!coverAudioPathName.endsWith(".nwr"))
    {
      System.out.println("cover audio file:"+welcomeFile+
        " extension is not .nwr - ignoring file" );
      return  bytesWritten;
    }

    // copy Welcome.new file to output
    byte[] welcomeHeader = new byte[9];
    byte[] welcomeMessage = null;
    int messageLength = 0;
    boolean firstWelcomeMessage = true;
    long firstWelcomeTimestamp = 0, welcomeTimestamp;

    try
    {
      while(true)
      {
        // read header
        if(welcomeFile.read(welcomeHeader) == -1)break;
        if(firstWelcomeMessage)firstWelcomeTimestamp = bytesToInt(welcomeHeader);
        welcomeTimestamp = bytesToInt(welcomeHeader);
        messageLength = bytesToShort(welcomeHeader[4], welcomeHeader[5]);
        if(messageLength > MAX_MESSAGE_LENGTH)
        {
          System.out.println("record overlength in Welcome.new file, length:"
            + messageLength);
          break;
        }

        // read and write message
        welcomeMessage = new byte[messageLength];
        if(welcomeFile.read(welcomeMessage,0,messageLength) == -1)
        {
          System.out.println("incomplete record in Welcome.nwr file");
          break;
        }
        if(!writeMessage(outFile,messageHeader,
                         10000 - (welcomeTimestamp - firstWelcomeTimestamp),
                         welcomeMessage,messageLength,firstWelcomeMessage,audio))
          return -2;
        firstWelcomeMessage = false;
      } // end while(true)
    }
    catch(IOException ioe)
    {
      System.out.println("IO Exception in Welcome.nwr copy:"+ioe);
      System.out.println("shutting down - output file incomplete");
      return -3;
    }

    // write a NewPageMessage to the output file, timestamped
    // at end of 10 second welcome window
    System.arraycopy(intToBytes(commandNumber),0,drawNewPageMessage,24,4);
    System.arraycopy(intToBytes(commandNumber),0,drawNewPageMessage,28,4);
    drawNewPageMessage[17] = pageBlock;
    drawNewPageMessage[19] = 0;
    RESEQ.updatePageNumber(drawNewPageMessage,16);
    if(!writeMessage(outFile,messageHeader,0,drawNewPageMessage,
                     drawNewPageMessage.length,true,whiteboard))return -1;
    return bytesWritten;
  } // end makeCoverSlide method

 /**
  *  emits DrawText messages in the output file, corresponding to input string
  *
  *  parameters:
  *    text        : string to be placed on WBD
  *    xBase       : float x postion of lower left corner of the string on page
  *    yBase       : float y position of lower left corner of the string on page
  *    textColor   : byte[4] color for text, from WBD palette
  *    font        : byte[4] style and size of font in WBD coding
  *    centerText  : boolean; if true (x,y) is bottom center of text;
  *                  otherwise, it is bottom left; if false case, string
  *                  is word-wrapped
  *
  *  returns true if successful, false otherwise
  *
  * NOTE: max characters in a drawText message for current WBD is 107;
  *       we avoid that limit by sending at most one line per message.
  */
  static boolean emitText(RandomAccessFile outFile, byte[] text,
                      byte[] recordingHeader, float x, float y,
                      byte[] textColor, byte[] font, boolean centerText)
  {
    int thisChar, numberOfChars = text.length, firstCommand = commandNumber;
    boolean result;
    float xStart, xBase = x, yBase = y;
    byte charIndex;

    // set up a message core
    int variablePart = 12*numberOfChars;
    int fixedPart = 48;
    int messageSize = fixedPart + variablePart;
    byte[] drawTextMessage= new byte[messageSize];
    System.arraycopy(drawTextCore,0,drawTextMessage,0,fixedPart);

    // insert values that do not change: page number, color and font
    RESEQ.updatePageNumber(drawTextMessage,16);
    System.arraycopy(textColor,0,drawTextMessage,40,4);
    System.arraycopy(font,0,drawTextMessage,44,4);

    // determine length on the screen
    //
    // make an array with cumulative length of string on screen
    float[] xDistance = new float[numberOfChars+1];
    xBase = 0;
    for(thisChar=0; thisChar<numberOfChars; ++thisChar)
    {
      charIndex = text[thisChar];
      if(charIndex < 32 || charIndex > 127)
      {
        System.out.println("unprintable code in whiteboard text data:"+charIndex);
        charIndex = 40;
      }
      xBase += ((float)font[3])*fontWidth[charIndex-32]/32f;
      xDistance[thisChar]= xBase;
    }
    xDistance[thisChar] = xBase;

    // for centered string, find its starting position;
    // for left-justified, figure out where word wraps are
    if(centerText)xStart = x - .5f*xBase;
    else
    {
      // scan across the text to locate the last blank that fits in each row
      xStart = x;
      int lastBlankThatFitsIndex = 0;

      // scan a new line
      while(lastBlankThatFitsIndex < numberOfChars)
      {
        xBase = xStart;
        float xNextLine = xDistance[lastBlankThatFitsIndex];
        thisChar = 0;

        // find position of the blank
        while((xBase+xDistance[thisChar]-xNextLine) < (screenXMax-leftXPosition))
        {
          if(text[thisChar] == blank[0])lastBlankThatFitsIndex = thisChar;
          if(++thisChar >= numberOfChars)break;
        }

        // insert 0 in its place and start scanning again
        if(lastBlankThatFitsIndex < numberOfChars)
          text[lastBlankThatFitsIndex++] = 0;

      } // end while(lastBlankThatFitsIndex < numberOfChars)
    } // end else/if(centerText)

    // insert the individual characters
    xBase = xStart;
    int startLine = 0, lineLength = 0;
    for(thisChar=0; thisChar<numberOfChars; ++thisChar)
    {
      charIndex = text[thisChar];
      if(charIndex == 0) // indicates new line when in wrap mode
      {
        // insert drawText header values based on amount of text presented
        variablePart = 12*lineLength;
        System.arraycopy(intToBytes(commandNumber),0,drawTextMessage,20,4);
        System.arraycopy(intToBytes(commandNumber++),0,drawTextMessage,24,4);
        drawTextMessage[32] = shortToByte1(variablePart+12);
        drawTextMessage[33] = shortToByte2(variablePart+12);
        drawTextMessage[36] = shortToByte1(lineLength);
        drawTextMessage[37] = shortToByte2(lineLength);

        // write message and start a new line
        if(!writeMessage(outFile,recordingHeader,10000,drawTextMessage,
                         fixedPart+variablePart,false,whiteboard))
          return false;
        xBase = xStart;
        yBase += (1.5f*(float)font[3]);
        lineLength = 0;
        startLine = thisChar+1;

      } // end if(charIndex == 0)
      else
      {
        // replace illegal characters
        if(charIndex < 32 || charIndex > 127)charIndex = blank[0];

        // insert character code and position
        lineLength++;
        int basePosition = fixedPart + 12*(thisChar-startLine);
        drawTextMessage[basePosition] = 0;
        drawTextMessage[basePosition+1] = charIndex;
        drawTextMessage[basePosition+2] = 0;
        drawTextMessage[basePosition+3] = 0;
        System.arraycopy(encodeFloat(xBase),0,drawTextMessage,basePosition+8,4);
        System.arraycopy(encodeFloat(yBase),0,drawTextMessage,basePosition+4,4);
        xBase += ((float)font[3])*fontWidth[charIndex-32]/32f;

      } // end else/if(charIndex == 0)
    } // end for( int thisChar...

    // insert final drawText message header values based on amount of text presented
    variablePart = 12*lineLength;
    System.arraycopy(intToBytes(commandNumber),0,drawTextMessage,20,4);
    System.arraycopy(intToBytes(commandNumber++),0,drawTextMessage,24,4);
    drawTextMessage[32] = shortToByte1(variablePart+12);
    drawTextMessage[33] = shortToByte2(variablePart+12);
    drawTextMessage[36] = shortToByte1(lineLength);
    drawTextMessage[37] = shortToByte2(lineLength);

    // final message is assembled - write it out
    if(lineLength > 0)
      return writeMessage(outFile,recordingHeader,10000,drawTextMessage,
                          fixedPart+variablePart,false,whiteboard);
    return true;
  } // end emitText method

 /**
  * Writes a message to the output file
  *
  * returns number or bytes written if successful, -1 otherwise
  */
  static boolean writeMessage(RandomAccessFile outFile,
                              byte[] recordingHeader,
                              long timestampBackoff,
                              byte[] message,
                              int messageLength,
                              boolean setReset,
                              byte medium)
  {
    // make a new header with timestamp from original offset by timestampBackoff
    byte[] outputHeader = new byte[9];
    if(timestampBackoff == 0)System.arraycopy(recordingHeader,0,outputHeader,0,4);
    else
    {
      long timestamp = bytesToInt(recordingHeader) - timestampBackoff;
      System.arraycopy(intToBytes(timestamp),0,outputHeader,0,4);
    }
    outputHeader[4] = shortToByte1(messageLength);
    outputHeader[5] = shortToByte2(messageLength);
    if(setReset)outputHeader[8] = 1;
    else outputHeader[6] = 0; // not a control message
    outputHeader[7] = medium; // audio, WBD or video message
    outputHeader[8] = 0;      // reset flag off
    try
    {
      outFile.write(outputHeader);
      outFile.write(message,0,messageLength);
    }
    catch(IOException ioe)
    {
      System.out.println("IOException writing cover slide:"+ioe.getMessage());
      return false;
    }
    bytesWritten += (9 + message.length);
    return true;
  } // end writeMessage method

 /**
  * Reads the Copyright.txt file to the next carriage Return character of eof
  */
  static byte[] readCoverSlideFile()
  {
    int byteCount, readInt = 0, lastReadByte = 0;
    byte[] buffer = new byte[500];
    for(byteCount=0; byteCount<500; ++byteCount)
    {
      try
      {
        // absorb extraneous Windows linefeeds if found
        readInt = coverSlideFile.read();
        if((byte)readInt == lineFeed[0])readInt = coverSlideFile.read();
      }
      catch(IOException ioe)
      {
        System.out.println("IOException reading Cover Slide file:"+ioe.getMessage());
        break;
      }
      if(readInt == -1)break;
      byte readByte = (byte)readInt;
      if(readInt == carriageReturn[0])break;
      buffer[byteCount] = readByte;
      //lastReadByte = readByte;
    }

    //if(readInt == -1)++byteCount;
    byte[] returnBytes = new byte[byteCount];
    System.arraycopy(buffer,0,returnBytes,0,byteCount);
    return returnBytes;
  }

/**
  *
  */
  static byte[] makeSessionMessage(byte[] sourceAddress)
  {
    byte[] sessionMessage = // IP address goes at [0], [16], [28]
  {127,  0,  0,  1,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  1,127,  0,  0,  1,
     0,  0,  0,  1,  0,  0,  0,  1,127,  0,  0,  1,  0,  0,  0,  1,  0,  1,  0,  0,
   127,  0,  0,  1,  0,  0,  0,  7,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,  0};

    // copy source address to output
    System.arraycopy(sourceAddress,0,sessionMessage,0,4);
    System.arraycopy(sourceAddress,0,sessionMessage,16,4);
    System.arraycopy(sourceAddress,0,sessionMessage,28,4);
    return sessionMessage;
  }

 /**
  * Encodes the specified value, returning the result as a byte array
  *
  * @param value the value to encode
  * @return a byte array containing the encoded value
  */
  static byte[] encodeFloat(float value)
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(4);
    DataOutputStream dos = new DataOutputStream(baos);
    try{dos.writeFloat(value);}catch(IOException ioe){}

    // swap to net order and return the result
    byte[] result = baos.toByteArray();
    return result;
  } // end netEncodeFloat method

 /**
  * Decodes and returns the value stored in the specified buffer
  *
  * @param buffer the buffer containing the encoded value
  * @return the decoded value
  */
  static float decodeFloat(byte[] buffer)
  {
    ByteArrayInputStream bais = new ByteArrayInputStream(buffer);
    DataInputStream dis = new DataInputStream(bais);
    try{return dis.readFloat();}catch(IOException ioe){}
    return 0;
  }

 /** utility method to convert two bytes to an int
  */
  public static int bytesToShort(byte byte1, byte byte2)
  {
    int high_byte = byte1;
    if( high_byte < 0 ) high_byte = 127 - high_byte;
    int low_byte = byte2;
    if( low_byte < 0 ) low_byte = 127 - low_byte;
    return high_byte*256 + low_byte;
  } // end bytesToShort method

 /** utility method to convert to first byte of an int
  */
  public static byte shortToByte1( int input )
  {
    byte byte1;
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);
    return byte1;
  } // end shortToByte1 method

 /** utility method to convert to create the second byte of two from int
  */
  public static byte shortToByte2( int input )
  {
    byte byte1, byte2;

    // recalculate the high byte- this is quicker than invoking the method
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);

    // now form the low byte
    int lowByte = input - 256*highByte;
    if( lowByte <= 127 )byte2 = (byte)lowByte;
    else byte2 = (byte)(int)(127-lowByte);
    return byte2;
  } // end shortToByte2 method

 /** utility method to convert four bytes to a long
  */
  public static long bytesToInt(byte[] bytes)
  {
    long long1 =
        ( bytes[0] >= 0 ? bytes[0] : 255 - ~bytes[0] ) * 16777216 +
        ( bytes[1] >= 0 ? bytes[1] : 255 - ~bytes[1] ) * 65536 +
        ( bytes[2] >= 0 ? bytes[2] : 255 - ~bytes[2] )  * 256 +
        ( bytes[3] >= 0 ? bytes[3] : 255 - ~bytes[3] );
    return long1;
  }

  public static long bytesToInt(byte[] bytes, int offset)
  {
    long long1 =
        ( bytes[offset] >= 0 ? bytes[offset] : 255 - ~bytes[offset] ) * 16777216 +
        ( bytes[offset+1] >= 0 ? bytes[offset+1] : 255 - ~bytes[offset+1] ) * 65536 +
        ( bytes[offset+2] >= 0 ? bytes[offset+2] : 255 - ~bytes[offset+2] )  * 256 +
        ( bytes[offset+3] >= 0 ? bytes[offset+3] : 255 - ~bytes[offset+3] );
    return long1;
  } // end bytesToInt method

 /** utility method to convert a long to four bytes
  */
  public static byte[] intToBytes(long long1)
  {
    byte[] result = new byte[4];
    result[0] = (byte ) (long1/16777216);
    long rem =  ( long1 - result[0] * 16777216);
    result[1] =  (byte) (rem  / 65536);
    rem =  ( rem - result[1] * 65536 );
    result[2] = (byte) (rem/256);
    rem =  ( rem - result[2] * 256) ;
    result[3] = (byte) rem;
    return result;
  } // end intToBytes method
public static void Print4(int ID, byte[]data, int offset)
{System.out.println(ID+":"+data[offset]+" "+data[offset+1]+" "+data[offset+2]+" "+data[offset+3]);}
} // end CoverSlide class