/*
 * DBQuery for use with MySQL + NEWDB database and NEWCM
 * Created on December 30, 2009
 * @author  mpullen & students
 * @version 5.1.1
 * @modified NKC 12/30/09
 * @modified JMP 1/5/11 to post user counts directly
 */
/*----------------------------------------------------------------*
|   Copyright 2001-2010 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/

import java.sql.*;
import java.util.Date.*;
import java.util.TimeZone.*;
import javax.management.timer.Timer;

/**
 *
 * @author nclark1
 */
class DBQuery {

    /**
     * Database connection
     */
    private Connection conn;
    /**
     * Do we have an active connection?
     */
    protected boolean isConnected;
    /**
     * Storage for user information obtained from a query
     */
    protected authenticatedUser user;
    /**
     * Database prefix to be appended to database table names
     */
    protected String prefix = "mdl_";

    /**
     * Constructor, create object and connect to database, reset user to null
     * @param url DB URL including hostname and database name
     * @param username Username to use for DB connection
     * @param password Password to use for DB connection
     * @throws ClassNotFoundException Could not find the database driver class
     * @throws SQLException Error connecting to the database
     */
    DBQuery(String url, String username, String password, String pre) throws ClassNotFoundException, SQLException {
        isConnected = connect(url, username, password);
        user = null;
        prefix = pre;
    }

    /**
     * Connect to a database
     * @param url DB URL including hostname and database name
     * @param username Username to use for DB connection
     * @param password Password to use for DB connection
     * @throws ClassNotFoundException Could not find the database driver class
     * @throws SQLException Error connecting to the database
     */
    protected boolean connect(String url, String username, String password) throws ClassNotFoundException, SQLException {
        isConnected = false;
        user = null;
        Class.forName("com.mysql.jdbc.Driver"); //throws ClassNotFoundException
        conn = DriverManager.getConnection(url+"?useUnicode=true&characterEncoding=utf-8", username, password); // throws SQLException
        isConnected = true;
        return isConnected;
    }

    /**
     * Close the DB connection if one is active, mask the exception and print
     * an error message
     */
    protected void close() {
        if (!isConnected) {
            return;
        }
        try {
            conn.close();
            user = null;
        } catch (SQLException ex) {
            System.out.println("Database Exception: " + ex.getMessage());
        }
    }

    /**
     * Semester = For MIST/C this is not used and is hardcoded
     * @return String representing current semester code
     */
    protected String getCurrentSemester() {
        return "MOO"; //Moodle/MISTC does not use semester
    }

    /**
     * Authenticate the user based on the auth token or session ID. If authentication
     * is successful, then the user object is populated with information about the user
     * Note: the last login timestamp is validated as part of this, see validateLastLogonTimestamp
     * @param t String representing the autentication token (or session ID)
     * #param id String the CID valiue from the conf file, used as the MISTC ID to limit login to specific instance
     * @return True if user is authenticated or false if not
     * @throws SQLException Problem performing the query or with database connection
     */
    protected boolean authUserWithToken(String t, String id) throws SQLException {
        user = null;
        boolean result = false;
        String query = "SELECT u.username,u.firstname,u.lastname,a.mistcid,a.courseid,a.permission,a.timestamp " +
                "FROM " + prefix + "mistc_auth a  " +
                "LEFT JOIN " + prefix + "user u  " +
                "ON a.userid=u.id " +
                "WHERE a.authtoken=? and a.mistcid=?";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, t); // authtoken/sessionID
        ps.setString(2,id); // id is actually a bigint in the mysql database but should work as string in query
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            user = new authenticatedUser();
            long ts = rs.getLong("timestamp");
            if (!validateLastLogonTimestamp(ts)) { // check that last logon is acceptable
                return false;
            }
            result = true;
            user.lastLogon = String.valueOf(ts);
            user.authtoken = t;
            user.username = rs.getString("username");
            user.courseID = rs.getString("courseid");
            user.semester = rs.getString("mistcid");
            user.fullname = rs.getString("lastname") + ", " + rs.getString("firstname");
            user.controller = rs.getInt("permission");
            user.instructor = rs.getInt("permission");
            user.showGrid = 1; // force showGrid to be 1 always
            if (rs.getInt("permission") > 0) {
               user.controller = 2;
            }

        }
        return result;
    }

    /**
     * NOTE: THIS CODE DOES NOT WORK IN MIST/C
     * Authenticate the user by retrieving the user password. Returns true if the requested
     * user is found in the database and false if not. If true, the user object is populated with
     * user information including the password. Actual password comparison is to be validated in Auth.java
     * @param usr username to query the database
     * @param crs courseID to query the database
     * @param sem semester code to query the database
     * @return True if user is found, false if not.
     * @throws SQLException Problem performing the query or with database connection
     */
    protected boolean authUserWithPassword(String usr, String crs, String sem) throws SQLException {
        user = null;
        boolean result = false;

        String authQuery = "SELECT password,fullname,IPAddress,instructor,controller,showGrid " +
                "FROM userinfo " +
                "WHERE username=? AND courseID=?";

        if (sem.trim().length() > 0) {
            authQuery = authQuery + " AND semester=?";
        }

        PreparedStatement ps = conn.prepareStatement(authQuery);
        ps.setString(1, usr); //username
        ps.setString(2, crs); //courseID
        if (sem.trim().length() > 0) {
            ps.setString(3, sem);
        }
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            user = new authenticatedUser();
            user.password = rs.getString("password)");
            user.username = usr;
            user.courseID = crs;
            user.semester = sem;
            user.fullname = rs.getString("fullname");
            user.controller = rs.getInt("controller");
            user.instructor = rs.getInt("instructor");
            user.showGrid = rs.getInt("showGrid");
            user.IPAddress = rs.getString("IPAddress");
            result = true;
        }
        return result;
    }

    /**
     * Not used in MIST/C
     * Validate the datetime string and check that it is not older than 60 seconds
     * Note: This code was pulled exactly from Auth.java and moved here
     * @param ts String timestamp to be validated, format: 2009-06-15 13:24:49
     * @return True if ts is not older than 60 seconds, False otherwise
     */
    protected boolean validateLastLogonTimestamp(String ts) {
        boolean result = false;

        // test for valid case
        String thisDateTime = TLM.dateTime();
        String searchYear = thisDateTime.substring(0, 4);
        String searchMonth = thisDateTime.substring(5, 7);
        String searchDay = thisDateTime.substring(8, 10);

        // check for date match, ignoring case for midnight
        // (that user can try again!)
        if (ts.substring(0, 4).equals(searchYear) &&
                ts.substring(5, 7).equals(searchMonth) &&
                ts.substring(8, 10).equals(searchDay)) {
            int thisTime, logonTime;
            try {
                // convert logon time and current system time to seconds
                logonTime =
                        3600 * (new Integer(ts.substring(11, 13))).intValue() +
                        60 * (new Integer(ts.substring(14, 16))).intValue() +
                        (new Integer(ts.substring(17, 19))).intValue();
                thisTime =
                        3600 * (new Integer(thisDateTime.substring(11, 13))).intValue() +
                        60 * (new Integer(thisDateTime.substring(14, 16))).intValue() +
                        (new Integer(thisDateTime.substring(17, 19))).intValue();
            } catch (NumberFormatException nfe) {
                System.out.println("internal error comparing dates " +
                        ts + " " + thisDateTime);
                thisTime = 100;
                logonTime = 0;
            }
            if (logonTime > (thisTime - 60)) // 60 seconds
            {
                result = true;
            }
        } // end if(readLastLogon...

        return result;
    }

    /**
     * Validate the datetime string and check that it is not older than 60 seconds
     * @param ts long timestamp to be validated, format in seconds since epoch
     * @return True if ts is not older than 60 seconds, False otherwise
     */
    protected boolean validateLastLogonTimestamp(long ts) {
        boolean result = false;
        Date now = new Date(System.currentTimeMillis() - 60L * Timer.ONE_SECOND);
        Date tsn = new Date(ts * 1000);
        if (tsn.after(now)) {
            return true;
        }
        return result;
    }

    /**
     * Used in Auth.java nameAvailable method, looks for username in database
     *  and returns true if found and also sets user object to the user's info
     * @param usr username to query for
     * @param crs courseID to query for
     * @return True if user is found, false otherwise
     * @throws SQLException Problem performing the query or with database connection
     */
    protected boolean nameAvailable(String usr, String crs) throws SQLException {
        boolean result = false;
        user = null;

        String query = "SELECT password " +
                "FROM userinfo " +
                "WHERE username=? AND courseID=? AND userDeleted='0'";

        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, usr); //username
        ps.setString(2, crs); //courseID
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            user = new authenticatedUser();
            user.password = rs.getString("password");
            user.username = usr;
            user.courseID = crs;
            result = true;
        }
        return result;
    }

    /**
     * NOTE: THIS CODE DOES NOT WORK IN MIST/C
     * Query the database with the session variable/authtoken provided and
     * return true if found (also set user with some user information)
     * @param sess SessionID to query for
     * @return True if found, false otherwise
     * @throws SQLException Problem performing the query or with database connection
     */
    protected boolean validateSessionVariable(String sess) throws SQLException {
        boolean result = false;
        user = null;

        String query = "SELECT courseID,lastLogin " + // need IPAddress here (?)...
                "FROM userinfo " +
                "where NEWSessID=? and userDeleted='0';";

        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, sess); //session id
        ResultSet rs = ps.executeQuery();
        while (rs.next()) {
            user = new authenticatedUser();
            user.courseID = rs.getString("courseID");
            user.lastLogon = rs.getString("lastLogin");
            if (validateLastLogonTimestamp(user.lastLogon)) {
                return true;
            }
        }
        return result;
    }

    /**
     * NOTE: THIS CODE DOES NOT WORK IN MIST/C
     * Retrieve certificate information from the database based on the certificate number
     * @param cnum Certificate number to query for
     * @return TLM.certificateInfo object that contains the certificate info (or null if not found)
     * @throws SQLException Problem performing the query or with database connection
     */
    protected TLM.certificateInfo getCertificateInfo(long cnum) throws SQLException {
        String query = "SELECT semester,courseID,username,downloaded,commonName," +
                "orgName,orgUnit,orgLoc,orgState,orgCountry,emailAddr,createDt,downloadDt,expiryDt " +
                "FROM certifinfo " +
                "WHERE serialNum=?";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setLong(1, cnum); //certificate number
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            TLM.certificateInfo c = new TLM.certificateInfo();
            c.semester = rs.getString("semester");
            c.courseID = rs.getString("courseID");
            c.username = rs.getString("username");
            c.downloadStatus = new Short(rs.getString("downloaded"));
            c.commonName = rs.getString("commonName");
            c.orgName = rs.getString("orgName");
            c.orgUnit = rs.getString("orgUnit");
            c.orgLoc = rs.getString("orgLoc");
            c.orgState = rs.getString("orgState");
            c.orgCountry = rs.getString("orgCountry");
            c.emailAddr = rs.getString("emailAddr");
            c.createDate = rs.getString("createDt");
            c.downloadDate = rs.getString("downloadDt");
            c.expiryDate = rs.getString("expiryDt");
            return c;
        }
        return null;
    }

    /**
     * NOTE: THIS CODE DOES NOT WORK IN MIST/C
     * Retrieve ResultSet containing course information
     * @param sem Semester code to query for
     * @param cid CID of server
     * @return ResultSet object of query results
     * @throws SQLException Problem performing the query or with database connection
     */
    protected ResultSet getCourseInfo(String sem, String cid) throws SQLException {
        String query = "SELECT courseID,courseNum,classDayID,classStartTime,hideName,classEndTime " +
                "FROM courselist " +
                "WHERE semester=? AND CID_live=?";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, sem);
        ps.setString(2, cid);
        ResultSet rs = ps.executeQuery();
        return rs;
    }

    /**
     * Retrieve the CourseID for this Room/MISTC ID instance
     * @param rid RoomID/MISTCID to search
     * @return String representing the numerical courseID
     * @throws SQLException Problem performing query or with database connection
     */
    protected String getCourseIDForRoomID(String rid) throws SQLException {
        String result=null;
        String query = "SELECT course " +
                "FROM " + prefix + "mistc " +
                "WHERE id=?";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, rid);
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            result = rs.getString("course");
        }
        return result;
    }

    /**
     * Retrieve the Moodle Short course name for this Room/MISTC ID instance
     * @param rid RoomID/MISTCID to search
     * @return String representing the short name for the requested room/mistc ID
     * @throws SQLException Problem performing query or with database connection
     */
    protected String getMoodleShortName(String rid) throws SQLException {
        String result=null;
        String query = "SELECT c.shortname " +
                "FROM " + prefix + "course c, " + prefix + "mistc m " +
                "WHERE c.id=m.course " +
                "AND m.id=?";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, rid);
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            result = rs.getString("shortname");
        }
        return result;
    }

    /**
     * Retrieve all users in a course, returns ResultSet of DB query.
     * This is used for usergroups in Auth.java
     * @param rid RoomID of this server
     * @param crs CourseID not actually used, look in DB based on roomid
     * @return ResultSet of query results
     * @throws SQLException Problem performing the query or with database connection
     */
    protected ResultSet getUsersInCourse(String rid, String crs) throws SQLException {
        // note set this to crs to use what is passed rather than looking it up based on the room ID??
        String cid = getCourseIDForRoomID(rid);
        String query = "SELECT u.username, CONCAT(u.lastname, \", \", u.firstname) AS fullname " +
                "FROM " + prefix + "role_assignments ra, " + prefix + "user u, " + prefix + "course c, " + prefix + "context cxt " +
                "WHERE ra.userid=u.id " +
                "AND ra.contextid=cxt.id " +
                "AND cxt.contextlevel=50 " +  //Context Level 50 = course
                "AND cxt.instanceid=c.id " +
                "AND c.id=? "; // courseID parameter
                //"AND roleid=5"; // roleid 5 = students only
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, cid);
        ResultSet rs = ps.executeQuery();
        return rs;
    }

    /**
     * Retrieve ResultSet of user group assignments
     * @param rid MISTC or Room ID to query for
     * @param crs CourseID, note this is not used right now
     * @return ResultSet of username,groupnum
     * @throws SQLException Problem performing the query or with database connection
     */
    protected ResultSet getUserGroupAssignments(String rid, String crs) throws SQLException {
        //String cid = getCourseIDForRoomID(rid);
        /*String query = "SELECT u.username, ug.groupnum " +
                "FROM " + prefix + "user u, " + prefix + "mistc_usergroups ug " +
                "WHERE u.id=ug.userid " +
                "AND ug.mistcid=?";*/
        String query = "SELECT username, groupnum " +
                "FROM " + prefix + "mistc_usergroups " +
                "WHERE mistcid=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, rid);
        ResultSet rs = ps.executeQuery();
        return rs;
    }

    /**
     * Search the Moodle usertable for the UserID matching this username
     * @param username String to use in the query
     * @return String represneting the userid for the username
     */
    protected String getIDForUser(String username) throws SQLException {
        String result = null;
        String query = "SELECT id " +
                "FROM " + prefix + "user " +
                "WHERE username=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, username);
        ResultSet rs = ps.executeQuery();
        if (rs.next()) {
            result = rs.getString("id");
        }
        return result;
    }

    /**
     * Add an entry to the usergroups table
     * @param rid MISTC/RoomID
     * @param crs CourseID of addition
     * @param usr Username of user to be added
     * @param gnum Group number for this user
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void insertToUserGroups(String rid, String crs, String usr, int gnum) throws SQLException {
        //String uid = getIDForUser(usr);
        String cid = getCourseIDForRoomID(rid);
        String query = "INSERT INTO " + prefix + "mistc_usergroups " +
                "(mistcid,courseid,username,groupnum)" +
                "VALUES(?,?,?,?)";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, rid);
        ps.setString(2, cid);
        ps.setString(3, usr);
        ps.setInt(4, gnum);
        ps.executeUpdate();
        ps.close();
    }

    /**
     * Update a record in usergroups
     * @param rid MISTC/RoomID
     * @param crs CourseID of entry to be updated
     * @param usr Username to be updated
     * @param gnum New group number for entry
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void updateUserGroups(String rid, String usr, int gnum) throws SQLException {
        //String uid = getIDForUser(usr);
        String cid = getCourseIDForRoomID(rid);
        String query = "UPDATE " + prefix + "mistc_usergroups " +
                "SET groupnum=? " +
                "WHERE mistcid=? " +
                "AND username=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, gnum);
        ps.setString(2, rid);
        ps.setString(3, usr);
        ps.executeUpdate();
        ps.close();
    }

    /**
     * Add a log entry to the session log
     * @param roomID
     * @param username
     * @param dateTime
     * @param IPAddress
     * @param semester Not used in this version
     * @param courseID
     * @param pathFileName
     * @param elapsedTime
     * @param controller
     * @param protocol
     * @param peakRTT
     * @param serverUtil
     * @param transactionType
     * @throws SQLException
     */
    protected void insertLog(String roomID, String username, String dateTime, String IPAddress,
            String semester, String courseID, String pathFileName, String elapsedTime,
            String controller, String protocol, String peakRTT, String serverUtil,
            String transactionType) throws SQLException {
        String query = "INSERT into " + prefix + "mistc_sessionlog " +
                "(username,timemodified,ipaddress,mistcid,courseid,pathfilename,elapsedtime," +
                "controller,protocol,peakrtt,serverutil,transactiontype) " +
                "VALUES(?,?,?,?,?,?,?,?,?,?,?,?)";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setString(1, username);
        ps.setString(2, dateTime);
        ps.setString(3, IPAddress);
        ps.setString(4, roomID);
        ps.setString(5, courseID);
        ps.setString(6, pathFileName);
        ps.setString(7, elapsedTime);
        ps.setString(8, controller);
        ps.setString(9, protocol);
        ps.setString(10, peakRTT);
        ps.setString(11, serverUtil);
        ps.setString(12, transactionType);
        ps.executeUpdate();
        ps.close();
    }

    /**
     * Post the number of users in a MIST/C room
     * @param rid MISTC/RoomID
     * @param count count of users
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void postNumUsers(String rid, int count) throws SQLException {
        //String uid = getIDForUser(usr);
        //String cid = getCourseIDForRoomID(rid);
        int id = Integer.valueOf(rid);
        String query = "UPDATE " + prefix + "mistc " +
                "SET numusers = " + count +
                " WHERE id=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, id);
        ps.executeUpdate();
        ps.close();
    }
    /**
     * Post the number of privileged/instructor/controller users in a MIST/C room
     * @param rid MISTC/RoomID
     * @param count count of such users
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void postNumPrivUsers(String rid,int count) throws SQLException {
        int id = Integer.valueOf(rid);
        String query = "UPDATE " + prefix + "mistc " +
                "SET numprivusers= " + count +
                " WHERE id=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, id);
        ps.executeUpdate();
        ps.close();
    }

    /**
     * Decrement the number of users in a MIST/C room
     * @param rid MISTC/RoomID
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void decrementNumUsers(String rid) throws SQLException {
		int id = Integer.valueOf(rid);
        String query = "UPDATE " + prefix + "mistc " +
                "SET numusers=numusers-1 " +
                "WHERE id=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, id);
        ps.executeUpdate();
        ps.close();
    }

    /**
     * Decrement the number of privileged/instructor/controller users in a MIST/C room
     * @param rid MISTC/RoomID
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void decrementNumPrivUsers(String rid) throws SQLException {
		int id = Integer.valueOf(rid);
        String query = "UPDATE " + prefix + "mistc " +
                "SET numprivusers=numprivusers-1 " +
                "WHERE id=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, id);
        ps.executeUpdate();
        ps.close();
    }

   /**
     * Zero the number of users in a MIST/C room
     * @param rid MISTC/RoomID
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void zeroNumUsers(String rid) throws SQLException {
		int id = Integer.valueOf(rid);
        String query = "UPDATE " + prefix + "mistc " +
                "SET numusers=0 " +
                "WHERE id=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, id);
        ps.executeUpdate();
        ps.close();
    }

    /**
     * Zero the number of privileged/instructor/controller users in a MIST/C room
     * @param rid MISTC/RoomID
     * @throws SQLException Problem performing the query or with database connection
     */
    protected void zeroNumPrivUsers(String rid) throws SQLException {
		int id = Integer.valueOf(rid);
        String query = "UPDATE " + prefix + "mistc " +
                "SET numprivusers=0 " +
                "WHERE id=? ";
        PreparedStatement ps = conn.prepareStatement(query);
        ps.setInt(1, id);
        ps.executeUpdate();
        ps.close();
    }


    /**
     * Storage of user information from a query
     */
    class authenticatedUser {

        String username;
        String password;
        String IPAddress;
        String semester;
        String courseID;
        String fullname;
        String firstname;
        String lastname;
        int instructor;
        int controller;
        int showGrid;
        String lastLogon;
        String authtoken;
    }
}

