/*
 * FCConnect.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author mpullen & students
 * @version 5.2.4
 * @modified JMP 12/28/12 to return floor automatically to original holder
 * @modified JMP 5/7/11 for server recording download
 * @modified JMP 4/24/11 for multicast tunnel
 * @modified JMP 1/7/10 for breakout groups
 * @modified JMP 5/21/10 to support fullname in FC grid
 * @modified Nick Clark 9/12/10 for MISTC database access
 * @modified JMP 1/14/13 to support auto-upload of recordings 
 */
/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
/******************************************************************************
  FCConnect class logic (in module TLM.java):
->Until stopTLM is signalled
-> loop forever
-> open a listening TCP server socket
-> accept connection from any host and spawn a FloorControlListen thread
-> end loop forever
 ******************************************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;

public class FCConnect extends TLM implements Runnable
{
  //public static ServerSocket fcConnectSocket = null;
  public Socket fcSocket;

  // Gets incremented whenever a request comes. The value is always unique.
  static int requestNumber = 0;

  // On grant next grant only the request number above this value.
  static int latestRequestGranted = 0;
  static Integer requestNumberObject;

  // objects used to synchronize changes of global variables
  // floorStatus and floorHolder
  static Object synchFloorStatus, synchFloorHolder, synchWriteToAll;

  // activeUsers hashtable contains details of all those users who have
  // connected successfully.
  static Hashtable activeUsers = new Hashtable(2*maxConnections);
  // for Java 1.5:
  // static Hashtable<String,UserDetails> activeUsers =
  //   new Hashtable<String,UserDetails>();

  // pendingRequests Hashtable holds all the requests that come when the floor
  // status is ASK ME. Key is sequence number of the request and the object stored
  // is the userRequest details whch contains the username, hostIndex and request
  // number.
  static Hashtable pendingRequests = new Hashtable();
  // for Java 1.5:
  // static Hashtable<Integer,UserFloorRequest> pendingRequests =
  //   new Hashtable<Integer,UserFloorRequest>();

  // This vector contains the already logged in users. Need not be active,
  // would contain even if they drop off. If they login with the same name,
  // the name won't be added to this else will be added. Dropping off and
  // logging in with different name is possible for wild card user names.
  // The purpose of this vector is, once the size of this vector exceeds the
  // maximum number of users .conf provides i.e. the maximum number of users
  // the client grid can support, any new user request would be rejected by
  // sending the response 4.
  static Vector loggedInUsers = new Vector();
  // for Java 1.5:
  // static Vector<String> loggedInUsers = new Vector<String>();
  static UserDetails userDetails;
  static byte[] pUsername;
  static byte[] pPassword;

  // messages to turn client audio on and off
  static byte[] audioOnBuffer = {69,0,0,1,0};
  static byte[] audioOffBuffer = {69,0,0,0,0};
  
  // data to log voting
  int voteSequence = 0, propositionLength = 0;
  String proposition = null;
  
  // thread to reset server state five minutes after last logout
  // unless autoscheduler has set classInSession
  CloseAfterFiveMinutes closeAfterFiveMinutes;
  Thread runAfterFiveMinutes;
  Boolean runningFiveMinutes = new Boolean(false);
  
  // floorHolder data
  int firstFloorHolder = -1;
  static String nobodyUsername = "********";
  static String nobodyFullname = "(nobody has the floor)"; 
  static String nobodyYetFullname = "(no controller is logged in)";
  static boolean inPAMode = false;
  
/***************************************/

  // constructor
  public FCConnect()
  {
	  // TLM.floorStatus variable can take one of the values {1,2,3}
	  // 1 => ALWAYS, 2 => NEVER (LOCKED), 3 => ASK ME
	  // When it is ALWAYS, whenever a user requests the floor the earlier
	  // one who already has will lose it; the user knows about this change in
	  // floor access looking at the color change on the panel
	  pUsername = new byte[8];
	  pPassword = new byte[8];
    
	  // synchronizing objects
    synchFloorStatus = new Object();
	  synchFloorHolder = new Object();
	  synchWriteToAll = new Object();
    
	  // create a dummy floor change hold thread so we
	  // won't have a nullpointer when testing isAlive()
	  FCHold fcHold = new FCHold(0,"--------",0,0);
	  floorChangeHoldThread = new Thread(fcHold);
  }
  
  public void run()
  {
	  InetAddress openingHost = null;
	  stopTLM = false;
	  boolean hostFound = false;
  
	  // create a server socket on the tcpPort+1
	  try
	  {
	    //fcConnectSocket = new ServerSocket( tunnelTCPPort+1, MAX_HOSTS );
	    System.out.println("\nAuthentication server waiting for client on TCP port "+
	    //fcConnectSocket.getLocalPort() );
	    NEWSocket.getControlSocketLocalPort() );
	  }
    catch ( Exception e )
	  {
	    System.out.println ("Server socket I/O error in FCConnect: " + e);
	    stopTLM = true;
	    System.exit(-1);
	  }
    
	  // accept connections from FC clients
	  while ( !stopTLM )
	  {
	    try
	    {
		    // open a dedicated socket for next connecting host
		    //fcSocket = fcConnectSocket.accept();
		    fcSocket = NEWSocket.getControlSocket();
		    if(fcSocket == null)continue;
        
		    // spawn a thread to service the FC connection
		    FCListen r1 = new FCListen(fcSocket);
		    Thread t1 = new Thread(r1);
		    t1.start();
	    }
	    catch ( Exception e )
	    {
		    System.out.print("FCConnect server socket: " + e + "\n" );
	    }
	  } // end while(!stopTLM)
  } // end run()
  
  
  // floor handoff logic:
  // - when there is no floorHolder, first user to take floor becomes firstFloorHolder
  // - after that, no matter who takes the floor, if when released there would be no 
  //   floorHolder, it reverts to the firstFloorHolder
  // - if the firstFloorHolder gives up the floor, status returns to no floorHolder
  // - when any floorHolder gives up the floor, the transition is delayed one second;
  //   during that time all floor requests are ignored in interest of stability
  //   (this is true also for transition from one floorHolder to another)
  
  // implementing one-second delay class
  boolean setFloorHolderPending = false;
  class delaySetFloorHolder implements Runnable
  {
    // parameters for the delayed setFloorHolder
    int hostIndex; 
    String setName; 
    int oldFloorHolder;
    int groupNumber;
    
    public delaySetFloorHolder(int hostIndexRef,
                               String setNameRef, 
                               int oldFloorHolderRef,
                               int groupNumberRef)
    {
      // only one setFloorHolder per second
      if(setFloorHolderPending)
      {
        sendRejection(hostIndexRef);
        return;
      }
      setFloorHolderPending = true;
      
      // copy out parameters
      hostIndex = hostIndexRef;
      setName = new String(setNameRef);
      oldFloorHolder = oldFloorHolderRef;
      groupNumber = groupNumberRef;
      
    }// end constructor
    
    public void run()
    {
      // sleep for one second, then setFloorHolder
      if(floorHolder >= 0)
        try
        {
          Thread.sleep(1000);
        }
        catch(InterruptedException ie){}
      setFloorHolder(hostIndex,setName,oldFloorHolder,groupNumber);
      setFloorHolderPending = false;
      
    }//end run
    
  }// end class delaySetFloorHolder
  
  
  // method to run above class
  void runDelaySetFloorHolder(int hostIndex,
                              String setName, 
                              int oldFloorHolder,
                              int groupNumber)
  {
    (new Thread(new delaySetFloorHolder(hostIndex,
                                        setName,
                                        oldFloorHolder,
                                        groupNumber))).start();
  }// end runDelaySetFloorHolder()
  
  
  // Reset the activeUsers hashtable
  static void clearActiveUsers()
  {
	  activeUsers = new Hashtable(2*maxConnections);
  }
  
  
  // The attempted connection is not in the hostAddr table;
  // it still could be a wildcard entry- check the name.
  static synchronized boolean nameAlreadyTaken(String pUsername)
  {
	  boolean nameUsed = false;
	  nameUsed = activeUsers.containsKey(pUsername);
	  return nameUsed;
  }
  
  
  // locate host table position corresponding to a username (the data
  // structure used here could be improved by ordering or hashing)
  // returns the table position, or -1 if not present
  static int getUserPosition(String findUsername)
  {
    int tablePosition = 0;
	  synchronized(updateUserlist)
	  {
	    for(tablePosition = 0; tablePosition < maxConnections; ++tablePosition)
	      if(username[tablePosition].equals(findUsername))
		      return tablePosition;
    }
	  return -1;
  }
  
  
  // This method is called by FCListen when a user's details are authenticated. That
  // authentic user's details are obtained from arrays using the hostindex and a
  // userdetails object is made with those details and put into this activeUsers hashtable.
  // activeUsers is the hashtable i.e. looked at whenever controlPrivilege or drawgrid
  // etc. of a user are needed
  //
  // This method returns:
  //   0 if it successfully adds the username
  //   1 if the number of users exceeds maxUsers
  //   2 if the name user tried logging in with already exists
  //
  static int addUser(int addHostIndex,String addUsername)
  {
    // get the user details from corresponding arrays using hostindex as the
    // index into those arrays
    byte[] bytesAddUsername = addUsername.getBytes();

    // Add to the vector if the size is less than the maxUsers and if the vector
    // does not contain the given username. Second case possible, if the user drops
    // out and logs back, then the name should not be added to the vector. However,
    // we need to add to the activeUsers hash table as we delete that name from it
    // on drop event but not from vector. The reason for not deleting from vector is
    // that the vector's count should give the number of users logged in, independent
    // of whether they are active are dropped off as they still hold a position on
    // client's grid and we need count for that maximum number of grids available.
    //
    // also, until one floor controller logs in we hold one grid space open to
    // avoid case where the controller is locked out any nobody can get the floor
	  synchronized(updateUserlist)
	  {
      if(!loggedInUsers.contains(addUsername))
        if(controllerLoggedIn || (initialFloorStatus == 1))
        {
          if(loggedInUsers.size() < maxConnections)
            loggedInUsers.add(addUsername);
          else
            return 1;
        }
        else
        {
          if(loggedInUsers.size() < (maxConnections-1))
            loggedInUsers.add(addUsername);
          else
            return 1;
        }

      // We don't need to check for the name in username array here;
      // that check is required only in second stage of nameExistsAgain
      // check and by then it is only activeUsers that might get modified
      // if during a context switch. If the name exists in te username array it
      // would be caught in the first check for nameExists itself.
      boolean nameExists = activeUsers.containsKey(pUsername);

      // If the name does not exist, proceed to add to user list; else return true
      // to FCListen which lets user know that the name he is trying to use
      // already exists
      if(!nameExists)
      {
        boolean draw = drawGrid[addHostIndex];
        userDetails = new UserDetails();
        userDetails.setUsername(bytesAddUsername);
        userDetails.setStatus(true);
        userDetails.setControl(controlPrivilege[addHostIndex]);
        userDetails.setDrawgrid(draw);

        // If floor rule = Always, the first user who logs in gets control.
        // If the user is the first, floorHolder will change from
        // -1 to user's index in the hostAddr[] array
        //
        // case occcurs if there is no controller in the .conf file.
        // equivalent to "NEVER"/"ASK ME" option by default
        if((floorStatus == 1) && (floorHolder == -1))
        {
          // uncomment to give controller the floor on login
          //setFloorHolder( addHostIndex, addUsername, -1, -1 );

          // If there is no controller in the conference, it implies by default
          // the first user who logged in will be treated by client (but not
          // server) as controller
          controllerLoggedIn = true;
        }
        activeUsers.put(addUsername,userDetails);
        return 0;
      } // end if(!nameExists)
    }// end synchronized(updateUserlist)
    // name is already in the table
    return 2;
  } // end addUser method

  synchronized void removeUser(int removeHostIndex, String removeUsername)
  {
    // Remove the corresponding user details from activeUsers hash table so
    // that in next initialization message this user details won't be sent.
    fcIsInitialized[removeHostIndex] = false;
    fcIsReadyForAppData[removeHostIndex] = false;
    synchronized(activeUsers)
    {
      activeUsers.remove(removeUsername);
      
      // If the number of active users falls to 0, this implies the end of session.
      // At this point, empty the logged in users list, so that dead names won't
      // block the service for the next session.
      if(activeUsers.size() == 0)
      {
        // stop recording
	      if(recordOnServerConfig != 0)
	      {
		      if(serverRecord != null)
			      serverRecord.shutdownRecorder();
          serverRecord = null;
		      recordOnServer = true;
        }
		
		    // clear out state
        loggedInUsers.removeAllElements();
        controllerLoggedIn = false;
        numberOfControllersLoggedIn = 0;
        floorStatus = initialFloorStatus;
        sendingSlide = false;
        setFloorHolder(-1, nobodyUsername, removeHostIndex, -1);
		    inBreakoutMode = false;
        if(autoSchedule && !classInSession)
   		    fcConnect.closeAllConnectionsAfterFiveMinutes("room empty");
      }
    } // end synchronized(activeUsers)
  
  } // end removeUser method


  // synchronized update to value of floorStatus
  void setFloorStatus(byte status)
  {
    synchronized(synchFloorStatus)
    {
      // Set the floor status and also send system message to all clients
      floorStatus = status;
      String message = null;
      switch(status)
      {
        case 1:
          message = new String("FLOOR CONTROL: OPEN TO ALL");
          break;

        case 2:
          message = new String("FLOOR CONTROL: LOCKED");
          break;

        case 3:
          message = new String("FLOOR CONTROL: MUST BE GRANTED");
          break;

        default:
          System.out.println("invalid floor status code received in FCConnect");
      }
      // send status to all users' floor controls
      if(message != null)
        fcConnect.writeMessageToAll(34,message.length(),message.getBytes(),false);
    }// end synchronized(synchFloorStatus)

  }// end setFloorStatus()

  // synchronized access to floorStatus (ALWAYS/NEVER/ASK ME)
  synchronized byte getFloorStatus()
  {
    return(floorStatus);
  }

  // Once the user is authenticated his control privilege is obtained by calling
  // this method, and this value is sent in initialization message to client.
  static synchronized char getControlPrivilege(int getHostIndex)
  {
    return TLM.controlPrivilege[getHostIndex];
  }

  // Once the user is authenticated the drawgrid requirement is obtained by calling
  // this method, and this value is sent in initialization message to client.
  static synchronized boolean getDrawGrid(int getHostIndex)
  {
    return TLM.drawGrid[getHostIndex];
  }

  // Whenever a client sends some message, this method is called to see if the
  // user is the floorHolder or not. This need not be a method and these 2 lines
  // of code can be embedded in FCListen class but for modularity it is
  // implemented as method
  //
  // returns true if the host has the floor (if inBreakoutMode,
  // it will be the floorHolder in its group) 
  static synchronized boolean getFloorAccess(int getHostIndex)
  {
    if(getFloorHolder(getHostIndex) == getHostIndex)return true;
    else return false;
  }

  // The below method adds the request to pending requests hash table when the
  // floor status option is Ask Me. When a floor request arrives, we increment
  // the request sequence number (which will always be unique) and store
  // sequence number and the request details : {username, hostIndex,
  // request number} in pendingRequests hashtable

  // If the request is granted by clicking on grid cell, we get the username in
  // the grantMessage, in that case, we search the pendingRequests hashtable for
  // the request detaisl object which has this username, and delete the
  // corresponding entry from hashtable, assigning latest floor to that user
  // and that hostIndex.

  // If the request is granted by clicking on grantNext button, nextKey in
  // pendingRequests hashtable returns the smallest sequence number present as
  // keys in hashtable are arranged in sorted order anyway. From sequence number,
  // we get the userRequest deatails object, its mapping to which stores
  // username and hostIndex.

  synchronized void addFloorRequest(int addHostIndex, byte[] addUsername)
  {
    boolean requestExists = false;
    String testUsername, stringAddUsername = new String(addUsername);
    UserFloorRequest userRequest;

    Enumeration e = pendingRequests.elements();

    while(e.hasMoreElements())
    {
      userRequest = (UserFloorRequest)e.nextElement();
      testUsername = userRequest.getUsername();
      if(testUsername.equals(stringAddUsername))
      {
        requestExists = true;
        break;
      }
    } // end while(e.hasMoreElements())

    // If request exists i.e. same user request for floor twice ignore request;
    // else add to both the hashtables and send request message.
    if(!requestExists)
    {
      requestNumber++;
      requestNumberObject = new Integer(requestNumber);
      userRequest = new UserFloorRequest(stringAddUsername, 
	                                       addHostIndex,
                                         requestNumber);
      pendingRequests.put(requestNumberObject, userRequest);
    } // end if(!requestExists)

    // send request to all clients
    fcConnect.writeMessageToAll(37,8,addUsername,false);
    
  }// end addFloorRequest()
  
  
  // combine a username and a userFullname to make byte[48] to transmit
  static byte[] namesToBytes(String userName,String userFullname)
  {
    byte[] bytesOut = new byte[48];
	  System.arraycopy(padBlanks(userName,8).getBytes(),0,
                     bytesOut,0,
                     8);
    System.arraycopy(padBlanks(userFullname,40).getBytes(),0,
	                   bytesOut,8,
                     40);
	  return bytesOut;
  
  }// end namesToBytes()
  
  // The below method removes the request from pendingRequests hashTable.
  // Returns true if there is a request to grant, false otherwise.
  // Floor is transferred to this username if transferFloor == true.
  synchronized boolean removeFloorRequest(byte[] removeUsername,
                                          boolean transferFloor)
  {
    boolean pending = false;

    // The request to be removed is stored in requestSequence;
    // initialize with -1 as there is no request with that sequence number.
    Integer requestSequence = new Integer(-1);
    int hostIndexValue = -9; // data value is never -9
    UserFloorRequest userRequest;
    boolean requestFound = false;
    boolean pendingRequestExists = false;
    String testUsername, stringRemoveUsername = new String(removeUsername);

    // If the username is "--------" => this is a grantNext; give
    // floor to the first pending request
    if(stringRemoveUsername.equals("--------"))
    {
      // Need to remove the first pending request.
      // First element of keys enumeration is the lowest requestNumber present.
      Enumeration e = pendingRequests.keys();

      // This while loop required as the hash table is storing the entries in
      // decreasing request number order. i.e. the lowest requestNumber is the
      // last element in the hashTable
      while(e.hasMoreElements())
      {
        pendingRequestExists = true;
        requestSequence = (Integer)e.nextElement();
      }
      if(pendingRequestExists)
      {
        userRequest = (UserFloorRequest)pendingRequests.get(requestSequence);
        pendingRequests.remove(requestSequence);
        stringRemoveUsername = userRequest.getUsername();
        hostIndexValue = userRequest.getHostIndex();
        requestFound = true;
      }
      else return false;
    } // end if(deleteUsername.equals("--------"))

    // If the grantMessage is sent by clicking on user grid
    // search the hash table for the element with this username entry
    // and delete that entry from the table.
    else
    {
      Enumeration e = pendingRequests.elements();
      while(e.hasMoreElements())
      {
        userRequest = (UserFloorRequest)e.nextElement();
        if(userRequest.getUsername().equals(stringRemoveUsername))
        {
          requestFound = true;
          requestSequence = new Integer(userRequest.getRequestNumber());
          pendingRequests.remove(requestSequence);
          stringRemoveUsername = userRequest.getUsername();
          hostIndexValue = userRequest.getHostIndex();
          break;
        } // end if(lUsername.equals(pUsername))
      } // end while(e.hasMoreElements())
    } // end else/if(pUsername.equals("--------"))
    if(transferFloor && requestFound)
    {
      if(inBreakoutMode)
//        setFloorHolder(hostIndexValue,stringRemoveUsername,floorHolder,breakoutGroup[hostIndexValue]);
        runDelaySetFloorHolder(hostIndexValue,
                               stringRemoveUsername,
                               floorHolder,
                               breakoutGroup[hostIndexValue]);
      else
//        setFloorHolder(hostIndexValue,stringRemoveUsername,floorHolder,-1);
        runDelaySetFloorHolder(hostIndexValue,
                               stringRemoveUsername,
                               floorHolder,
                               -1);
      return true;
    } // end if(requestFound)
    return false;
  } // end removeFloorRequest method
  

  // method to find the fullname associated with a username
  // if no match, returns null string ""
  static synchronized String getFullname( String testUsername )
  {
    int userIndex = 0;
    for( ; userIndex < numberOfHosts; ++ userIndex)
      if( testUsername.equals(username[userIndex]) )
        return fullname[userIndex];
    return "";
	
  }// end getFullname()


  // method to set host table index and username of floorholder
  // oldFloorHolder is used to identify group in breakoutMode 
  // returns true if successful
  // hostIndex      - of client to become floorholder
  // setName        - username of client to become floorholder
  // oldFloorHolder - hostIndex of client giving up floor
  // groupNumber    - group of client giving up floor (-1 if none)
  synchronized boolean setFloorHolder(int hostIndex, 
                                      String setName, 
                                      int oldFloorHolder,
                                      int groupNumber)
  {
    if(hostIndex >= 0)
    {
      // if we are currently sending a slide, postpone floor change
      // until it is all queued or slideSendTimeout occurs
      if (sendingSlide && !floorChangeHoldThread.isAlive())
      {
        FCHold fcHold = new FCHold(hostIndex, setName, oldFloorHolder, groupNumber);
        floorChangeHoldThread = new Thread(fcHold);
        floorChangeHoldThread.start();
        return false;
      } // end if(sendingSlide)
    } // end if(hostIndex >= 0)
  
    // if nobody had floor, this is new firstFloorHolder
    if(floorHolder < 0)firstFloorHolder = hostIndex;
     
    // turn off floor holder's audio (if in breakout and sender is 
    // breakoutController and not in a group, turn off all group floorHolders)
    int groupsToTurnoff = 1;
    if(inBreakoutMode && (hostIndex == breakoutController))
      if(hostIndex >= 0)
        if(breakoutGroup[hostIndex] <= 0)
          groupsToTurnoff = numberOfBreakoutGroups + 1;
    for(int turnoffIndex = 0;
        turnoffIndex < groupsToTurnoff;
        turnoffIndex++)
    {
      int testFloorHolder = oldFloorHolder;
      if(groupsToTurnoff  > 1)// for group turnoff case
        testFloorHolder = breakoutFloorHolder[turnoffIndex];
      if(testFloorHolder < 0)continue;
      try
      {
        if(tunnelIsActive[testFloorHolder])
          if(testFloorHolder != hostIndex)
          {
            // TCP case
            if(tcpIsOpen[testFloorHolder])
            {
              queueTunnel(2,testFloorHolder,audioOffBuffer,5);
              countTcpSend++;
            }
            
            // UDP case
            else if(uRecvTime[ testFloorHolder ] != 0)
            {
              DatagramPacket sendMessage =
                new DatagramPacket(audioOffBuffer, 
                                   5, 
                                   hostInetAddr[floorHolder],
                                   tunnelUdpPort+1 );
              tunnelMessageUDPSocket.send( sendMessage );
              countUdpSend++;
            } // end else if(uRecvTime...
          } // end if((testFloorHolder != hostIndex)...
      } // end try
      catch( IOException e2 )
      {
        if((hostIndex >= 0) && e2.getMessage().equals("Broken pipe"))
        {
          // lost connection
          closeTunnel( testFloorHolder, true, "set latest floor broken pipe" );
          return false;
        }
      } // end catch
    }// end for(int turnoffIndex = 0;

    // multiple simultaneous change requests possible; 
    // synchronize them
    synchronized(synchFloorHolder)
    {
      // changing group floorholder
      if(inBreakoutMode)
      {        
        // going to no floorholder
        if(hostIndex < 0)
        {
          // group floorHolder giving up floor
          if(groupNumber > 0)
          {
            breakoutFloorHolder[groupNumber] = -1;
            breakoutFloorHolderName[groupNumber] = nobodyUsername;
          }
          
          // breakoutController giving up floor
          else 
          {
            floorHolder = -1;
            floorHolderName = nobodyUsername;
          }
          
          // breakoutController leaving "PA mode" 
          if((oldFloorHolder == breakoutController) && 
             (oldFloorHolder >= 0) && inPAMode)
          {
            inPAMode = false;
            for(int nextGroup = 1;
                nextGroup <= numberOfBreakoutGroups;
                ++nextGroup)
            {
              breakoutFloorHolder[nextGroup] = -1;
              breakoutFloorHolderName[nextGroup] = nobodyUsername;
            }
          }
        }
        else //!if(hostIndex < 0)
        {
          // going to all groups ("PA mode")
          if((hostIndex == breakoutController) &&
             (breakoutGroup[hostIndex] <= 0))
          {
            inPAMode = true;
            floorHolder = hostIndex;
            floorHolderName = setName;
            for(int nextGroup = 1;
                nextGroup <= numberOfBreakoutGroups;
                ++nextGroup)
            {
              breakoutFloorHolder[nextGroup] = breakoutController;
              breakoutFloorHolderName[nextGroup] = breakoutControllerUsername;
            }
          }
          else
          {
            // leaving PA mode - set floorholder to nobody
            if((oldFloorHolder == breakoutController) && inPAMode)
            {
              inPAMode = false;
              for(int nextGroup = 1;
                  nextGroup <= numberOfBreakoutGroups;
                  ++nextGroup)
              {
                breakoutFloorHolder[nextGroup] = -1;
                breakoutFloorHolderName[nextGroup] = nobodyUsername;
              }
            }
            // going to one group
            else
            {
              breakoutFloorHolder[groupNumber] = hostIndex;
              breakoutFloorHolderName[groupNumber] = setName;
            }
          }            
        }// end else/if(hostIndex < 0)
      }
      // not in breakout mode
      else
      {
        floorHolder = hostIndex;
        floorHolderName = setName;
        oldInstructorBreakoutGroup = 0;
      }

      // turn on new floorHolder's audio, if there is one
      if (hostIndex >= 0) 
      {
        try 
        {
          // TCP case
          if (tcpIsOpen[hostIndex]) 
          {
            queueTunnel(2,hostIndex,audioOnBuffer,5);
            countTcpSend++;
          }
          // UDP case
          else if (uRecvTime[hostIndex] != 0) 
          {
            DatagramPacket sendMessage =
                new DatagramPacket(audioOnBuffer, 5, hostInetAddr[hostIndex],
                                   tunnelUdpPort + 1);
            tunnelMessageUDPSocket.send(sendMessage);
            countUdpSend++;
          } // end else if(uRecvTime...
        } // end try
        catch (IOException e3) 
		    {
          // lost connection
          closeTunnel(hostIndex, true, "set latest floor turn on audio lost connection" );
          return false;
        } // end catch
      } // end if(hostIndex >= 0)
      
      // notify all FC clients
      if(inBreakoutMode)
      { 
        // breakoutController not in group and giving up floor:
        // if oldFloorHolder is breakout controller and is a member of
	      // default group 0, direct all FC to set oldFloorHolder grid cell white, 
	      // then group 0 to set it green (group 0 is breakout controller only)
        if(oldFloorHolder == breakoutController)
	        if(breakoutGroup[breakoutController] == 0)		  
	        {
	          fcConnect.writeMessageToAll(27,// participant left group
                                        8,
                                        breakoutControllerUsername.getBytes(),
                                        false);
            byte[] usernameMessage = buildBreakoutControllerUsernameMessage(0);
            fcConnect.writeMessageToGroup(29,// recolor group cells 
                                          usernameMessage.length,
                                          usernameMessage,
                                          false,
                                          0);
          }
            		  
        // breakoutController is changing groups: 
        // change breakoutController's cell
        // to white in all, then repaint all groups
        if(((hostIndex == breakoutController) ||
            (oldFloorHolder == breakoutController)) &&
           (instructorBreakoutGroup != oldInstructorBreakoutGroup))
        {
          // set breakoutController's cell to white
          fcConnect.writeMessageToAll(27,
                                      8,
                                      breakoutControllerUsername.getBytes(),
                                      false);
          
          // if breakoutController is giving up floor
          // tell everybody there is no non-breakout floorHolder
          if(hostIndex == -1)
            fcConnect.writeMessageToAll(38,
                                        48,
                                        namesToBytes(nobodyUsername,
                                                     nobodyFullname),
                                        false);
                                        
          // repaint all the FC cells
          for(int repaintGroup = 1;
              repaintGroup <= numberOfBreakoutGroups;
              ++repaintGroup)
          {
            byte[] usernameMessage = 
              buildUsernameMessage(repaintGroup);
            fcConnect.writeMessageToGroup(29,
                                          usernameMessage.length,
                                          usernameMessage,
                                          false,
                                          repaintGroup);
          }
          oldInstructorBreakoutGroup = instructorBreakoutGroup;
        }
        if(hostIndex >= 0)
        {
          // breakout controller requesting to send to all groups
          if((hostIndex == breakoutController) && (breakoutGroup[hostIndex] <= 0))
          {
            fcConnect.writeMessageToAll(38,
                                        48,
                                        namesToBytes(breakoutControllerUsername,
                                          fullname[hostIndex]),
                                        false);
            byte[] usernameMessage = 
              buildUsernameMessage(-1);                              
            fcConnect.writeMessageToGroup(29,
                                          usernameMessage.length,
                                          usernameMessage,
                                          false,
                                          0);                                        
          }
          
          // other user is requesting floor in a group
          else if(breakoutGroup[hostIndex] > 0)
            fcConnect.writeMessageToGroup(38,
                                          48,
                                          namesToBytes(username[hostIndex],//breakoutControllerUsername,
                                            fullname[hostIndex]),//getFullname(breakoutControllerUsername)),
                                          false,
                                          groupNumber);
        }//end if(hostIndex >= 0)
        else if(groupNumber >= 0)
        
          // floorHolder is giving up the floor
          fcConnect.writeMessageToGroup(38,
                                        48,
                                        namesToBytes(nobodyUsername,nobodyYetFullname),
                                        false,
                                        groupNumber);
      }// end if(inBreakoutMode)
	    else
	      fcConnect.writeMessageToAll(38,
                                    48,
                                    namesToBytes(getFloorHolderName(hostIndex),
                                      getFullname(getFloorHolderName(hostIndex))),
                                    false);
                                            
    }// end synchronized(synchFloorHolder)
    
    // change the floorholder in each affected TCPTunnel
    int groupToChange = 0;
    if(hostIndex >= 0)groupToChange = breakoutGroup[hostIndex];
    for(int scanIndex=0; scanIndex<numberOfHosts; ++scanIndex)
    {
      try
      {
        if(tcpTunnelInstance[scanIndex] == null)continue;
        
        // this is hosts that are not breakout controller
        // and not leaving PA mode
        if(inBreakoutMode &&
           ((hostIndex != breakoutController) &&
            (oldFloorHolder != breakoutController) && 
            (groupToChange > 0)))
        {
          // breakout mode; change only for hosts where group is matched unless
          // hostIndex or oldHostIndex is breakoutController who is not in a group 
          if(hostIndex >= 0)
          {
            if((hostIndex != breakoutController) || (groupToChange > 0))
            {
              if(breakoutGroup[scanIndex] == groupToChange)
                tcpTunnelInstance[scanIndex].setTCPTunnelFloorHolder(hostIndex);
            }
            else tcpTunnelInstance[scanIndex].setTCPTunnelFloorHolder(hostIndex);
          }
          // not in breakout mode - set floorHolder to nobody
          else tcpTunnelInstance[scanIndex].setTCPTunnelFloorHolder(-1);
        }
        else // normal mode or breakoutController "PA", change all
          tcpTunnelInstance[scanIndex].setTCPTunnelFloorHolder(hostIndex);
      }
      catch(Exception e)
      {
        System.out.println("*** Exception in breakout scan tunnel:" +
                           e +" index "+scanIndex);
      }
    }// end for(int scanIndex...
    
    return true;

  }// end setFloorHolder()
  

  // gets floorholder name, global or by group, depending on breakout mode
  static String getFloorHolderName(int hostIndex)
  {
    if(inBreakoutMode)
    {
      if(hostIndex >= 0)
        if(breakoutGroup[hostIndex] >= 0)
          return breakoutFloorHolderName[breakoutGroup[hostIndex]];
        else return nobodyUsername;
      else return nobodyUsername;
    }
    return floorHolderName;
    
  }// end getFloorHolderName()
  
  
  // sets floorHolderName properly depending on breakoutMode
  void setFloorHolderName(int hostIndex,String newName)
  {
    if(inBreakoutMode)
    {
      if(hostIndex >= 0)
        if(breakoutFloorHolder[breakoutGroup[hostIndex]] >= 0 )
          breakoutFloorHolderName[breakoutGroup[hostIndex]] = newName;
    }
    else floorHolderName = newName;
    
  }// end setFloorHolderName()
  
  
  // method to broadcast client text messages
  // if controllerOnly, messages only go to controllers
  // This method is called by FCListen and TLM.
  void writeMessageToAll(int messageType,
                         int messageLength,
                         byte[] sendMessage,
                         boolean controllersOnly)
  {
    // build message for direct FC connections
    int fcMessageLength = messageLength+3;
    byte[] fcMessage = new byte[fcMessageLength];
    fcMessage[0] = (byte)messageType;
    fcMessage[1] = intToByte1(messageLength);
    fcMessage[2] = intToByte2(messageLength);
	  if(sendMessage != null)
	    System.arraycopy(sendMessage,0,fcMessage,3,messageLength);

    // build message for data tunnel FC connections
    int tunnelMessageLength = fcMessageLength+5;
    byte[] tunnelMessage = new byte[tunnelMessageLength];
    tunnelMessage[0] = 66;
    tunnelMessage[1] = intToByte1(fcMessageLength);
    tunnelMessage[2] = intToByte2(fcMessageLength);
    tunnelMessage[3] = 0;
    tunnelMessage[4] = 0;
    System.arraycopy(fcMessage,0,tunnelMessage,5,fcMessageLength);
    
    // send one or the other to all active FC including multicast
    if(usingMcFC)mcfcListen.mcFcSend(fcMessage);
    synchronized(synchWriteToAll)
    {
      for (int hostIndex = 0; hostIndex < numberOfHosts; hostIndex++)
      {
        // there are three cases; connection through data tunnel, direct, and MC
        // only send where one of them is active
        if ( fcIsInitialized[hostIndex] &&
            (tcpIsOpen[hostIndex] || uRecvTime[hostIndex] != 0))
        {
          // direct case:
          if(fcOutStreams[hostIndex] != null)
          {
            try
            {
              if( (controlPrivilege[hostIndex] > '0') ||
                  !controllersOnly)
              {
                fcOutStreams[hostIndex].write(fcMessage,0,fcMessageLength);
                fcOutStreams[hostIndex].flush();
              }
            }
            catch (IOException ioe) 
            {
              // disconnected
			        if(sendMessage != null)
                closeTunnel(hostIndex, true,
                            "disconnect in writeMessageToAll for command:" +
                            sendMessage[0]);
			        else
			          closeTunnel(hostIndex, true, "disconnect in writeMessageToAll");
            }
          }// end if(fcIsInitialized[hostIndex] && ...
          else
          {
            // tunnel case:
            if((controlPrivilege[hostIndex] > '0') || !controllersOnly)
              queueTunnel(2,hostIndex,tunnelMessage,tunnelMessageLength);
              
          }// end else/if(fcOutStreams[hostIndex] != null).
        } // end if(fcIsInitialized...
      } // end for(int hostIndex=0...
    }//end synchronized(synchWriteToAll)
  }// end writeMessageToAll()
  
  
  // method to send to group of clients text messages to TCP/UDP (no multicast)
  // for system messages (code 34) inserts host's group number into last position 
  // of message; messages are sent to hosts by toGroup (to all if toGroup is -1)
  public void writeMessageToGroup(int messageType,
                                  int messageLength,
                                  byte[] sendMessage,
                                  boolean controllersOnly,
                                  int toGroup)
  { 
    // build message for direct FC connections
    int fcMessageLength = messageLength+3;
    byte[] fcMessage = new byte[fcMessageLength];
    fcMessage[0] = (byte)messageType;
    fcMessage[1] = intToByte1(messageLength);
    fcMessage[2] = intToByte2(messageLength);
    if(sendMessage != null)
      System.arraycopy(sendMessage,0,fcMessage,3,messageLength);
    
    // build message for data tunnel FC connections
    int tunnelMessageLength = fcMessageLength+5;
    byte[] tunnelMessage = new byte[tunnelMessageLength];
    tunnelMessage[0] = 66;
    tunnelMessage[1] = intToByte1(fcMessageLength);
    tunnelMessage[2] = intToByte2(fcMessageLength);
    tunnelMessage[3] = 0;
    tunnelMessage[4] = 0;
    System.arraycopy(fcMessage,0,tunnelMessage,5,fcMessageLength);

    synchronized(synchWriteToAll)
    {
      for (int hostIndex = 0; hostIndex < numberOfHosts; hostIndex++)
      {
        // if toGroup is negative send to all
        if(toGroup >= 0)
          if(breakoutGroup[hostIndex] != toGroup)continue; 
         
        // special treatment for system message (code 34): insert groupNumber in message
        if(messageType == 34)
        {
          String groupNumber;
          if(breakoutGroup[hostIndex] <= 0)continue;
          groupNumber = (new Integer(breakoutGroup[hostIndex])).toString().substring(0,1);
          byte[] groupNumberByte = groupNumber.getBytes();
          fcMessage[fcMessageLength-1] = groupNumberByte[0];
          tunnelMessage[tunnelMessageLength-1] = groupNumberByte[0];
        }
	    
        // there are three cases; connection through data tunnel, direct, and MC
        // only send where one of them is active
        if ( fcIsInitialized[hostIndex] &&
            (tcpIsOpen[hostIndex] || uRecvTime[hostIndex] != 0))
        {
          // direct case:
          if(fcOutStreams[hostIndex] != null)
          {
            try
            {
              if( (controlPrivilege[hostIndex] > '0') ||
                  !controllersOnly)
              {
                fcOutStreams[hostIndex].write(fcMessage,0,fcMessageLength);
                fcOutStreams[hostIndex].flush();
              }
            }
            catch (IOException ioe) 
            {
              // disconnected
			        if(sendMessage != null)
                closeTunnel(hostIndex, true,
                            "disconnect in writeMessageToGroup for command:" +
                            sendMessage[0]);
			        else
			          closeTunnel(hostIndex, true, "disconnect in writeMessageToGroup");
            }
          }// end if(fcOutStreams[hostIndex] != null)
          else
          {
            // tunnel case:
            if( (controlPrivilege[hostIndex] > '0') || !controllersOnly)
              queueTunnel(2,hostIndex,tunnelMessage,tunnelMessageLength);
              
          }// end else/if(fcOutStreams[hostIndex] != null)
        } // end if(fcIsInitialized...
      } // end for(int hostIndex=0...
    }//end synchronized(synchWriteToAll)
  }// end writeMessageToGroup()
  

  // deal with incoming session variable from floor control via tunnel
  // (same code is in FCListen for direct-connected FC)
  // returns true if error; false otherwise
  boolean processSessionVariable(int hostIndex,
                                 byte[] fcSessionVariable,
                                 String remoteHost,
                                 int tlmcTcpPort)
  {
    // read the session variable sent first through the connection
    if (fcSessionVariable.length != 64)
    {
      internalErrorMessage(63,hostIndex);
      return true;
    }
    String offeredSessionVariable = new String(fcSessionVariable);
    
    // unlike the case with separate FC connection, here we know the
    // hostIndex to be associated with this FC; just need to confirm
    // that the sessionVariable and port we just got matches
    if((!offeredSessionVariable.equals(sessionVariable[hostIndex]) &&
        !offeredSessionVariable.equals(sixtyFourZeros))
         || (tlmcTcpPort != otherEndPort[hostIndex]) )
    {
      closeTunnel(hostIndex,true,"FC does not match for " + remoteHost);
      return true;
    }
    return false;
    
  }// end processSessionVariable


  // deal with login message arriving from floor control via tunnel
  // returns true if error; false otherwise
  boolean processLoginMessage(int hostIndex,
                              byte[] loginMessage,
                              byte[] sessionVariable)
  {
    byte[] name = new byte[8];
    byte[] passwd = new byte[8];
    byte[] courseID = new byte[11];
    byte[] semester = new byte[3];
    String offeredUsername = "";
    String offeredPassword = "";
    String offeredCourseID = "";
    String offeredSessionVariable = "";
    boolean foundMatch = false;
    byte[] readUsername;
    int addUserReturnValue = 0; // if unsuccessful gets set to 1 or 2
    boolean nameExists = false; // if exists gets set true

    // confirm messageLength
    if(loginMessage.length != 30)
    {
	    tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
      closeTunnel(hostIndex, true, "Invalid login messageLength " +
                  loginMessage.length + " received from host");
      return true;
    }
    // pull name/passwd/courseID/semester out of buffer
    System.arraycopy(loginMessage,0,name,0,8);
    System.arraycopy(loginMessage,8,passwd,0,8);
    System.arraycopy(loginMessage,16,courseID,0,11);
    System.arraycopy(loginMessage,27,semester,0,3);
    offeredUsername = new String(name);
    offeredPassword = new String(passwd);
    offeredCourseID = (new String(courseID)).toUpperCase();
    offeredSessionVariable = new String(sessionVariable);
    
    // Check for allowAllUsers case (anonymousHost = true)- user
    // does not need to authenticate, only to provide unique username
    if( anonymousHost[ hostIndex ] )
    {
      // username is unique if its not in the database/file for this course
      if( TLM.auth.nameAvailable( offeredUsername, offeredCourseID ) )
      {
        // and it is not already logged in
        nameExists = (fcConnect.activeUsers).containsKey(offeredUsername) &&
                     !userIsConnected(offeredUsername);

        // accept the login if the offered username
        // is not already taken and force drawGrid to true
        // since we can't look it up in the database/file
        if (!nameExists)
        {
          foundMatch = true;
          checkForDeadHostTableEntry(offeredUsername);
          username[hostIndex] = offeredUsername;
          drawGrid[hostIndex] = true;
        }
      } // end if( TLM.authnameAvailable...
    } // end if( anonymousHost[ thisHostIndex ] )

    // if authentication is required, check the quadruple
    // {username, password, IPaddress, courseID} to authenticate
    else
    {
      // authenticate via database or .auth file
      loginData loginResult = null;
      loginResult =
        auth.authenticate(hostIndex,
                          offeredUsername,
                          offeredPassword,
                          hostAddr[hostIndex],
                          sourcePort[hostIndex],
                          offeredCourseID,
                          offeredSessionVariable);
      if(loginResult != null)
      {
        offeredUsername = loginResult.username;
        offeredCourseID = loginResult.courseID;
        userCourseID[hostIndex] = offeredCourseID.trim();
        foundMatch = true;
        nameExists = (fcConnect.activeUsers).containsKey(offeredUsername) &&
                     userIsConnected(offeredUsername);
      }
      else
      {
        nameExists = false;
        if(webAuthenticate)
        {
          tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
          incrementUserCounts(hostIndex);  // bump count up so it can be reduced
          closeTunnel(hostIndex, false, "Web authentication failed");
          return true;
        }
      }
      // verify courseID
      if(foundMatch)
      {
        // if no class in schedule or privileged user, don't verify
        if(autoSchedule && classInSession &&
           (instructorPrivilege[hostIndex] < 2))
        {
          // check that courseID matches a course in nextCourseID[]
          int test;
          for (test = 0; test < numberOfCourseIDAliases; ++test)
            if (nextCourseID[test].trim().equals(offeredCourseID.trim()))
              break;
          if (test >= numberOfCourseIDAliases)foundMatch = false;
          
        } // end if(autoSchedule...
      } // end if(foundMatch)
    } // end else/if( anonymousHost[ hostIndex ] )
     
    // scan host table for inactive tunnels with this username and kill them
    String matchUsername = username[hostIndex];
    for(int testIndex=0; testIndex < maxConnections; ++testIndex)
    {
      if(username[testIndex].equals(matchUsername))
        if(testIndex != hostIndex)
        {
          tunnelIsActive[testIndex] = true;// force active status for closeTunnel()
          closeTunnel(testIndex,false,
                      "FCConnect found inactive tunnel with same username:" + offeredUsername);
          initializeAutoReconnect(testIndex);
          clearHostDataFromHostTable(testIndex);
        }
    }
    if(!foundMatch)
    {
      // the offered name/password does not authenticate
      if (++numAuthenticateTries[hostIndex] >= authenticateTimes)
      {
        if (sendToFC(61, 0, null, hostIndex))
          return true; // breaking connection
        numAuthenticateTries[hostIndex] = 0;
        System.out.println("host# " + hostIndex +
                           " failed three times to authenticate.");
		    tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
        closeTunnel(hostIndex, false,
                    "FCListen client failed to authenticate three times");
        return true;
      } // end if(numAuthenticateTries...
      else
      {
        if (sendToFC(32, 0, null, hostIndex))
          return true; // not accepted try again
      } // end else/if(numAuthenticateTries...
      return false;
    }// end if(!foundMatch)
     
    // the user has authenticated...
    // we continue to treat all Exceptions as lost connection, but after this
    // point upon exception we must clear out user state by dropParticipant()
	    
    // Add the user to activeUsers list in FCConnect.
    // addUser returns 0 if user successfully added
    //                 1 if no more room
    //                 2 if username already exists
    addUserReturnValue = fcConnect.addUser(hostIndex,offeredUsername);
    nameExistsAgain[hostIndex] = (addUserReturnValue != 0);
    if(nameExistsAgain[hostIndex]) // => user is not authenticated
    {
      if(addUserReturnValue == 1) // table full - do not accept another
      {
        if(sendToFC(59,0,null,hostIndex))return true;// server full
		      tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
        closeTunnel(hostIndex, false, "User:" + offeredUsername +
                    " rejected after valid login because server is full");
      }
      return false;
      
    } // end if(nameExistsAgain[hostIndex])
     
    // since we're sending FC through the data tunnel,
    // set its timer negative indicating no heartbeat
    // (it would duplicate the data heartbeat)
    fRecvTime[ hostIndex ] = -1;
    uRecvTime[ hostIndex ] =
      System.currentTimeMillis() + 2*tlmcDisconnectTime;
		
    // reset auto-tableClearing if pending
	  if(fcConnect.runningFiveMinutes)
	    runAfterFiveMinutes.interrupt();
    
    // set authenticated status
	  synchronized(authenticatedHost)
	  {
      authenticatedHost[hostIndex] = true;
    }
    return false;

  }// end processLoginMessage()
  
  // make the intialization messages and sends to client.
  // returns false on success, true on exception
  boolean sendFcInitMessage(int hostIndex)
  {
    // wait for tunnel to become active before sending FC init
    try
    {
      int tunnelActivateCounter = 0;
      while(!tunnelIsActive[ hostIndex ] &&
            (tunnelActivateCounter++ < TUNNEL_ACTIVATION_TIMEOUT) &&
            (tcpIsOpen[ hostIndex ] ||
            (fRecvTime[ hostIndex ] > System.currentTimeMillis() - 5000)) )
      {
        Thread.sleep(TUNNEL_ACTIVATION_SLEEP_MS);
        if(stopTLM)return true;
      }
      if(!tunnelIsActive[ hostIndex])
      {
        tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
        closeTunnel( hostIndex, false,
                     "timeout waiting for tunnel activation");
        return true;
	    }
	    if(!(tcpIsOpen[ hostIndex ] ||
         (fRecvTime[hostIndex] > System.currentTimeMillis() - 5000)) )
	    {
        tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
        closeTunnel( hostIndex, false,
                    "disconnected while waiting for tunnel activation");
	      return true;
	    }
    }
    catch(InterruptedException ie)
    {
      tunnelIsActive[hostIndex] = true;// force active status for closeTunnel()
	    closeTunnel(hostIndex, false,
	                "sleep interrupted while waiting for tunnel activation)");
	    return true;
    }

	  // determine floor control status
    byte hasControlPrivilege, drawGrid, thisIsTheController = 0;

    // check floor control privileges
    if(getControlPrivilege(hostIndex) > '0')
    {
      hasControlPrivilege = 1;
      controllerLoggedIn = true;

      // if session starts in controlled status, make it so
      if(initialFloorStatus > 1)
      {
        localFloorStatus[hostIndex] = initialFloorStatus;

        // uncomment to give controller the floor at login
        // fcConnect.setFloorHolder( hostIndex, offeredUsername, -1, -1 );
        thisIsTheController = 1;
		  
      } // if if(initialFloorStatus > 1)
    } // end if(FCConnect.getControlPrivilege
    else
    {
      hasControlPrivilege = 0;
      thisIsTheController = 0;
    } // end else/if(FCConnect.getControlPrivilege
	
	  // lookup username in groupData and set breakoutGroup
	  if(thisIsTheController == 1)
	    breakoutGroup[hostIndex] = 0;
	  else
	    breakoutGroup[hostIndex] = auth.getGroup(username[hostIndex]);
    
    // check whether to draw participant grid
    if(FCConnect.getDrawGrid(hostIndex))drawGrid = 1;
    else drawGrid = 0;

    // build initial FC message;
    // if no controller has logged in, send ********
    // else send floorholder name
    int userCount = (fcConnect.activeUsers).size();
    String thisUsername = username[hostIndex];
    Enumeration active = (fcConnect.activeUsers).elements();
    int initMessageSize = 61;
    if(drawGrid == 1)initMessageSize += userCount*48;
    byte[] initMessage = new byte[initMessageSize];
    initMessage[0] = (byte)initialFloorStatus;
    initMessage[1] = hasControlPrivilege;
    initMessage[2] = drawGrid;
    initMessage[3] = fcMaxUsers;
    
    // init message also contains username for floorHolder and user count
    System.arraycopy(getFloorHolderName(hostIndex).getBytes(),0,initMessage,4,8);
	  byte userCountByte = -1;
    if(drawGrid == 1)initMessage[12] = unsignedByte(userCount);
    else initMessage[12] = 1;
	
	  // plus username + fullname for this user
    System.arraycopy(thisUsername.getBytes(),0,
	                   initMessage,13,
					           8);
	  System.arraycopy(padBlanks(getFullname(thisUsername),40).getBytes(),0,
                     initMessage,21,
					           40);
    int offset = 61;

    // If drawGrid is true the init message also contains count of users
    // and usernames for all other users.
    if(drawGrid == 1)
    {
System.out.print("FCINIT:");//debugx
      // extract user and move into initial FC message
      for(int i=0; i<userCount; ++i)
      {
        String nextUsername = ((UserDetails)active.nextElement()).getUsername();
System.out.print(nextUsername+"|");//debugx
        if(!nextUsername.equals(thisUsername) && !nextUsername.equals("        "))
        {
          System.arraycopy(nextUsername.getBytes(),0,
		                       initMessage,offset,
						               8);
          offset += 8;
		      System.arraycopy(padBlanks(getFullname(nextUsername),40).getBytes(),0,
                           initMessage,offset,
						               40);
          offset += 40;
        }
      } // end for(int i=0; i<userCount; ++i)
System.out.println(" ");//debugx
    } // end if(drawGrid == 1)
    
    // send initMessage to client
    if(sendToFC(33,initMessageSize,initMessage,hostIndex))return true;
	   
    // start the server=side recorder if configured
    if(recordOnServer && (serverRecord == null) &&
	   ((initialFloorStatus <= 1) || (thisIsTheController == 1)))
	  {
      serverRecord = new ServerRecord();
	  	
      // if startup failed, disable recorder
      if(!recordOnServer)serverRecord = null;
	    else
	      System.out.println("started recorder at " + dateTime());
	  }
    
    // update all FC with changes
	  return refreshFloorControlStatus(hostIndex);
    
  }// end sendFcInitMessage()
  
  
  // send status update for a newly connected FC
  // returns false if messages sent successfully true if error
  boolean refreshFloorControlStatus(int hostIndex)
  {
    // when a user logs in, send all clients the floorControlRule
    // with floorRule set to the existing status
    byte[] floorRuleMessage = new byte[1];
    floorRuleMessage[0] = (byte)fcConnect.getFloorStatus();
    fcConnect.writeMessageToAll(41,1,floorRuleMessage,false);
	
	  // send vote panel setup and ongoing vote if they currently exist
	  if(openVoteDialog != null)
	  {
  	  if(sendToFC(20,openVoteDialog.length,openVoteDialog,hostIndex))
	    return true;
      if(startAVote != null)
	    {
	      if(sendToFC(22,startAVote.length,startAVote,hostIndex))
		    return true;
      }
	  }// end if(openVoteDialog)
	
    // if setup for breakout, re-announce breakout status
	  if(breakoutController >= 0)
	    announceBreakoutStatus();
	  
    // flag this FC as initialized
	  fcIsInitialized[hostIndex] = true;
    
    // if breakout has been initialized, 
    // send breakout group to the new client
    if(breakoutController >= 0)
    {
      if(auth.sendFCBreakoutMessage(hostIndex))
        return true;
      announceBreakoutStatus();
    }
    
    // send a newParticipant message to everybody
    sendUserStatus(hostIndex,(byte)0);
    
    return false;

  }// end refreshFloorControlStatus()
  

  // send a newParticipantMessage to all
  // currently connected users  
  void sendUserStatus(int hostIndex,byte status)
  {
    byte thisIsTheController = 0;
	  if(getControlPrivilege(hostIndex) > '0')
	    thisIsTheController = 1;
    byte[] newParticipantMessage = new byte[50];
    newParticipantMessage[0] = thisIsTheController;
    System.arraycopy(username[hostIndex].getBytes(),0,
	                   newParticipantMessage,1,
					           8);
	  System.arraycopy(padBlanks(getFullname(username[hostIndex]),40).getBytes(),0,
	                   newParticipantMessage,9,
					           40);
    newParticipantMessage[49] = status;
    fcConnect.writeMessageToAll(39,50,newParticipantMessage,false);
    
  }// end sendUserStatus()
  

  // set values associate with overall and group floorHolder to 'nobody'
  void setFloorHoldersToNobody()
  {
    breakoutGroup[breakoutController] = 0;
    setFloorHolder(-1,nobodyUsername,floorHolder,-1);
    for(int groupNumber=1; groupNumber<=numberOfBreakoutGroups; ++groupNumber)
      setFloorHolder(-1,nobodyUsername,breakoutFloorHolder[groupNumber],groupNumber);
          
  }// end setFloorHoldersToNobody
  
  // inform requester floor is not available
  byte[] rejectionMessage = ("CANNOT RELEASE FLOOR NOW").getBytes();
  boolean sendRejection(int thisHostIndex)
  {
    if(thisHostIndex < 0)return false;
    return sendToFC(34,
                    rejectionMessage.length,
                    rejectionMessage,
                    thisHostIndex);
                
  }// end sendRejection
  

  // process a received FC message
  // returns true if exception; otherwise false
  byte[] appFileResponse = new byte[1];
  long thisAppFileLength = 0, thisAppFileMtime = 0;
  boolean processFCMessage(int transactionType,
                           byte[] fcMessage,
                           int fcMessageLength,
                           int thisHostIndex)
  {
    // for heartbeat message, just continue read loop
    if(transactionType == 19)
    {
      if(fRecvTime[thisHostIndex] >= 0)
      {
        fRecvTime[thisHostIndex] = System.currentTimeMillis();
        if(testMessageLength(fcMessageLength,0,19,thisHostIndex))return true;
      }
      return false;
    }
//System.out.println("processFC TC:"+transactionType+" host:"+//debugx
//thisHostIndex+" TCP:"+connectedByTCP[thisHostIndex]);//debugx   
	  // ensure that voting setup comes only from controllers
	  if((transactionType >= 20) && (transactionType <= 23))
	    if(!fcConnect.getFloorAccess(thisHostIndex))
	      return true;
    
    // parse the incoming message depending on its transaction type
    switch(transactionType)
    {
      case 4: // Message type url
        if(fcConnect.getFloorAccess(thisHostIndex)) // client has the floor
        {
          byte[] sendTLMCMessage;
          int urlLength = 1;

          //  URL length of < 4 not permissible; or purposeful null URL
          if(fcMessageLength <= 4)
          {
            // format message telling TLMC to kill browser
            sendTLMCMessage = new byte[6];
            sendTLMCMessage[0] = 71;
            sendTLMCMessage[1] = intToByte1(0);
            sendTLMCMessage[2] = intToByte2(1);
            sendTLMCMessage[3] = 0;
            sendTLMCMessage[4] = 127;
            sendTLMCMessage[5] = 0;

            // if a short URL, inform the client
            // (except for length zero, used to kill browser)
            if(fcMessageLength >= 0)
            {
              if(sendToFC(52,0,null,thisHostIndex))return true;
            }
          }
          else // otherwise we have a URL to send
          {
            String URL = new String(fcMessage);

            // filter for URLs: insert http:// unless a .mov or .jpg
            try
            {
              // Check if the URL provided ends with .MOV or .JPEG substring.
              // In that case, or if http:/// is already there, don't insert http://
              String endString = URL.substring(fcMessageLength-4);
              if(!URL.startsWith("http://") && !URL.startsWith("https://")
                && endString.compareToIgnoreCase(".MOV")!=0
                && endString.compareToIgnoreCase(".JPG")!=0 )
              {
                URL = "http://" + URL;
              } // end if( startString...
            } // end try
            catch( Exception e )
            {
              System.out.println("Exception in URL process:" + e.getMessage());
            }
            // build message to go to each TLMC via tunnel
            urlLength = URL.length();
            if(urlLength > 127)urlLength = 127;
            sendTLMCMessage = new byte[urlLength+5];
            sendTLMCMessage[0] = 71;
            sendTLMCMessage[1] = intToByte1(urlLength);
            sendTLMCMessage[2] = intToByte2(urlLength);
            sendTLMCMessage[3] = 0;
            sendTLMCMessage[4] = 127;
            System.arraycopy(URL.getBytes(),0,sendTLMCMessage,5,urlLength);

            // send the system message (url being accessed) to all active FC
            fcConnect.writeMessageToAll(36,urlLength,URL.getBytes(),false);

          } // end else/if(messageLength<=4)

          // send to all active TCP tunnels
          for(int tunnelIndex = 0;
            tunnelIndex < numberOfHosts; ++tunnelIndex )
          {
            if( !connectedByTCP[tunnelIndex] )continue;
            queueTunnel( 1, tunnelIndex, sendTLMCMessage, urlLength+5 );
          } // end for(int sendTunnelIndex...
        } // end if(hasFloor)
        else // If user doesn't have floorAccess, send nofloor message.
        {
          if(sendToFC(42,0,null,thisHostIndex))return true; // you don't have the floor
        } // end else/if(fcConnect.getFloorAccess(thisHostIndex))
        break;

      case 5: // incoming message type floorRequest
        if(testMessageLength(fcMessageLength,8,5,thisHostIndex))return true;
        
        // If the user who requested floor already holds it, send
        // reminder it is not needed then reset it anyhow
		    int currentFloorHolder = getFloorHolder(thisHostIndex);
        if(currentFloorHolder == thisHostIndex)
          if(!inBreakoutMode)
            if(sendToFC(43,0,null,thisHostIndex))return true; // request floor by floorholder
        
        // send any grantMessage using refreshed floor status from global value
		    if(thisHostIndex >= 0)
          localFloorStatus[thisHostIndex] = fcConnect.getFloorStatus();
          
        // if we're in breakout mode and breakoutController has the floor
        // in "PA" mode, refuse any other requests for the floor
        if(inBreakoutMode && inPAMode && (thisHostIndex != breakoutController))
          sendRejection(thisHostIndex);

        // if the user is a controller or we're in breakoutMode without 
		    // controller send-to-all-groups floorlock, treat floorStatus as "ALWAYS"
        if(((controlPrivilege[thisHostIndex] > '0') && !inBreakoutMode) || 
		        (inBreakoutMode && 
		        ((floorHolder != breakoutController) || (breakoutGroup[breakoutController] > 0))))
          localFloorStatus[thisHostIndex] = 1;
        switch(localFloorStatus[thisHostIndex])
        {
          case 1:
            // if in breakoutMode and this is breakout controller
            // in "PA mode" (breakoutController not in a group), set
            // grid cells for group 0 (i.e. breakoutController) to green
            if(inBreakoutMode && inPAMode)
			      {
			        byte[] controllerUserMessage = buildUsernameMessage(-1);
			        fcConnect.writeMessageToGroup(29,
				                                    controllerUserMessage.length,
                                            controllerUserMessage,
								                            false,
								                            0);
			      }
            
            // The floorStatus is ALWAYS; take the floorAccess from
            // the earlier user and give it to the one who requested it
            if(inBreakoutMode)
              runDelaySetFloorHolder(thisHostIndex,
                                     new String(fcMessage),
                                     breakoutFloorHolder[breakoutGroup[thisHostIndex]],
                                     breakoutGroup[thisHostIndex]);
            else
              runDelaySetFloorHolder(thisHostIndex,
                                     new String(fcMessage),
                                     currentFloorHolder,
                                     -1);
            break;

          case 2:
            // The floor Status is NEVER; reject any floorRequest message
            // not from a controller.
            if(sendToFC(44,8,fcMessage,thisHostIndex))return true;
            break;

          case 3:
            // The floor status is ASK ME; send a request message to all the
            // controllers to raise flag telling that request is registered;
            // then add the request to pending requests and send a request
            // message to all clients so that those for controllers
            // can turn the requesting user's grid to pink.
            if(sendToFC(46,0,null,thisHostIndex))return true;
            fcConnect.addFloorRequest(thisHostIndex,fcMessage);
            
        } // end switch(localFloorStatus)
/*debugx        
        // if configured to support appDataZipFile download, check to see if we need
        // to offer floor controller a chance to accept/designate download file
        if((appFileDownloadLimit > 0) && !downloadBuffering) 
          if(controlPrivilege[thisHostIndex] > '0')
            if(!downloadOfferedToController)
            {
              downloadOfferedToController = true;
              hostToWhichDownloadOffered = thisHostIndex;
              fcConnect.sendZipfileList(thisHostIndex);
		    }
debugx*/
        break;

      case 6: // incoming message type floorGrant
        // This message occurs only when controller grants the floor in
        // ASK ME mode; in this case, remove the pending request of
        // this user, send grant message to all clients, and set the latest
        // floor user to this user name. All this is done in
        // removeFloorRequest synchronized method.
        if(testMessageLength(fcMessageLength,8,6,thisHostIndex))return true;

        // confirm the requester has privilege to ask
        if(!fcConnect.getFloorAccess(thisHostIndex))
        {
          if (sendToFC(42, 0, null, thisHostIndex))return true;
        }
        else
        {
          // post message to inform controller if there
          // was no request to grant
          if(!fcConnect.removeFloorRequest(fcMessage,true))
            if(sendToFC(45,0,null,thisHostIndex))return true;
        }
        break;

      case 10: // incoming message type dropParticipant
        // send message to other clients and kill this thread
        if(testMessageLength(fcMessageLength,8,10,thisHostIndex))return true;
        dropParticipant(thisHostIndex);
        closeTunnel(thisHostIndex, false, "dropParticipant received in FCListen");
        return true;

      case 7: // message type floor status (floor rule) change request
        // Process only if the user who has sent the message has the floor;
        // otherwise send nofloorMessage as byproduct of sending the
        // systemMessage.
        if(testMessageLength(fcMessageLength,1,7,thisHostIndex))return true;
        if(fcConnect.getFloorAccess(thisHostIndex))
        {
          localFloorStatus[thisHostIndex] = fcMessage[0];
          fcConnect.setFloorStatus((byte)localFloorStatus[thisHostIndex]);

          // When floorStatus changes, send floorControlRule message
          byte[] floorRuleMessage = new byte[1];
          floorRuleMessage[0] = (byte)fcConnect.getFloorStatus();
          fcConnect.writeMessageToAll(41,1,floorRuleMessage,false);
        }
        break;

      case 8: // incoming message type floorYield
        if(testMessageLength(fcMessageLength,8,8,thisHostIndex))return true;

        // only the floorholder can yield
        if(!inBreakoutMode)
          if(thisHostIndex != getFloorHolder(thisHostIndex))
            break;
           
        // reset firstFloorHolder to none
        if(floorHolder == firstFloorHolder)firstFloorHolder = -1;

        // send grantMessage to give the floor to nobody (username 8 stars)
        if(inBreakoutMode)
          fcConnect.runDelaySetFloorHolder(-1,
                                           nobodyUsername,
                                           thisHostIndex,
                                           breakoutGroup[thisHostIndex]);
        // or to firstFloorHolder
        else
          fcConnect.runDelaySetFloorHolder(firstFloorHolder,
                                           (firstFloorHolder < 0)?
                                             nobodyUsername:username[firstFloorHolder],
                                           thisHostIndex,
                                           -1);
        break;

      case 9: // incoming message type request fullname

        // don't respond if name is restricted
        if(testMessageLength(fcMessageLength,8,9,thisHostIndex))return true;
        if(hideFullname)break;

        // reply with the fullname if one is found
        byte[] fullnameMessage = new byte[48];
        System.arraycopy(fcMessage,0,fullnameMessage,0,8);
        String sendFullname = fcConnect.getFullname(new String(fcMessage));

        // pad the name out to 40 bytes and send it
        System.arraycopy(padBlanks(sendFullname,40).getBytes(),0,
                         fullnameMessage,8,40);
        if(sendToFC(53,48,fullnameMessage,thisHostIndex))return true;
        break;

      case 12: // incoming message type withdraw floor request
        // remove the pending request of this user; then
        // change the user's grid square back to greeen
        // in all connected clients
        if(testMessageLength(fcMessageLength,8,12,thisHostIndex))return true;
        fcConnect.removeFloorRequest(fcMessage,false);
        fcConnect.writeMessageToAll(55,8,fcMessage,false);
        break;

      case 13:// incoming message type launchApp
        // client must have the floor and also be controller
        if(fcConnect.getFloorAccess(thisHostIndex) && (controlPrivilege[thisHostIndex] > 0))
        {
          // send the appLaunch message (launchApp string) to all active FC
          fcConnect.writeMessageToAll(63,fcMessageLength,fcMessage,false);
        } // end if(hasFloor)
        else // If user doesn't have floorAccess, send noFloor message.
        {
          if(sendToFC(42,0,null,thisHostIndex))return true; // you don't have the floor
        } // end else/if(fcConnect.getFloorAccess(thisHostIndex))
        break;

      case 14:// incoming message identifies FCis ready to receive appDataZipFile
      
        // post status change
        fcIsReadyForAppData[thisHostIndex] = true;

        // if an addDataZipFile is buffered, start it downlaoding
        if(downloadIsReady)
        {
          try
          {
            if(tunnelSends[thisHostIndex] != null)
              tunnelSends[thisHostIndex].startDownloading();
            else if(!downloadBuffering)
              System.out.println("*** couldn't start download for host:" +
                                 thisHostIndex + " TunnelSend not ready"); 
          }
          catch(Exception e)
          {
            System.out.println("*** failed to start download for host:" +
                               thisHostIndex + " exception:" + e);
          }
        }
        else 
          System.out.println("*** host:" + thisHostIndex +
                             " requested appDataZipfile but none is buffered");
        break;
              
      case 18:// incoming message ACKs FC initialize message (33)
        if(refreshFloorControlStatus(thisHostIndex))return true;
        announceBreakoutStatus();
        break;
		    
      case 23://incoming message shuts off and contains results of a vote
        // individual FCs will have tallied the votes as they arrived
        // here we need only to log it
        if(testMessageLength(fcMessageLength,8,23,thisHostIndex))return true;
        int yesVotes = bytesToUshort(fcMessage[0],fcMessage[1]);
        int noVotes = bytesToUshort(fcMessage[2],fcMessage[3]);
        int abstainVotes = bytesToUshort(fcMessage[4],fcMessage[5]);
        int possibleVotes = bytesToUshort(fcMessage[6],fcMessage[7]);
        startAVote = null;
        	
        // log the vote results
        System.out.println("Course " + thisCourseID + " completed vote at " +
                           dateTime() + " sequence " + voteSequence + 
                           " results yes:" + yesVotes + " no:" + noVotes + 
                           " abstain:" + abstainVotes + " possible:" + possibleVotes + 
                           " proposition:\n" + proposition);
        fcConnect.writeMessageToAll(transactionType,
                                    fcMessageLength,
                                    fcMessage,
                                    false);
        break;
        
      case 22:// incoming message starts a vote; save its proposition for log
        if(testMessageLength(fcMessageLength,-4,22,thisHostIndex))return true;
        voteSequence = bytesToUshort(fcMessage[0],fcMessage[1]);
        propositionLength = bytesToUshort(fcMessage[2],fcMessage[3]);
        proposition = new String(fcMessage,4,propositionLength);
        startAVote = new byte[fcMessageLength];
        System.arraycopy(fcMessage,0,startAVote,0,fcMessageLength);

        // send the message to all connected FC
        fcConnect.writeMessageToAll(transactionType,
                                    fcMessageLength,
                                    fcMessage,
                                    false);
        break;
											
      case 20:// open a vote panel
        if(testMessageLength(fcMessageLength,10,20,thisHostIndex))return true;
	      openVoteDialog = new byte[fcMessageLength];
        openVoteFloorholder = thisHostIndex;
        System.arraycopy(fcMessage,0,openVoteDialog,0,fcMessageLength);
        		
        // send the message to all connected FC
        fcConnect.writeMessageToAll(transactionType,
                                    fcMessageLength,
                                    fcMessage,
                                    false);
        break;
        
      case 21:// kill a vote panel
        if(testMessageLength(fcMessageLength,0,21,thisHostIndex))return true;
        openVoteDialog = null;
        startAVote = null;
        fcConnect.writeMessageToAll(transactionType,
                                    fcMessageLength,
                                    fcMessage,
                                    false);
        break;
        	
      case 24:// submit an individual vote
        // send the message to all connected FC
        if(testMessageLength(fcMessageLength,2,24,thisHostIndex))return true;
        fcConnect.writeMessageToAll(transactionType,
                                    fcMessageLength,
                                    fcMessage,
                                    false);
        break;
		
      // Breakout messages
      //
      // notes on floorHolder for breakout:
      //
      // 1. when not inBreakoutMode, floorholder management
      //    is unchanged from previous versions
      // 2. whenever changing to/from inBreakoutMode, direct all FC clients
      //    to clear floorholder status, then set all cells white except
      //    those with active connections who are in the FC's breakout group
      // Wnen in breakout mode:
      // 3. floor rule is effectively 'always' so any requester gets the floor
      // 4. any user not placed in group by controller is in group 1
      // 5. until controller with breakout control panel joins a group,
      //    that controller's FC receives from no group but when it takes
      //    the floor it sends to all groups in "PA" mode; when this is 
      //     happening:
      //    a. the controller shows up as floorholder in all groups
      //    b. nobody else can take the floor until controller releases
      //       (to communicate, others use chat)
      //    c. all other connected hosts show green on controller's grid
      //    d. when controller releases, that cell goes back to white
      // 6. when controller with breakout panel joins a group, the controller's
      //    grid box shows up green in that group only, and the group only
      //    shows up in controller grid
		
      case 25:// request breakout data
        // confirm sender status
        if(testMessageLength(fcMessageLength,1,25,thisHostIndex))return true;
        if(controlPrivilege[thisHostIndex] == '0')
        {
          String noControl = "YOU'RE NOT AUTHORIZED";
          sendToFC(34,noControl.length(),noControl.getBytes(),thisHostIndex);
          break;
        }
        if((breakoutController > -1) && (fcMessage[0] == 1))
        {
          String alreadyActive = "BREAKOUT CONTROL ALREADY ACTIVE";
          sendToFC(34,alreadyActive.length(),alreadyActive.getBytes(),thisHostIndex);
          break;
        }
        if(fcMessage[0] == 1)
        {
          // select this host as breakoutController
          breakoutController = thisHostIndex;
          breakoutControllerUsername = padBlanks(username[thisHostIndex],8);
           
          // send breakout data; if database error set number of groups -1
          if(auth.sendBreakoutData())
            numberOfBreakoutGroups = -1;          
        }
        else resetBreakoutMode(thisHostIndex);
        break;
		
      case 26:// assign user to breakout group
	      // confirm sender status, then invoke assignment method in Auth
        if(testMessageLength(fcMessageLength,9,26,thisHostIndex))return true;
        if(controlPrivilege[thisHostIndex] == '0')break;
        if(breakoutController < 0)break;// self-protection
        if(fcMessageLength <= 8)break;  // self-protection
        String assignUsername = new String(fcMessage,1,8);
        int assignIndex = getUserPosition(assignUsername);
        if(assignIndex < 0)break;       // self-protection
               
        // if user being assigned is breakoutController leaving
        // 'PA mode', move overall and group floor holders to 'nobody'
        if(inBreakoutMode && inPAMode)
          if(assignIndex == breakoutController)
            setFloorHoldersToNobody(); 
         
        // if in breakout mode and this is the breakoutController and is not 
        // in active group (i.e. is in group 0), or is floorHolder in its group,
        // take floor away from this one before it joins a group
        if(inBreakoutMode)
          if(((assignIndex == breakoutController) && (breakoutGroup[assignIndex] == 0)) ||
             ((breakoutGroup[assignIndex] > 0) &&
              (breakoutFloorHolder[breakoutGroup[assignIndex]] == assignIndex)))
                runDelaySetFloorHolder(-1,
                                       nobodyUsername,
                                       assignIndex,
                                       breakoutGroup[assignIndex]);
          
        // set status in database and announce it to clients
        auth.setBreakoutGroup(fcMessage[0],assignUsername);
        announceBreakoutStatus();
        break;
		
      case 27:// update breakout mode and breakout controller groupNumber
	      // confirm sender status, then forward go-to-breakout to all connected
        if(testMessageLength(fcMessageLength,2,27,thisHostIndex))return true;
        if(breakoutController < 0)break;// self-protection
        if(controlPrivilege[thisHostIndex] == '0')break;
        fcConnect.writeMessageToAll(30,// goto breakout mode
                                    fcMessageLength,
                                    fcMessage,
                                    false);
        
        // send system message if breakout status changed
        boolean oldBreakoutMode = inBreakoutMode;
        inBreakoutMode = (fcMessage[0] == 1);
        instructorBreakoutGroup = fcMessage[1];
        
        // if moving from breakout mode to normal mode, 
        // reset controller's group to -1
        if(!inBreakoutMode)
          if(oldBreakoutMode ^ inBreakoutMode)
          {
            byte[] controllerLeaveBreakout = new byte[9];
            controllerLeaveBreakout[0] = -1;
            System.arraycopy(breakoutControllerUsername.getBytes(),0,
                             controllerLeaveBreakout,1,
                             8);
            fcConnect.writeMessageToAll(26,
                                        9,
                                        controllerLeaveBreakout,
                                        false);
          }
        
        // if in breakout mode and breakoutController is not in
        // a group, or if changing modes, move 
        // breakoutController's host to group 0
        if(inBreakoutMode)
          sendBreakoutControllerGroup((byte)0);
          
        // set overall and each group floorHolder status to 
        // nobody, every time we make transition in/out of breakout 
        if(inBreakoutMode ^ oldBreakoutMode)
          setFloorHoldersToNobody();
          
        // initialize breakoutController
        if((breakoutGroup[thisHostIndex] != instructorBreakoutGroup) ||
           (inBreakoutMode ^ oldBreakoutMode))
        {
          breakoutGroup[thisHostIndex] = instructorBreakoutGroup;
          announceBreakoutStatus();
        }
        
        // in breakoutMode, if breakoutController changes groups withdraw floorholder status
        // (this includes changing to/from sentToAll "PA" status which is group 0)
        if(inBreakoutMode)
          if(thisHostIndex == breakoutController)
            if(oldInstructorBreakoutGroup != instructorBreakoutGroup)
              if(oldInstructorBreakoutGroup >= 0)
                if(breakoutFloorHolder[oldInstructorBreakoutGroup] == breakoutController)
                  runDelaySetFloorHolder(-1,
                                        nobodyUsername,
                                        breakoutController,
                                        oldInstructorBreakoutGroup);
                  
        // if in breakoutMode we must update grid status on all users' FC
        if(inBreakoutMode)
          updateGridBreakoutStatus();
          
        // just left breakout; send usernameMessage to reset all active users
        else
          sendNonbreakoutUsernameMessage();
        break;
       	
      case 28:// set number of groups
        // confirm sender status, then invoke numbering method in Auth
        if(breakoutController < 0)break;//self-protection
        if(testMessageLength(fcMessageLength,1,28,thisHostIndex))return true;
        if(controlPrivilege[thisHostIndex] == '0')break;
        auth.setNumberOfGroups(fcMessage[0]);
        break;
        
      case 101:// request list of server recordings
        sendRecordingsList(thisHostIndex,fcMessageLength);
        break;
        
      case 102:// transmit requested server recording to requesting host
        recordingDownloadFilename = new String(fcMessage);
        tunnelSends[thisHostIndex].startRecordingDownload();
        break;
        
      case 107:// FC requests to upload new appDataZipFile
        if(controlPrivilege[thisHostIndex] == 0)
        {
          System.err.println("*** host:" + thisHostIndex + 
                             " non-controller attempted appDataZipFile upload");
          break;
        }
        thisAppFileLength = bytesToLong(fcMessage,0);
        thisAppFileMtime = bytesToLong(fcMessage,8);
        String appFileName = new String(fcMessage,16,20);
         
        // confirm zipfile name is reasonable
        if(!appFileName.substring(0,moodleFileName.length()).equals(moodleFileName) ||
           !appFileName.trim().endsWith(".zip"))
        {
          System.out.println("*** host:" + thisHostIndex + 
            " bad appDataZipFile name attempted upload:" + appFileName +
            " correct name is:" + moodleFileName);
          break;
        }
            
        // check whether we already have that file, based on length and Mtime
        String appFilePathName = 
          new String(appFilePath + moodleFileName + "/" + moodleFileName + ".zip");
  	    File appZipFile = new File(appFilePathName);
        long currentFileLength = 0, currentFileMtime = 0;
  	    if(appZipFile.exists())
  	    {
  	      currentFileLength = appZipFile.length();
  	      currentFileMtime = appZipFile.lastModified();
        }
        appFileResponse[0] = 0; 
        if((thisAppFileLength != currentFileLength) || (thisAppFileMtime != currentFileMtime))
        {
          appFileResponse[0] = 1;
          
          // send respond to system messages window on FC
          byte[] uploadMessage = ("uploading appData " + moodleFileName).getBytes();
          fcConnect.sendToFC(49,
                             uploadMessage.length,
                             uploadMessage,
                             thisHostIndex);
        }
          
        // send response to FC
        sendToFC(108,1,appFileResponse,thisHostIndex);
        break;
      
      case 109:// FC sends new appDataZipFile segment
        if(appFileResponse[0] == 1)
          writeAppDataZipFile(fcMessageLength,fcMessage,thisHostIndex);
        break;
        
      // recording upload messages
      
      case 119:// REC request to upload recording, relayed through RC & FC
      
        // make sure user has upload privilege
        if(controlPrivilege[thisHostIndex] == 0)
        {
          System.err.println("*** host:" + thisHostIndex + 
                             " non-controller attempted recording upload");
          break;
        }
        
        // extract filename and use it to make a thread to handle the upload
        String uploadFilename = new String(fcMessage);
        UploadRecording uploadRecording = 
          new UploadRecording(uploadFilename,thisHostIndex);
        thisHostUpload[thisHostIndex] = uploadRecording;
        break;
        
      case 121:// REC provides a byte array to add to the recording
        
        // make sure user has upload privilege
        if(controlPrivilege[thisHostIndex] == 0)
        {
          System.err.println("*** host:" + thisHostIndex + 
                             " non-controller attempted recording segment upload");
          break;
        }
        try
        {
          thisHostUpload[thisHostIndex].queueMessage(fcMessage);
          thisHostUpload[thisHostIndex].notifyWait();
        }
        catch(NullPointerException npe)
        {
          System.err.println(
            "*** received code 121 for host:" + thisHostIndex +
            " without previous code 119 upload acceptance");
        }
        break;
        
      case 124:// FC provide a filename in recordingsDirectory to delete
        
        // make sure user has upload privilege
        if(controlPrivilege[thisHostIndex] == 0)
        {
          System.err.println("*** host:" + thisHostIndex + 
                             " non-controller attempted recording upload");
          break;
        }
        
        // delete the file
        String fileToDelete = new String(fcMessage);
        String filePathToDelete = 
          recordingsSubdirectory + '/' + fileToDelete;
        File testExistence = new File(filePathToDelete);
        if(!testExistence.exists())
        {
          System.err.println(
            "*** client sent invalid filename to delete:" +
            filePathToDelete);
          break;
        }
        if(!testExistence.delete())
          System.err.println(
            "*** failed to delete server recording as requested by host:" +
            thisHostIndex + "\nfilename:" + filePathToDelete);
        break;
                	
      default:
        // If the message type is none of the above, then it is to be
        // forwarded to all clients (text, systemMessage types).
        //
        // Check whether the user sending messages needs to have the floor
        // i.e. when the message sent is text, it is not required.
        int responseType;
        switch(transactionType)
        {
          case 2:
            responseType = 51;
            break;

          case 3:
            responseType = 35;
            break;

          default:
            responseType = 49;
            break;
        }
        if((transactionType == 2) || (thisHostIndex == getFloorHolder(thisHostIndex))
           || (transactionType == 3)) // text messages
        {
          // read message and send to all FC clients
          int thisMessageLength = fcMessageLength;
          if(thisMessageLength > 127)thisMessageLength = 127;
          fcConnect.writeMessageToAll(responseType,
                                      thisMessageLength,
                                      fcMessage,
                                      (transactionType == 2));

          // also print text message to log
          if(thisMessageLength != 0)
          {
            if(transactionType == 3)
               System.out.println(dateTime() + " " + hostAddr[thisHostIndex] +
                                  "\nCHAT:" +
                                  new String(fcMessage));
            if(transactionType == 2)
              System.out.println(dateTime() + " " + hostAddr[thisHostIndex] +
                                 "\nTO PROF:" +
                                 new String(fcMessage));
          }
        } // end if((transactionType == 2)... // text message

        // If user doesn't have floorAccess, send nofloor message.
        else
        {
          if(sendToFC(42,0,null,thisHostIndex))return true; // you don't have floor
        } // end else/if(hasFloor...
    } // end switch(transactionType)
    
    return false;

  }// end processFCMessage
  
  
  // reset all breakout mode state
  // when thisHostIndex exits breakout
  void resetBreakoutMode(int thisHostIndex)
  {
    inBreakoutMode = false;
    breakoutGroup[thisHostIndex] = 0;
    setFloorHoldersToNobody();
    breakoutController = -1;
    breakoutControllerUsername = "        ";
    setFloorHolder(thisHostIndex,username[thisHostIndex],thisHostIndex,-1);
    announceBreakoutStatus();
    
  }//end resetBreakoutMode()
  
  
  // sends individual FC cell change message for breakout controller
  void sendBreakoutControllerGroup(byte groupNo)
  {
    // set values in Auth and breakoutFloorHolder
    if(breakoutController < 0)return;
    auth.setBreakoutGroup(groupNo,breakoutControllerUsername);
    breakoutFloorHolder[groupNo] = -1;
    
    // make and send message
    byte[] setBreakoutControllerMessage = new byte[9];
    setBreakoutControllerMessage[0] = groupNo;
    System.arraycopy(breakoutControllerUsername.getBytes(),0,
                     setBreakoutControllerMessage,1,
                     8);
    fcConnect.writeMessageToAll(26,
                                9,
                                setBreakoutControllerMessage,
                                false);
                                
  }// end sendBreakoutControllerGroup()
  
  
  // update grid status for all FC
  void updateGridBreakoutStatus()
  {
	  byte[] usernameMessage;
    if(inBreakoutMode)
    {
      // iterate over all groups, sending usernameMessages
      for (int groupNumber = 0;
           groupNumber <= numberOfBreakoutGroups;
           groupNumber++)
      {
        // refresh all connected usernames
        usernameMessage = buildUsernameMessage(groupNumber);
        fcConnect.writeMessageToGroup(29,
                                      usernameMessage.length,
                                      usernameMessage,
                                      false,
                                      groupNumber);
         
        // refresh floorHolder
        int setGroupFloorHolder = breakoutFloorHolder[groupNumber];
        if(setGroupFloorHolder >= 0)
          fcConnect.writeMessageToGroup(38,
                                        48,
                                        namesToBytes(username[setGroupFloorHolder],
                                          fullname[setGroupFloorHolder]),
                                        false,
                                        groupNumber);
        else 
          fcConnect.writeMessageToGroup(38,
                                        48,
                                        namesToBytes(nobodyUsername,
                                                     nobodyFullname),
                                        false,
                                        groupNumber);
      }
    }
    else
    {
      // not in breakout; send one big message
      usernameMessage = buildUsernameMessage(-1);
      fcConnect.writeMessageToAll(29,
                                  usernameMessage.length,
                                  usernameMessage,
                                  false);
    }
  }// end updateGridBreakoutStatus()
  
  
  // send system message with breakout status of each host
  void announceBreakoutStatus()
  {
    String breakoutMessage;
    if(inBreakoutMode)
	  {
      breakoutMessage = "YOU ARE IN BREAKOUT GROUP  ";
      fcConnect.writeMessageToGroup(34,
						                        breakoutMessage.length(),
						                        breakoutMessage.getBytes(),
						                        false,
						                        -1);
	  }
	  else
    {
      breakoutMessage = "IN WHOLE CLASS MODE";
      fcConnect.writeMessageToAll(34,
					                        breakoutMessage.length(),
						                      breakoutMessage.getBytes(),
						                      false);
	  }
	  
	  // redistribute grid status to all FC
	  updateGridBreakoutStatus();
	
  }// end announceBreakoutStatus


  // send a message to Floor Control client
  // parameters:
  //  messageType: byte 1 of message indicating its format (see TLM comments)
  //  messageLength: length of the message contents, excluding header (can be 0)
  //  message: stuff to send
  //  hostIndex: global index of sending host, used for exception handling
  // returns true on exception, false otherwise
  boolean sendToFC(int messageType,int messageLength,byte[]message,int hostIndex)
  {
    BufferedOutputStream out;
     
    // don't send anything but setup traffic before setup is complete
    if(!fcIsInitialized[hostIndex])
    if(!((messageType == 32) || (messageType == 33) || (messageType == 58) ||
         (messageType == 59) || (messageType == 61) ))
    {
      System.err.println("*** premature TC in sendToFC:"+messageType);
      return true;
    }
    
    // if called with non-host, reply error
    if(hostIndex < 0)
    {
      System.err.println("*** invalid hostIndex:" + hostIndex + 
                         " in sendToFC with messageType:" + messageType + 
                         " MessageLength:" + messageLength);
      return true;
    }
    
    // make and send write header; then write message, if any, to tunnel
    try
    {
      byte[] headerBuffer = new byte[8];
      headerBuffer[0] = 66;
      headerBuffer[1] = intToByte1(messageLength+3);
      headerBuffer[2] = intToByte2(messageLength+3);
      headerBuffer[3] = 0;
      headerBuffer[4] = 0;
      headerBuffer[5] = (byte)messageType;
      headerBuffer[6] = intToByte1(messageLength);
      headerBuffer[7] = intToByte2(messageLength);
      if(fcOutStreams[hostIndex] != null)// non-null when direct connected
      {
        out = fcOutStreams[hostIndex];
        out.write(headerBuffer, 5, 3);
        if (messageLength > 0)
          out.write(message, 0, messageLength);
        out.flush();
      }
      else
      {
        if(usingMcFC && (hostInetAddr[hostIndex] != null))
        {
          byte[] udpBuffer = new byte[messageLength+3];
          System.arraycopy(headerBuffer,5,udpBuffer,0,3);
          if(messageLength > 0)
            System.arraycopy(message,0,udpBuffer,3,messageLength);
          mcfcListen.ucFcSend(udpBuffer,hostInetAddr[hostIndex]);
        }
        if(messageLength == 0)
          queueTunnel(2,hostIndex,headerBuffer,8);
        else
        {
          byte[] rebuffer = new byte[messageLength+8];
          System.arraycopy(headerBuffer,0,rebuffer,0,8);
          System.arraycopy(message,0,rebuffer,8,messageLength);
          queueTunnel(2,hostIndex,rebuffer,messageLength+8);
        }
        
      }// end else/if(fcOutStreams[hostIndex] != null)
    }
    catch (IOException ioe)
    {
      closeTunnel(hostIndex, true,
                  "FCListen disconnected:" + ioe.getMessage());
      return true;
    }
    return false;
    
  } // end sendToFC


  // drop this participant from the session
  void dropParticipant(int thisHostIndex)
  {
    // if the floorholder disconnected, insert a placeholder
    if(thisHostIndex == getFloorHolder(thisHostIndex))
      setFloorHolderName(thisHostIndex,nobodyUsername);
          
    // if the user dropping is the only controller, set the new floor status
    // back to initial floor status and setFloorStatus to inform all the users
    if((controlPrivilege[thisHostIndex] > '0') && 
       (numberOfControllersLoggedIn < 2))
    {
      fcConnect.setFloorStatus(initialFloorStatus);

      // when floorStatus changes, send the floorControlRule message to all FC
      byte[] floorRuleMessage = new byte[1];
      floorRuleMessage[0] = (byte)fcConnect.getFloorStatus();
      fcConnect.writeMessageToAll(41,1,floorRuleMessage,false);
    }
    
    // if user had authenticated, send delete message to all FC
    if(authenticatedHost[thisHostIndex] && !username[thisHostIndex].equals("        "))
    {
      fcConnect.writeMessageToAll(40,8,username[thisHostIndex].getBytes(),false);
      
      // remove this from activeUsers hashtable
/*debugx
      if( fcConnect.getControlPrivilege(thisHostIndex) > '0' )
      {
        // if this is the only controller and no class in
        // session, clear the download option
        // (this allows controller a retry by logout/login)
        if(numberOfControllersLoggedIn <= 1)
        if(downloadOfferedToController)
          if(hostToWhichDownloadOffered == thisHostIndex)
          {
            downloadOfferedToController = false;

            // if there are zero TCP users,
            // also reset overall download status
            if((numberOfTunnelsActive  == 0) && !classInSession)
              downloadBuffering = false;//debugx
          }
      }
debugx*/
      fcConnect.removeUser(thisHostIndex,username[thisHostIndex]);

      // remove this from the validatedHosts hashtable
      if(usingMcFC)mcfcListen.removeHost(thisHostIndex);

    } // end if(authenticatedHost[thisHostIndex])

    // close connection associated with this thread
    closeTunnel(thisHostIndex,false,"drop participant by floor control");

  } // end dropParticipant method
  
  
  // build a usernameMessage just for the breakoutController
  byte[] buildBreakoutControllerUsernameMessage(int groupNumber)
  {
    byte[] usernameMessage = new byte[9];
    usernameMessage[0] = (byte) groupNumber;
    System.arraycopy(breakoutControllerUsername.getBytes(),0,
                     usernameMessage,1,
                     8);
    return usernameMessage;
    
  }// end buildBreakoutControllerUsernameMessage()
  
  
  // build a message consisting of active usernames in a group
  // (group -1 means all active usernames)
  byte[] buildUsernameMessage(int groupNumber)
  {
    // side-effect: update counts of active users and controllers
    int countTunnelsActive = 0;
    int countControllersLoggedIn = 0;
    int activeUsersInGroup = 0;
    int scanUsers;
  
    synchronized(authenticatedHost)
    {
      // determine how many active users are in the group
      if(groupNumber < 0)
        activeUsersInGroup = loggedInUsers.size();
      else
      {
        for(scanUsers = 0;
            scanUsers < maxConnections;
            scanUsers++)
         if(authenticatedHost[scanUsers])
           if(breakoutGroup[scanUsers] == groupNumber)
             activeUsersInGroup++;
      }
       
      // build the message 
      byte[] usernameMessage = new byte[1+8*(activeUsersInGroup+1)];
      usernameMessage[0] = (byte)groupNumber;
	    int copyIndex = 0;
	    for(scanUsers = 0;
	        scanUsers < maxConnections;
		      scanUsers++)
      {
        if(authenticatedHost[scanUsers])
        {
          countTunnelsActive++;
          if(controlPrivilege[scanUsers] != '0')
            countControllersLoggedIn++;
        }
	      if((groupNumber >= 0) &&
	         (breakoutGroup[scanUsers] != groupNumber))
	        continue;
        if(!authenticatedHost[scanUsers] ||
           (!connectedByTCP[scanUsers] && (uRecvTime[scanUsers] != 0)))
	      continue;
	      String copyUsername = username[scanUsers];
		    if(copyUsername.trim().length() > 0)
		    {
           
          // pad username out to 8 characters and add to buffer
	        System.arraycopy(padBlanks(copyUsername,8).getBytes(), 0,
                           usernameMessage, 1+8*copyIndex,
                           8);
	        copyIndex++;
          if((9+8*copyIndex) >= usernameMessage.length)break;
        }
	    }//end for(scanUsers...
       
      // refresh the counts
      setUserCounts(countTunnelsActive,countControllersLoggedIn);
      
      return usernameMessage;
	  
	  }// end synchronized(authenticatedHost)
  }// end buildUsernameMessage()
  
  // test messageLength and kill tunnel if it is corrupt
  // (there is no way to resynchronize it)
  // returns false if they match, true if not
  // if expectLength is negative, its absolute value is length
  // of shortest allowable message
  boolean testMessageLength(int messageLength,
                            int expectedLength,
                            int typeCode,
                            int thisHostIndex)
  {
    if(expectedLength < 0)
	{
	  if(messageLength >= (-expectedLength))
	    return false;
          closeTunnel(thisHostIndex, true,
                "expected messageLength at least " + (-expectedLength) +
                " but  received " + messageLength +
                " for message type " + typeCode +
                " on hostIndex " + thisHostIndex);
	}
    if(messageLength == expectedLength)
	  return false;
    closeTunnel(thisHostIndex, true,
                "expected messageLength " + expectedLength +
                " but  received " + messageLength +
                " for message type " + typeCode +
                " on hostIndex " + thisHostIndex);
    return true;
  }
  
  
  // post internal error message and kill this connection
  // returns true if successful
  boolean internalErrorMessage(int messageCode,int hostIndex)
  {
    if(hostIndex >= 0)
      closeTunnel(hostIndex, true,
                  "FCConnect internal error processing message code:" +
                  messageCode);
    else System.err.println("FCConnect internal error processing message code:" +
                            messageCode);
    return true;
  }
  
  
  // sends invitation to select with list of available appDataZipfiles
  // returns true on exception, false otherwise
  boolean sendZipfileList(int hostIndex)
  {
    // file list up to 60 files max 24 chars delimited by "\"
    byte[] zipFileListMessage;
    File checkFiles[], zipFiles[];
    File zipfilePath = new File(appFilePath);
    checkFiles = zipfilePath.listFiles();

    // extract .zip files
    if(checkFiles == null)return false;
    zipFiles = new File[checkFiles.length];
    String strFiles = "";
    int checkListNumber, zipListNumber = 0;
    for(checkListNumber = 0;
        checkListNumber < checkFiles.length;
        checkListNumber++)
    {
	  // accept only zipfiles
      String testFileName = checkFiles[checkListNumber].getName();
      if((testFileName.length() > 24) || !testFileName.endsWith(".zip"))
        continue;
      zipFiles[zipListNumber++] = checkFiles[checkListNumber];
      if(zipListNumber >= 60)break;
    }
    // sort the list using bubble sort
    for(int highEnd = zipListNumber;
        highEnd > 0;
        --highEnd)
    {
      for(int bubbleUp = 1;
          bubbleUp < highEnd;
          ++bubbleUp)
        if(zipFiles[bubbleUp-1].getName().
           compareTo(zipFiles[bubbleUp].getName())>0)
        {
          File swap = zipFiles[bubbleUp-1];
          zipFiles[bubbleUp-1] = zipFiles[bubbleUp];
          zipFiles[bubbleUp] = swap;
        }
      }// end for(int bubbleUp = 1;

      // move the filenames to a string, separated by /
      for(checkListNumber = 0;
          checkListNumber < zipListNumber;
          checkListNumber++)
      {
         String testFileName = zipFiles[checkListNumber].getName();
         if(strFiles.length() == 0)strFiles = testFileName;
         else strFiles = strFiles + "/" + testFileName;
      }

      // send the list in a message to FC, starting with invite code
      int liststringLength = strFiles.length();
      byte[] inviteMessage = new byte[liststringLength+2];
      inviteMessage[0] = 1;
      inviteMessage[1] = 0;
      System.arraycopy(strFiles.getBytes(),0,
                       inviteMessage,2,
                       liststringLength);
      return sendToFC(54,liststringLength+2,inviteMessage,hostIndex);

  } // end sendZipfileList
  
  
  // gets value of floorholder; if in 
  // breakout mode, specific to user's group
  // static to force serial use
  static int getFloorHolder(int hostIndex)
  {
    if(hostIndex < 0)return -1;
    if(inBreakoutMode)
	  {
      // in breakout "PA" mode breakoutController holds floor
      if(floorHolder == breakoutController)return floorHolder;
	    if(breakoutGroup[hostIndex] < 0)return -1;
      return breakoutFloorHolder[breakoutGroup[hostIndex]];
    }
    else
      return floorHolder;
	  
  }// end getFloorHolder()
  
  
  // resets floorHolder status in all FC after reconnect
  void reconnectFloorHolderStatus(int hostIndex)
  {
    // reset auto-tableClearing if pending
    if(fcConnect.runningFiveMinutes)
      runAfterFiveMinutes.interrupt();  
    
    // restore the user to hashtable activeUsers
    fcConnect.addUser(hostIndex,username[hostIndex]);
         
    // on disconnect server should have dropped floorHolder;
    // if somebody else took it, make sure reconnected user's
    // floor control is informed; otherwise just set floorholder 
    // back to reconnected user
    if(inBreakoutMode)
    {
       // this does not deal with breakout floorholders debugx
       if(refreshFloorControlStatus(hostIndex))return;
       updateGridBreakoutStatus();
    }
    else // whole-group mode
    {
      if(refreshFloorControlStatus(hostIndex))return;
      // v5.1.0 had bad bug makes client shutdown when they receive this: 
	    sendNonbreakoutUsernameMessage();
          
      // this host didn't have the floor when disconnected - set "nobody" - needs fixed
 //debugx  if(hostIndex != previousFloorHolder)
 //     {
 //       setFloorHolder(-1,nobodyUsername,hostIndex,-1);	  
 //       return;
 //     }
    }
/* disabled 8-31-12 JMP debugx 
    // floor was not taken during disconnect - give it back to this user
    setFloorHolderName(hostIndex,username[hostIndex]);
    if(inBreakoutMode)
    {
      if(floorHolder >= 0)
        setFloorHolder(hostIndex,getFloorHolderName(hostIndex),floorHolder,breakoutGroup[floorHolder]);
      else
         setFloorHolder(hostIndex,getFloorHolderName(hostIndex),floorHolder,-1);
    }
    else
      setFloorHolder(hostIndex,getFloorHolderName(hostIndex),floorHolder,-1);
//8-31-12 */
  }//end reconnectFloorHolder()
  
  
  // sends a username message to update all cells in the grid
  void sendNonbreakoutUsernameMessage()
  {
	  byte[] usernameMessage = buildUsernameMessage(-1);
	  fcConnect.writeMessageToAll(29,
                                usernameMessage.length,
                                usernameMessage,
					                      false);
                                
  }// end sendNonbreakoutUsernameMessage()
  
  
  // close all connections five minutes after last disconnect
  void closeAllConnectionsAfterFiveMinutes(String comment)
  {
    synchronized(fcConnect.runningFiveMinutes)
	  {
      if(fcConnect.runningFiveMinutes)return;
      closeAfterFiveMinutes = new CloseAfterFiveMinutes(comment);
	    runAfterFiveMinutes = new Thread(closeAfterFiveMinutes);
	    runAfterFiveMinutes.start();
	  }//end synchronized()
  }// end closeAllConnectionsAfterFiveMinutes()
  
  // invokes closeAllConnections after five minutes if not interrupted
  class CloseAfterFiveMinutes implements Runnable
  {
    String reason;
	
    public CloseAfterFiveMinutes(String reasonParm)
	  {
	    reason = reasonParm;
    }
  
    public void run()
    {
      fcConnect.runningFiveMinutes = true;
      try
      {
        Thread.sleep(reconnectTimeoutMillisec);// five minutes
      }
      catch (InterruptedException ie)
      {
        fcConnect.runningFiveMinutes = false;
        return;
      }
      closeAllConnections(reason);
      fcConnect.runningFiveMinutes = false;
      	  
    }// end run()
	  
  }// end class CloseAfterFiveMinutes
  
  // returns a String right-passed with blanks to paddedSize
  static String padBlanks(String sourceString,int paddedSize)
  {
    // if it's already over target size, truncate
    if(sourceString.length() > paddedSize)
	  return sourceString.substring(paddedSize);
	  
	// make a new string and tack on blanks to achieve target
	String paddedString = new String(sourceString);
	while(paddedString.length() < paddedSize)
	  paddedString += ' ';
	return paddedString;
  }
  
  // returns an "unsigned byte" in range 0 to 255
  byte unsignedByte(int sourceInt)
  {
    if(sourceInt > 255)return 11;
	if(sourceInt < 128)return (byte)sourceInt;
	return (byte)(256 - sourceInt);
  }
  
  
  // adds to appDataZipFile; creating the file if necessary
  // when file is complete, indicated by new segment length 0,
  // starts buffering process
  // returns true if any problem, false otherwise
  File uploadFile = null;
  FileOutputStream uploadFileStream;
  String uploadFilePathName, oldFilePathName;
  int copyCount;
  boolean writeAppDataZipFile(int bufferLength, byte[] buffer, int thisHostIndex)
  {
    try
    {
      // if the file doesn't exist, start it
      // (already cleared that is should be written)
      if(uploadFile == null)
      {
        // setup filenames
        uploadFilePathName = 
          appFilePath + moodleFileName + "/" + moodleFileName + "^^^^.zip";
        oldFilePathName = 
          appFilePath + moodleFileName + "/" + moodleFileName + ".zip";
          
        // if a previous upload is left over, delete it
        File deleteFile = new File(uploadFilePathName);
        if(deleteFile.exists())
          if(!deleteFile.delete())
          {
            System.err.println("*** appDataZipFile upload unable to delete previous file:" +
                               uploadFilePathName);
            return true;
          }
          
        // if loadDownloadBuffer is buffering, wait (this is on main
        // FCConnect receive loop so we hope it doesn't take long!)
        int waitCount = 0;
        copyCount = 0; 
        try
        {
          while(downloadBuffering && (waitCount++ < 50))
            Thread.sleep(100);
        }
        catch(InterruptedException ie)
        {
          appFileResponse[0] = 0;
          uploadFile = null;
          return false;        
        }
        if(waitCount >= 50)
        {
          System.err.println("*** downloadBuffering timeout for file:" +
                             uploadFilePathName);
          appFileResponse[0] = 0;
          uploadFile = null;
          return false;
        }
        
        // setup new file to write
        uploadFile = new File(uploadFilePathName);
        uploadFileStream = new FileOutputStream(uploadFile);
        System.out.println("\nuploading appDataZipFile:" + uploadFilePathName +
                           " from host:" + thisHostIndex);
      }
      
      // segment length zero indicates file is complete
      if(bufferLength == 0)
      {
        uploadFileStream.close();
        uploadFile = new File(uploadFilePathName);
        
        // delete the previous appDataZipFile
        File oldFile = new File(oldFilePathName);
        if(oldFile.exists())
          if(!oldFile.delete())
          {
            System.err.println("*** appDataZipFile upload unable to delete previous:" + 
                               oldFilePathName);
            return true;
          }
          
        // rename new file to old name and set its Mtime
        if(!uploadFile.renameTo(new File(oldFilePathName)))
        {
          System.err.println("*** appDataZipFile upload unable to rename:" + 
                             uploadFilePathName + " to:" + oldFilePathName );
          return true;
        }
        if(!(new File(oldFilePathName)).setLastModified(thisAppFileMtime))
        {
          System.err.println("*** appDataZipFile upload unable to setLastModified:" + 
                             oldFilePathName + " to:" + thisAppFileMtime);
          return true;
        }        
        appFileResponse[0] = 0;
        uploadFile = null;
        System.out.println("\nuploaded appDataFile " + copyCount +
                           " bytes:" + oldFilePathName + 
                           " from host:" + thisHostIndex);
                           
        // send status to system messages window on FC
        byte[] finishedMessage = ("finished upload " + moodleFileName).getBytes();
        fcConnect.sendToFC(49,
                           finishedMessage.length,
                           finishedMessage,
                           thisHostIndex);
        
        // start a thread to buffer the download file
        ReBuffer reBuffer = new ReBuffer();
        reBuffer.start();
        return false;
        
      }// end if(bufferLength == 0)
      
      // write segment to file
      uploadFileStream.write(buffer,0,bufferLength);
      copyCount += bufferLength;
      return false;
    }
    catch(IOException ioe)
    {
      System.err.println("*** IOException writing appDataZipFile:" + 
                         uploadFilePathName + " from host:" + thisHostIndex);
      appFileResponse[0] = 0;
      uploadFile = null;
      return true;
    }
    
  }// end writeAppDataZipFile()
  
  
} // end class FCConnect
