/*
 * FCListen.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author mpullen & students
 * @version 4.5.1
 & @modified JMP 12/7/00
 */
/*----------------------------------------------------------------*
|   Copyright 2001-2009 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/

/*******************************************************************************
FCListen.java class logic:
-> read one message: the authentication data handle, key and implicitly host address
-> if the host address is present in hostAddr[]
->   check the handle for match
->     if there is a key for that host it must match also
-> else if an address with value one and matching handle is present in hostAddr[]
->   also check if the entry has a key it must match too
-> if either of the above conditions holds, set "authenticated" bit for the host
-> send one-byte message to connecting client: 1 for authenticated, 0 for failed
-> if failed, disconnect
-> loop forever
->   listen for additional messages from TLMFC and respond on a case basis
-> end loop forever
The time of authentication is stored in the uRecvTime value for the host; when sending
UDP data or accepting a TCP connection, if the time since authentication is over w
time units, the action is not taken.
******************************************************************************/
//package TLM;
import java.io.*;
import java.util.*;
import java.net.*;

public class FCListen extends TLM implements Runnable
{
  // constants
  static int retryTimes = 60;       // times to try reading connection
                                    // at 1000 ms/time, gives 60 seconds/login

  // network variables
  private InetAddress remoteHost;
  private Socket thisFcSocket;
  int thisHostIndex = -1;

  BufferedInputStream in; //PM
  BufferedOutputStream out; //PM

  private byte[] flrStatus; // floor status code 1=automatic; 2=never; 3=ask me
  private byte[] noFloor;
  private byte[] initMessage;
  private byte[] floorMessage;
  private byte[] floorRuleMessage;
  private byte[] floorNotAvailableMessage;
  private byte[] newParticipantMessage;
  private byte[] dropParticipantMessage;
  private String floor_status, offeredSessionVariable;
  private int offeredTlmcDataPort;
  private String offeredUsername = "";
  private String offeredPassword = "";
  private String offeredCourseID = "";
  private boolean nameExists = false;      // if exists gets set true
  private int addUserReturnValue = 0; // if unsuccessful gets set to 1 or 2
  private byte[] readUsername, loginSessionVariable;
  private byte thisIsTheController = 0;

  // constructor
  public FCListen(Socket fcSocketParm)
  {
    floor_status = new String();
    thisFcSocket = fcSocketParm;
    try
    {
      out = new BufferedOutputStream(thisFcSocket.getOutputStream());
      out.flush();
      in = new BufferedInputStream(thisFcSocket.getInputStream());
    }
    catch(Exception e)
    {
      System.out.println(e);
      e.printStackTrace(System.out);
    }
  }

  public void run()
  {
    // arrive here with a new thisFcSocket opened and ready to read;
    // to avoid a possible denial of service, look ahead to
    // see if authentication message from the client is available;
    // allow it retryTimes increments of 1000 ms to arrive, then
    // disconnect at server end
    int readTries, readIn, transactionType;
    boolean foundMatch = false;
    byte[] readBuffer = new byte[ MAX_MESSAGE ], sendClientMessage;
    byte[] name = new byte[8];
    byte[] passwd = new byte[8];
    byte[] courseID = new byte[11];
    byte[] semester = new byte[3];
    int localFloorStatus = initialFloorStatus;
    remoteHost = thisFcSocket.getInetAddress();

    // if this is loopback address (running client on server machine)
    // convert it to local address
    if(remoteHost.equals(loopbackAddress))
      remoteHost = localAddress;

    // treat all Exceptions alike: lost connection
    try
    {
      // read the session variable sent first through the connection
      if((transactionType = receiveFromFC()) != 11)
      {
        System.out.println("invalid FC transaction code for " + remoteHost);
        thisFcSocket.close();
        return;
      }
      if(messageLengthFromFC() != 18)
      {
        internalErrorMessage(11);
        return;
      }
      byte[] loginSessionData = new byte[18];
      if(getBytesFromFC(11,18,loginSessionData))return;
      offeredSessionVariable = new String(loginSessionData,0,16);
      offeredTlmcDataPort = bytesToUshort(loginSessionData[16],loginSessionData[17]);

      // lookup the host table index; try for up to 20 seconds
      // in case the TLMC TCP connection is delayed
      int testTimer = 0;
      while(( testTimer++ < 40) &&
            ((thisHostIndex =
              lookupPendingConnection(remoteHost,
                                      offeredSessionVariable,
                                      offeredTlmcDataPort))
             < 0))
        try{Thread.sleep(500);}catch(InterruptedException ie){}
      if(thisHostIndex < 0)
      {
        thisFcSocket.close();
        System.out.println("FC nomatch disconnect for " + remoteHost);
        return;
      }
      fcOutStreams[thisHostIndex] = out;
      fcInStreams[thisHostIndex] = in;
      fcSockets[thisHostIndex] = thisFcSocket;

      // loop allowing attempts to authenticate
      // (available() is not compatible with SSL)
      while(!stopTLM) // accept authentication loop
      {
        if(!useSSL)
        {
          for(readTries = 0; readTries < retryTimes; ++readTries)
          {
            if(in.available() >= 33)break;
            try{Thread.sleep(1000);}catch( InterruptedException e ){}
          } // endfor(readTries...

          // did read loop timeout?
          if(readTries >= retryTimes)
          {
            // lookup the host table index and use it to disconnect
            closeTunnel(thisHostIndex, false,
                        "FCListen timed out reading authentication from client");
            return;
          } // end if(readTries...
        } // end if(!useSSL)

        // At this point we know (or, with SSL, hope)
        // there is enough data available to authenticate.
        if((transactionType = receiveFromFC()) != 0)
        {
          closeTunnel(thisHostIndex, true,
                      "Invalid connect transactionCode:" +
                      transactionType + " received from host");
          return;
        }
        if(messageLengthFromFC()!= 30)
        {
          closeTunnel(thisHostIndex, true,
                      "Invalid messageLength " + messageLengthFromFC() + " received from host");
          return;
        }
        if(getBytesFromFC(0,8,name) ||
           getBytesFromFC(0,8,passwd) ||
           getBytesFromFC(0,11,courseID) ||

           // semester not used in TLM - included for compatibility
           getBytesFromFC(0,3,semester))
        {
          internalErrorMessage(0);
          return;
        }
        offeredUsername = new String(name);
        offeredPassword = new String(passwd);
        offeredCourseID = (new String(courseID)).toUpperCase();

        // Check for allowAllUsers case (anonymousHost = true)- user
        // does not need to authenticate, only to provide unique username
        if( anonymousHost[ thisHostIndex ] )
        {
          // username is unique if its not in the database/file for this course
          if( TLM.auth.nameAvailable( offeredUsername, offeredCourseID ) )
          {
            // and it is not already logged in
            nameExists = (fcConnect.activeUsers).containsKey(offeredUsername) &&
                         !userIsConnected(offeredUsername);

            // accept the login if the offered username
            // is not already taken and force drawGrid to true
            // since we can't look it up in the database/file
            if (!nameExists)
            {
              foundMatch = true;
              username[thisHostIndex] = offeredUsername;
              drawGrid[thisHostIndex] = true;
            }
          } // end if( TLM.authnameAvailable...
        } // end if( anonymousHost[ thisHostIndex ] )

        // if authentication is required, check the quadruple
        // {username, password, IPaddress, courseID} to authenticate
        else
        {
          // wait two seconds to let TCPTunnel or UDPTunnel get setup
          try{Thread.sleep(2000);}catch(InterruptedException ie){}

          // authenticate via database or .auth file
          loginData loginResult = null;
          loginResult =
            auth.authenticate(thisHostIndex,
                              offeredUsername,
                              offeredPassword,
                              hostAddr[thisHostIndex],
                              sourcePort[thisHostIndex],
                              offeredCourseID,
                              offeredSessionVariable);
          if(loginResult != null)
          {
            offeredUsername = loginResult.username;
            offeredCourseID = loginResult.courseID;
            userCourseID[thisHostIndex] = offeredCourseID.trim();
            foundMatch = true;
            nameExists = (fcConnect.activeUsers).containsKey(offeredUsername) &&
                         userIsConnected(offeredUsername);
          }
          else
          {
            nameExists = false;
            if(webAuthenticate)
            {
              closeTunnel(thisHostIndex, false, "Web authentication failed");
              return;
            }
          }
          // verify courseID
          if(foundMatch)
          {
            // if no class in schedule or privileged user, don't verify
            if(autoSchedule && classInSession &&
              (instructorPrivilege[thisHostIndex] < 2))
            {
              // check that courseID matches a course in nextCourseID[]
              int test;
              for (test = 0; test < numberOfCourseIDAliases; ++test)
                if (nextCourseID[test].trim().equals(offeredCourseID.trim()))
                  break;
              if (test >= numberOfCourseIDAliases)foundMatch = false;
            } // end if(autoSchedule...
          } // end if(foundMatch)
        } // end else/if( anonymousHost[ thisHostIndex ] )

        // name duplicates are not allowed
        if(nameExists)
        {
          if(fcConnect.sendToFC(58,0,null,thisHostIndex))return;// username already taken
          closeTunnel(thisHostIndex, false, "FCListen found username already taken");
          return;
        }
        if(foundMatch)break;

        // the offered name/password does not authenticate
        if(++numAuthenticateTries[thisHostIndex] >= authenticateTimes)
        {
          if(fcConnect.sendToFC(61,0,null,thisHostIndex))return;// breaking connection
          numAuthenticateTries[thisHostIndex] = 0;
          System.out.println("host# " + thisHostIndex
            + " failed three times to authenticate.");
          closeTunnel( thisHostIndex, false,
            "FCListen client failed to authenticate three times");
          return;
        } // end if(numAuthenticateTries...
        else
        {
          if(fcConnect.sendToFC(32,0,null,thisHostIndex))return; // not accepted
        } // end else/if(numAuthenticateTries...
      } // end while(!stopTLM) // accept authentication loop

      // user has authenticated; verify conditions for connection
      // is there is a TCP connection or UDP flow to match?
      int connectTries;

      // watch the connection indicators for 10 seconds
      for(connectTries=0; connectTries<10; ++connectTries)
      {
        if (tcpIsOpen[thisHostIndex] || (uRecvTime[thisHostIndex] != 0))
          break;
        try{Thread.sleep(1000);}catch(InterruptedException ie){}
      } // int for(int connectTries...
      if(connectTries >= 10)
      {
        if(fcConnect.sendToFC(57,0,null,thisHostIndex))return; // TLMC not connected
        closeTunnel(thisHostIndex, false,
                    "FCListen found no data connection after authentication");
        return;
      } // end if(connectTries >= 10)
    } //end try
    catch(IOException e)
    {
      // lost connection
      closeTunnel( thisHostIndex, true,
                   "Exception in FCListen during authentication:" + e );
      return;
    } // end catch

    // we continue to treat all Exceptions as lost connection, but after this
    // point upon exception we must clear out user state by dropParticipant()
    try
    {
      // Add the user to activeUsers list in FCConnect.
      // addUser returns 0 if user successfully added
      //                 1 if no more room
      //                 2 if username already exists
      addUserReturnValue = fcConnect.addUser(thisHostIndex,offeredUsername);
      nameExistsAgain[thisHostIndex] = (addUserReturnValue != 0);
      if(nameExistsAgain[thisHostIndex]) // => user is not authenticated
      {
        if(addUserReturnValue == 1) // table full - do not accept another
        {
          if(fcConnect.sendToFC(59,0,null,thisHostIndex))return;// server full
          closeTunnel(thisHostIndex, false, "User:" + offeredUsername +
            " rejected after valid login because server is full");
        }
        else // addUserReturnValue == 2  means the offered name is
             // already in the system - do not accept again
        {
          if(fcConnect.sendToFC(58,0,null,thisHostIndex))return;// already logged in
          closeTunnel(thisHostIndex, false, "User:" + offeredUsername +
            " rejected because that name is logged in or reserved");
        }
        return;
      } // end if(nameExistsAgain[thisHostIndex])

      // set timers to watch for disconnect
      fRecvTime[ thisHostIndex ] = System.currentTimeMillis();
      uRecvTime[ thisHostIndex ] =
      fRecvTime[ thisHostIndex ] + 2*tlmcDisconnectTime;

      // set authenticated status and save thread info
	  synchronized(authenticatedHost)
	  {
        authenticatedHost[ thisHostIndex ] = true;
      }
      
      // all setup completed; send user accepted setup
      if(fcConnect.sendFcInitMessage(thisHostIndex))return;

      // once the user is authenticated, continuously read messages from client
      // and process them as long as TLMFC stays connected
      while(foundMatch & !stopTLM)
      {
        // read a message from client, using 'available' to sense
        // lost connection at one-second interval
        if(!useSSL)
        {
          try
          {
            while (in.available() < 3)
              Thread.sleep(1000);
          }
          catch (InterruptedException ie) {}
          if (in == null)
          {
            // lost connection
            closeTunnel(thisHostIndex, true, "FCListen lost input stream from client");
            return;
          }
        }
        if((transactionType = receiveFromFC()) == -1)
        {
          // lost connection
          closeTunnel(thisHostIndex, true, "FCListen disconnected from client");
          return;
        }
        int messageLength = messageLengthFromFC();
        if(messageLength < 0)
        {
          testMessageLength(messageLength,0,0);
          return;
        }
        byte[] fcMessage = new byte[messageLength];
        getBytesFromFC(transactionType,messageLength,fcMessage);
        fcConnect.processFCMessage(transactionType,
                                   fcMessage,
                                   messageLength,
                                   thisHostIndex);
      } //end while(foundMatch & !stopTLM)
    } // end try
    catch(IOException ioe)
    {
      // lost connection
      closeTunnel(thisHostIndex, true, "FCListen disconnected in main loop");
    }
    catch(Exception e)
    {
      // lost connection
      closeTunnel(thisHostIndex, true, "FCListen Exception in main loop:" +
                  e.getMessage());
      e.printStackTrace(System.out);
      return;
    }
  } // end run

  // receive a message from Floor Control client
  // buffers the result for access by getBytesFromFC and getByteFromFC
  // returns messageType received or -1 if IOException
  byte[] receiveBuffer;
  int receiveIndex = -1, receiveLength = -1, receiveCode = -1;
  int receiveFromFC()
  {
    try
    {
      // no input available, reset and bail out
      if ( (receiveCode = in.read()) == -1)
      {
        receiveIndex = -1;
        receiveLength = -1;
        if(thisHostIndex >= 0)
          closeTunnel(thisHostIndex, true, "FCListen disconnected");
        return -1;
      }
      // read the length field
      int lengthByte1, lengthByte2;
      lengthByte1 = in.read();
      lengthByte2 = in.read();
      if((lengthByte1 == -1) || (lengthByte2 == -1))
      {
        receiveIndex = -1;
        receiveLength = -1;
        if(thisHostIndex >= 0)
          closeTunnel(thisHostIndex, true, "FCListen disconnected");
        return -1;
      }
      receiveLength = bytesToUshort((byte)lengthByte1,(byte)lengthByte2);
      receiveIndex = 0;
      if(receiveLength == 0)return receiveCode;

      // read and buffer the data
      receiveBuffer = new byte[receiveLength];
      if(in.read(receiveBuffer,0,receiveLength) == -1)
      {
        receiveIndex = -1;
        receiveLength = -1;
        closeTunnel(thisHostIndex, true, "FCListen disconnected in main loop");
        return -1;
      }
    }
    catch(IOException ioe)
    {
      receiveIndex = -1;
      receiveLength = -1;
      receiveCode = -1;
      closeTunnel(thisHostIndex, true,
                  "FCListen disconnected in main loop:" + ioe.getMessage());
      return -1;
    }
    return receiveCode;
  } // end receiveFromFC

  // returns length of buffer read in latest receiveTLM
  int messageLengthFromFC()
  {
    return receiveLength;
  }

  // moves numberOfBytes from latest message from TLM to target array
  // confirms expected transactionCode matches the one read unless it is -1
  // returns true if error (no data, bad code, etc) false otherwise
  boolean getBytesFromFC(int transactionCode,
                         int numberOfBytes,
                         byte[] target)
  {
    if(transactionCode != receiveCode)
      if(transactionCode != -1)
        return internalErrorMessage(transactionCode);

    // error if buffer was not loaded
    if(receiveIndex < 0)
      return internalErrorMessage(transactionCode);

    // error trying to read more than buffe rholds
    if((receiveIndex+numberOfBytes) > receiveLength)
      return internalErrorMessage(transactionCode);

    // move requested data to return array
    System.arraycopy(receiveBuffer,receiveIndex,target,0,numberOfBytes);
    receiveIndex += numberOfBytes;
    return false;
  }
  // same as above but returns int value of a
  // single byte from buffer or -1 for error
  int getByteFromFC(int transactionCode)
  {
    int numberOfBytes = 1;
    if(transactionCode != receiveCode)
    {
      internalErrorMessage(transactionCode);
      return -1;
    }
    if(receiveIndex < 0)
    {
      internalErrorMessage(transactionCode);
      return -1;
    }
    if((receiveIndex+numberOfBytes) > receiveLength)
    {
      internalErrorMessage(transactionCode);
      return -1;
    }
    return (int)receiveBuffer[receiveIndex++];
  }// end getByteFromTLM

  boolean internalErrorMessage(int messageCode)
  {
    if(thisHostIndex >= 0)
      closeTunnel(thisHostIndex, true,
                  "FCListen internal error processing message code:" +
                  messageCode);
    else System.err.println("FCListen internal error processing message code:" +
                            messageCode);
    return true;
  }

  // test messageLength and kill tunnel if it is corrupt
  // (there is no way to resynchronize it)
  // returns false if they match, true if not
  boolean testMessageLength(int messageLength,int expectedLength,int typeCode)
  {
    if(messageLength == expectedLength)return false;
    closeTunnel(thisHostIndex, true,
                "expected messageLength " + expectedLength +
                " but  received " + messageLength +
                " for message type " + typeCode +
                " on hostIndex " + thisHostIndex);
    return true;
  }
} // end TLM.FCListen class
