/*----------------------------------------------------------------*
|    Copyright 2001-2008 Networking and Simulation Laboratory     |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

// Network EducationWare recording resequence launch GUI
// by Rama Venna 8-20-03
// revised by JMP 9-10-03
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;


/*******************************************************************************
GUI FOR NEW RESEQ

This GUI provides parameters to RESEQ, which reads a NEW class recording and
updates page numbers as they would have been created if the inputs were from one
continuous session of WBD, starting at 1. It handles multiple input files in the
same run, and  also will segment the output file, and will place a cover slide
with copyright statement and an optional audio clip on each segment, if copyright
file is present. The segments start at the next new slide after the segmentation
count is exceeded. (As a consequence, a purely audio recording cannot be
segmented.)

Input Parameters Required to run the Program:
-OutputFile   : full path of output file (or filename in execution directory)
-Messages/Seg : (optional) number of records per segment (if 0, file is not to
                be segmented) (default 20000)
-CoverSlide   : full path of cover slide information file (or filename in
                execution directory);
-CoverAudio   : full path of cover audio information file (or filename in
                execution directory);
-OmitVideo    : (optional) passed to RESEQ as 1 if block is checked
-InputFiles   : full path of input files (or filenames in execution directory).
                At least one input file must be entered. Max 9 may be entered.

The user can use the Browse button to select the files from the local machine,
then click on the RunResequence button to perform the segmentation.
NOTE: According to Sun's documentation, the FilenameFilter function used here
does not work on Sun's SDK for Windows. Thus all files will be displayed.

Optional command-line arguments to invoke GUI at convenient starting point
for browsing (all optional, all positional):
1. starting outputPath (0 for execution path)
2. starting coverSlidePath (0 for execution path)
3. starting inputPath (0 for execution path)
4. starting messages per segment (positive integer)
5. starting omit video (0 or 1) (default is 0)
6 to 11. filenames to load in GUI area

NOTE: audio cover file is assumed to be in the same directory as the
cover slide file that is selected by browsing, and have name Welcome.mwr.

See comments in CoverSlide for format of the optional input file.

/******************************************************************************/
public class GUI extends Frame implements ActionListener
{
  static String[] inputFiles = new String[9];
  static int numberOfInputFiles = 1;

  public static void main(String[] args)
  {
    // test and set initial parameters
    int argsLength = args.length;
    if(argsLength > 3)argsLength = 3;
    for(int index=0; index<argsLength; ++index)
      if(!args[index].equals("0"))
        if(!(new File(args[index])).isDirectory())
          System.out.println("GUI found invalid directory on command line");
    if(args.length > 0)
      if(!args[0].equals("0"))startingOutputPath = args[0];
    if(args.length > 1)
      if(!args[1].equals("0"))startingCoverSlidePath = args[1];
    if(args.length > 2)
      if(!args[2].equals("0"))startingInputPath = args[2];
    if(args.length > 3)startingMessagesPerSegment = args[3];
    runMessagesPerSegment = startingMessagesPerSegment;
    if(args.length > 4)startingOmitVideo = args[4];
    runOmitVideo = startingOmitVideo;

    // move input files to array for use in GUI
    for(; numberOfInputFiles < 10; ++numberOfInputFiles)
    {
      if(args.length <= (numberOfInputFiles+4))break;
      inputFiles[numberOfInputFiles-1] = args[numberOfInputFiles+4];
    }
    numberOfInputFiles--;

    // start the program
    GUI reseqWindow = new GUI();
    RESEQ reseqCommand = new RESEQ();
    reseqWindow.inAnApplet = false;
    reseqWindow.setVisible(true);
    reseqWindow.setTitle("NEW");
  }
  // components
  TextField outputText,messageNumberText,coverSlidePathNameText,videoFlagText;
  TextArea inputText;
  Label title,outputFile,messageCount,videoFlag,coverSlideName,inputFile;
  Button browseOutput,browseCoverSlide,browseInput,runReseq,quitReseq;
  TextArea outputTextArea = new TextArea("Output",10,10,TextArea.SCROLLBARS_BOTH);
  String browseOutputPath="";
  String[] reseqArgs, inputPathArgs = new String[9];
  static String startingOutputPath="", startingCoverSlidePath="";
  static String startingInputPath="";
  static String startingMessagesPerSegment = "20000", runMessagesPerSegment;
  static String startingOmitVideo = "0", runOmitVideo;
  Frame fileFrameNwr, fileFrameNwt;
  FileDialog fileDialogNwr, fileDialogNwt;

  boolean omitVideo = false, dataHasBeenModified = false;

  Font f1 = new Font("Serif",Font.BOLD,20);
  Font f2 = new Font("Serif",Font.BOLD,14);
  Font f3 = new Font("Serif",Font.PLAIN,14);
  boolean inAnApplet = true;
  int clickCount = 0;

  // constructor
  public GUI()
  {
    setBackground(Color.lightGray);
    setSize(700,400);
    setLocation(320,0);

    //component initialization
    title  = new Label("Network EducationWare Recording Resequence",Label.CENTER);
    outputFile = new Label("OutputFile:",Label.LEFT);
    messageCount = new Label("Messages/Segment:",Label.LEFT);
    videoFlag = new Label("Omit video:",Label.RIGHT);
    coverSlideName   = new Label("CoverSlide:",Label.LEFT);
    inputFile  = new Label("InputFiles:",Label.LEFT);
    browseOutput = new Button("Browse");
    browseOutput.setActionCommand("Output");
    browseCoverSlide = new Button("Browse");
    browseCoverSlide.setActionCommand("CoverSlide");
    browseInput = new Button("Browse");
    browseInput.setActionCommand("Input");
    runReseq = new Button("Run Resequence");
    quitReseq = new Button("Quit");
    runReseq.addActionListener(this);
    quitReseq.addActionListener(this);
    browseInput.addActionListener(this);
    browseOutput.addActionListener(this);
    browseCoverSlide.addActionListener(this);
    fileFrameNwr = new Frame("NEW RESEQ");
    fileDialogNwr = new FileDialog(fileFrameNwr,"File Dialog");
    fileFilterAccept nwrFilter = new fileFilterAccept(".nwr");
    fileDialogNwr.setFilenameFilter(nwrFilter);
    fileFrameNwt = new Frame("NEW RESEQ");
    fileDialogNwt = new FileDialog(fileFrameNwt,"File Dialog");
    fileFilterAccept nwtFilter = new fileFilterAccept(".nwt");
    fileDialogNwt.setFilenameFilter(nwrFilter);
    outputText = new TextField(60);
    messageNumberText = new TextField(startingMessagesPerSegment,20);
    videoFlagText = new TextField(startingOmitVideo,1);
    coverSlidePathNameText = new TextField(60);
    inputText= new TextArea(10,61);

    // move the initial input files to inputText
    for(int i=0; i<numberOfInputFiles; ++i)
      inputText.append(inputFiles[i]+"\n");

    //add the components to the frame at specified constraints
    setConstraints();

    // add a listener to shutdown gracefully
    addWindowListener
    (
      new WindowAdapter()
      {
        public void windowClosing(WindowEvent we)
        {
          if(inAnApplet)
          {
            dispose();
          }
          else
          {
            System.exit(0);
          }
        }
      }
    );
  } // end GUI constructor

  /**
   * takes action on click of control buttons
   */
  public void actionPerformed(ActionEvent ae)
  {
    String pathName = "";
    Frame outputDialogFrame = new Frame("");

    // processing Browse button action
    if(((Button)ae.getSource()).getLabel().equals("Browse"))
    {
      if(ae.getActionCommand().equals("Input"))
      {
        // actions if the input text area is empty
        if   (inputText.getText().equals("")
          || (clickCount == 0)
          || inputText.getText().equals("\n"))
        {
          if(inputText.getText().equals("\n"))inputText.setText("");
          if(startingInputPath.length()!=0)
            fileDialogNwr.setDirectory(startingInputPath);
          fileDialogNwr.setVisible(true);
          pathName = fileDialogNwr.getDirectory() + fileDialogNwr.getFile();
          if(!pathName.endsWith("null"))
          {
            inputText.append(pathName+"\n");
            dataHasBeenModified = true;
          }
          if(clickCount == 1)clickCount = 0;
        }
        else
        {
          errorMessage("please clear the InputFile Area including white space " +
            "if you want to enter only new files; otherwise continue");
        }
      }// end if(ae.getActionCommand().equals("Input")

      // action when output file is entered
      if(ae.getActionCommand().equals("Output"))
      {
        if(startingOutputPath.length()!=0)
          fileDialogNwr.setDirectory(startingOutputPath);
        fileDialogNwr.setVisible(true);
        pathName = fileDialogNwr.getDirectory() + fileDialogNwr.getFile();
        if(!pathName.endsWith("null"))
        {
          outputText.setText(pathName);
          dataHasBeenModified = true;
        }
      }// end getActionCommand for Output

      // action when CoverSlide file is entered
      if(ae.getActionCommand().equals("CoverSlide"))
      {
        if(startingCoverSlidePath.length()!=0)
          fileDialogNwt.setDirectory(startingCoverSlidePath);
        fileDialogNwt.setVisible(true);
        pathName = fileDialogNwt.getDirectory() + fileDialogNwt.getFile();
        if(!pathName.endsWith("null"))
        {
          coverSlidePathNameText.setText(pathName);
          dataHasBeenModified = true;
        }
      }// end getActionCommand for CoverSlide
    }// end Browse Action Command

    // processing Quit action command
    if(ae.getActionCommand().equals("Quit"))
    {
      if(dataHasBeenModified ||
         !messageNumberText.getText().equals(runMessagesPerSegment) ||
         !videoFlagText.getText().equals(runOmitVideo))
      {
        if(displayMessage(
          "Resequence not run since last input; quit anyhow?").equals("Yes"))
          System.exit(0);
      }
      else System.exit(0);
    }

    // processing Run Resequence button action
    if(ae.getActionCommand().equals("Run Resequence"))
    {
      // pick up data from fields and validate it
      int i = 0;
      boolean flag = false;
      clickCount++;
      OutputFrame outputFrame =
        new OutputFrame(outputDialogFrame,"A Frame Window");

      // network and file variables
      String outputPath  = outputText.getText();
      String strSegSize  = messageNumberText.getText();
      String coverSlidePathName  = coverSlidePathNameText.getText();
      String inputPathArea = inputText.getText();
      int segmentSize = 0;
      int appToOmit = -1;
      RandomAccessFile testFile;

      // check if any of the four input fields not entered
      if( (outputPath.equals(""))
        ||(strSegSize.equals(""))
        ||(coverSlidePathName.equals(""))
        ||(inputPathArea.equals("")))
      {
        errorMessage("you must provide input in each space");
        clickCount = 0;
        return;
      }
      try
      {
        segmentSize = new Integer(strSegSize).intValue();
      }
      catch(NumberFormatException ne)
      {
        errorMessage("number of records per segment must be a positive integer");
        clickCount = 0;
        return;
      }

      // verify output path/file
      try
      {
        testFile = new RandomAccessFile(outputPath,"r");
        /* output file now is template for name containing datetime
        if(displayMessage("Output file exists; overwrite?").equals("No"))
        {
          clickCount=0;
          errorMessage("please change the output file name");
          return;
        }
        try{testFile.close();}catch(IOException ioe){}
        */
      }
      catch(FileNotFoundException fnfe){}
      if(!outputPath.endsWith(".nwr"))
      {
        errorMessage("output file name must have .nwr extension");
        clickCount = 0;
        return;
      }

      // verify CoverSlide path/file
      if(!coverSlidePathName.endsWith(".nwt"))
      {
        errorMessage("cover slide must have .nwt extension");
        clickCount = 0;
        return;
      }
      try
      {
        File testCover = new File(coverSlidePathName);
        if(!testCover.canRead())
        {
          errorMessage("cover slide contents (.nwt) file not readable");
          clickCount = 0;
          return;
        }
      }
      catch(NullPointerException npe)
      {
        errorMessage("cover slide contents (.nwt) file not readable");
        clickCount = 0;
        return;
      }
      catch(SecurityException se)
      {
        errorMessage("cover slide contents (.nwt) file not readable");
        clickCount = 0;
        return;
      }

      // extract path from coverSlide string and use it to build
      // the coverAudio string
      String coverAudioPath = "./";
      if(!coverSlidePathName.equals("0"))
      {
        // if there is not a \ or /, path remains "./"
        int backSlashIndex = coverSlidePathName.lastIndexOf("\\");
        int slashIndex = coverSlidePathName.lastIndexOf("/");
        if(backSlashIndex > 0)slashIndex = backSlashIndex;
        if(slashIndex > 0)
          coverAudioPath = coverSlidePathName.substring(0,slashIndex+1);
      }

      // loop through all input files in InputFiles Text Area
      StringTokenizer st = new StringTokenizer(inputPathArea,"\n");
      int inputPathNumber = 0;
      while(st.hasMoreTokens())
      {
        if(inputPathNumber > inputPathArgs.length)
        {
          errorMessage("too many input files - max " + inputPathArgs.length);
          clickCount = 0;
          return;
        }
        String inputPath = st.nextToken();
        int start = inputPath.lastIndexOf("\\");

        // test input file for format and presence
        if(!inputPath.endsWith(".nwr"))
        {
          errorMessage("input files must have .nwr extension");
          clickCount = 0;
          return;
        }
        try
        {
          testFile = new RandomAccessFile(inputPath,"r");
        }
        catch(FileNotFoundException fe)
        {
          errorMessage("input file:" + inputPath + " not found"+"\n");
          clickCount = 0;
          return;
        }
        inputPathArgs[inputPathNumber++] = inputPath;
      }// end while(st.hasMoreTokens())

      // load the arguments as they would have come from command line to RESEQ
      reseqArgs = new String[5+inputPathNumber];
      reseqArgs[0] = outputPath;
      reseqArgs[1] = messageNumberText.getText();
      runMessagesPerSegment = reseqArgs[1];
      reseqArgs[2] = coverSlidePathName;
      reseqArgs[3] = coverAudioPath + "Welcome.nwr";
      reseqArgs[4] = videoFlagText.getText();
      runOmitVideo = reseqArgs[4];
      System.arraycopy(inputPathArgs,0,reseqArgs,5,inputPathNumber);

      // run RESEQ - results will display in command window
      RESEQ.main(reseqArgs);
      dataHasBeenModified = false;

    }// end if(ae.getActionCommand("Run Resequence");
  }//end actionPerformed Method

  /**
   * if user enters wrong input data
   */
  public void errorMessage(String message)
  {
    Frame errorDisplay = new Frame("");
    int x = (message.length()*4) + 200;
    errorDisplay.setVisible(false);
    ErrorFrame errorFrame = new ErrorFrame(errorDisplay,"Message",message);
    errorFrame.setBounds(350,400,x,100);
    errorFrame.setVisible(true);
    System.out.println(message);
  }

  /**
   * if user must make a choice yes or no0
   */
  public String displayMessage(String message)
  {
    Frame messageDisplay = new Frame("");
    int x = (message.length()*4) + 200;
    messageDisplay.setVisible(false);
    MessageFrame messageFrame = new MessageFrame(messageDisplay,"AlertMessage",message);
    messageFrame.setBounds(350,400,x,100);
    messageFrame.setVisible(true);
    return messageFrame.YesOrNo;
  }

  /**
   *  set the constraints of the components and add them to the Frame
   */
  public void setConstraints()
  {
    GridBagLayout gridBag = new GridBagLayout();
    setLayout(new GridBagLayout());

    // title bar
    GridBagConstraints c;
    c = new GridBagConstraints();
    c.gridx = 0;c.gridy =0;c.insets = new Insets(5,0,0,0);
    c.gridwidth = 5;c.anchor = GridBagConstraints.SOUTH;
    c.fill = GridBagConstraints.BOTH;
    title.setFont(f1);title.setForeground(Color.blue);
    add(title,c);

    // output file
    c = new GridBagConstraints();
    c.gridx = 1;c.gridy = 2;c.insets = new Insets(5,0,0,0);
    c.anchor = GridBagConstraints.WEST;outputFile.setFont(f2);
    add(outputFile,c);
    c = new GridBagConstraints();
    c.gridx=2;c.gridy=2;c.insets = new Insets(5,0,0,0);
    c.gridwidth = 3;c.anchor = GridBagConstraints.WEST;
    add(outputText,c);
    c = new GridBagConstraints();
    c.gridx=0;c.gridy=2;c.insets = new Insets(5,0,0,5);
    c.anchor = GridBagConstraints.WEST;
    add(browseOutput,c);

    // message count
    c = new GridBagConstraints();
    c.gridx = 1;c.gridy = 3;c.insets = new Insets(5,0,0,0);
    c.anchor = GridBagConstraints.WEST;messageCount.setFont(f2);
    add(messageCount,c);
    c = new GridBagConstraints();
    c.gridx=2;c.gridy=3;c.insets = new Insets(5,0,0,0);
    c.anchor = GridBagConstraints.WEST;
    add(messageNumberText,c);

    // omit video
    c = new GridBagConstraints();
    c.gridx=3;c.gridy=3;c.insets = new Insets(5,0,0,5);
    c.anchor = GridBagConstraints.EAST;videoFlag.setFont(f2);
    add(videoFlag,c);
    c = new GridBagConstraints();
    c.gridx = 4;c.gridy = 3;c.insets = new Insets(0,0,0,5);
    c.anchor = GridBagConstraints.WEST;
    add(videoFlagText,c);

    // cover slide file
    c = new GridBagConstraints();
    c.gridx = 1;c.gridy = 4;c.insets = new Insets(5,0,0,0);
    c.anchor = GridBagConstraints.WEST;coverSlideName.setFont(f2);
    add(coverSlideName,c);
    c = new GridBagConstraints();
    c.gridwidth = 3;c.anchor = GridBagConstraints.WEST;
    c.gridx=2;c.gridy=4;c.insets = new Insets(5,0,0,0);
    add(coverSlidePathNameText,c);
    c = new GridBagConstraints();
    c.anchor = GridBagConstraints.WEST;
    c.gridx=0;c.gridy=4;c.insets = new Insets(5,0,0,5);
    add(browseCoverSlide,c);

    // input files
    c = new GridBagConstraints();
    c.gridx = 1;c.gridy = 5;c.insets = new Insets(5,0,0,0);
    c.anchor = GridBagConstraints.WEST;inputFile.setFont(f2);
    add(inputFile,c);
    c = new GridBagConstraints();
    c.gridx=2;c.gridy=5;c.insets = new Insets(5,0,0,0);
    c.gridwidth = 3;c.anchor = GridBagConstraints.WEST;
    add(inputText,c);
    c = new GridBagConstraints();
    c.gridx=0;c.gridy=5;c.insets = new Insets(5,0,0,5);
    c.anchor = GridBagConstraints.WEST;
    add(browseInput,c);

    // run and quit buttons
    c = new GridBagConstraints();
    c.gridx = 2;c.gridy =7;c.gridwidth = 2;
    c.insets = new Insets(5,0,0,0);
    c.anchor = GridBagConstraints.WEST;runReseq.setFont(f2);
    add(runReseq,c);
    c = new GridBagConstraints();
    c.gridx = 4;c.gridy =7;c.gridwidth = 2;
    c.insets = new Insets(5,0,0,5);
    c.anchor = GridBagConstraints.EAST;quitReseq.setFont(f2);
    add(quitReseq,c);

  }// end setConstraints method

}// end GUI class


 /**
  *  FileDialog FilenameFilter implementation
  */
class fileFilterAccept implements FilenameFilter
{
  String acceptExtension;

  public fileFilterAccept(String extension)
  {
    acceptExtension = extension;
  }

  /**
   * filter out all but specific extension files from FileDialog
   */
  public boolean accept(File dir, String name)
  {
System.out.println(acceptExtension+" "+name);
    return name.endsWith(acceptExtension);
  }
}

/**
 * display some output in an OutputFrame
 */
class OutputFrame extends Dialog
{
  OutputFrame(Frame parent,String title)
  {
    super(parent,title,false);
    addWindowListener(
      new WindowAdapter()
      {
        public void windowClosing(WindowEvent we)
        {
          setVisible(false);
        }
      }
    );
  }//OutFrame Constructor
}//end class OutputFrame

/**
 *  displays error message
 */
class ErrorFrame extends Dialog implements ActionListener
{
  ErrorFrame(Frame parent,String title,String errorMesg)
  {
    super(parent,title,true);
    addWindowListener
    (
      new WindowAdapter()
      {
        public void windowClosing(WindowEvent we)
        {
          setVisible(false);
        }
      }
    );
    setLayout(new FlowLayout(FlowLayout.CENTER));
    setTitle(title);
    Label error = new Label(errorMesg);
    Button ok = new Button("OK");
    add(error);
    add(ok);
    ok.addActionListener(this);
  }// end ErrorFrame Constructor

  public void actionPerformed(ActionEvent ae)
  {
    dispose();
  }
}// end class ErrorFrame

/**
 * displays message requiring yes or no response
 */
class MessageFrame extends Dialog implements ActionListener
{
  String YesOrNo;
  MessageFrame(Frame parent,String title,String message)
  {
    super(parent,title,true);
    addWindowListener
    (
      new WindowAdapter()
      {
        public void windowClosing(WindowEvent we)
        {
          setVisible(false);
        }
      }
    );
    setLayout(new FlowLayout(FlowLayout.CENTER));
    setTitle(title);
    Label yesNoMessage = new Label(message);
    Button Yes = new Button("Yes");
    Button No = new Button("No");
    add(yesNoMessage);
    add(Yes);
    add(No);
    Yes.addActionListener(this);
    No.addActionListener(this);
  }// end ErrorFrame constructor

  public void actionPerformed(ActionEvent ae)
  {
    if(ae.getActionCommand().equals("Yes"))YesOrNo = "Yes";
    else if(ae.getActionCommand().equals("No"))YesOrNo = "No";
    dispose();
  }
}// end class MessageFrame
