/*----------------------------------------------------------------*
|   Copyright 2001-2008 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

 /*
  * Title:       IRMP.java
  * Description: listener for IRMP over UDP/IP
  */

/**
  * IRMP is the Improved Reliabilty Multicast Protocol. It
  * provides reliability nearly as good as that of TCP over
  * a best-effort multicast group.
  *
  * This program implements IRMP in a multicast environment.
  * Each Javaprocess in the multicast group needs to have an
  * an IRMP class to receive and send messages on one group
  * and port. Multiple instances of this class can be created
  * to support different multicast groups and ports.
  */

import java.io.*;
import java.net.*;
import java.util.*;

public class IRMP extends TLM implements Runnable
{
  /*
   * maximum sequence number in IRMP. In the current version
   * of IRMP SN is a 22 bit sequence
   */
  private final int maxSequenceNumber=4194303;
  /*
   * The default multicast group address
   */
  private String groupAddress="230.0.0.1";
  /*
   * The default port Number
   */
  private int portNumber=4446;

  /*The sequence number of the message.Keeps incrementing for every
    data message
   */
  private static int sequenceNumber=0;
  /*
   * The maximum time to wait before sending out the ACK/NACK
   */
  private int tmax=1000;
  private final  int recvdBufferSize = 50; //size fo the received buffer
  private final  int ackBufferSize=50;     //the size of the ACK BUFFER
  private final  int sentBufferSize=30;    //the size of the SENT BUFFER
  private final  int nackBufferSize=50;    //the size of the NACK BUFFER

  /*
   * The four buffers:
   * Received Buffer- stores information of recieved messages
   * ACK Buffer-stores information of ACK's being created
   * NACK Buffer- stores information of NACK's being created
   * SENT Buffer- stores information of messages Sent by Local
   *              machine
   */
  //  | SourceIP | SourcePort | SequenceNumber|
  private String[][] receivedBuffer = new String[recvdBufferSize][4];

  //  | SourceIP | SourcePort | SequenceNumber|Thread Number|
  private String[][] ackBuffer      = new String[ackBufferSize][4];

  //  | SourceIP | SourcePort | SequenceNumber|Thread Number|
  private String[][] nackBuffer     = new String[nackBufferSize][4];

  //  | Sent Data | Sent SequenceNumber
  private String[][] sentBuffer     = new String[sentBufferSize][3];

  /*
   * ROUND TRIP TIME- Time to wait before resending the data message
   * if an acknowldgement is not received
   */
  private int rtt=1000;

  private int recvdBufferIndex=0;          //index of the received mesasge buffer to write into
  private int ackBufferIndex=0;            //index of NACK Buffer to write into
  private int nackBufferIndex=0;           //index of the NACK buffer to write into
  private int sentBufferIndex=0;           //index of the SENT buffer to write into

  MulticastSocket multicastSocket = null;
  //testing for duplicate packets- class uses this variable to resend packet with same SN
  private int tempseq=1;
  private IRMPCreateSent cs=new IRMPCreateSent();//object to create and stop sent threads
  private static int tempNackIndex;        //used in the Acknowledgement scheme
  private Thread runThread;                //Thread to start IRMp
  //private IRMPInterface irmpInterface;     //Interface object to call IRMPReceive
  private int MAX_MESSAGE=1455;            //default max received buffer size

  /** This function initializes the IRMP client to the port
    * and address specified
    * @param port The port to initialize the client on
    * @param groupAddress1 The multicast group address to which to subscribe
    */

  public IRMP(int port,
                 InetAddress inetGroupAddress1)
  //               IRMPInterface i1)
  {
  //	irmpInterface=i1;               //initialize IRMPInterface object
    portNumber=port;                //initialize port number
    try
    {
      multicastSocket=new MulticastSocket(portNumber);
    }
    catch(Exception se)
    {
      System.out.println(" Could not initialize irmp client ");
      se.printStackTrace();
    }
    try
    {
	  //initialize the multicast group address
      groupAddress=inetGroupAddress1.getHostAddress();
      multicastSocket.joinGroup(inetGroupAddress1);

    }
    catch(Exception u)
    {
	  System.out.println("UnknownHostException in IRMP constructor:"+u.getMessage());
    }
  }

  /**
   * This method closes the IRMP socket client
   */
  protected void irmp_close()
  { try
    { multicastSocket.close();
    }
    catch(Exception e)
    { System.out.println("Unable to close Socket");
    }
  }


  /**
   * The irmp_send(String data) sends the data to the configured
   * group address. It then adds an entry to the Sent Buffer
   * and starts a thread which waits for a time RTT. If an
   * acknowledgement is not received before time RTT, the
   * message is resent.
   * @param String data The data that needs to be sent
   * */
  protected void irmp_send(String data)
  {
	byte[] buf=new byte[1000]; //The byte buffer to store data to be sent
	String versionType="0000000100"; //the protocol version+ type of data which is 'data'
	//Create the header which is the protocol version + data type + sequence number
	/*The method addLeadingZeros returns a string after adding leading zeros
	  to the String specified (1st parameter) to make the total length of the String
	  to the value in the 2nd parameter. This method is used here to make the SN
	  a 22 bit sequence and therefore the header 32 bits
	*/

	String sequenceNumberWithZeros=addLeadingZeros(Integer.toBinaryString(sequenceNumber),22);
	String header=versionType+sequenceNumberWithZeros;

    if(sequenceNumber>maxSequenceNumber)
	sequenceNumber=0; //make sequence number back to 0 if it has reached the max. value

	String packet=header+data;//create the irmp packet which is header
	buf=packet.getBytes();    //convert into bytes to send data

	try
	{ InetAddress group = InetAddress.getByName(groupAddress);
      DatagramPacket irmpPacket = new DatagramPacket(buf, buf.length, group, portNumber);
      multicastSocket.send(irmpPacket); //send the data
      //createSentThread creates a thread for the sent message
      //and returns the Thread number (index)
      int sentThreadIndex=cs.createSentThread(rtt,
                                              data,sequenceNumberWithZeros);
      sentBuffer[sentBufferIndex][0]=sequenceNumberWithZeros;
      sentBuffer[sentBufferIndex][1]=data;
      sentBuffer[sentBufferIndex][2]=Integer.toString(sentThreadIndex);
	  sentBufferIndex++;
	  sequenceNumber++; //increment sequence number for the next message.
      if(sentBufferIndex>=sentBufferSize)
        sentBufferIndex=0;

    }
	catch(Exception e)//error while sending the data
	{ System.out.println("Could not send data");
      System.out.println("Error: "); e.printStackTrace();
	}
  }

  /**
   * This method re-sends the data with the sequence number
   * to the configured group address. This metod is either
   * called when an ACK is not received within the RTT
   * or when a NACK is received
   *
   * @param data The data to be resent
   * @param sequencenumber2 The sequence number of the data
   * @param index The index of the sent thread that caused the resend
   */
  protected void irmp_send(String data, String sequencenumber2,int index)
  {
	byte[] buf=new byte[1000]; //The byte buffer to store data to be sent
   	String versionType="0000000100"; //the protocol version+ type of data which is 'data'
   	//Create the header which is the protocol version + data type + sequence number
   	/*
   	   The method addLeadingZeros returns a string after adding leading zeros
   	   to the String specified (1st parameter) to make the total length of the String
   	   to the value in the 2nd parameter. This method is used here to make the SN
   	   a 22 bit sequence and therefore the header 32 bits
    */
   	String header=versionType+sequencenumber2;
   	String packet=header+data;//create the irmp packet which is header
   	buf=packet.getBytes(); //convert into bytes to send data
   	try
   	{ InetAddress group = InetAddress.getByName(groupAddress);
   	  DatagramPacket irmpPacket = new DatagramPacket(buf, buf.length, group, portNumber);
      multicastSocket.send(irmpPacket); //send the data
   	}
   	catch(Exception e)//error while sending the data
   	{ System.out.println("Could not re-send data");
      System.out.println("Error: "); e.printStackTrace();
   	}
  }

  /**
   * This function sends an ACK, NACK or MESGU message.
   * depending upon the type of the message. Th parameters are:
   * type= type of the mesage- ACK:"01", NACK:"10", MESGU:"11"
   * 'sourceIpAddress' is the ipaddress of the source to which this ACK/NACK/MESGU is intended
   * The funtion first generates the header. The source sequence Number is sent as the sequence
   * number of the ACK message
   *
   * @param messageType The type of the message ("01"-ACK, "10"-NACK, "11-MESGU)
   * @param sourceIpAddress The IP Address of the sender, will be added into Data of packet
   * @param sourcePortNumber The port number of the sending machine
   * @param sequenceNumber The sequence number of the packet received
   * @param index The index of the thread (of ACK/NACK) (used for debugging)
   */

  protected void irmp_send(String messageType,
                           String sourceIpAddress,
                           String sourcePortNumber,
                           String sequenceNumber1,
                           int index)
  {

	 //convert the String to int to inturn make it a binary string
     int sequenceNumber2=Integer.parseInt(sequenceNumber1,2);
	 byte[] buf=new byte[1000];

	 //the protocol version+ type of data which is either ACK, NACK or MESGU
	 String versionType="00000001"+messageType;

	 /* Create the header which is the protocol version + data type + sequence number
	    The function addLeadingZeros(String data, int size) adds leading zeros
	    to make the number of bits in data equal to 'size'
	  */
	 String header=versionType+addLeadingZeros(Integer.toBinaryString(sequenceNumber2),22);

	 //System.out.println("The header for the ACK is: "+header+" , length = :"+header.length());

     /* In IRMP the payload user data contains the source IP Address
	    and the source Port number. The String literal "port" is added between
	    the IP address and Port Number to help in extracting the information
	  */ // JMP comment: a seriously rotten programming practice!!!!!
	 String data=sourceIpAddress+"port"+sourcePortNumber;

	 String packet=header+data;//create the irmp packet which is header
	 buf=packet.getBytes(); //convert into bytes to send data
	 try
	 { InetAddress group = InetAddress.getByName(groupAddress);
	   DatagramPacket irmpPacket = new DatagramPacket(buf, buf.length, group, portNumber);
	   /* since its a ACK/NACK/MESGU it can be sent from any port
	      The following statement create a Datagram socket on any available port
	      on the machine and sends the ACK/NACK/MESGU
	    */
	   if(multicastSocket==null)
	   multicastSocket=new MulticastSocket();
	   multicastSocket.send(irmpPacket); //send message
/* debugx
	   if(messageType.equals("01"))
	   System.out.println(" ACK Sent: index:"+index+" data: "+data);
	   else if(messageType.equals("10"))
	   System.out.println(" NACK Sent: index:"+index+" data: "+data);
debugx */
	 }
	 catch(Exception e)//error while sending ACK/NACK/MESGU
	 { System.out.println("Could not send ACK data");
	   System.out.println("Error: "); e.printStackTrace();
	 }
   }

  /**
   * The run method that iterates for the times specified in the while loop
   * to receive messages. The received data is extracted by separating the header.
   * The message is then categorized into a Data Message, ACK, NACK, MESU depending
   * upon the messageType. Depending upon the kind of messsage, the appropraite
   * functions are called.
   */

  public void run()
  {
	DatagramPacket packet=null;

	//an object of CreateAck.java, will be used to create and stop threads
	IRMPCreateAck ca = new IRMPCreateAck();

	//an object of CreateNack.java. object will be used to create and stop threads
	IRMPCreateNack cn = new IRMPCreateNack();

    int i=0;
    int nackThreadIndex;
    while(true)//loop to receive messages
    { // loop to get the messages that are being sent

      // temporary storage of the received Data
      byte receivedByteBuffer[] = new byte[MAX_MESSAGE];

      try //
      {
        packet = new DatagramPacket(receivedByteBuffer, receivedByteBuffer.length);
        multicastSocket.receive(packet); //receive packet
        receivedByteBuffer=packet.getData();
      }
      catch(Exception e)
      {
        System.out.println("Error while receiving data");
      }

      // extract header and data information from the packet
      String receivedStringBuffer=new String(receivedByteBuffer); //convert message into String

      //  The first 8 bits constitute the protocol version
      String protocolVersion=receivedStringBuffer.substring(0,8);

      //  The next two bits denote the messageType
      String messageType=receivedStringBuffer.substring(8,10);

      //  SN- the next 22 bits
      String recvdSequenceNumber=receivedStringBuffer.substring(10,32);

      //  The remaining is the data
      String receivedData=receivedStringBuffer.substring(32);

      //Get the Source IP and SourcePort of the packet
      String sourceIpAddress = packet.getAddress().getHostAddress();
      String sourcePort = Integer.toString(packet.getPort());

      //Check if the received message is an IRMP message
      if(!(protocolVersion.equals("00000001")))
      { //message received is not a IRMP message
  	    continue;
	  }

	  //received message is an IRMP message so continue
      //check what kind of a message it is
      if(messageType.equals("00"))
      {
	    /**
	     check if there is a NACK thread running for this message
	     If a NACK thread is running, it needs to be stopped
	     */
	    int nackIndex= checkNackThread(sourceIpAddress,
	                                   sourcePort,
			 		                   recvdSequenceNumber);
	    if(nackIndex!=-1)
	    { //NACK thread exists, need to be stopped
		  cn.decideAction(nackIndex,3);
		  //make the Thread Index Entry in the NACK Buffer =-1
	      nackInvalidate(Integer.toString(nackIndex));
	    }

	    //check if this is a duplicate message
	    if(hasReceived(sourceIpAddress,sourcePort,
			          recvdSequenceNumber)==0)
	    { //message has already been received- ignore message
		  continue;
	    }
	    else //not a duplicate message
	    {  //update received buffer
		   receivedBuffer[recvdBufferIndex][0] = sourceIpAddress;
		   receivedBuffer[recvdBufferIndex][1] = sourcePort;
		   receivedBuffer[recvdBufferIndex][2] = recvdSequenceNumber;
		   receivedBuffer[recvdBufferIndex][3] = receivedData;
		   recvdBufferIndex++; //increment index of recvd buffer
		   i++;
           if(recvdBufferIndex>=recvdBufferSize)
           { //recvdBufferIndex is beyond limit make it back to '0'
             recvdBufferIndex=0;
           }

           // deliver the message to application JMP 3-2-07
           int payloadLength = packet.getLength()-32;
           byte[] payloadBuffer = null;
           if(payloadLength > 0)
           {
             payloadBuffer = new byte[payloadLength];
             System.arraycopy(receivedByteBuffer,32,payloadBuffer,0,payloadLength);
             //irmpInterface.IRMPReceive(payloadBuffer,
             mcfcListen.MCFCReceive(payloadBuffer,
                                    payloadLength,
                                    packet.getAddress());
           }

           // Create an ACK thread for the received message

           // index of ACK thread that is going to be created
           int threadIndex=0;
           try// create ACK thread by creating object of CreatAck
           {
             threadIndex=ca.createAcknowledgement(tmax,
                                                  sourceIpAddress,
                                                  sourcePort,
                                                  recvdSequenceNumber);
           }
           catch(Exception e)//error creating ACK
           {
             System.out.println("Error creating Ack");
             e.printStackTrace();
           }

           //update ACK BUffer when ACK is created
           ackBuffer[ackBufferIndex][0] = sourceIpAddress;
           ackBuffer[ackBufferIndex][1] = sourcePort;
           ackBuffer[ackBufferIndex][2] = recvdSequenceNumber;
           ackBuffer[ackBufferIndex][3] = Integer.toString(threadIndex);
           ackBufferIndex++; //increment ack buffer index
           if(ackBufferIndex>=ackBufferSize)
           ackBufferIndex=0;
          }//end else (not duplicate message)
        }  //end if messageType.equals("00")

      if(messageType.equals("01"))
      { // The messsage is a ACK, do the ACK procedure
	     /*'j' is a temprary variable used to extract the
	  	   source IP and source Port from the User Data
	   	 */
	    int j=0;
	    String ackSourceIpAddress="";
	    //System.out.println("AcK received: "+receivedData); checking
	    //User Data is of the form "<SourceIP>port<Port>"
	    //extract sourceIP from user Data
	    while(j<receivedData.length())
	    { if(receivedData.substring(j,j+1).equals("p"))
	  	  break;
		  ackSourceIpAddress=ackSourceIpAddress+receivedData.substring(j,j+1);
		  j++;
	    }

	    //extract Port from User Data
		String ackPortNumber=receivedData.substring((j+4)).trim();

	    /*
	      Check if a Sent Thread exists that is waiting for this ACK
	      If a Sent thread is found, then the Sent thread needs to be
	      stopped. The method getSentIndex returns the Sent Array
	      index if it finds a match in the sent Buffer
	     */
		int sentIndex=getSentIndex(recvdSequenceNumber);
	    if(sentIndex!=-1)
	    { //there exists an entry for this in the SENT Buffer
		  //check if it has not already been invalidated
		  if(Integer.parseInt(sentBuffer[sentIndex][2])!=-1)
	      { //has not been invalidated
			//stop the Thread by sending the threadIndex to StopThread
			cs.stopThread(Integer.parseInt(sentBuffer[sentIndex][2]));
		    //invalidate the thread
		    sentBuffer[sentIndex][2]="-1";
	      }//end "if(Integer.parseInt(..."
        }//end "if(sentIndex..."

		//check if there is already an ACK thread running for the messge
		int index=isAckThreadRunning(ackSourceIpAddress,
				                     ackPortNumber,
				                     recvdSequenceNumber);
        if(index!=-1)
		{ // a thread is running for this thread- needs to be stopped
          ca.stopThread(index); //stop thread

          //invalidate the entry in the ACK Buffer (make the thread number '-1'
		  ackInvalidate(Integer.toString(index));
		}//end "if(index!=-1..."
		else
		{ /* ACK thread does not exist. IRMP now checks if
		     the message has already been received. If the message
		     has not received, then a NACK will be generated
		  */

	      //check if message has already been recived
		  if(hasReceived(ackSourceIpAddress,ackPortNumber,
		    recvdSequenceNumber) !=0 )
		  { /*
		      no entry in the received buffer- this implipes
		      that the client hasn't received this message.
		      NACK needs to be generated
		    */

	        //create the NACK thread
		    int threadIndex = cn.createNack(tmax,
		                    ackSourceIpAddress,ackPortNumber,
		                    recvdSequenceNumber);
	        //Update NACK buffer
		    nackBuffer[nackBufferIndex][0] = ackSourceIpAddress;
		    nackBuffer[nackBufferIndex][1] = ackPortNumber;
	        nackBuffer[nackBufferIndex][2] = recvdSequenceNumber;
	        nackBuffer[nackBufferIndex][3] = Integer.toString(
	      		                           threadIndex);
		    nackBufferIndex++; //increment NACK buffer index
		    if(nackBufferIndex>=nackBufferSize)
		      nackBufferIndex=0;
	      }//end "if hasReceived(..."
         }//end "Else{.."
      }
	  if(messageType.equals("10"))
	  { // System.out.println("Nack Received");
	    //It is a NACK, do the NACK procedure

	    int j=0;
		String nackSourceIpAddress="";
		/*
		   Data in the NACK is of the form
		    ____________________________
		   |<ipaddress>PORT<portnumber> |
		    ____________________________
		 */
		 //extract sourceIP from user Data
		 while(j<receivedData.length())
		 { if(receivedData.substring(j,j+1).equals("p"))
		   break;
		   nackSourceIpAddress=nackSourceIpAddress+receivedData.substring(j,j+1);
		   j++;
	     }

	     //extract Port from User Data
	     String nackPortNumber=receivedData.substring((j+4)).trim();

	     //check if this NACK is for a packet that the local system has sent

	     //determine Local IP
	     String localIpAddress="";
	     try
	     { localIpAddress = InetAddress.getLocalHost().getHostAddress();
		 }
		 catch(UnknownHostException uhe)
		 { System.out.println("Error determining Local IP address");
		   uhe.printStackTrace();
		 }
         String tempPortNumber=""+portNumber;

	     if(nackSourceIpAddress.equals(localIpAddress) &&
	        nackPortNumber.equals(tempPortNumber))
	     { //NACK for a Packet sent by the Local machine
		   //Packet needs to be resent
           int sentIndex=getSentIndex(recvdSequenceNumber);
		   /*
			 CHECK IF sentIndex ==-1 and if it is then send a MESU message
			*/
            if(sentIndex==-1)
            { //message unavailable in SENT BUFFER
              //Send Message Unavilable Error
			  irmp_send("11",nackSourceIpAddress,nackPortNumber,recvdSequenceNumber,0);
            }
			else
			{ //message available in Sent Buffer
              irmp_send(sentBuffer[sentIndex][1],sentBuffer[sentIndex][0],1);
			}
		 }

		 /*
		   Check if there is a NACK thread running for this message. If yes, IRMP
		   does not stop the thread but restarts the timer for the Thread. This is to
		   overcome a situation where the NACK sent is lost and the local machine
		   never gets the data message.
		 */

		 //check if the NACK thread is running
         int index=isNackThreadRunning(nackSourceIpAddress,
                                       nackPortNumber,
                                       recvdSequenceNumber);
         if(index!=-1) // a thread is running- needs to be restarted
		 { //making the action "1" restarts the thread
		   cn.decideAction(index,1); //restart NACK
		 }
		 if(hasReceived(sourceIpAddress,sourcePort,
		 			    recvdSequenceNumber)==0)
		 { //message has already been received- ignore message
		   System.out.println("Duplicate message received:"+receivedByteBuffer[32]);
		   continue;
	    }
      }//end "if messageType.equals("10")

	  if(messageType.equals("11"))
	  {
		/*
		Message Unavilable message has been received. This implies that the
		sender no longer has the message that has been NACK'ed. If there is
		a NACK thread running on the local machine, it needs to stop
		*/
	    System.out.println("Message Unavilable received");

	    //Extract MESGU sourceIP and source Port
	    String MESGUSourceIpAddress="";
	    /*
	       Data in the NACK is of the form
		    ____________________________
		   |<ipaddress>PORT<portnumber> |
		    ____________________________
	     */

	     //extract sourceIP from user Data
	     int j=0;
		 while(j<receivedData.length())
	     { if(receivedData.substring(j,j+1).equals("p"))
		   break;
		   MESGUSourceIpAddress=MESGUSourceIpAddress+receivedData.substring(j,j+1);
		   j++;
	     }
		 //extract Port from User Data
		 String MESGUPortNumber=receivedData.substring((j+4)).trim();

	     //check if there is a thread running
	     int nackIndex=isNackThreadRunning(MESGUSourceIpAddress, MESGUPortNumber,
			 			 		         recvdSequenceNumber);

		 if(nackIndex!=-1)
	     { //There is a NACk thread running. Needs to be stopped
		   //stop the thread by making action "3"
		   cn.decideAction(nackIndex,3);

		   //invalidate the entry in the NACK buffer
		   nackInvalidate(Integer.toString(nackIndex));
		   System.out.println("Stopped Nack because of MESGU");
		 }
       }//end "if messageType.equals"11")

 	   }


    }//end run method


/**
  * This method that adds the leading zeros to the sequence number to make it a
  * string of total_length (22 in the case of the IRMP v 1.0) number of bits
  *
  * @param data The string to to which leading zeros have to be added
  * @param total_length The final length of the string after prefixing '0''s
  * @return The string after adding the leading zeros
  */
protected String addLeadingZeros(String data, int total_length)
{ int length=data.length();
  String leadingZeros="";
  for(int i=0;i<(total_length-length);i++)
  { leadingZeros=leadingZeros+"0";
  }
  data=leadingZeros+data;
  return(data);
}


/**
  * This method that checks if there exists an entry in the received buffer
  * that matches the parameters specified.
  *
  * @param ipAddress The IP Address that needs to be matched
  * @param port      The port that needs to be matched
  * @param sequenceNumber The Sequenece Numbre that needs to be matched
  *
  * @return Returns 0 if there is a match, 1 otherwise
  */

protected int hasReceived(String ipAddress, String port, String sequenceNumber)
{
  for(int i=0;i<recvdBufferSize;i++)
  {
	 if(receivedBuffer[i][0]==null)
	 break;
	 if(    receivedBuffer[i][0].equals(ipAddress)
        &&  receivedBuffer[i][1].equals(port)
        &&  receivedBuffer[i][2].equals(sequenceNumber)
      )
      { return(0);
	  }
  }
  return(1);
}

/**
  * This method checks the ACK Buffer to see if there is a thread running
  * with the parameters specified
  *
  * @param ipAddress      The IP Address that needs to be matched in ACK Buffer
  * @param port           The port that needs to be matched
  * @param sequencenumber The sequence number to be matched
  *
  * @return returns the threadIndex (fourth enrty in the byffer) if there is
  *         a match, '-1' otherwise
  */

protected int isAckThreadRunning(String ipAddress, String port, String sequenceNumber)
{
  for(int i=0;i<ackBufferSize;i++)
  {
	 if(ackBuffer[i][0]==null)
	 break;
    if(   ackBuffer[i][0].equals(ipAddress)
       && ackBuffer[i][1].equals(port)
	   && ackBuffer[i][2].equals(sequenceNumber)
	  )
    {
	      if(ackBuffer[i][3].equals("-1"))//ACK has already been invalidated
		  return(-1);
		  return(Integer.parseInt(ackBuffer[i][3]));
    }
  }
   return(-1);
}


/**
  * This method invalidates an ACK entry in the ACK Buffer. This is
  * done by making the Thread index value equal to '-1' for that
  * particular entry in the ACK Buffer
  *
  * @param index The Thread number in the ACK buffer that will be made '-1'
  */

protected void ackInvalidate(String index)
{ for(int i=0;i<ackBufferSize;i++)
  {
	 if(ackBuffer[i][0]==null)
	 break;
	if(ackBuffer[i][3].equals(index))
    { ackBuffer[i][3]="-1";
    }
  }
}


/**
  * This method checks the NACK Buffer to see if there is a thread running
  * with the parameters specified
  *
  * @param ipAddress      The IP Address that needs to be matched in NACK Buffer
  * @param port           The port that needs to be matched
  * @param sequencenumber The sequence number to be matched
  *
  * @return returns the threadIndex (fourth enrty in the byffer) if there is
  *         a match, '-1' otherwise
  */

protected synchronized int isNackThreadRunning(String ipAddress, String port, String sequenceNumber)
{
  for(int i=0;i<nackBufferSize;i++)
  {
     if(nackBuffer[i][0]==null)
	 break;
    if(   nackBuffer[i][0].equals(ipAddress)
	   && nackBuffer[i][1].equals(port)
	   && nackBuffer[i][2].equals(sequenceNumber)
      )
    {
		  if(nackBuffer[i][3].equals("-1"))
		    return(-1);
		  return(Integer.parseInt(nackBuffer[i][3])) ;
    }

  }
  return(-1);
}



/**
  * Method That invalidates a NACK thread. This can be done by making the
  * Thread number in the NACK Buffer to -1
  * A NACK is invalidated only when a corresponsing Data message for the NACK
  * or a corresponding MESGU message is recived
  *
  * @param index The Thread number that needs to be invalidated
  */

protected void nackInvalidate(String index)
{ for(int i=0;i<nackBufferSize;i++)
   {
	   if(nackBuffer[i][0]==null)
	   break;
	  if(nackBuffer[i][3].equals(index))
      nackBuffer[i][3]="-1";
   }
}// end method nackInvalidate(...


/**
  *
  * This method returns the index in the Sent Buffer corresponding to a Match
  * with the parameter specified (sequencenumber)
  * The method takes the sequence number as the parameter and checks if there is
  * an entry corresponding to sequencenumber specified as the parameter
  * If an entry is present, the method returns the index of the Entry.
  * This is done because the local host may want to resend the Data and will need
  * both the sequence number and the Data of the message. And because two values
  * cannot be returned by the method, the method returns the index of the entry so that the
  * local host can retrive the values from the Sent Buffer itself
  *
  * @param sequencenumber The sequence number to be matched in the Sent Buffer
  */

protected int getSentIndex(String sequencenumber)
{ for(int j=0;j<sentBufferSize;j++)
  {
	if(sentBuffer[j][0]==null)
	break;
	if(sentBuffer[j][0].equals(sequencenumber))
    {  //match in Sent Buffer
         return(j);//return Index of the Entry in the SENT Buffer
      }
   }
   return(-1); //return '-1' if there is no match in the Sent Buffer
}//end Method getSentIndex(...


/*
 * This method is used by IRMP to check if a NACK thread is running
 * when a data message is received.
 *
 * @param ip   IP Address to be matched
 * @param port port to be matched
 * @param sn   Sequence number to be matched
 *
 * @return The thread index if there is a match, -1 otherwise
*/
public int checkNackThread(String ip, String port, String sn)
{ int match=0;
  for(int countNack=0;countNack<nackBufferIndex;countNack++)
  {
	 if(nackBuffer[countNack][0]==null)
	 break;
    if(   nackBuffer[countNack][0].equals(ip)
	   && nackBuffer[countNack][1].equals(port)
	   && nackBuffer[countNack][2].equals(sn)
	  )
    {
	  match=1;
	  return(Integer.parseInt(nackBuffer[countNack][3]));
    }
    else
    {
    }
  }
  if(match!=1)
  {
    return(-1);
  }
return(-1);
}//end checkNackMethod




/**
  * Title:       Create Acknowledgement Thread
  * Description: creates and stops a Acknowledgement thread
  *
*/
/*
 * This class creates an Acknowledgement thread for every Data message received.
 * When a Acknowldgement is received, this class takes the id of the acknowldgement
 * and stops the thread. The  thread is created by creating an object of the
 * Create_Ack_Thread class. The Create_Ack_thread class has the implementaion
 * of the random backoff and the sending fo the ACK
 *
*/
class IRMPCreateAck
{ protected int threadIndex;
  protected int maxNumberOfAcks=20;
  protected int tmax;
  protected String sourceIpAddress="";
  protected String sourcePortNumber="";
  protected String sequenceNumber="";
  Create_Ack_Thread[] ack = new Create_Ack_Thread[maxNumberOfAcks];
  /**
    * constructor that instantiates the values source IP Address
    * port number and sequence number. These values are passed on to the
    * Create_Ack_Thread.java class to generate the ACK
    */
  public void IRMPCreateAck()
  { //System.out.println("In the create ack constructor");
    ack[threadIndex]=new Create_Ack_Thread();
    //instantiate the vlues in the current thread
    ack[threadIndex].instantiate(tmax, sourceIpAddress, sourcePortNumber, sequenceNumber,threadIndex);
    //start the current thread
    ack[threadIndex].start();
    threadIndex++;
  }

  /*
   * This method instantiates the values of the sourceIpAddress, sourcePort and sequenceNumber
   * These values are then passed over to Create_Ack_Thread.java to send an Acknowledgement
   *
   * @param ipaddress the sourceIpAddress of the packet
   * @param portnumber the source port of the packet
   * @param tmax1 the value of tmax (maximum wait time before sending the Ack
   */

  public int createAcknowledgement(int tmax1, String ipaddress, String portnumber, String sequencenumber)
  {
	sourceIpAddress="";
    sourcePortNumber="";
    sequenceNumber="";
    tmax=tmax1;
    sourceIpAddress=ipaddress;
    sourcePortNumber=portnumber;
    sequenceNumber=sequencenumber;
    ack[threadIndex]=new Create_Ack_Thread();
    //instantiate the vlues in the current thread
    ack[threadIndex].instantiate(tmax, sourceIpAddress, sourcePortNumber, sequenceNumber, threadIndex);
    //start the current thread
    ack[threadIndex].start();
    threadIndex++;
    if(threadIndex>=maxNumberOfAcks)
    threadIndex=0;
    return((threadIndex-1));
  }

  /**
    * This method stops the thread indexed by the parameter 'index'.
    * This is done by calling the 'stopThread()' method of the corresponding thread
    *
    * @param index The index of the thread to be stopped.
    */
    protected void stopThread(int index)
    { try
      {
		  ack[index].stopThread(index);//stop the thread-calls CreateAck.stopThread();
      }
      catch(Exception e)//error while trying to stop thread
      { System.out.println("Unable to stop sending of Acknowledgement");
        e.printStackTrace();
      }
    }

    /** Title:       Create_Ack_Thread.java
	  * Description: class that creates a thread for an Acknowledgement
	  */

	/**
	  *Class that creates the Acknowledgement thread for each message recieved
	  * When a data message is received by IRMP, the class IRMP_IMPLEMENTATION.java
	  * calls this class (Create_Ack_Thread.java) to create a ACK thread for each message received.
	  * This class then creates a ACK thread and waits for a random time between 0 and tmax
	  * before it sends out the acknowledgement. It saves the thread information
	  * in the ACK buffer. The thread can be destroyed by IRMP_IMPLEMENTATION.java if an
	  * aknowledgement is received for the same message before the elapsed time.
	  */

	class Create_Ack_Thread extends Thread
	{ private long tmax; // The parameter tmax - is received from the IRMPImplementation class
	  private String ipaddress="";
	  private String port_number="";
	  private String sequenceNumber="";
	  private boolean isAlive=true;
	  private int index;

	 /** This method instantaites the the variables of the ACK thread
	   * @param tmax The value of tmax (maximum amount of time to wait before sending ACK
	   * @param ipaddress The source IP address of the packet
	   * @param port_number1 The source port number of the packet
	   * @param sequqencenumber The sequence number (SN) of the packet that is to be acknowledged
	   * @param index1 The index of the thread- used for debugging and knowing the index of thread
	  **/

	public void instantiate(int tmax1,String ipaddress1,
	                        String port_number1, String sequencenumber,int index1)
	   //instantiates the values for tmax, ipaddress and port number
	{ this.tmax=tmax1;             //instantiate the value of tmax,
	  this.ipaddress=ipaddress1;   //ipaddress and port number to send the ACK
	  this.port_number=port_number1;//values are received from IRMPImplemenation.java
	  this.sequenceNumber=sequencenumber;
	  this.index = index1;
	}

	/** This method Stops the thread from sending the acknowledgement
	  * The thread sends the acknowledgement only if the value of isAlive is 'true;
	  * The acknowledgement can therefore be stopped if the value of isAlive is made 'false'
	  * @param index The idnex of the thread that is being stopped (used for debugging)
	  */

	void stopThread(int index)
	{ this.isAlive=false; //stops the thread from sending the Acnowledgement
	  //System.out.println("Stopped ACK thread "+index);//debugx
	}

	/** The method that creates the ACK thread and waits for a random amount of
	  * time before sending the ACK
	  *
	  */

	public void run()
	{ try{ double random=Math.random();  //generate random number
		   double waitTime2=random*tmax; //make it < tmax
		   sleep((long)(waitTime2));      // sleep for a time that is uniformly random
	     }//end try                       //between 0 and tmax
	     catch(InterruptedException e)
	     { e.printStackTrace();
		 }
	     if(this.isAlive)      //thread hasn't been stopped - which means the ACK hasn't
			                  //been sent by anyone else
		 {
		   try
		   {
			 irmp_send("01",ipaddress,port_number,sequenceNumber,index);
		   }
		   catch(Exception e)//error while sending ACK
		   { System.out.println("Error sending the ACK message Checking");
			 e.printStackTrace();
		   }
		 }//end if(this.isAlive
	   }
	}//end class Create_Ack_Thread

}//end class IRMPCreateAck



/**
  * Title:       Create N- Acknowledgement Thread
  * Description: creates and stops a N- Acknowledgement thread
  *
*/
/*
 * This class creates a NACK when an ACK for a Data message is received and
 * the local machine does not have that Data message.
 * The maximum number of NACK threads that can be created is specified
 * by the 'maxNumberOfNacks' variable.A Nack can be created by calling the
 * createNack() method. The class then creates a NACK thread with the
 * index as 'threadIndex'. The action of the NACK is decided by the variable
 * 'toDo' as specified below:
 *
 *   toDo          Action                       Reason
 *     1       Restart the NACK thread    NACK already received
 *     2       Send the NACK			  NO NACK receievd
 *     3       Stop the NACK			  Data message/MESGU received
*/

class IRMPCreateNack
{ protected int threadIndex;
  protected int maxNumberOfNacks=20;
  protected int tmax;
  protected String sourceIpAddress="";
  protected String sourcePortNumber="";
  protected String sequenceNumber="";
  Create_Nack_Thread[] Nack = new Create_Nack_Thread[maxNumberOfNacks];

  /*
   * Instantiates the values of the sourceIpAddress, sourcePort and sequenceNumber
   * These values are then passed over to Create_Nack_Thread.java to send an N- Acknowledgement
   *
   * @param ipaddress the sourceIpAddress of the Packet
   * @param portnumber the source port of the Packet
   * @param tmax1 the value of tmax (maximum wait time before sending the Nack
   */

  public int createNack(int tmax1, String ipaddress, String portnumber, String sequencenumber)
  {
	sourceIpAddress="";
	sourcePortNumber="";
    sequenceNumber="";
    tmax=tmax1;
    sourceIpAddress=ipaddress;
    sourcePortNumber=portnumber;
    sequenceNumber=sequencenumber;
	Nack[threadIndex]=new Create_Nack_Thread();
    //instantiate the values in the current thread
    Nack[threadIndex].instantiate(tmax, sourceIpAddress, sourcePortNumber, sequenceNumber, threadIndex);
    //start the current thread
    Nack[threadIndex].start();
    threadIndex++;
    if(threadIndex>=maxNumberOfNacks)
    threadIndex=0;
    return((threadIndex-1));
  }

  /*
   * Stops the thread indexed by the parameter
   * @param index The index of the thread to be stopped.
   */
  protected void decideAction(int index,int toDo)
  { try
    {
      Nack[index].decideAction(toDo);//stop the thread-calls CreateNack.stopThread();
    }
    catch(Exception e)//error while trying to stop thread
    { System.out.println("Unable to stop sending of N- Acknowledgement");
      e.printStackTrace();
    }
  }

  /** Title:       Create_Nack_Thread.java
    * Description: Class that creates a thread for a NACK
    */

  /* Class that creates the N- Nacknowledgement thread for each message recieved
   * When a data message is received by IRMP, the class CreateNack.java
   * calls this class (Create_Nack_Thread.java) to create a Nack thread for each message received.
   * This class then creates a Nack thread and waits for a random time between 0 and tmax
   * before it sends of the NACK. If a NACK for the same message is received before the timer
   * goes off, then the timer is restrated. If a Data message or a Message Unavailable message
   * is received corresponding to this NACK then this NACK is stopped by the IRMP
   * class through the CreateNack class
  */

  class Create_Nack_Thread extends Thread
  { private long tmax; // The parameter tmax is received from the IRMP class
    private String ipaddress="";
    private String port_number="";
    private String sequenceNumber="";
    private boolean isAlive=true;
    private int index;
    // The max number to times to restart the NACK
    private int maxRestartTimes=5;
    //variable that decides what the NACK thread does after the wait time
    private int nextAction=2;


  /** instantaites the parameters of the Nack thread
    * @param tmax The value of tmax (maximum amount of time to wait before sending Nack
    * @param ipaddress The source IP address of the pNacket
    * @param port_number1 The source port number of the pNacket
    * @param sequqencenumber The sequence number (SN) of the pNacket that is to be Nacknowledged
    * @param index1 The index of the thread- used for debugging and knowing the index of thread
    **/

    public void instantiate(int tmax1,String ipaddress1,
                          String port_number1, String sequencenumber,int index1)
    { this.tmax=tmax1;             //instantiate the value of tmax,
      this.ipaddress=ipaddress1;   //ipaddress and port number to send the Nack
      this.port_number=port_number1;//values are received from IRMPImplemenation.java
      this.sequenceNumber=sequencenumber;
      this.index = index1;

    }

    /* This method sets the 'nextAction' varaible whichDecides what needs to be
     * done to the NACK thread once the wait time
     * is over. The action is according to the value of toDo as follows
     * '1' -- > NACK already received, restart waittime
     * '2' --> No NACK received, send the NACK
     * '3' --> Data or MESGU received don't send NACK
     */

     public void decideAction(int nextAction1)
     { nextAction=nextAction1;
     }

     /**
       * The method that creates the Nack thread and waits for a random amount of
       * before sending out the NACK. If another NACK for the same message has been
       * received, IRMP makes the value of nextAction as '1', and so the
       * NACK thread is restarted for another random amount of time. The thread can be stopped
       * if the value of nextAction is made '3'. This is done by IRMP
       * (through CreateNAck.java) when a corresponding Data Message or a Message Unavilable
       * message is received
       *
       */

     public void run()
     { int count=0;
  	 while(count<maxRestartTimes)
  	 { try{ double random=Math.random(); //generate random number
              double waitTime2=random*tmax; //make it < tmax
  		    sleep((long)(waitTime2));// sleep for a time that is uniformly random
  		  }//end try                  //between 0 and tmax
  		  catch(InterruptedException e)
  		  {  e.printStackTrace();
  	      }
  		  if(this.nextAction==1)//received a NACK, thread needs to restart
  	      { /*
  	            make the NextAction as 2(Send) to send Nack
  	            after the restart of the therad
  	         */
  			 nextAction=2;
  			 count++;
  			 continue;
  	      }
  		  else if(this.nextAction==2)
  		  { // The thread hasn't been stopped. The NACK can be sent
  		    try
  		    {
  			  //send the Nack
  			  irmp_send("10",ipaddress,port_number,sequenceNumber,index);
  		    }
  		    catch(Exception e)
  		    { System.out.println("Error Sending NACK Message");
  		    }
  		  }
  		  else
  		  {
  	        break;
  	      }
  		  count++;
  	}//end while
    }//end method run
  }//end Class Creat_NACK_Thread



}//end class IRMPCreateNack


/**
  * Title:       Create Sent Thread
  * Description: creates and stops a Sent Message thread
  *
*/
/*
 * This class creates an 'Sent' thread for every Data message that is sent.
 * The thread waits for a time RTT (sent as a parameter to createSentThread()
 * before resending the message. The thread can be stopped when the client
 * receives an ACK for the message.The number of times that the message is resent is
 * specified by the attribute -'maxResendTimes'.
*/
class IRMPCreateSent
{ protected int threadIndex;
  protected int maxNumberOfSentThreads=20;
  /*
   * ROUND TRIP TIME- Time to wait before resending the message
   */
  protected int rtt;
  protected String sequenceNumber="";
  protected String data="";
  Create_Sent_Thread[] sent = new Create_Sent_Thread[maxNumberOfSentThreads];


  /*
   * This method instantiates the values of the sourceIpAddress, sourcePort and sequenceNumber
   * These values are then passed over to Create_Ack_Thread.java to send an Acknowledgement
   *
   * @param ipaddress the sourceIpAddress of the packet
   * @param portnumber the source port of the packet
   * @param tmax1 the value of tmax (maximum wait time before sending the Ack
   */
  public int createSentThread(int rtt1, String data1,String sequencenumber)
  { rtt=rtt1;
    sequenceNumber=sequencenumber;
    data=data1;
    sent[threadIndex]=new Create_Sent_Thread();
    sent[threadIndex].instantiate(rtt,data,sequenceNumber,threadIndex);
    //start the current thread
    sent[threadIndex].start();
    threadIndex++;
    if(threadIndex>=maxNumberOfSentThreads)
    threadIndex=0;
    return((threadIndex-1));
  }

  /**
    * This method stops a SENT thread by calling the stopThread()
    * method of the corresponding thread
    *
    * @param index The index of the thread that needs to be stopped
    */

  protected void stopThread(int index)
  { try
     {
	   sent[index].stopThread();//stop the thread-calls CreateAck.stopThread();
     }
     catch(Exception e)//error while trying to stop thread
     { System.out.println("Unable to stop resending of Data");
       e.printStackTrace();
     }
  }//end method StopThread()

  /** Title:       Create_Sent_Thread.java
    * Description: Class that creates a thread for a Data Messsage to be resent
    */

  /* Class that creates a Sent thread for each Data message sent
   * When a data message is sent by IRMP, the class IRMP_IMPLEMENTATION.java
   * calls this class (Create_Snet_Thread.java through createSent.java) to create a
   * Sent thread for each Data message sent.
   * This class then creates a Sent thread and waits for a time rtt
   * before it resends the Data message. It saves the thread information
   * The thread can be destroyed (stopped) by IRMP_IMPLEMENTATION.java(through CreateSent.java) if a
   * ACK corresponding to this Data Message
   * If the thread is not stopped and the timer goes off, then the thread restarts itself
   * for a maximum number of times- 'maxResendingTimes'
   */

  class Create_Sent_Thread extends Thread
  { private long rtt; // The parameter tmax - is received from the IRMP class
    private String data="";
    private String sequenceNumber="";
    private boolean isAlive=true;
    private int index;
    // The max number to times to restart the Resending Thread
    private int maxResendingTimes=3;

    /** instantaites the parameters of the Sent thread
      * @param rtt1 The value of RTT (time to wait before resending Data Messsage)
      * @param data1 The data to be resent
      * @param sequqencenumber The sequence number (SN) of the packet that is to be resent
      * @param index1 The index of the thread- used for debugging and knowing the index of thread
      */

    public void instantiate(int rtt1,String data1,
                            String sequencenumber, int index1)
    { this.rtt=rtt1;       //instantiate the value of rtt,
  	this.data=data1;       //instantiate the data to be resent
  	this.sequenceNumber=sequencenumber; //instantiate the value of sequence Number
  	this.index = index1;   //instantiate the value of the index of the thread
    }

    /**
     * This method Decides what needs to be done to the NACK thread once the wait time
     * is over. The action is according to the value of toDo as follows
     * '1' -- > NACK already received, restart waittime
     * '2' --> No NACK received, send the NACK
     * '3' --> Data or MESGU received don't send NACK
     */

     public void stopThread()
     {
  	   this.isAlive=false;
     }

     /**
       * This method creates the Sent thread and waits for a random amount of
       * time before resending the message
       *
       */
     public void run()
     { int count=0;
  	 while(count<maxResendingTimes)
  	 {
	   try{
  		   sleep((long)(rtt));// sleep for a time that is uniformly random
  		  }//end try          //between 0 and tmax
  		  catch(InterruptedException e)
  		  { e.printStackTrace();
  	      }
  		  if(this.isAlive==false)//received a ACK, thread needs to stop
  	      { break;
  	      }
  		  else
  		  { // The thread hasn't been stopped. message needs to be resent
  		    try
  		    {
  			  irmp_send(data,sequenceNumber,index);
            }
  		    catch(Exception e)
  		    { System.out.println("Error resending Data Message");
  		    }
  		    count++;
  	      }
       }//end run
    }//end emthod run
  }//end class Create_Sent_Thread

}//end class IRMPCreateSent


/**
  * This method sets the maximum length of the received Buffer
  * @param messageLength The maximum received buffer size
  */
public void setMaxMessage(int messageLength)
{
  this.MAX_MESSAGE=messageLength;

}

}//end class IMRPRun

