//package tlm;

/**
 * <p>Title: NEWSocketChannel</p>
 * <p>Description: Transport Layer Multicaster secure tunnel non-blocking</p>
 * <p>Copyright: Copyright (c) 2013</p>
 * <p>Company: GMU NETLAB</p>
 * @author Charles Snow & Mark Pullen
 * @version 5.2.4
 */

/********************************************
 * REVISION HISTORY:
 *
 * 1.0.0 Sept 2004 original release with NEWv4.0
 * 1.0.1 Oct  2004 extends certificate verification features
 * 1.1.1 Oct  2005 stopped crash when sockets are null - JMP
 * 5.2.4 MAr  2013 converted NEWSocket to NEWSocketChannel
 ********************************************/

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/

/**
 *  Generates SocketChannels for both in-the-clear and stunnel TCP connections.
 *
 *  These are general-purpose Java SocketChannels so tehy can have a Socket 
 *  extracted, which supports methods such as getInetAddress().
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;
import java.text.*;

import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

import javax.net.ssl.*;
import javax.net.ssl.SSLServerSocket;
import javax.security.cert.*;

public class NEWSocketChannel
{
	static ServerSocketChannel dataSocketChannel;
	static boolean usingSSL;
	//static SSLServerSocket dSocket, cSocket;
	static int VerifyFlags = 1; // assume certificate vfy everything
	static SimpleDateFormat msgTimeFormat = new SimpleDateFormat("yyyy MM dd HH:mm:ss");

	/*   constants for conditions that may be verified in SafeSocket()
	 *
	 *   be sure to set MAX_VFY to the value of the highest valued check flag: it is
	 *   used as an upper bound on loop checks
	 */
	static final int SEMESTER_ID =	  1;
	static final int COURSE_ID   =	  2;
	static final int USERNAME    =	  4;
	static final int COMMONNAME  =	  8;
	static final int EMAIL_ADDR  =	 16;
	static final int ORG_NAME    =	 32;
	static final int ORG_UNIT    =	 64;
	static final int ORG_LOC     =	128;
	static final int ORG_STATE   =	256;
	static final int ORG_COUNTRY =	512;
	static final int MAX_VFY     =  ORG_COUNTRY;

	static final int VERIFY_REF  = (EMAIL_ADDR | COMMONNAME | ORG_UNIT | ORG_NAME | ORG_LOC | ORG_STATE | ORG_COUNTRY);
	static final int X509_MAXATTR =   6;

	/* extended constructor: 4 parms: port number, queue length, useSSL flag, vfy checks flag (bits). */

	public NEWSocketChannel (int tunnelTCPPort, int queueLength, boolean useSSL, int vfymask)
    throws SocketException
	{
		usingSSL = useSSL;	// capture state (private tunnels or not)

		VerifyFlags = vfymask;	// select which values to verify in SafeSocket()

		if (!usingSSL) 
    {	// use in-the-clear connections
			try 
      {
				// create, configure and bind the SocketChannel
				dataSocketChannel = ServerSocketChannel.open();
        dataSocketChannel.configureBlocking(true);// block on capp to acccept()
        dataSocketChannel.socket().bind(
          new java.net.InetSocketAddress(
            InetAddress.getLocalHost(),
            tunnelTCPPort));

			} catch (Exception e) {
				System.out.println (
          msgTime() + "NEWSocketChannel(): exception creating non-secured SocketChannel: " + e);
        e.printStackTrace(System.out);
        throw new SocketException();
			}
		}
		else {		// use SSL authenticated + encrypted connection

			// be pedantic: ensure these are bogus until we know they aren't

			dataSocketChannel = null;
      System.out.println(
        "*** attempt to create secure SocketChannel which is not supported by NEWSocketChannel");
      throw new SocketException();

/*
			SSLServerSocketFactory ssf = (SSLServerSocketFactory) (SSLServerSocketFactory.getDefault ());

			// do these separately in case we want to handle the exceptions separately

			try {
				dSocket = (SSLServerSocket) (ssf.createServerSocket (tunnelTCPPort, queueLength));
			}
			catch (IOException e) {
				System.out.println (msgTime() + "NEWSocket(): exception creating SSL dsocket: " + e);
			};

			try {
				cSocket = (SSLServerSocket) (ssf.createServerSocket (tunnelTCPPort + 1, queueLength));
			}
			catch (IOException e) {
				System.out.println (msgTime() + "NEWSocket(): exception creating SSL csocket: " + e);
			};

			if (dSocket != null) {
				dSocket.setNeedClientAuth (true);	// server must authenticate client
			}

			if (cSocket != null) {
				cSocket.setNeedClientAuth (true);	// server must authenticate client
			}

			dataSocket = (ServerSocket) (dSocket);
*/
		}
	} /* end NEWSocket new constructor */

	/* basic constructor: 3 parms: port number, queue length, useSSL flag; compatible with older version of NEWSocket() */

	public NEWSocketChannel (int tunnelTCPPort, int queueLength, boolean useSSL)
    throws SocketException
	{
		this(tunnelTCPPort,queueLength,useSSL,1);
	}

  /**
   * Sets the connection timeout on both sockets
   * @return
   * @throws Exception
   */
	public static void setSoTimeout (int timeoutValue) throws SocketException {
		if (usingSSL) {
			dataSocketChannel.socket().setSoTimeout (timeoutValue);
		}
		else {
//			dataSocket.setSoTimeout (timeoutValue);
//			controlSocket.setSoTimeout (timeoutValue);
		}
	}

  /**
   * Waits for and accepts a new TCP data connection
   * @returns socket for a new connection
   * @throws Exception
   */
	public static SocketChannel getDataSocketChannel () throws Exception {

		if (!usingSSL) {

      // block waiting new connection
      SocketChannel newSocketChannel = dataSocketChannel.accept();
      if(newSocketChannel == null)
      {
        System.out.println("*** SocketChannel accept() failed");
        return null;
      }
      
      // make the SocketChannel non-blocking and return it
      newSocketChannel.configureBlocking(false);
			return newSocketChannel;
		}

		// want to catch and handle certain SSL exceptions here
		// letting other exceptions be visible to invoker
/*
		Socket privsock = null;
		try {
			privsock = dSocket.accept();
		}
		catch (SSLException sslex) {
			System.out.println ("NEWSocket.getDataSocket(): ssl exception \"" + sslex.getMessage () + "\"");
			System.exit(1);
		}
		if (SafeSocket (privsock)) {
			return (privsock);
		}
		else {
			return (null);
		}
*/
    return null;
	}

  /**
   * Accesses the InetAddress for a TCP
   * @returns the address
   */
	public static InetAddress getDataSocketChannelInetAddress ()
	{
		return dataSocketChannel.socket().getInetAddress ();
	}

  /**
   * Accesses the local port for a TCP data socket
   * @returns the port
   */
	public static int getDataSocketChannelLocalPort ()
	{
		return dataSocketChannel.socket().getLocalPort ();
	}

  /**
   * Closes the server socket (socket factory)
   * @throws IOException
   */
	public static void closeDataSocketChannel () throws IOException
	{
		if(dataSocketChannel != null)dataSocketChannel.socket().close();
	}


  /**
   * checks a connection to declare the Socket 'safe' to use
   * @param s the socket to test
   * @returns true if the socket is safe
   * @throws Exception
   */
/*
	public static boolean SafeSocket (Socket s) throws Exception
	{
		int chainlength = 0;
		int i,j,k;
		int invalid_dates = 0;
		int vfytest = 0;
		boolean retval = false;
		long serial_number = -1;
		String cert_subjectDN = "";
		StringTokenizer cert_tokens;
		String[] cert_fields = new String[4*X509_MAXATTR];

		X509Certificate[] xcert;

		if (s == null) {
			System.out.println("SafeSocket(): can't verify null socket");
			return (false);
		}

		SSLSession sess = ((SSLSocket) (s)).getSession ();
		String clientName = sess.getPeerHost ();
		System.out.println("SafeSocket(" + VerifyFlags + "): clientName is \"" + clientName + "\" from session" +
			sess);

		try {
			xcert = sess.getPeerCertificateChain ();
		} catch (SSLPeerUnverifiedException e) {	// or do we let it blow up?
			System.out.println (msgTime() + "NEWSocket(): exception creating privacy socket: " + e);
			return (false);
		}

		chainlength = xcert.length;	// should be 1 for our first simple case
*/
     /************************************************************************
      *
      * Begin validity checks on certificates in the cert chain received from
      * the client. This section can grow to include any number of such
      * checks; this version implements only a few rudimentary ones, and
      * largely assumes there is only one certificate in the chain, that being
      * the one provided to the student by the server at start of course.
      *
      ************************************************************************/
/*
		// check validity of dates on certs: count number of out-of-range

		for (i = 0; i < chainlength; i++) {
			try {
				xcert[i].checkValidity ();
			}
			catch (CertificateExpiredException cnv) {
				invalid_dates++;
			}
			catch (CertificateNotYetValidException cnv) {
				invalid_dates++;
			}
		}
		if (invalid_dates > 0) {	// no point doing more work
			return (false);
		}


		if (VerifyFlags == 0) { // there are no further checks to perform
			return(true);
		}
*/
		/*  we're here because there are additional verification checks
		 *  to perform on the certificate before we're willing to accept it.
		 *
		 *  all subsequent checks depend on getting the cert serial number
		 *  of the cert we issued so can look it up in server (acting as issuer) db
		 */

		/*  ensure that we are only going to check for things we have: the NEW db
		 *  entries contain a superset of what the cert gives us
		 */

		/* VerifyFlags &= VERIFY_REF;  */
/*
		for (i = 0; i < chainlength; i++ ) {

			vfytest = 0;
			retval = false; // pessimistic default: cert will fail tests

			serial_number = (xcert[i].getSerialNumber()).longValue();
			cert_subjectDN = (xcert[i].getSubjectDN()).toString();
			System.out.println("SN[" + i + "]:" + serial_number + " has DN \"" + cert_subjectDN + "\"");

			// tokenize the distinguished name string from the cert
			// for ease of use in tests to come, below

			cert_tokens = new StringTokenizer(cert_subjectDN,"=,");
			j = 0;
			while (cert_tokens.hasMoreTokens()) {
				cert_fields[j++] = cert_tokens.nextToken();
				if (j > 20 ) {
					System.out.println("SafeSocket(): cert has too many attributes " + j);
					return(false);
				}
			}

			// throw-away diagnostic dumps out the array from the tokenizing
			// for (k = 0; k < j; k++) {
			//	System.out.println("[" + k + "] is \"" + cert_fields[k] + "\"");
			//	System.out.flush();
			// }

			// get the certificate info from the db for comparison

			TLM.certificateInfo ci = Auth.getCertificateInfo(serial_number);
			// System.out.println("ci returned for sernum " + serial_number + " is: " + ci);

			if (ci == null) {
				System.out.println("SafeSocket(): can't get db info for cert " + serial_number);
				// return(true); %% ONLY DO THIS FOR DEVEL/DEBUG !!!
				continue; // in case cert chain length > 1, keep looking for our's
			}

			String cert_db_course      = ci.getcourseID();
			String cert_db_username    = ci.getUsername(); // not in cert
			String cert_db_commonname  = ci.getCommonName();
			String cert_db_emailaddr   = ci.getEmailAddr();
			String cert_db_orgunit     = ci.getOrgUnit();
			String cert_db_orgname     = ci.getOrgName();
			String cert_db_orglocation = ci.getOrgLoc();
			String cert_db_orgstate    = ci.getOrgState();
			String cert_db_orgcountry  = ci.getOrgCountry();

			if (cert_db_emailaddr.compareTo(cert_fields[1]) == 0) vfytest |= EMAIL_ADDR;
			if (cert_db_commonname.compareTo(cert_fields[3]) == 0) vfytest |= COMMONNAME;
			if (cert_db_orgunit.compareTo(cert_fields[5]) == 0) vfytest |= ORG_UNIT;
			if (cert_db_orgname.compareTo(cert_fields[7]) == 0) vfytest |= ORG_NAME;
			if (cert_db_orglocation.compareTo(cert_fields[9]) == 0) vfytest |= ORG_LOC;
			if (cert_db_orgstate.compareTo(cert_fields[11]) == 0) vfytest |= ORG_STATE;
			if (cert_db_orgcountry.compareTo(cert_fields[13]) == 0) vfytest |= ORG_COUNTRY;

			System.out.println("vfy result is " + vfytest + " out of " + VERIFY_REF);

			if ((vfytest | VERIFY_REF) != VERIFY_REF) { // should xor so result would be mismatch
				return(false);
			}
			else {
				retval = true;
				break;
			}

		}

		if (retval == true) {
			System.out.println(msgTime() + "NEWSocket(): socket is safe/client cert valid"); // debug msg
		}

		return (retval);
	}
*/

	private static String msgTime()
	{
		Date dnow = new Date();
		String nowstr = msgTimeFormat.format(dnow);

		return(nowstr + " | ");

	}

}	// end NEWSocketChannel class
