/*
 * <p>Title: ParseConfig</p>
 * <p>Description: Parses XML configuration file for Transport Layer Multicaster</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: GMU NETLAB</p>
 * @author JMP
 * @version 5.2.4
 * @modified JMP 3/1/13
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;
import java.sql.*;
import java.nio.channels.SocketChannel;

// the XML schema is embedded in method calls below; which also
// allocates the host table (all per-host variables)
public class ParseConfig extends TLM
{
  /**
   * max number of characters allowed in a token (including comments)
   */
  static final int MAX_TOKEN = 1000;

  /**
   * max number of characters allowed in a hostname
   */
  static final int MAX_HOST = 255;

  /**
   * max number of characters allowed in a directory pathname
   */
  static final int MAX_PATH = 255;

  /**
   * max number of characters allowed in a token
   */
  static final int MAX_CMD = 100;

  /**
   * max number of characters in a generic name
   */
  static final int MISC_NAME_MAX = 30;

  /**
   * max number of characters in a course ID
   */
  static final int COURSE_NAME_MAX = 11;

  /**
   * symbolic failure and sucesss codes
   */
  static final boolean ERROR_CODE = true, OK_CODE = false;

  /**
   *  holds the String value associated with latest leaf node scanned
   */
  static String pcData;

  /**
   *  holds the value of the latest tag scanned
   */
  static String parsedTag;

  /**
   *  the configuration file
   */
  static BufferedReader configFile;

  /**
   * this is an adaptation of the simple XML parser in TLMC; each node
   * of the valid XML configuration tree is parsed with a call to parseTag
   *
   * @parameters: none
   *
   * @returns: false if no parse error; true otherwise
   */

  static boolean parseTlmConfig(String configFileName)
  {
    // open the configuration file
    try
    {
      configFile = new BufferedReader(new FileReader(configFileName));
    }
    catch(FileNotFoundException fnfe)
    {
      System.out.println("configuration file " + configFileName + " not found");
      stopTLM = true;
    }

    // <?xml version="1.0"?>
    // <!-- This .conf file configures a live conference/course server for the
    //      Network EducationWare (NEW) Transport Layer Multicaster (TLM)
    //      http://netlab.gmu.edu/NEW -->
    if(parseTag("<?xml version=\"1.0\"?>",
                 "XML header ",
                 false, false, MAX_CMD))
      return ERROR_CODE;

    // <NEW-Server-Config>
    if(parseTag("<NEW-Server-Config>",
                "TLM header ",
                true, false, MAX_CMD))
      return ERROR_CODE;

    // <!-- next line identifies this configuration file version and is not processed  --->
    // <version>040000</version>
    if(parseTag("<version>",
                "<NEW-Server-Config>",
                true, true, 6))
       return parseError("missing version number");
    configurationVersion = new String(pcData);
    System.out.println("TLM running configuration file version:" +
                       configurationVersion);

    // <!-- next line is mandatory and defines classrom/conference ID -->
    // <CID>DEMO</CID>
    if(parseTag("<CID>",
                "<NEW-Server-Config>",
                true, true, MISC_NAME_MAX))
       return parseError("classroom/conference ID is required");
    CID = convertPcData();
    roomID = new String(pcData);
   
    // pad roomID to 11 chars with leading zeros
    while(roomID.length() < 11)roomID = '0' + roomID;
    roomIDBytes = roomID.substring(0,11).getBytes();
    System.out.println("conference/classroom ID:" + roomID);

    // <!-- next line is mandatory; 0 or 1 indicates whether privacy feature is used -->
    // <privateTunnels>0</privateTunnels>
    if(parseTag("<privateTunnels>",
                "<NEW-Server-Config>",
                true, true, 1))
      return parseError("privacy status is required");
    useSSL = !pcData.equals("0");
    if(useSSL)System.out.println("using only private tunnels to clients");
    else System.out.println("not using private tunnels to clients");
    	
    // <!-- next line is mandatory:
    //      "none" for no authentication
    //      "file" for file-based passwords, default filename CID.auth
    //      "database" for database passwords -->
    // <authenticationSource>database</authenticationSource>
    if(parseTag("<authenticationSource>",
                "<NEW-Server-Config>",
                true, true, 8))
      return parseError("authentication source is required");
    if(pcData.equals("none"))
    {
      System.out.println("no authentication will be used");
      checkPassword = false;
    }
    else if(pcData.equals("file"))
    {
      // flag operation without database
      noDatabase = true;

      // set default authfileName, which might change below
      authFileName = roomID + ".auth";

      // <!-- next line is optional and used with file-based authentication;
      //      path-file-name or file-name in startup directory;
      //      default CID.auth -->
      // <authenticationFile>/var/config/foo.txt</authenticationFile>
      if(parseTag("<authenticationFile>",
                  "<NEW-Server-Config>",
                  false, true, MAX_HOST))
        return true;
      if(!pcData.equals(""))authFileName = new String(pcData);
      System.out.println("authentication will be based on file " +
                         authFileName);
    }
    else if(pcData.equals("database"))
    {
      // <!-- next line is mandatory if using database authentication; 
      //      it is used by Moodle -->
      // <databasePrefix>iMDL_</databasePrefix>
      if(parseTag("<databasePrefix>",
                  "<NEW-Server-Config>",
                  false, true, MISC_NAME_MAX))
          return parseError("databasePrefix too long");
        databasePrefix = new String(pcData);
      
      // <!-- next line is optional and used with database authentication;
      //      DNS name of host where database runs; default localhost -->
      // <databaseHost>foobar.stateu.edu</databaseHost>
      if(parseTag("<databaseHost>",
                  "<NEW-Server-Config>",
                  false, true, MAX_HOST))
        return true;
      if(pcData.length() > 0)databaseHost = new String(pcData);

      // <!-- next line is mandatory if using database authentication -->
      // <databaseUsername>username</databaseUsername>>
      if(parseTag("<databaseUsername>",
                  "<NEW-Server-Config>",
                  true, true, MISC_NAME_MAX))
        return parseError("databaseUsername required");
      databaseUsername = new String(pcData);
      
      // <!-- next line is mandatory if using database authentication -->
      // <databasePassword>password</databasePassword>
      if(parseTag("<databasePassword>",
                  "<NEW-Server-Config>",
                  false, true, 255))
        return parseError("databasePassword required");
      databasePassword = new String(pcData);
      
      // <!-- next line is mandatory if using database authentication -->
      // <databaseName>
      if(parseTag("<databaseName>",
                  "<NEW-Server-Config>",
                  true, true, MISC_NAME_MAX))
        return parseError("databaseName required");
      databaseName = new String(pcData);
            
      // <!-- next line is mandatory if using database passwords:
      //      FC for lookup of passwords provided by floor control
      //      WEB for webserver session variables imbedded in nwc filenames -->
      // <authenticationMode>WEB</authenticationMode>
      if(parseTag("<authenticationMode>",
                  "<NEW-Server-Config>",
                  true, true, 3))
        return parseError("authenticationMode required");
      if(pcData.toUpperCase().equals("FC"))
      {
        webAuthenticate = false;
        System.out.println("authentication to come from floor control panel");
      }
      else if(pcData.toUpperCase().equals("WEB"))
      {
        webAuthenticate = true;
        System.out.println(
            "authentication to come from Webserver via session variable");
      }
      else return parseError("authenticationMode must be FC or WEB");
      System.out.println("authentication will be based on database " +
                         databaseName);
    }
    // <!-- next line is optional; 3-character semester code -->
    // <semesterCode>CFN</semesterCode>
    if(parseTag("<semesterCode>",
                "<NEW-Server-Config>",
                false, true, 3))
      return true;
    if(pcData.length() > 0)
      thisSemester = new String(pcData.toUpperCase());

    // convert thisSemester to bytes even if its blank
    thisSemesterBytes = thisSemester.getBytes();
    System.out.println("semester code set to " + thisSemester);

    // <!-- next line is optional; max application data download zipfile
    //      server will support; default is 0  -->
    // <appFileDownloadMaxMB>0</appFileDownloadMaxMB>
    if(parseTag("<appFileDownloadMaxMB>",
                "<NEW-Server-Config>",
                false, true, MAX_HOST))
      return ERROR_CODE;
    if(parsedTag.equals("<appFileDownloadMaxMB>"))
    {
	    long inputNumber = convertPcData()*1000000;
	    if(inputNumber > DOWNLOAD_MAX_SIZE)
	    {
		    System.out.println("<appFileDownloadMaxMB> exceeds allowable range (0 to " +
		                       DOWNLOAD_MAX_SIZE + ") - set to zero");
		    inputNumber = 0;
      }
      appFileDownloadLimit = (int)inputNumber;
      System.out.println("max application data zipfile download:" + appFileDownloadLimit);
      appFileDownload = new byte[appFileDownloadLimit];

      // <!-- next line is required if and only if previous line is present;
      //   path to COURSES directory on server (to locate zipfile, append to this
      //   CouseID/APPDATA/appZipFile.zip  -->
      // <appFileCoursesDirectory></appFileCoursesDirectory>
      if(parseTag("<appFileCoursesDirectory>",
                  "<NEW-Server-Config>",
                  true, true, MAX_PATH))
      return parseError(
		    "directory path of appDownloadFile is required if download is supported" +
		    " or server-side recording is configured");
	      appFilePath = new String(pcData);

	    // if final slash is missing from path, add it
	    if(!appFilePath.endsWith("/") && !appFilePath.endsWith("\\"))
  	    appFilePath += "/";
	    System.out.println("application file download path:" + appFilePath);
    }

    // <!-- next line is mandatory; integer indicates number of applications supported -->
    // <numberOfApplications>3</numberOfApplications>
    if(parseTag("<numberOfApplications>",
                "<NEW-Server-Config>",
                true, true, 1))
      return parseError("count of applications supported is required");
    if((numberOfApplications = convertPcData()) < 0)return ERROR_CODE;
    System.out.println("number of applications to multicast:" + numberOfApplications);

    // allocate arrays based on numberOfApplications
    // make one extra space in uPort for rpTCPPort
    udpPort = new int[numberOfApplications+1];
    mcPort = new int[numberOfApplications];
    mcGroup = new InetAddress[numberOfApplications];
    udpSocket1Save = new DatagramSocket[numberOfApplications];
    udpSocket2Save = new DatagramSocket[numberOfApplications];
    mcSocket1Save = new MulticastSocket[numberOfApplications];
    mcSocket2Save = new MulticastSocket[numberOfApplications];

    // allocate and zero floor control arrays 1 to 9 (0 for orphans)
    breakoutFloorHolder = new int[MAX_BREAKOUT_GROUPS+1];
    breakoutFloorHolderName = new String[MAX_BREAKOUT_GROUPS+1];
    for(int i=0; i<=MAX_BREAKOUT_GROUPS; ++i)
	  {
	    breakoutFloorHolder[i] = -1;
	    breakoutFloorHolderName[i] = "********";
	  }

    // <!-- next line is mandatory; integer indicates max user connections including
    //      instructor, controller, and any slots reserved for specific IP address -->
    // <maxConnections>10</maxConnections>
    if(parseTag("<maxConnections>",
                "<NEW-Server-Config>",
                true, true, 2))
      return parseError("missing max connections count supported <maxConnections>");
    if((maxConnections = convertPcData()) < 0)return ERROR_CODE;
    if(maxConnections> MAX_HOST)
      return parseError("maximum connections " + maxConnections +
                        " exceeds server capacity");
    System.out.println("max simultaneous connections supported:" + maxConnections);

    // allocate arrays based on maxConnections
    hostInetAddr = new InetAddress[ maxConnections ];
    hostAddr = new String[ maxConnections ];
    sourceVideoPort = new int[ maxConnections ];
    sessionVariable = new String[ maxConnections ];
    otherEndPort = new int[ maxConnections ];
    udpActive = new boolean[ maxConnections ];
    anonymousHost = new boolean[ maxConnections ];
    setupReceived = new boolean[ maxConnections ];
    authenticatedHost = new boolean[ maxConnections ];
//debugx    previouslyAuthenticatedHost = new boolean[ maxConnections ];
    loopbackMinRTT = new int[ maxConnections ];
    latestRTTIncrease = new int[ maxConnections ];
    audioDropped = new int[ maxConnections ];
    whiteboardDropped = new int[ maxConnections ];
    lastErrorMessageType = new byte[ maxConnections ];
    videoDropped = new int[ maxConnections ];
    multicastDropped = new int[ maxConnections ];
    username = new String[ maxConnections ];
    sourcePort = new int[ maxConnections ];
    sourceVideoPort = new int[ maxConnections ];
    fullname = new String[ maxConnections ];
    userCourseID = new String [ maxConnections ];
    loginTime = new long[ maxConnections ];
    controlPrivilege = new char[ maxConnections ];
    instructorPrivilege = new int[ maxConnections ];
    drawGrid = new boolean[ maxConnections ];
    killConnection = new boolean[ maxConnections ];
    tunnelIsActive = new boolean[ maxConnections ];
    tcpIsOpen = new boolean[maxConnections];
    connectedByTCP = new boolean[ maxConnections ];
    connectionTestPassed = new boolean[ maxConnections ];
    connectedVideoByTCP = new boolean[ maxConnections ];
    udpLoopbackCount = new int[ maxConnections ];
    udpSetupTime = new long[ maxConnections ];
    uRecvTime = new long[ maxConnections ];
    fRecvTime = new long[ maxConnections ];
    fcIsMc = new boolean[ maxConnections ];
    peakRTT = new int[ maxConnections ];
    lastWarningTime = new long[ maxConnections ];
    tcpSocketChannels = new SocketChannel[ maxConnections ];
    tcpTunnelInstance = new TCPTunnel[ maxConnections ];
	  tcpTunnelThreads = new Thread[ maxConnections ];
    videoThreads = new Thread[ maxConnections ];
	  for(int i=0; i<maxConnections; ++i)
	  {
	    tcpTunnelInstance[i] = null;
	    tcpTunnelThreads[i] = null;
      videoThreads[i] = null;
    }
    udpSockets = new DatagramSocket[ maxConnections ];
    tcpVideoSocketChannels = new SocketChannel[ maxConnections ];
    sendThreads = new Thread[ maxConnections ];
    tunnelSends = new TunnelSend[ maxConnections ];
	  tcpHandoffs = new TCPHandoff[ maxConnections ];
    fcSockets = new Socket[ maxConnections ];
    fcOutStreams = new BufferedOutputStream[ maxConnections ];
    fcInStreams = new BufferedInputStream[ maxConnections ];
    numAuthenticateTries = new int[ maxConnections ];
    audioQueues = new byte[ maxConnections ][ audioRingSize ][ MAX_MESSAGE+5 ];
    fcQueues = new byte[ maxConnections ][ fcRingSize ][ MAX_MESSAGE+5 ];
    videoQueues = new byte[ maxConnections ][ videoRingSize ][ MAX_MESSAGE+5 ];
    whiteboardQueues = new byte[ maxConnections ][ whiteboardRingSize ][ MAX_MESSAGE+5 ];
    multicastQueues = new byte[ maxConnections ][ multicastRingSize ][ MAX_MESSAGE+5 ];
    audioLengths = new int[ maxConnections ][ audioRingSize ];
    fcLengths = new int[ maxConnections ][ fcRingSize ];
    videoLengths = new int[ maxConnections ][ videoRingSize ];
    whiteboardLengths = new int[ maxConnections ][ whiteboardRingSize ];
    multicastLengths = new int[ maxConnections ][ multicastRingSize ];
    audioWindowStart = new int[ maxConnections ];
    audioWindowEnd = new int[ maxConnections ];
    fcWindowStart = new int[ maxConnections ];
    fcWindowEnd = new int[ maxConnections ];
    videoWindowStart = new int[ maxConnections ];
    videoWindowEnd = new int[ maxConnections ];
    whiteboardWindowStart = new int[ maxConnections ];
    whiteboardWindowEnd = new int[ maxConnections ];
    multicastWindowStart = new int[ maxConnections ];
    multicastWindowEnd = new int[ maxConnections ];
    udpSignalSlideDone = new boolean[ maxConnections ];
    tunnelClosing = new boolean[ maxConnections ];
    localFloorStatus = new int[ maxConnections ];
    netCapacitykbps = new int[ maxConnections ];
    nameExistsAgain = new boolean[ maxConnections ];
    fcIsInitialized = new boolean[ maxConnections ];
    fcIsReadyForAppData = new boolean[ maxConnections ];
    appFileMtime = new int[ maxConnections ];
    appFileSize = new int[ maxConnections ];
    appFileStatNew = new boolean[ maxConnections ];
    unzipIsReady = new boolean[ maxConnections ];
    breakoutGroup = new int[ maxConnections ];
	  reconnected = new boolean[ maxConnections ];
    protocolIsTCP = new boolean[ maxConnections ];
    thisHostUpload = new UploadRecording[ maxConnections ];
    thisHostUploadThread = new Thread[ maxConnections ];
    recordDownloadMessagesSent = new int[ maxConnections ];
    recordDownloadAcksReceived = new int[ maxConnections ];

    // <!-- next line is optional; 0 or 1 (default 0) indicates whether TLM turns service
    //      off and on to individual classes, based on database table "courselist" -->
    // <autoSchedule>1</autoSchedule>
    if(parseTag("<autoSchedule>",
                "<NEW-Server-Config>",
                false, true, 1))
      return true;
    if(parsedTag.equals("<autoSchedule>"))
      autoSchedule = pcData.equals("1");
    if(autoSchedule)System.out.println("server will schedule sessions using database");
    else System.out.println("server running single continuous session (no autoschedule)");

    // <!-- next line is optional; 0 or 1 (default 0) indicates whether TLM listens on the
    //      application UDP ports and sends received messages to all configured receivers
    //      except the sender (value 1 is inconsistent with privateTunnels value 1) -->
    // <listenOnApplicationUDP>1</listenOnApplicationUDP>
    if(parseTag("<listenOnApplicationUDP>",
                "<NEW-Server-Config>",
                false, true, 1))
      return true;
    if(parsedTag.equals("<listenOnApplicationUDP>"))
      retransmitUDP = pcData.equals("1");
    else retransmitUDP = false;
    if(useSSL & retransmitUDP)
      parseError("privateTunnels and listenOnApplicationUDP are incompatible options");
    if(retransmitUDP)System.out.println("listening to/forwarding local UDP traffic");
    else System.out.println("not listening to/forwarding local UDP traffic");

    // <!-- next line is optional; value 0 or 1 indicates whether password is required;
    //      default 1 -->
    // <passwordRequired>0</passwordRequired>
    checkPassword = true;
    if(parseTag("<passwordRequired>",
                "<NEW-Server-Config>",
                false, true, 1))
      return ERROR_CODE;
    if(!parsedTag.equals("<passwordRequired>"))
      checkPassword = !pcData.equals("0");
    if(!checkPassword)System.out.print("not ");
    System.out.println("checking passwords against database or .auth file");

    // <!-- next line is optional code for control status of floor at startup
    //      values 1 (ALWAYS YIELD) 2 (NEVER YIELD) 3 (ASK CONTROLLER) -->
    // <initialFloorStatus>3</initialFloorStatus>
    initialFloorStatus = 3;
    if(parseTag("<initialFloorStatus>",
                "<NEW-Server-Config>",
                false, true, 1))
      return ERROR_CODE;
    if(parsedTag.equals("<initialFloorStatus>"))
    {
      int convertFloorStatus;
      if((convertFloorStatus = convertPcData()) < 0)return ERROR_CODE;
      initialFloorStatus = (byte)convertFloorStatus;
    }
    floorStatus = initialFloorStatus;
    switch(initialFloorStatus)
    {
      case 1:System.out.println("initial floor status 1 (always release floor on request)");
        break;
      case 2:System.out.println("initial floor status 2 (never release floor on request)");
        break;
      case 3:
      default:System.out.println("initial floor status 3 (ask controller on floor request)");
    }

    // <!-- next line is optional value in ms of timeout on floor control delay when
    //      slide is being sent; default is 15000 ms  -->
    // <floorControlSlideDelay>15000</floorControlSlideDelay>
    fcSlideSendTimeout = 15000;
    if(parseTag("<floorControlSlideDelay>",
                "<NEW-Server-Config>",
                false, true, 6))
      return ERROR_CODE;
    if(parsedTag.equals("<floorControlSlideDelay>"))
    {
      int convertSlideTimeout;
      if((convertSlideTimeout = convertPcData()) < 0)return ERROR_CODE;
      fcSlideSendTimeout = convertSlideTimeout;
    }
    System.out.println("max floor control delay when sending a slide:" +
                       fcSlideSendTimeout);

    // <!-- next line is optional; integer between 1 and 50 defines number of grid rows
    //      optional, if present must occur in this position; defaults to maxConnections -->
    // <fcMaxUsers>20</fcMaxUsers>
    fcMaxUsers = (byte)maxConnections;
    if(parseTag("<floorControlMaxUsers>",
                "<NEW-Server-Config>",
                false, true, 2))
      return ERROR_CODE;
    if(parsedTag.equals("<floorControlMaxUsers>"))
    {
      int convertMaxUsers;
      if((convertMaxUsers = convertPcData()) < 0)return ERROR_CODE;
      fcMaxUsers = (byte)convertMaxUsers;
    }
    if(fcMaxUsers < 1)fcMaxUsers = 1;
    if(fcMaxUsers > 50)fcMaxUsers = 50;
    System.out.println("number of users row in floor control:" + fcMaxUsers);

    // <!-- next line is optional; provides sending rate in kbps for network test;
    //      user will be disconnected by TLMC if this rate can't be sustained in test;
    //      default is 25 kbits/sec; range 25 to 500 (the associated time interval is
    //      1.25*50000/minimumRate = 2500 ms; this is called "connectionThreshold") -->
    // <minRatekbps>40</minRatekbps>
    minimumRate = 25;
    if(parseTag("<minRatekbps>",
                "<NEW-Server-Config>",
                false, true, 3))
      return ERROR_CODE;
    if(parsedTag.equals("<minRatekbps>"))
      if((minimumRate = convertPcData()) < 0)return ERROR_CODE;
    if(minimumRate < 25)minimumRate = 25;
    if(minimumRate > 500)minimumRate = 500;
    connectionThreshold = (int)((60000)/(double)minimumRate);
    System.out.println("minimum acceptable server transmit rate:" + minimumRate +
                       " kb/s => connection threshold:" + connectionThreshold + " ms");

    // <!-- next line is optional; provides length of time between received
    // data heartbeats that triggers disconnect -->
    // <disconnectTimeout>10000</disconnectTimeout>
    tlmcDisconnectTime = 10000;
    if(parseTag("<disconnectTimeout>",
                "<NEW-Server-Config>",
                false, true, 6))
      return ERROR_CODE;
    if(parsedTag.equals("<disconnectTimeout>"))
      if((tlmcDisconnectTime = convertPcData()) < 0)return ERROR_CODE;
    if(tlmcDisconnectTime < 2000)tlmcDisconnectTime = 2000;
    if(tlmcDisconnectTime > 30000)tlmcDisconnectTime = 30000;
    System.out.println("data disconnectTimeout:" + tlmcDisconnectTime + " ms");
    if(fcDisconnectTime < tlmcDisconnectTime)fcDisconnectTime = tlmcDisconnectTime;
    System.out.println("floor control disconnect timeout:" + fcDisconnectTime + " ms");
	  
	// <!-- next line is optional; provides length of time in ms server will hold state
	//      waiting for autoReconnect of normal user e.g. student (default 5 minutes) -->
    normalReconnectTimeout = 5*60*1000;// 5 minutes
    if(parseTag("<normalReconnectTimeout>",
				"<NEW-Server-Config>",
                false, true,10))
      return ERROR_CODE;
    if(parsedTag.equals("<normalReconnectTimeout>"))
      if((normalReconnectTimeout= convertPcData()) < 0)return ERROR_CODE;
	System.out.println("normal reconnect timeout:"+normalReconnectTimeout);
	
    // <!-- next line is optional; provides length of time in ms server will hold state
	//      waiting for autoReconnect of instructor (default 3 hours) -->
	instructorReconnectTimeout = 3*60*60*1000;// 3 hours
    if(parseTag("<instructorReconnectTimeout>",
                "<NEW-Server-Config>",
                false, true,10))
	   return ERROR_CODE;
	if(parsedTag.equals("<instructorReconnectTimeout>"))
	  if((instructorReconnectTimeout= convertPcData()) < 0)return ERROR_CODE;
	System.out.println("instructor reconnect timeout:"+instructorReconnectTimeout);
    
    // <!-- next line is optional; provides number of milliseconds increase in RTT that
    //      will trigger throttling of video traffic (at twice this value audio traffic
    //      also is throttled); default is 500 ms; range 100 to 2000 -->
    // <congestionThreshold>300</congestionThreshold>
    congestionThreshold = 500;
    if(parseTag("<congestionThreshold>",
                "<NEW-Server-Config>",
                false, true, 4))
      return ERROR_CODE;
    if(parsedTag.equals("<congestionThreshold>"))
      if((congestionThreshold = convertPcData()) < 0)return ERROR_CODE;
    if(congestionThreshold < 100)congestionThreshold = 100;
    if(congestionThreshold > 2000)congestionThreshold = 2000;
    System.out.println("congestion threshold:" + congestionThreshold + " ms");

    // <!-- next line is optional; defines interval is ms between keep-alive probes;
    //      default 1000 ms; range 100 to 5000 -->
    // <probeInterval>1000</probeInterval>
    probeInterval = 1000;
    if(parseTag("<probeInterval>",
                "<NEW-Server-Config>",
                false, true, 4))
      return ERROR_CODE;
    if(parsedTag.equals("<probeInterval>"))
      if((probeInterval = convertPcData()) < 0)return ERROR_CODE;
    if(probeInterval < 100)probeInterval = 100;
    if(probeInterval > 5000)probeInterval = 5000;
    System.out.println("probe interval:" + probeInterval + " ms");

    // <!-- next line is mandatory port number for TCP data connection
    //      (TCP control port is one higher) -->
    // <dataTcpPort>62000</dataTcpPort>
    if(parseTag("<dataTcpPort>",
                "<NEW-Server-Config>",
                true, true, 5))
      return parseError("data tunnel TCP port is required");
    if((tunnelTcpPort = convertPcData()) < 0)return ERROR_CODE;
    System.out.println("data tunnel TCP port:" + tunnelTcpPort);

    // <!-- next line is optional port number for UDP data connections -->
    // <dataUdpPort>62000</dataUdpPort>
    if(parseTag("<dataUdpPort>",
                "<NEW-Server-Config>",
                false, true, 5))
      return ERROR_CODE;
    if(parsedTag.equals("<dataUdpPort>"))
    {
      if ((tunnelUdpPort = convertPcData()) < 0)
        return ERROR_CODE;
      System.out.println("data tunnel UDP port:" + tunnelUdpPort);
    }
    // <!-- next line is optional multicast group for UDP floor control (if not
    //      present floor control will attempt to connect via TCP at dataTcpPort+1) -->
    // <fcMulticastAddress>225.10.10.10</fcMulticastAddress>
    if(parseTag("<fcMulticastAddress>",
                "<NEW-Server-Config>",
                false, true, MAX_HOST))
      return ERROR_CODE;
    if(parsedTag.equals("<fcMulticastAddress>"))
    {
      if ( (mcFcInetAddr = convertInetAddress()) == null)
        return ERROR_CODE;
      System.out.println("Floor Control multicast address:" + mcFcInetAddr);
    }

    // <!-- next line is optional UDP port number for multicast floor control; default
    //      dataTcpPort+1 or dataUdpPort+1;ignored if fcMulticastAddress is not defined -->
    // <fcMulticastPort>22001</fcMulticastPort>
    if(parseTag("<fcMulticastPort>",
                "<NEW-Server-Config>",
                false, true, 5))
      return ERROR_CODE;
    if(parsedTag.equals("<fcMulticastPort>"))
    {
      if ((mcFcPort = convertPcData()) < 0)
        return ERROR_CODE;
      System.out.println("floor control port:" + mcFcPort);
    }
    else
    {
      if( tunnelTcpPort != 0 )mcFcPort = tunnelTcpPort+1;
      else if( tunnelUdpPort != 0 )mcFcPort = tunnelUdpPort+1;
      System.out.println("default multicast floor control port:" + mcFcPort);
    }
	
	//   <!-- next line 1 or 0 indicating whether to run a recorder
    //         as part of the server; if configured it will run only
    //         during auto-scheduled classes -->
	if(parseTag("<recordOnServer>",
				"<NEW-Server-Config>",
				false, true, 5))
      return ERROR_CODE;
	if(parsedTag.equals("<recordOnServer>"))
	{
	  if((recordOnServerConfig = convertPcData()) < 0)
	    return ERROR_CODE;
	  System.out.println("server-side recording configured");
          recordOnServer = true;	  
          
	  // <!-- next line is mandatory if but only used if recordOnServer 
	  //      is present; it indicates the directory where serverside 
	  //      recordings are created -->
	  // <serverRecordingsDirectory>
	  //   /var/NEW/S10/serverRecordings
	  // </serverRecordingsDirectory>
      if(parseTag("<serverRecordingsDirectory>",
                  "<NEW-Server-Config>",
                  true, true, MAX_PATH))
      return parseError(
		"directory path of serverRecordingsDirectory is required if server recording is supported");
	  recordingsDirectory = new String(pcData);
    recordingsDownloadDirectory = recordingsDirectory + '/' + (new Integer(CID)).toString();
    System.out.println("recordingsDownloadDirectory is:" + recordingsDownloadDirectory);

    // <!-- next line is optional and only used if recordOnServer 
    //      is present; it indicates that the recording is to include
    //      video. Default is 0 (do not record) -->
	  if(parseTag("<recordVideo>",
	              "<NEW-Server-Config>",
                      false, true, 5))
        return ERROR_CODE;
	  if(parsedTag.equals("<recordVideo>"))
	  {
	    int dataValue = convertPcData();
	    if(dataValue < 0)
	      return ERROR_CODE;
            recordVideo = (dataValue == 1);
	  }
	  else
	    System.out.println("server-side video recording not configured");
  }
    
    
    // <multicastTunnelGroup>225.0.0.1</multicastTunnelGroup>
    if(parseTag("<multicastTunnelGroup>",
           "<NEW-Server-Config><applications>",
           false, true, MAX_HOST))
      return ERROR_CODE;
    if(parsedTag.equals("<multicastTunnelGroup>"))
    {
      if ( (multicastTunnelGroup = convertInetAddress()) == null)
        return ERROR_CODE;      
	    System.out.println("multicast application tunnel group address:" +
						 multicastTunnelGroup);

      // <multicastTunnelPort>7777</multicastTunnelPort>
      if (parseTag("<multicastTunnelPort>",
                   "<NEW-Server-Config>",
                   true, true, 5))
      {
        System.out.println(
            "<multicastTunnelPort> is mandatory with <multicastTunnelGroup>");
          return ERROR_CODE;
      }
      if ( (multicastTunnelPort = convertPcData()) < 0)
		    return ERROR_CODE;
       System.out.println("multicast application tunnel port:" +
                           multicastTunnelPort);
       multicastTunnel = true;

       // <multicastTunnelServerSubnet>0</multicastTunnelServerSubnet>
       if(parseTag("<multicastTunnelServerSubnet>",
                   "<NEW-Server-Config>",
                   false,true, 1))
         return ERROR_CODE;
       if(parsedTag.equals("<multicastTunnelServerSubnet>"))
       {
         int dataValue = convertPcData();
         if(dataValue < 0)
           return ERROR_CODE;
         multicastTunnelServerSubnet = (dataValue == 1);
       }
       else
         System.out.println("multicast tunnel on server subnet not configured");
      
       // <multicastTunnelFloorControl>0</multicastTunnelFloorControl>
              if(parseTag("<multicastTunnelFloorControl>",
                   "<NEW-Server-Config>",
                   false,true, 1))
         return ERROR_CODE;
       if(parsedTag.equals("<multicastTunnelFloorControl>"))
       {
         int dataValue = convertPcData();
         if(dataValue < 0)
           return ERROR_CODE;
         multicastTunnelFloorControl = (dataValue == 1);
       }
       if(multicastTunnelFloorControl)
         System.out.println("multicast tunnel is floor-controlled");
       else
         System.out.println("multicast tunnel is not floor-controlled");
 
    } // end if(parsedTag.equals("<multicastTunnelGroup>"))
    else System.out.println("multicast application tunnel not configured");
	
    // <!-- the next tag is mandatory -->
    // <applications>
    // <!-- the application groups must match numberOfApplications in count; each
    //     must have a unicast UDP port designated; the multicast address/port is
    //     optional; floorControlled is optional and defaults to 1 -->
    if(parseTag("<applications>",
               "<NEW-Server-Config>",
               true, false, 0))
      return ERROR_CODE;
    System.out.println("application configurations:");

    // <audioTool>
    if(parseTag("<audioTool>",
               "<NEW-Server-Config><applications>",
               true, false, 0))
      return ERROR_CODE;
    System.out.println("audio:");
    int applicationNumber = 0;

    // <unicastPort>62004</unicastPort>
    if(parseTag("<unicastPort>",
                "<NEW-Server-Config><applications><audio>",
                true, true, 5))
      return ERROR_CODE;
    if((audioPort = convertPcData()) < 0)return ERROR_CODE;
    udpPort[applicationNumber] = audioPort;
    System.out.println("audio client UDP port:" + audioPort);

    // <multicastAddress>225.10.10.4</multicastAddress>
    mcGroup[applicationNumber] = null;
    mcPort[applicationNumber] = 0;
    if(parseTag("<multicastAddress>",
                "<NEW-Server-Config><applications><audio>",
                false, true, MAX_HOST))
      return ERROR_CODE;
    if(!parsedTag.equals("<multicastAddress>"))
      System.out.println("audio multicast not configured");
    else
    {
      if((mcGroup[applicationNumber] = convertInetAddress()) == null)
        return ERROR_CODE;
      System.out.println("audio multicast address:" + mcGroup[applicationNumber]);
      multicastIsConfigured = true;

      // <multicastPort>62104</multicastPort>
      if (parseTag("<multicastPort>",
                   "<NEW-Server-Config><applications><audio>",
                   true, true, 5))
        return ERROR_CODE;
      if ( (mcPort[applicationNumber] = convertPcData()) < 0)
        return ERROR_CODE;
      System.out.println("audio client multicast port:" +
                         mcPort[applicationNumber]);
    }

    // <floorControlled>1</floorControlled>
    audioFloorControl = true;
    if(parseTag("<floorControlled>",
                "<NEW-Server-Config><applications><audio>",
                false, true, 1))
      return ERROR_CODE;
    if(parsedTag.equals("<floorControlled>"))
    {
      audioFloorControl = !pcData.equals("0");
      if (audioFloorControl)
        System.out.println("audio is floor-controlled");
      else
        System.out.println("audio is not floor-controlled");
    }

    // </audioTool>
    if(parseTag("</audioTool>",
                "<NEW-Server-Config><applications>",
                true, false, 0))
      return ERROR_CODE;

    // <whiteboardTool>
    if(parseTag("<whiteboardTool>",
           "<NEW-Server-Config><applications>",
           false, false, 0))
      return ERROR_CODE;
    if(parsedTag.equals("<whiteboardTool>"))
    {
      applicationNumber++;
      System.out.println("whiteboard:");

      // <unicastPort>62008</unicastPort>
      if(parseTag("<unicastPort>",
                  "<NEW-Server-Config><applications><whiteboard>",
                  true, true, 5))
        return ERROR_CODE;
      if((wbdPort = convertPcData()) < 0)return ERROR_CODE;
      udpPort[applicationNumber] = wbdPort;
      System.out.println("whiteboard client UDP port:" + wbdPort);

      // <multicastAddress>225.10.10.5</multicastAddress>
      mcGroup[applicationNumber] = null;
      mcPort[applicationNumber] = 0;
      if(parseTag("<multicastAddress>",
                  "<NEW-Server-Config><applications><whiteboard>",
                  false, true, MAX_HOST))
        return ERROR_CODE;
      if(!parsedTag.equals("<multicastAddress>"))
        System.out.println("whiteboard multicast not configured");
      else
      {
        if((mcGroup[applicationNumber] = convertInetAddress()) == null)
          return ERROR_CODE;
        if(mcGroup[applicationNumber] == null)
        System.out.println("whiteboard multicast address:" +
                           mcGroup[applicationNumber]);
        multicastIsConfigured = true;

        // <multicastPort>62108</multicastPort>
        if(parseTag("<multicastPort>",
                    "<NEW-Server-Config><applications><whiteboard>",
                    true, true, 5))
          return ERROR_CODE;
        if((mcPort[applicationNumber] = convertPcData()) < 0)
          return ERROR_CODE;
        System.out.println("whiteboard client multicast port:" +
                           mcPort[applicationNumber]);
      }

      // <floorControlled>1</floorControlled>
      wbdFloorControl = true;
      if(parseTag("<floorControlled>",
                   "<NEW-Server-Config><applications><whiteboard>",
                  false, true, 1))
        return ERROR_CODE;
      if(parsedTag.equals("<floorControlled>"))
        wbdFloorControl = !pcData.equals("0");
      if(wbdFloorControl)System.out.println("whiteboard is floor-controlled");
      else System.out.println("whiteboard is not floor-controlled");

      // </whiteboardTool>
      if(parseTag("</whiteboardTool>",
                "<NEW-Server-Config><applications>",
                true, false, 0))
      return ERROR_CODE;

    } // end if(parsedTag.equals("<whiteboardTool>"))
    else System.out.println("whiteboard not configured");

    // <videoTool>
    if(parseTag("<videoTool>",
           "<NEW-Server-Config><applications>",
           false, false, 0))
      return ERROR_CODE;
    if(parsedTag.equals("<videoTool>"))
    {
      applicationNumber++;
      System.out.println("video:");

      // <unicastPort>62012</unicastPort>
      if(parseTag("<unicastPort>",
                  "<NEW-Server-Config><applications><video>",
                  true, true, 5))
        return ERROR_CODE;
      if((videoPort = convertPcData()) < 0)return ERROR_CODE;
      udpPort[applicationNumber] = videoPort;
      System.out.println("video client UDP port:" + videoPort);

      // <multicastAddress>225.10.10.6</multicastAddress>
      mcGroup[applicationNumber] = null;
      mcPort[applicationNumber] = 0;
      if(parseTag("<multicastAddress>",
                  "<NEW-Server-Config><applications><video>",
                  false, true, MAX_HOST))
        return ERROR_CODE;
      if(!parsedTag.equals("<multicastAddress>"))
        System.out.println("video multicast not configured");
      else
      {
        if((mcGroup[applicationNumber] = convertInetAddress()) == null)
          return ERROR_CODE;
        System.out.println("video multicast address:" +
                           mcGroup[applicationNumber]);
        multicastIsConfigured = true;

        // <multicastPort>62112</multicastPort>
        if (parseTag("<multicastPort>",
                     "<NEW-Server-Config><applications><video>",
                     true, true, 5))
          return ERROR_CODE;
        mcPort[applicationNumber] = 0;
        if ( (mcPort[applicationNumber] = convertPcData()) < 0)
          return ERROR_CODE;
        System.out.println("video client multicast port:" +
                           mcPort[applicationNumber]);
      }

      // <floorControlled>1</floorControlled>
      videoFloorControl = true;
      if(parseTag("<floorControlled>",
                   "<NEW-Server-Config><applications><video>",
                  false, true, 1))
        return ERROR_CODE;
      if(parsedTag.equals("<floorControlled>"))
        videoFloorControl = !pcData.equals("0");
      if(videoFloorControl)System.out.println("video is floor-controlled");
      else System.out.println("video is not floor-controlled");

      // </videoTool>
      if(parseTag("</videoTool>",
                 "<NEW-Server-Config><applications>",
                 true, false, 0))
        return ERROR_CODE;

    } // end if(parsedTag.equals("<videoTool>"))
    else System.out.println("video not configured");

    // </applications>
    if(parseTag("</applications>",
                "<NEW-Server-Config>",
                true, false, 0))
      return ERROR_CODE;

    // <!-- next section is optional; defines unicast IP addresses with special roles -->
    // <addresses>
    if (parseTag("<addresses>",
                 "<NEW-Server-Config>",
                 false, false, 0))
      return ERROR_CODE;
    if(parsedTag.equals("<addresses>"))
    {
      // <!-- next line is optional; it provides an address from which the
      //      one and only UDP video input will come -->
      //<soleVideoAddress>129.174.161.241</soleVideoAddress>
      if (parseTag("<soleVideoAddress>",
                   "<New-Server-Config><addresses>",
                   false, true, MAX_HOST))
        return ERROR_CODE;
      if(parsedTag.equals("<soleVideoAddress>"))
      {
        if ( (videoInetAddress = convertInetAddress()) == null)
          return ERROR_CODE;
        System.out.println("sole video source address:" + videoInetAddress);
        initializeHostTableAt(numberOfHosts,true);
        anonymousHost[numberOfHosts] = false;
        hostInetAddr[numberOfHosts] = videoInetAddress;
        hostAddr[numberOfHosts] = videoAddress;
        numberOfHosts++;
        videoAddressConfigured = true;
      }
      else System.out.println("sole video source address not configured");
      // <!-- next line is optional; defines an IP address of another server
      //      chained by UDP to this one; multiple instances allowed -->
      // <chainedServer>128.182.86.113</chainedServer>
      numberOfChainedServers = 0;
      chainedServer = new InetAddress[MAX_CHAINED_SERVERS];
      if (parseTag("<chainedServer>",
                   "<NEW-Server-Config><addresses>",
                   false, true, MAX_HOST))
        return ERROR_CODE;
      if (parsedTag.equals("<chainedServer>"))
          System.out.println("other servers chained to this one:");
      while (parsedTag.equals("<chainedServer>"))
      {
        if((chainedServer[numberOfChainedServers] = convertInetAddress()) == null)
          return ERROR_CODE;
        System.out.println("  " + chainedServer[numberOfChainedServers]);
        numberOfChainedServers++;

        // <chainedServer>131.120.7.41</chainedServer>
        if (parseTag("<chainedServer>",
                     "<NEW-Server-Config><addresses>",
                     false, true, MAX_HOST))
          return ERROR_CODE;
      } // end while (parsedTag.equals("<chainedServer>"))

      // <!-- next line is optional; defines an IP address for which a position in
      //      the multicasting table will be held open, whether used or not;
      //      multiple instances allowed; Usernames listed in the authentication file
      //      (see below) will not be allowed to login to these reserved addresses;
      //      any username not in the authentication file will be accepted, without
      //      checking the password, when connecting from the reserved IP addresses. -->
      //<reservedAddress>129.174.65.118</reservedAddress>
      numberOfReservedAddresses = 0;
      if (parseTag("<reservedAddress>",
                   "<NEW-Server-Config><addresses>",
                   false, true, MAX_HOST))
        return ERROR_CODE;
      if(parsedTag.equals("<reservedAddress>"))
        System.out.println("addresses with reserved host table positions:");
      else System.out.println("no host table positions reserved");
      while(parsedTag.equals("<reservedAddress>"))
      {
        if( numberOfHosts >= maxConnections )
        {
          System.out.println("too many reserved hosts for table (max " +
                             maxConnections + ")");
          return ERROR_CODE;
        }
        if ( (hostInetAddr[numberOfHosts] = convertInetAddress()) == null)
          return ERROR_CODE;
        hostAddr[numberOfHosts] = hostInetAddr[numberOfHosts].getHostAddress();
        System.out.println("  " + hostAddr[numberOfHosts]);
        numberOfHosts++;
        numberOfReservedAddresses++;

        // <reservedAddress>129.174.65.55</reservedAddress>
        if (parseTag("<reservedAddress>",
                     "<NEW-Server-Config><addresses>",
                     false, true, MAX_HOST))
          return ERROR_CODE;
      } // end  while (parsedTag.equals("<reservedAddress>"))

      // </addresses>
      if (parseTag("</addresses>",
                   "<NEW-Server-Config>",
                   true, false, 0))
        return ERROR_CODE;
   } // end if(parsedTag.equals("<addresses>")

    // </<NEW-Server-Config>
    if(parseTag("</NEW-Server-Config>",
                 "",
                 true, false, 0))
      return ERROR_CODE;
    System.out.println("server configuration completed");
    return OK_CODE;
  } // end parseTLMConfig

  /**
   *  holds the current tag being scanned
   */
  static String tagBuffer = "";

  /**
   *  if positive, holds value of a character read but not scanned
   */
  static int unreturnedChar = -1;

  /**
   *  indicates endfile has been read on configFile
   */
  static boolean configFileEOF = false;

  /**
   *  reads in the next token (tag or PCDATA) from configFile
   *
   *  @param   seekingTag indicates the token should be an XML tag
   *
   *  @returns token if found, otherwise nullstring
   */
  static String readToken(boolean seekingTag)
  {
    String token = "";

    // check for disallowed conditions
    if(configFileEOF)return "";

    // start token buffer with unread character
    // (skip whitespace if seeking tag)
    if((unreturnedChar >= 0) && (!seekingTag || ((char)unreturnedChar == '<')))
      token += (char)unreturnedChar;
    try
    {
      while ( (unreturnedChar = configFile.read()) != -1)
      {
        if (token.length() >= MAX_TOKEN)
        {
          System.out.println("configuration file token too long:" + token);
          return "";
        }
        // could be in one of three states here:
        //  1. scanning in a tag
        //  2. scanning to find a tag
        //  3. scanning in pcData
        if (seekingTag)
        {
          // case 1: scanning in a tag
          if(token.length() > 0)
          {
            if (token.charAt(0) == '<')
            {
              token += (char)unreturnedChar;
              if ((char)unreturnedChar == '>')
              {
                unreturnedChar = -1;

                // check for XML comment
                if (token.length() < 7)
                  return token;
                if (!token.startsWith("<!--"))
                  return token;
                if (token.endsWith("-->"))
                  token = "";
              } // end if(unreturnedChar...
            } // end if(token.charAt(0)...
          } // end if(token.length() > 0)

          // case 2: scanning to find a tag
          else
          {
            // seeking tag; found it
            if ( (char) unreturnedChar == '<') token = "<";
            // seeking tag, empty token; keep scanning
            else;
          }
        } // end if(seekingTag)

        // case 3: not seeking tag; scanning in pcData
        else
        {
          if (unreturnedChar != '<')
            token += (char)unreturnedChar;
          else
            return token;
        }
      } // end while((unreturnedChar
    }
    catch(IOException ioe)
    {
      System.out.println("IOException reading configuration file");
      return "";
    }

    // if we come here, have reached EOF without end of last tag
    System.out.println("found end of configuration file before closing tag");
    return "";
  } // end readToken()


  /**
   *  parses a tag in the XML stream
   *
   *  @params templateTag  XML tag to match e.g. <foo>
   *          parentTags   contenation of tags above this one in the XML tree
   *          isMandatory  means the tag is required at this point in a valid config
   *          isLeafNode   means this tag must have an associated value
   *          maxChars     applies to the length of the value
   *
   *  @returns false if the template is parsed successfully, i.e. it finds
   *           <tag>value</tag> if the tag is mandatory; true if error
   *
   *  @side-effect: places the value from a leaf node in pcData when scanned
   */
  static boolean parseTag(String templateTag,
           String parentTags,
           boolean isMandatory,
           boolean isLeafNode,
           int maxChars)
  {
    String compareTag, closingTag;

    // initialize results expected by ParseConfig
    parsedTag = "";
    pcData = "";

    // if tag buffer is empty, read a tag
    if(tagBuffer.length() == 0)
      if((tagBuffer = readToken(true)).length() == 0)
      {
        System.out.println("configuration file is not valid XML");
        return true;
      }

    // if sought tag is mandatory, raise error if not matched
    if(!tagBuffer.equals(templateTag))
    {
      if (isMandatory)
      {
        // can't find the tag;
        // case 1: it is not in the file
        if (tagBuffer.length() == 0)
          return parseError("missing mandatory tag " + templateTag);

        // case 2: the tag in the tagBuffer is bogus
        else
          return parseError("expecting tag " + templateTag +
                            " found unexpected tag " + tagBuffer);
      } // end if(isMandatory)
      return OK_CODE;
    } // end if(!tagBuffer.equals()templateTag))

    // tag just read matches the templateTag - confirm length is OK
    if (tagBuffer.length() >= MAX_CMD)
      return parseError("overlength token in configuration file:" + tagBuffer);

    // tag matches - accept it
    parsedTag = new String(tagBuffer);
    tagBuffer = "";

    // if this tag is not a leaf node or the tag is empty element, just return
    if (!isLeafNode || parsedTag.endsWith("\\>"))
      return OK_CODE;

    // otherwise read the PCDATA associated with the tag
    if ( (pcData = new String(readToken(false))).length() == 0)return ERROR_CODE;

    // check for overlength result
    if (pcData.length() > maxChars)
      return
        parseError("overlength data value in configuration file for tag " +
                   parentTags + templateTag);

    // look for closing tag to the one sought
    closingTag = new String(readToken(true));
    if (closingTag.length() >= MAX_CMD)
      return parseError("overlength tag in configuration file:" + closingTag);
    compareTag = "</";
    if( templateTag.length() != 0)compareTag += templateTag.substring(1);
    if(!compareTag.equals(closingTag))
      return parseError("no closing tag found for " + parentTags + templateTag);
    return OK_CODE;
  } // end parseTag()

  /**
   * prints parameter as error message and returns true to indicate error
   * @returns true
   */
  static boolean parseError(String errorMessage)
  {
    System.out.println("server configuration error:" + errorMessage);
    return ERROR_CODE;
  } // end parseError()

  /**
   * converts current pcData to int; prints message if not possible
   * @returns -1 if convert error; otherwise the converted value
   */
  static int convertPcData()
  {
    try
    {
      return Integer.parseInt(pcData);
    }
    catch(NumberFormatException nfe)
    {
      parseError("invalid numeric data " + pcData);
      return -1;
    }
  } // end convertPcData()

  /**
   * converts current pcData {S,M,T,W,R,F,A} to day-of-week number
   * @returns the number (from 0 to 6) or -1 if not found
   */
  static int dayOfWeekNumber()
  {
    switch(pcData.toUpperCase().charAt(0))
    {
      case 'S':
        return 0;
      case 'M':
        return 1;
      case 'T':
        return 2;
      case 'W':
        return 3;
      case 'R':
        return 4;
      case 'F':
        return 5;
      case 'A':
        return 6;
      default:
        System.out.println("invalid day-of-week code:" + pcData);
        return -1;
    }
  } // end dayOfWeekNumber()

  /**
   * converts current pcData to InetAddress; prints message if not possible
   * @returns the InetAddress; null if error
   */
  static InetAddress convertInetAddress()
  {
    InetAddress convertAddress = null;
    try
    {
      convertAddress = InetAddress.getByName(pcData);
    }
    catch(UnknownHostException uhe)
    {
      parseError("invalid Internet address:" + pcData);
    }
    return convertAddress;
  } // end convertInetAddress()

} // end class ParseConfig
