// NetWork EducationWare Record module
// by Babu Shanmugam 2001-2002
// revised by J M Pullen 1-15-03, 3-25-04, 1-2-06
// revised by Sairam 10-3-06 to work with NEWRE
// revised by JMP 1-5-08 to work with appFileDownloads
// revised by JMP 4-21-09 to add external shutdown command
// version 2.7.3
/*----------------------------------------------------------------*
|    Copyright 2001-2008 Networking and Simulation Laboratory     |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
import java.net.*;
import java.io.*;
import java.util.*;

// NEW REC - records files to be played by NEW PLAY
//
// Classes:
//  REC                  - main executive
//  Recorder             - thread that starts a single record process and
//                         its controls
//  StatusThread         - sends a stream of status messages to the controller
//  DirectController     - accept control messages straight from control
//  IndirectController   - accepts control messages via multicast server
/////////////////////////////////////////////////////////////////////////////////
// main executive class for TLM recording
//
// this is the NEW Transport Layer Multicaster (TLM) recorder (REC)
//
// LIMITATIONS:
// (1) this program is capable of running only one Recorder process, thus unlike
// PLAY there is synchronization and the code does not support multiple instances
// (2) the generated filename uses date/time which only changes once each minute,
// thus the program can start a new file only once per minute; if more frequent
// starts are attempted, only the last one will remain
// (3) original spec for REC included a cheksum on the control messages,
// which is present in the messages as 0 because it has not been imlemented
// (4) there is provision for passing the control stream through the TLM: simply
// connect RC to the control port number on the TLM, and the control stream will
// pass through TLMC to REC; however because this arrangement uses the same
// port on TLM and REC, it is not possible to run REC and TLM on the same host.
// This does not generally cause a problem, as TLM normally runs on a separate
// server.
//
// args[0] = IP address of TLM client (0 if on same host as REC)
// args[1] = IP address of TLM REC control client (0 if on same host as REC)
// args[2] = TCP data port for TLM REC to connect (control connects to port+1)
// args[3] = subdirectory name (or full pathname) in which recording file is created
//           (the .nwt files are expected to be in subdirectory COVERSLIDES of
//           this directory)
// args[4] = 1 if RESEQ is to be invoked when REC terminates (optional; default 0)
// args[5] = when args[4]==1; max number of messages in a RESEQ output file segment,
//           if 0, output will not be segmented (optional; default 15000)
// args[6] = when args[4]==1; path of directory for resequenced files
//           (optional; defaults to args[3]+"-OUT")
//
// The addDataDownloadZipfile is assumed to be in the current directory.
//
class REC
{
  final static String version = "2.7.3";
  final static int MAX_MESSAGE = 1455;
  final static int MAX_DOWNLOAD_FILE = 10485760;
  final static int DOWNLOAD_MESSAGE_SIZE = 1250;// bytes in slice for application zipfile download
  public static Recorder R;
  public static int controlMode = 0;         // 0 for indirect, 1 for direct
                                             // (implicit based on connection)
  public static String strTLMCIPAddress;     // TLM Client IP address
  public static String strTLMRCIPAddress;    // Record Control IP address
  public static int recordDataPort;          // recordControlPort = recordDataPort+1
  public static String RecordingsDirectory;  // directory where output file will be made\
  public static boolean runRESEQ = false;    // run RESEQ at termination of REC
  public static String segmentSize = "0";    // default max messages in RESEQ segment
  IndirectController IDCR;
  DirectController DCR = null;
  Thread indirectControl, directControl;
  public static boolean directConnectionMade = false;
  public static boolean indirectConnectionMade = false;
  public static boolean audioMessageReceivedThisCycle = false;
  public static boolean audioWarningSignIsUp = false;
  public static boolean slideMessageReceivedOnce = false;
  public static boolean slideWarningSignIsUp = false;
  public static String[] outputFiles = new String[9];
  public static int numberOfOutputFiles = 0;
  public static boolean stop = false, appDataFileWritten = false, closingOutFile = false;
  static String localHostAddress = null;
  static String appDataDownloadFile = null;
  static byte[] appDataDownloadFileBytes;

  public static void main(String args[])
  {
    FileOutputStream recordingFile = null;
    strTLMCIPAddress = new String(args[0]);
    strTLMRCIPAddress = new String(args[1]);

    // add a shutdownhook to capture appDownloadFile on closing
    Runtime.getRuntime().addShutdownHook
      (new Thread()
    {
      public void run()
      {
        try
        {
		  copyAppDataFile();
        }
        catch (Exception e)
        { System.out.println("NEW REC Exception in ShutdownHook:" + e.getMessage());}
        } // end try
      } // end run()
    ); // end Shutdown Hook

    // parameter 0 stands for local address in either args[0] or args[1]
    if( strTLMCIPAddress.equals("0") )
    {
/*
      // we want to use the localHost Address, but the method to get
      // that will hang for timeout if it's not available, so do that
      // in a thread that we can bypass if it hangs
      lookForLocalHost look = new lookForLocalHost();
      Thread runLook = new Thread(look);
      runLook.start();

      // now wait one second to see if the getLocalHost works
      try{Thread.sleep(1000);}catch(InterruptedException ie){}

      // if the value is still null, the method is hung so bypass it
      // the thread will terminate by itself after timeout
*/
      if(localHostAddress == null)
        localHostAddress = "localhost";
      strTLMCIPAddress = localHostAddress;
    }
    if( strTLMRCIPAddress.equals("0") )
    {
      // we want to use the localHost Address, but the method to get
      // that will hang for timeout if it's not available, so do that
      // in a thread that we can bypass if it hangs
      lookForLocalHost look = new lookForLocalHost();
      Thread runLook = new Thread(look);
      runLook.start();

      // now wait one second to see if the getLocalHost works
      try{Thread.sleep(1000);}catch(InterruptedException ie){}

      // if the value is still null, the method is hung so bypass it
      // the thread will terminate by itself after timeout
      if(localHostAddress == null)
        localHostAddress = "/127.0.0.1";
      strTLMRCIPAddress = localHostAddress;
      strTLMRCIPAddress =
        strTLMRCIPAddress.substring(strTLMRCIPAddress.indexOf('/') + 1);
    }

    // args[2] is the data port number (one less than control port number)
    recordDataPort = new Integer(args[2]).intValue();

    // args[3] is the recording directory
    if(args.length > 3)RecordingsDirectory = args[3];
    else RecordingsDirectory = "c:\\PROGRA~1\\NEW\\RECORDINGS";
    if(!(new File(RecordingsDirectory)).exists())
      new File(RecordingsDirectory).mkdir();
    if(!(new File(RecordingsDirectory)).isDirectory())
    {
      System.err.println("NEW REC error in output recording directory");
      System.exit(-1);
    }
    System.out.println("NEW REC version " +  version +
                       " accepting connection from " +
                       strTLMRCIPAddress + "/" + recordDataPort);

    // create the recorder and both direct and indirect controls
    // also start the control threads
    if(args.length > 4)runRESEQ = args[4].equals("1");
    R = (new Recorder());
    Thread runRecorder = new Thread(R);
    runRecorder.start();
    try{runRecorder.join();}catch(InterruptedException ie){System.out.println("NEW REC "+ie);}

    // args[4] tells whether to run RESEQ when recording is done
    if(args.length > 4)
      if(args[4].equals("1"))
      {
        if(args.length > 5)segmentSize = args[5];
        // prepare the args[] list
        if(numberOfOutputFiles == 0)System.exit(0);
        String[] reseqGuiArgs = new String[numberOfOutputFiles+5];
        if(args.length > 6)
          reseqGuiArgs[0] = args[6];
        else
          reseqGuiArgs[0] = RecordingsDirectory;
        if(!(new File(reseqGuiArgs[0])).exists())
          new File(reseqGuiArgs[0]).mkdir();
        if(!(new File(reseqGuiArgs[0])).isDirectory())
        {
          System.err.println("NEW REC error in intermediate recording directory");
          System.exit(-1);
        }
        reseqGuiArgs[1] = RecordingsDirectory + "/COVERSLIDES";
        reseqGuiArgs[2] = RecordingsDirectory;
        reseqGuiArgs[3] = segmentSize;
        reseqGuiArgs[4] = "0";
        for(int i=1; i<=numberOfOutputFiles; ++i)
          reseqGuiArgs[4+i] = outputFiles[i-1];

        // run the RESEQ GUI
        GUI.main(reseqGuiArgs);
      }

  } // end REC main method

  // utility method to convert four bytes to a long
  public long bytesToInt(byte[] bytes)
  {
    long long1 =
        ( bytes[0] >= 0 ? bytes[0] : 255 - ~bytes[0] ) * 16777216 +
        ( bytes[1] >= 0 ? bytes[1] : 255 - ~bytes[1] ) * 65536 +
        ( bytes[2] >= 0 ? bytes[2] : 255 - ~bytes[2] )  * 256 +
        ( bytes[3] >= 0 ? bytes[3] : 255 - ~bytes[3] );
    return long1;
  }

  // utility method to convert a long to four bytes
  public byte[] intToBytes(long long1)
  {
    byte[] result = new byte[4];
    result[0] = (byte ) (long1/16777216);
    long rem =  ( long1 - result[0] * 16777216);
    result[1] =  (byte) (rem  / 65536);
    rem =  ( rem - result[1] * 65536 );
    result[2] = (byte) (rem/256);
    rem =  ( rem - result[2] * 256) ;
    result[3] = (byte) rem;
    return result;
  }

  // utility method to convert two bytes to a short
  public int bytesToShort(byte byte1, byte byte2)
  {
    int high_byte = byte1;
    if( high_byte < 0 ) high_byte = 127 - high_byte;
    int low_byte = byte2;
    if( low_byte < 0 ) low_byte = 127 - low_byte;
    return high_byte*256 + low_byte;
  } // end bytesToShort method

  // utility method to create the first byte of two from a short
  public static byte shortToByte1( int input )
  {
    byte byte1;
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);
    return byte1;
  } // end shortToByte1 method

  // method to create the second byte of two from a short
  public static byte shortToByte2( int input )
  {
    byte byte1, byte2;
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);
    int lowByte = input - 256*highByte;
    if( lowByte <= 127 )byte2 = (byte)lowByte;
    else byte2 = (byte)(int)(127-lowByte);
    return byte2;
  } // end shortToByte2 method

  // returns true for exception, otherwise false
  public boolean processControlCommand(int controlCode,
                                       BufferedOutputStream controlStream)
  {
    byte[] responseHeader = new byte[5];
    byte[] ackMessage = new byte[15];
    byte[] responseMessage = new byte[20];
    byte checksum = 0; // checksum not implemented

//  record and playback control code values:
//  0 stop
//  1 play
//  2 record
//  3 pause
//  4 fast forward
//  5 fast reverse
//  6 jump forward one slide (play only)
//  7 jump backward one slide (play only)
//  8 exception occurred
//  9 close record or playback server
// 10 packet counter value
// 11 file list request
// 12 RESEQ will be run at shutdown

    switch(controlCode)
    {
      case 0: // stop recording
        if(R.getRecordingFlag())
        {
		  copyAppDataFile();
          R.setRecordingFlag(false);
          outputFiles[numberOfOutputFiles++] = R.CloseFile();
        }
        break;

      case 2: // start recording
        if(!R.getRecordingFlag())
        {
          R.setRecordingFlag(true);
          R.setFirstMessage();
        }
        break;

      case 3: // pause
        if(R.getRecordingFlag())R.setRecordingFlag(false);
        else controlCode = 0;
        break;

      case 8: // exception
        try
        {
          System.out.println("NEW REC received exception message from record control - exiting");
          if( R.isFileOpen()) R.outFile.close();
        }
        catch(Exception e)
        {
          System.out.println("NEW REC exception closing file in process_control_command:"+e.getMessage());
        }
        return true;
		
     case 9: // shutdown
        try
        {
          System.out.println("NEW REC exiting for file " + R.outFileName);
          if( R.isFileOpen()) R.outFile.close();
        }
        catch(Exception e)
        {
          System.out.println("NEW REC exception closing file in process_control_command:"+e.getMessage());
        }
        return true;

      default: // not a valid controlCode - print error and skip processing
        String modeName = "indirect mode";
        if(controlMode == 1)modeName = "direct mode";
        System.out.println("NEW REC received invalid control code:" + controlCode
          + " in " + modeName);
        return false;
    }

    // Send ACK to record client TLMC
    ackMessage[0] = 75; // transaction code
    ackMessage[1] = (byte)controlCode;

    // insert next message number to play in bytes 10 to 13
    System.arraycopy(intToBytes(R.getNextMessageNumber()),0,ackMessage,2,4);

    // build header for reply to record control
    responseHeader[0] = 83;
    responseHeader[1] = 0;
    responseHeader[2] = 21;
    responseHeader[3] = 0;
    responseHeader[4] = 0;
    System.arraycopy(responseHeader,0,responseMessage,0,5);
    System.arraycopy(ackMessage,0,responseMessage,5,15);

    // send the response message
    try
    {
      if(controlMode == 0) // indirect
      {
        controlStream.write(responseMessage);
        controlStream.flush();
      }
      else // direct
      {
        controlStream.write(ackMessage);
        controlStream.flush();
      }
    }
    catch( Exception e)
    {
      // lost connection
      return true;
    }
    return false;
  } // end process_control_command method

  // thread class to getLocalHost in a way that can be
  // bypassed if it hangs waiting on local timeout
  static class lookForLocalHost implements Runnable
  {
	String testHostAddress = null;
    public void run()
    {
      try
      {
        testHostAddress = InetAddress.getLocalHost().getHostAddress();
      }
      catch (UnknownHostException uhe){}

      // if we got a result and the main thread didn't already
      // give up and provide a default, use our result
      if(testHostAddress != null)
        if(localHostAddress == null)
          localHostAddress = testHostAddress;
    }// end run()

  } // end class lookForLocalHost

  // if there is an appDataZipFile for the FC appLauncher,
  // append its name and contents to the end of the recording
  //
  // appDataZipFile segment format:
  // {
  //  transaction code byte[1] value "85"
  //  timestamp byte[4]
  //  dataLength unsigned short // 2 bytes
  //  controlCode byte[1] value "0"
  //  appNumber byte[1] value "125"
  //  fileSegment byte[dataLength]
  // }
  static void copyAppDataFile()
  {
	// only do this once
	if(appDataFileWritten)return;
	appDataFileWritten = true;

	// if we didn't receive a filename, this
	// does not apply
    if(appDataDownloadFile == null)return;

    // lock out further writing from input stream
    closingOutFile = true;

    // write the addDataDownloadFile message to the record file
    try
    {
      R.outFile.write(appDataDownloadFileBytes);
    }
    catch(IOException ioe)
    {
	  System.out.println(
		"\nNEW REC IOException writing application download filename to recording:" +
  	    ioe);
    }

    // find the appDataDownloadFile
    File appZipFile = new File(appDataDownloadFile);
  	if(!appZipFile.exists())
  	{
	  System.out.println("\nNEW REC application download zipfile \"" + appDataDownloadFile +
	                     "\" not accessible ");
	  return;
    }
    // obtain file length
  	long appZipFileLength = appZipFile.length();
  	if(appZipFileLength > MAX_DOWNLOAD_FILE)
  	{
      System.out.println("\nNEW REC application download zipfile \"" + appDataDownloadFile +
                         "\" + length " + appZipFileLength + " exceeds configured max "
                         + MAX_DOWNLOAD_FILE);
      return;
    }
    // open a stream on the file
  	FileInputStream appZipStream = null;
  	try
  	{
      appZipStream = new FileInputStream(appZipFile);
    }
    catch(FileNotFoundException fnfe)
    {
 	  System.out.println("\nNEW REC application download zipfile \"" + appDataDownloadFile +
                         "\" not found");
      return;
    }

    // read the file in segments of DOWNLOAD_MESSAGE_SIZE bytes
    // add a recording header and write them to the recording file
    long appFileDownloadSize = 0;

    // make a header for segment using timestamp from filename
	byte[] timestampBytes = new byte[4];
	System.arraycopy(appDataDownloadFileBytes,0,timestampBytes,0,4);
	byte[] lengthBytes = new byte[2];
    lengthBytes[0] = shortToByte1(DOWNLOAD_MESSAGE_SIZE);
    lengthBytes[1] = shortToByte2(DOWNLOAD_MESSAGE_SIZE);
	byte[] portBytes = new byte[2];
	portBytes[0] = 0;   // control
    portBytes[1] = 125; // appDataFile
    byte[] restartByte = new byte[1];
    restartByte[0] = 0;

    // install header values in the buffer
    byte[] fileSegment = new byte[timestampBytes.length + lengthBytes.length +
			             portBytes.length + restartByte.length + DOWNLOAD_MESSAGE_SIZE];
    System.arraycopy(timestampBytes,0,fileSegment,0,timestampBytes.length);
	System.arraycopy(lengthBytes,0,fileSegment,4,lengthBytes.length);
	System.arraycopy(portBytes,0,fileSegment,6,portBytes.length);
    System.arraycopy(restartByte,0,fileSegment,8,restartByte.length);

    // copy the appData file
    try
    {
      for (long byteNumber = 0;
           byteNumber < appZipFileLength;
           byteNumber += DOWNLOAD_MESSAGE_SIZE)
      {
		// all segments but last are DOWNLOAD_MESSAGE_SIZE
        long bytesToRead = appZipFileLength - appFileDownloadSize;
        if(bytesToRead > DOWNLOAD_MESSAGE_SIZE)bytesToRead = DOWNLOAD_MESSAGE_SIZE;

        // last segment only
        else
        {
		  fileSegment[4] = shortToByte1((int)bytesToRead);
		  fileSegment[5] = shortToByte2((int)bytesToRead);
	    }

        // read a segment, offset into buffer by size of header
        long bytesRead = appZipStream.read(fileSegment,9,(int)bytesToRead);
        if(bytesRead < 0)break;
        appFileDownloadSize += bytesRead;

        // write the header and segment to recording file
        R.outFile.write(fileSegment,0,(int)bytesToRead+9);

      }// end for (long byteNumber = 0;

      // close the output file
      R.outFile.close();
    }
    catch(IOException ioe)
    {
      System.out.println("\nNEW REC IOException copying application download zipfile \"" +
  	                     appDataDownloadFile + "\" to recording:" + ioe);
      appFileDownloadSize = 0;
    }
  }// end copyAppDataFile()

} // end REC class

////////////////////////////////////////////////////////////////////////
// A Direct Controller object reads the control messages
// from the Record control, when it has been connected to directly
// to its input port.
//
// For a direct connection, the control message will not have any header
// and so the control message can be processed without needing to strip
// the header.
//
// After the control message is processed the ACK is written directly on
// TCP connection to the RC client.
////////////////////////////////////////////////////////////////////////
class DirectController extends REC implements Runnable
{
  public BufferedInputStream inFromRC;
  public BufferedOutputStream outToRC;
  public FileOutputStream recordingFile;
  String strRCAddress=null;
  String RCInetAddress = null;
  byte restartTimeBase;
  Recorder recorder;

  // constructor
  public DirectController(Recorder recorderRef)
  {
    recorder = recorderRef;

    // perform DNS lookup on configured RC (record control) client
    try
    {
      RCInetAddress = InetAddress.getByName(strTLMRCIPAddress).getHostAddress();
    }
    catch(Exception e)
    {
      System.out.println("NEW REC GetHostByName exception in DirectController:"+e.getMessage());
    }
  } // end constructor

  // direct connection may never happen; must open server socket and see if it is used
  public void WaitForConnection()
  {
    try
    {
      int recordControlPort = recordDataPort + 1;
      ServerSocket s = new ServerSocket(recordControlPort,1);
      while( !directConnectionMade )
      {
        System.out.println("NEW REC listening for control connection from RC on " +
                           RCInetAddress + "/" + recordControlPort);
        Socket workingSocket = s.accept();
        System.out.println("NEW REC received connection from:" + workingSocket.getInetAddress());
        if(indirectConnectionMade)
        {
          System.out.println("NEW REC indirect control previously established- closing this connection");
          return;
        }
        String testHostAddress = workingSocket.getInetAddress().getHostAddress();
        if( !(RCInetAddress.equals(testHostAddress)) )
        {
          System.out.println("NEW REC connection host " + testHostAddress +
                             " does not match config -- closing direct connection");
          workingSocket.close();
        }
        else
        {
          directConnectionMade = true;
          System.out.println("NEW REC direct connection from RC accepted");
          inFromRC = new BufferedInputStream(workingSocket.getInputStream());
          outToRC = new BufferedOutputStream(workingSocket.getOutputStream());
          if(runRESEQ)
          {
            byte[] reseqMessage = new byte[15];
            reseqMessage[0] = 75; // command
            reseqMessage[1] = 12; // shows REC is configured to run RESEQ
            for(int i=2; i<15; ++i)reseqMessage[i] = 0; // fill
            outToRC.write(reseqMessage);
            outToRC.flush();
          }
        } // end else/if(!RCInetAddress...
      } // end while(!directConnectionMade)
    } // end try
    catch(IOException ioe)
    {
      System.out.println("NEW REC IOException in WaitForConnection:" + ioe.getMessage());
      System.exit(-1);
    }
  } // end WaitForConnection method

  // main run method of DirectController class
  public void run()
  {
    int controlCode, transactionCode = 0;
    byte[] timestampBytes = new byte[4];
    byte[] lengthBytes = new byte[2];
    byte[] rcPort = new byte[2];
    byte[] buffer = new byte[15];
    int times;
        
    // wait until connected
    WaitForConnection();
    controlMode = 1;
    
    // create and start a thread to update recorder from control messages
    StatusThread updater = new StatusThread(outToRC,recorder);
    Thread update = new Thread(updater);
    update.start();

    // loop until disconnected or terminated
    while( !recorder.done )
    {
      // Read the first byte from the RC connection.
      // If transaction code is bad, scan forward
      int badData = 0;
      try
      {
        transactionCode = inFromRC.read();
        int printCode = transactionCode;
        while((transactionCode!=74) && (transactionCode!=-1))
        {
          badData++;
          transactionCode = inFromRC.read();
        }
        if(badData > 0)System.out.println("NEW REC bad transaction code " + printCode
          + " encountered in input stream - skipped " + badData + " bytes");
      }
      catch(IOException ioe)
      {
        // broken connection
        return;
      }
      switch(transactionCode)
      {
        case 74: // regular control message
          try
          {
            inFromRC.read(buffer);
          }
          catch(IOException ioe)
          {
            // broken connection
            return;
          }
          rcPort[0] = buffer[0];
          rcPort[1] = buffer[1];
          controlCode = buffer[2];
          if(processControlCommand(controlCode,outToRC))return;
          break;

        case -1: // disconnected from input
          recorder.done = true;
          System.out.println("NEW REC controller disconnected - stopping recorder");
          try
          {
            recordingFile.close();
          }
          catch(IOException ioe)
          {
            System.out.println("NEW REC exception closing output file in DirectController:"
              + ioe.getMessage());
            return;
          }
          break;

        default: // error in message

      } // end switch(transactionCode)
    } // end while(!done)
  } // end DirectorController run method
} // end DirectController class

////////////////////////////////////////////////////////////////////////
// This class encapsulates the control message processor that reads the
// control messages on an indirect connection from the record control
// via TLM and TLMC.
//
// Indirect connection :
// TLMRC              ->TLM       ->TLMC               ->REC
// (playback control)   Multicaster Multicaster Client   Record Server
//
// The message processor processes the control message and writes the
// ACK packet to TLMRC indirectly. The ACK packet is written with a
// header.
////////////////////////////////////////////////////////////////////////
class IndirectController extends REC implements Runnable
{
  public  byte[] recordBuffer;
  public BufferedInputStream inFromTLMC;
  public BufferedOutputStream outToTLMC;
  Recorder recorder;

  // constructor
  IndirectController(BufferedInputStream inStream,
                     BufferedOutputStream outStream,
                     Recorder recorderRef)
  {
    inFromTLMC = inStream;
    outToTLMC = outStream;
    recorder = recorderRef;
  } // end constructor

  // main run method of IndirectController class
  public void run()
  {
    int controlCode;
    int transactionCode = 0;
    byte[] timestampBytes = new byte[4];
    byte[] lengthBytes = new byte[2];
    byte[] ipAddress1 = new byte[4];
    byte[] ipAddress2 = new byte[4];
    byte[] rcPort = new byte[2];
    byte[] portBytes = new byte[2];
    byte[] restartByte = new byte[1];
    byte[] dataBuffer;
    byte[] buffer = new byte[21];
    long currentTimestamp;
    byte restartTimeBase;
    boolean updaterRunning= false;
    int times, payloadLength = 0;
    
    // loop until stopped by control or shutdwon
    while( !recorder.done )
    {
      // Read the first byte from the TLMC connection.
      // If transaction code is bad, scan forward
      int badData = 0;
      try
      {
        transactionCode = inFromTLMC.read();
        int printCode = transactionCode;
        while((transactionCode!=80) && (transactionCode!=82) && (transactionCode!=-1))
        {
          badData++;
          transactionCode = inFromTLMC.read();
        }
        if(badData > 0)System.out.println("NEW REC bad transaction code " + printCode
          + " encountered in input stream - skipped " + badData + " bytes");
      }
      catch(IOException ioe)
      {
        // broken connection
        return;
      }

      // the message could be 82 (imbedded control) or 80
      switch( transactionCode )
      {
        case 82: // Control message imbedded in the data field
          try
          {
            // read the header
            inFromTLMC.read(timestampBytes);
            inFromTLMC.read(lengthBytes);

            // this type of message always has length 21
            int lengthValue = bytesToShort(lengthBytes[0],lengthBytes[1]);
            if(lengthValue != 21)
            {
              System.out.println(
                "NEW REC skipping invalid control message in IndirectController;"
                + " length should be 21 but is " + lengthValue);
              break;
            }
            inFromTLMC.read(portBytes);
            inFromTLMC.read(); // open position for restartTimebase
            // Now we have reached the control message
            inFromTLMC.read(buffer);
          }
          catch(IOException ioe)
          {
            // broken connection
            return;
          }

          // if direct control connection exists, ignore controls from TLMC
          if(directConnectionMade)
          {
            System.out.println("NEW REC.IndirectController ignoring control message in data path "
              + "because direct control has connected");
            break;
          }
          indirectConnectionMade = true;

          // extract and interpret data from the imbedded control message
          transactionCode = buffer[0];
          controlCode = buffer[11];
          if(processControlCommand(controlCode,outToTLMC))return;
          break;

        case 80: // data packet from TLMC - record or discard it based on flags
          // read header: timestamp, length of packet, port
          try
          {
            // wait until the header is available, then read its fields
            try
            {
              while( inFromTLMC.available() < 9 ) Thread.sleep(10);
		        }catch(InterruptedException ie){}
            inFromTLMC.read(timestampBytes,0,4);
            currentTimestamp = bytesToInt(timestampBytes);
            inFromTLMC.read(lengthBytes,0,2);
            payloadLength = bytesToShort(lengthBytes[0],lengthBytes[1]);
            inFromTLMC.read(portBytes,0,2);

            // read the unused restartTimeBase byte
            inFromTLMC.read(restartByte,0,1);

            // check for invalid length in header
            if(payloadLength > MAX_MESSAGE)
            {
              System.out.println("REC.IndirectController found data message (transaction code 80) "
              + "with header message length " + payloadLength + " - did not read past header");
              break;
            }

            // wait until the indicated amount of data is available then read it
            dataBuffer = new byte[payloadLength];
            try
            {
              while( inFromTLMC.available() < payloadLength ) Thread.sleep(10);
            }catch(InterruptedException ie){}
            inFromTLMC.read(dataBuffer,0,payloadLength);

            // if the message is an appDownloadFile name, capture iname and its mTime but don't record
            if((portBytes[0] == 0) && (portBytes[1] == 126) && (dataBuffer[0] == 84))
            {
			        appDataDownloadFile = new String(dataBuffer,5,19);
			        int appDataMessageLength = timestampBytes.length + lengthBytes.length +
			                                   portBytes.length + restartByte.length + payloadLength;

			        appDataDownloadFileBytes = new byte[appDataMessageLength];
			        System.arraycopy(timestampBytes,0,appDataDownloadFileBytes,0,timestampBytes.length);
              System.arraycopy(lengthBytes,0,appDataDownloadFileBytes,4,lengthBytes.length);
              System.arraycopy(portBytes,0,appDataDownloadFileBytes,6,portBytes.length);
              System.arraycopy(restartByte,0,appDataDownloadFileBytes,8,restartByte.length);
              System.arraycopy(dataBuffer,0,appDataDownloadFileBytes,9,dataBuffer.length);
			        break;
              
		        }// end if((portBytes[0] == 0)...

            // record the data in a file, if recording has not been stopped or paused
            if( R.getRecordingFlag() )
            {
              // preface recorded data with header: {timeStamp,
              // message length, port number} before data bytes
              R.currentMessageNumber++;
              recordBuffer = new byte[9+payloadLength];
              System.arraycopy(timestampBytes,0,recordBuffer,0,4);
              System.arraycopy(lengthBytes,0,recordBuffer,4,2);
              System.arraycopy(portBytes,0,recordBuffer,6,2);

              // check for audio message
              if((portBytes[0] == 0) && (portBytes[1] == 0))
                audioMessageReceivedThisCycle = true;

              // check for WBD NewPage message
              if((portBytes[0] == 0) && (portBytes[1] == 1))
                if((dataBuffer[10] == 0) && (dataBuffer[11] == 1))
                  if((dataBuffer[34] == 0) &&
                    ((dataBuffer[35] == 16) || (dataBuffer[35] == 17)))
                    slideMessageReceivedOnce = true;

              // at first message after a pause, restart the time base
              if( R.getFirstMessage() ) restartTimeBase = 1;
              else restartTimeBase = 0;
              recordBuffer[8] = restartTimeBase;
              System.arraycopy(dataBuffer,0,recordBuffer,9,payloadLength);

              // don't do any more with file here if in process of copy/close
              if(closingOutFile)break;

              // copy message to file, creating it first if necessary
              if( R.isFileOpen() )
                R.outFile.write(recordBuffer);
              else
              {
                // if the file has not been opened, create a
                // unique name for it based on date and time
                GregorianCalendar gcalendar = new GregorianCalendar();
                int month = gcalendar.get(Calendar.MONTH) + 1 ;
                int date = gcalendar.get(Calendar.DATE);
                int hour = gcalendar.get(Calendar.HOUR_OF_DAY);
                int minute = gcalendar.get(Calendar.MINUTE);
                String strmonth = "" + ( month < 10 ? "0" + month : "" + month);
                String strdate = "" + ( date < 10 ? "0" + date : "" + date);
                String strhour = "" + ( hour < 10 ? "0" + hour : "" + hour );
                String strminute = "" + ( minute < 10 ? "0" + minute : "" + minute );
                String strRecFile =  strmonth + strdate + strhour + strminute + "S.nwr";
                if(!R.OpenFile(strRecFile))return;
                R.currentMessageNumber = 0;
                R.outFile.write(recordBuffer);

                //SAIRAM.CH  10-3-2006.
                // write the tempeorary file name to another file of known name
                // in order to pass the filename to the client
                try
                {
                    BufferedWriter out =
                      new BufferedWriter(new FileWriter(
                        "C:\\Program Files\\NEW\\RECORDINGS\\^outfilename"));
                    out.write(strRecFile);
                    out.close();
                } catch (IOException e) {
                        System.out.println("Error in writing the file.");
                }

              }

              // sleep a minimal time, to ensure recording is not keeping
              // audio process from using the processor
              try
              {
				Thread.sleep(1);
		      }catch(InterruptedException ie){}

              // clear FirstMessage status flag and start StatusThread
              R.clearFirstMessage();
              if( !updaterRunning && (controlMode == 0) )
              {
                StatusThread updater = new StatusThread(outToTLMC,recorder);
                Thread update = new Thread(updater);
                update.start();
                updaterRunning = true;
              } // end if( !updaterRunning )
            } //  end if( R.getRecordingFlag() )
          } // end try
          catch(Exception e)
          {
            System.out.println("NEW REC exception in IndirectController:" + e.getMessage());
            System.out.println("skipping data message of length " + payloadLength);
          }
          break;

        case -1: // disconnected from input
          System.out.println("NEW REC data source disconnected- shutting down recorder");
          recorder.done = true;

      } // end switch( transactionCode )
    } // end end while(!done)
  } // end IndirectController run method
} // end IndirectController class

class StatusThread extends REC implements Runnable
{
  BufferedOutputStream outToRC;
  byte[] ackMessage = new byte[15];
  byte[] responseMessage = new byte[20];
  byte[] responseHeader = new byte[5];
  byte checksum = 0;
  byte[] bZeros = ("00000000000000").getBytes();
  Recorder recorder;

  // constructor - save parameters
  StatusThread(BufferedOutputStream outToRCRef, Recorder recorderRef)
  {
    outToRC = outToRCRef;
    recorder = recorderRef;
  }

  // run method of StatusThread class
  public void run()
  {
    while( !recorder.done )
    {
      // Send an ACK packet to the control client to update the
      // status of the recording
      ackMessage[0] = 75; // transaction code
      ackMessage[1] = 10;

      // copy numberOfPackets in file to positions 2-5
      System.arraycopy(intToBytes(R.getNextMessageNumber()),0,ackMessage,2,4);
      System.arraycopy(bZeros,0,ackMessage,6,9);

      // check for need to send audio or slide warning
      // check only if recorder is running
      if(R.getRecordingFlag())
      {
        // audio check repeats each StatusThread update cycle
        // (this assumes the recording should have constant audio input)
        if(audioMessageReceivedThisCycle)
        {
          audioMessageReceivedThisCycle = false;
          if(audioWarningSignIsUp)
          {
            ackMessage[1] = 2;
            audioWarningSignIsUp = false;
          }
          else
          {
            // slide check only happens once per recording
            // (no way to know how often a slide should be entered)
            if(slideMessageReceivedOnce)
            {
              if(slideWarningSignIsUp)
              {
                ackMessage[1] = 2;
                slideWarningSignIsUp = false;
              }
            }
            else
            {
              // popup slide warning sign
              ackMessage[1] = 5;
              slideWarningSignIsUp = true;
            } // end else/if(slideMessageReceivedOnce)
          } // end else/if(audioWarningSignIsUp)
        } // end if(audioMessageReceivedThisCycle)
        else
        {
          // popup audio warning sign
          ackMessage[1] = 4;
          audioWarningSignIsUp = true;
        } // end else/if(audioMessageReceivedThisCycle)
      } // end if(R.getRecordingFlag())

      // this is an indirect connection - insert five-byte header
      responseHeader[0] = 83;
      responseHeader[1] = 0;
      responseHeader[2] = 21;
      responseHeader[3] = 0;
      responseHeader[4] = 0;
      System.arraycopy(responseHeader,0,responseMessage,0,5);
      System.arraycopy(ackMessage,0,responseMessage,5,15);

      // send the response
      try
      {
        if( controlMode == 0 ) outToRC.write(responseMessage);
        else outToRC.write(ackMessage);
        outToRC.flush();
      }
      catch(IOException ioe)
      {
        // broken connection
        System.out.println("NEW REC broken stream connection");
        recorder.done = true;
        return;
      }
      try
      {
        Thread.sleep(2000);
      }
      catch(InterruptedException ie){ System.out.println(ie); }
    } // end while(!done)
  } // end run method
}

///////////////////////////////////////////////////////////////////////////////
// This class contains all the main functionality for the Record Server.
// Upon creation of the Recorder object, it connects to the TLM Client (TLMC).
// Player run pseudo code:
// loop forever
//   listen to command on TCP connection ( sent from playback control via TLMC )
//   process the command (start playback, stop playback, fast forward etc )
//   acknowledge the command
//////////////////////////////////////////////////////////////////////////////
class Recorder extends REC implements Runnable
{
  public ServerSocket listenSocket;
  public Socket workingSocket;
  public String tlmcHostAddress;
  public BufferedInputStream in;
  public  BufferedOutputStream out;
  public  byte[] recordBuffer;
  public  boolean recordFlag = false; // true while recorder is running
  boolean restartFlag = true;
  boolean firstMessage;               // true while recording first message
  public FileOutputStream outFile = null;
  public boolean outFileOpen = false;
  public String outFileName = "";
  public boolean bConnectDirect= false;
  public long currentMessageNumber = 0;
  public boolean notConnected = true;
  public boolean done = false;

  // constructor
  Recorder()
  {
    System.out.println("NEW Recorder version " + version +
                       " listening for connection from TLMC");
    if( acceptTLMCConnection() )
      System.out.println("NEW REC connected to TLMC " + strTLMCIPAddress +
                         "\nrecording will be placed in " + RecordingsDirectory);
    else
    {
      System.out.println("could not connect to TLMC due to exception");
      return;
    }
    
    // create and start threads for indirect and direct controllers
    // as we do not know which one will be used
    IDCR = new IndirectController(in,out,this);
    indirectControl = new Thread(IDCR);
    indirectControl.start();
    DCR = new DirectController(this);
    directControl = new Thread(DCR);
    directControl.start();
  } // end Recorder contructor

  // main run method of Recorder class
  public void run()
  {
    // wait for Indirect and Direct control threads to stop
    try
    {
      indirectControl.join();
      directControl.join();
    }
    catch(InterruptedException ie){System.out.println("NEW REC exception:"+ie);return;}
    System.out.println("Recorder terminating");
  } // end Record run method

  // methods to obtain and set status variables
  public long getNextMessageNumber()
  {
    return(currentMessageNumber) ;
  }

  public synchronized boolean getFirstMessage()
  {
    return(firstMessage);
  }

  public synchronized void setFirstMessage()
  {
     firstMessage =   true;
  }

  public synchronized void clearFirstMessage()
  {
     firstMessage = false;
  }

  public synchronized void setRestartFlag()
  {
     restartFlag = true;
  }

  public synchronized void clearRestartFlag()
  {
     restartFlag = false;
  }

  public synchronized boolean getRestartFlag()
  {
     return(restartFlag);
  }

  public synchronized void setRecordingFlag(boolean recFlag)
  {
     recordFlag = recFlag;
  }

  public synchronized boolean getRecordingFlag()
  {
     return(recordFlag);
  }

  // returns true on connection, false on exception
  public boolean acceptTLMCConnection()
  {
    try
    {
      while(notConnected)
      {
        // listen for connection from configured host
        tlmcHostAddress = InetAddress.getByName(strTLMCIPAddress).getHostAddress();
        System.out.println("NEW REC waiting for data connection by " + tlmcHostAddress + "/"
          + recordDataPort);
        listenSocket = new ServerSocket(recordDataPort,1);
        workingSocket = listenSocket.accept();
        String acceptHostAddress = workingSocket.getInetAddress().getHostAddress();
        System.out.println("NEW REC accepted connection from " + acceptHostAddress);
        if( !(tlmcHostAddress.equals(workingSocket.getInetAddress().getHostAddress())) &&
            !(acceptHostAddress.equals("127.0.0.1")))
        {
          System.out.println(
			"NEW REC connecting host does not match config- closing connection");
           workingSocket.close();
           try{Thread.sleep(10000);}catch(InterruptedException ie){}
           System.exit(-1);
        }
        notConnected = false;

        // set up input and output streams on the socket
        in = new BufferedInputStream(workingSocket.getInputStream());
        out = new BufferedOutputStream(workingSocket.getOutputStream());
      }
    }
    catch(IOException ioe)
    {
      System.out.println("NEW REC IOException in AcceptTLMCConnection:" + ioe.getMessage());
      return false;
    }
    return true;
  } // end acceptTLMCConnection method


  public synchronized boolean isFileOpen()
  {
    return(outFileOpen);
  }

  // returns true if file opened, false if exception
  public synchronized boolean OpenFile(String strFileName)
  {
    outFileName = RecordingsDirectory + "/" + strFileName;
    try
    {
      outFile = new FileOutputStream(outFileName);
    }
    catch(IOException ioe)
    {
      System.out.println("NEW REC exception opening output file " + outFile
        + ":" + ioe.getMessage());
      return false;
    }
    outFileOpen = true;
    return true;
  } // end OpenFile method

  public synchronized String CloseFile()
  {
    String fileName = outFileName;
    if(outFile != null)
    {
      try
      {
        outFile.close();
      }
      catch (IOException ioe) {
        System.out.println("NEW REC exception closing output file " + outFile
                           + ioe.getMessage());
      }
    }
    outFileOpen = false;
    outFileName = "";
    return fileName;
  }// end CloseFile method

} // end Recorder class
