// NetWork EducationWare record editing module to reqsequence recording pages
// based on EDIT by J M Pullen
// version 3.2.0  4/26/04
/*----------------------------------------------------------------*
|    Copyright 2001-2008 Networking and Simulation Laboratory     |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
import java.io.*;
import java.util.*;
/*******************************************************************************
NEW RESEQ

This program reads a NEW class recording and updates page numbers as they would
have been created if the inputs were from one continuous session of WBD, starting
at 1. Handles multiple input files in the same run. Also will segment the output
file, and will place a cover slide with copyright statement and an audio clip on
each segment if copyright file is present. The segments start at the next new
slide after the segmentation count is exceeded. (As a consequence, a purely audio
recording cannot be segmented.)

KNOWN BUGS: RESEQ may not be able to run with its own output as input.

command line parameters:
- full path of output file (or filename in execution directory) including .nwr
- number of records per segment (if 0, file is not to be segmented)
- full path of cover slide information file (or filename in execution
  directory); 0 if no cover slide is to be generated
- full path of cover audio information file (or filename in execution
  directory); 0 if no cover audio is to be generated
- 1 to omit video from the copy; 0 to retain it (default: do not omit)
- full path of first input file (or filename in execution directory)
optional:
- full path of subsequent input files (or filenames in execution directory)

Coverslide file (optional, use 0 if none):
- line 1: URL for cover slide
- line 2: subtitle text for cover slide (semester etc)
- line 3: copyright line for cover slide
- line 4: copyright conditions (max 500 bytes; no hard returns)

Cover slides will be generated only if the copyright file is present.
They will include the filename from first input file (usually the lesson number)
and segment number. If file Welcome.nwr is present in the execution directory,
its contents will be copied directly after the cover slide of each segment. This
file should be chosen so as to require more than 10 seconds to play.  Its play
time will determine how long the copyright notice is displayed.
/******************************************************************************/

public class RESEQ
{
  // global control variables and constants
  CoverSlide coverSlide = new CoverSlide();
  protected static final String version = "3.1";
  static boolean done = false;    // emergency brake - shuts down
  protected static final int MAX_MESSAGE_LENGTH = 1455;
  protected static final byte audio = 0, whiteboard = 1, video = 2;

  // global network and file variables
  static String inputPath, outputPath, inputDate;

  // file data
  public static RandomAccessFile inFile, outFile = null;
  public static RandomAccessFile newSlideFile;
  long recordPosition[];
  boolean newSlideMessage[];
  static long bytesIn = 0, bytesOut = 0, totalBytesOut = 0, bytesInThisInputFile;
  static int inputRecordNumber = 0, outputRecordNumber = 0, totalOutputRecordNumber = 0;
  static int outputSegmentNumber = -1, inputSegmentNumber = 0;
  static int slidesResequenced = 0, segmentSize, messageLength = 0, wbeVersion = 0;
  static long slideNumber = 1, lastSlideNumber = 0;
  static String coverSlidePathName, coverAudioPathName;
  byte[] timestampBytes = new byte[4];
  byte[] lengthBytes = new byte[2];
  byte[] portCodeBytes = new byte[2];
  byte[] pauseCodeByte = new byte[1];
  byte[] wbdHeaderBytes = new byte[12];
  byte[] wbdDrawBytes = new byte[24];
  byte[] playbackControlPort = new byte[2];
  byte[] messagesToJump = new byte[4];
  int port_value,length_value;
  static byte[] header = new byte[9];
  static final String[] segmentSuffixes = {"A","B","C","D","E","F","G","H","J"};
  static final int maxSegments = 9; // = segmentSuffixes.size() if such exists

  // page number data
  static long[] oldSlideNumbers = new long[1000];
  static long[] newSlideNumbers = new long[1000];
  static int slideNumberCount;
  static byte[] udpMessage = new byte[MAX_MESSAGE_LENGTH];
  static byte appToOmit = -1;

  static RESEQ reseq;

  public static byte[] senderAddress = {127,0,0,1}, loopbackAddress = {127,0,0,1};

  static byte[] firstPageNum = {0,1,0,1};

  public void RESEQ()
  {
    // nothing to initialize
  } // end RESEQ constructor

  // main method
  public static void main(String args[])
  {
    reseq = new RESEQ();
    System.out.println("RESEQ v1.0 run at:" + dateTime());

    // load command line parameters
    if(args.length < 6)
    {
      System.out.println("RESEQ requires at least 6 command line arguments:");
      System.out.println(
        "output path-file, segment size, cover slide path-file," +
          " cover audio path-file, omitVideo flag, input path-file");
      System.out.println("shutting down");
      System.exit(1);
    }
    System.out.println("NEW recording resequencer version " + version);
    outputPath = args[0];
    System.out.println("output file:"+outputPath);
    try{segmentSize = new Integer(args[1]).intValue();}
    catch(NumberFormatException nfe)
    {
      System.out.println("error interpreting integer segmentSize:"+args[1]);
      System.exit(2);
    }
    coverSlidePathName = args[2];
    System.out.println("cover slide file:"+coverSlidePathName);
    coverAudioPathName = args[3];
    System.out.println("cover audio file:"+coverAudioPathName);
    System.out.println("output segment size:"+segmentSize);
    if(args[4].equals("1"))appToOmit = 2;
    if(appToOmit == 2)System.out.println("deleting video messages");

    // loop through all input files on command line
    int fileNumber = 1;
    long previousTimeStamp = 0;
    while((++fileNumber+3)<args.length)
    {
      bytesInThisInputFile = 0;
      slideNumberCount = 0;
      inputPath = args[fileNumber+3];
      byte[] previousHeader = new byte[9];
      int previousLength=0;
      byte[] previousMessage = new byte[MAX_MESSAGE_LENGTH];
      byte[] newPageMessage = new byte[59];
      byte[] newPageHeader = new byte[9];
      byte[] slideNumberBytes = new byte[4];

      // test input file for format and presence
      System.out.println("input file:" + inputPath);
      int startDot = inputPath.length()-4;
      if(startDot <= 0)
      {
        System.out.println("input path must contain a file name with .nwr extension");
        System.exit(10);
      }
      if(!inputPath.substring(startDot,startDot+4).equals(".nwr"))
      {
        System.out.println("input files must have extension nwr");
        System.exit(6);
      }
      try
      {
        inFile = new RandomAccessFile(inputPath,"r");
      }
      catch(FileNotFoundException f)
      {
        System.out.println("input file:" + inputPath +
                           " not found - can't create output file");
        break;
      }
      int lastSlashPos = inputPath.lastIndexOf("/");
      int lastBackslashPos = inputPath.lastIndexOf("\\");
      if(lastBackslashPos > lastSlashPos)lastSlashPos = lastBackslashPos;
      inputDate = dateTime().substring(2,4)+inputPath.substring(lastSlashPos+1,lastSlashPos+5);
      System.out.println("first slide number:" + (slideNumber--));

      // copy file
      try
      {
        while(!done)
        {
          // read header
          if(inFile.read(header) == -1)
          {
            System.out.println(inputPath + " bytes read:" + bytesInThisInputFile);
            break;
          }
          messageLength = bytesToShort(header[4], header[5]);

          // set restart flag for timestamps that are not close together
          // so they don't hang the playback
          long timeStamp = bytesToInt(header);
          if((timeStamp < previousTimeStamp) ||
             ((timeStamp - previousTimeStamp) > 10000))
            header[8] = 1;
          previousTimeStamp = timeStamp;
          if(messageLength>MAX_MESSAGE_LENGTH)
          {
            System.out.println("messageLength=" + messageLength
              + " out of range at record:" + outputRecordNumber);
            System.out.println("terminating RESEQ - output file incomplete");
            break;
          }

          // read message
          if(inFile.read(udpMessage,0,messageLength) == -1)break;
          bytesIn += (messageLength+9);
          bytesInThisInputFile += (messageLength+9);
          inputRecordNumber++;
          boolean thisIsANewPage = false;
          boolean thisIsADrawMessage = false;

          // check for app to be omitted
          if(header[7] == appToOmit)continue;

          // check for new slide - advance sequence
          if((header[6] == 0) && (header[7] == 1))                 // WBD message
          {
            wbeVersion = bytesToShort(udpMessage[8],udpMessage[9]);
            if((udpMessage[10]==0) && (udpMessage[11]==1))         // draw command
            {
              thisIsADrawMessage = true;
              if((udpMessage[34] == 0) &&                          // new page or
                ((udpMessage[35] == 16) || (udpMessage[35] == 17)))// next slide
              {
                thisIsANewPage = true;
                System.arraycopy(udpMessage,16,slideNumberBytes,0,4);

                // if this is the first time RESEQ has been run against this recording,
                // it might contain multiple segments starting with slide 1; we
                // distinguish these by incrementing the second byte of the four-byte
                // slide number so it comes out different from each previous segment
                // (this assumes no more than 127 segments with 65535 slides - not
                // likely to be a problem! - and would not work if it were possible
                // to use "PREV PAGE" across the segement boundary)
                //
                // special actions for old WBD protocol:
                // if RESEQ has been run before, the slide number high order
                // bytes will not be zero
                if(wbeVersion == 0)
                {
                  long thisSlideNumber = bytesToInt(slideNumberBytes);
                  if (thisSlideNumber < 256) {
                    if ( (bytesToInt(slideNumberBytes) == 1) &&
                        (slidesResequenced > 2))
                      ++
                        inputSegmentNumber;
                    slideNumberBytes[1] = (byte) inputSegmentNumber;
                  }
                } // end if(wbeVersion == 0)
              }
            }
          }

          // write this message to the output file
          String newOutputPath = null;

          // check for need to start a new segment
          if((outputRecordNumber == 0) || ((segmentSize > 0) &&
                                           (outputRecordNumber >= segmentSize) &&
                                           thisIsANewPage))
          {
            if(outputRecordNumber > 0)
            {
              // insert a final session message with reset code
              byte[] sessionMessage = makeSessionMessage(udpMessage);
              byte[] sessionHeader = makeSessionHeader(header,sessionMessage);
              sessionHeader[8] = 1;
              outFile.write(sessionHeader);
              outFile.write(sessionMessage);
              bytesOut += (9+sessionMessage.length);
              outputRecordNumber++;

              // close the file, making sure its length is what we just wrote
              System.out.println("segment " + segmentSuffixes[outputSegmentNumber]
                + " completed; file bytes written:" + bytesOut
                + " records/messages written:" + outputRecordNumber);
              totalBytesOut += bytesOut;
              totalOutputRecordNumber += outputRecordNumber;
              bytesOut = 0;
              outputRecordNumber = 0;
              try{outFile.close();}
              catch(IOException ioe)
              {
                System.out.println("IOException closing output file:"+ioe.getMessage());
                done = true;
                break;
              }
              outFile = null;
            } // end if(outputRecordNumber > 0)

            // open the output file if it has not been opened
            if(outFile == null)
            {
              startDot = outputPath.length()-4;
              if(startDot <= 0)
              {
                System.out.println(
                  "output path must contain a file name with .nwr extension");
                System.exit(11);
              }
              if(!outputPath.substring(startDot,startDot+4).equals(".nwr"))
              {
                System.out.println("output file must have extension nwr");
                System.exit(4);
              }
              newOutputPath = outputPath;
              if(++outputSegmentNumber >= maxSegments)
              {
                System.out.println("number of segments exceeded program capacity");
                done = true;
                break;
              }
              newOutputPath = outputPath.substring(0,startDot) + // "-" + inputDate +
                              segmentSuffixes[outputSegmentNumber]+".nwr";
              try
              {
                outFile = new RandomAccessFile(newOutputPath,"rw");
                outFile.setLength(0);
              }
              catch(FileNotFoundException f)
              {
                System.out.println("can't create output file:" + newOutputPath);
                System.exit(1);
              }
            } // end if(outFile == null)

            // make a new cover slide
            int returnCode = 0;
            if(!coverSlidePathName.equals("0") && !coverAudioPathName.equals("0"))
              returnCode = CoverSlide.writeCoverSlide(outFile,newOutputPath,
                                                      coverSlidePathName,
                                                      coverAudioPathName,
                                                      header,
                                                      (byte)(++inputSegmentNumber));
            if(returnCode < 0)
            {
              System.out.println("terminating RESEQ - output file incomplete");
              System.exit(returnCode);
            }

            // account for changes due to the cover slide
            bytesOut += returnCode;
            if(returnCode > 0)
            {
              totalOutputRecordNumber += 15;
              System.arraycopy(loopbackAddress,0,senderAddress,0,4);
            }

            // if this changes senders, increment the slide number and create a new
            // session message, in the process adding the page and updating the message
            if(!bytesEqual(udpMessage,senderAddress,4))
            {
              System.arraycopy(udpMessage,0,senderAddress,0,4);

              // special actions for old WBD protocol
              if(wbeVersion == 0)
              {
                byte[] newSenderPage = new byte[4];
                System.arraycopy(udpMessage,16,newSenderPage,0,4);
                newSenderPage[1] = (byte)(++inputSegmentNumber);
                addPageNumber(newSenderPage,0);
                byte[] sessionMessage = makeSessionMessage(udpMessage);
                outFile.write(makeSessionHeader(header,sessionMessage));

                // hack for older WBD that did not distinguish pages by sender
                sessionMessage[26] = 1;
                outFile.write(sessionMessage);
                bytesOut += (9+sessionMessage.length);
                outputRecordNumber++;
              } // end if(wbeVersion == 0)
              thisIsANewPage = false;
              thisIsADrawMessage = false;
            }

            // flag to restart the play timer, thus ignoring the
            // inevitable timestamp discrepancy with cover slide
            header[8] = 1;
          } // end if((segmentSize>0)...

          // update page_num if needed
          //if(thisIsANewPage)addPageNumber(slideNumberBytes,0);
          //if(thisIsADrawMessage)updatePageNumber(udpMessage,16);

          // write out the message
          try
          {
            outFile.write(header);
            outFile.write(udpMessage,0,messageLength);
          }
          catch(IOException ioe)
          {
            System.out.println("IOException writing file:"+ioe.getMessage());
            System.exit(7);
          }
          bytesOut += (messageLength+9);
          outputRecordNumber++;
        } // end while(!done)

        // close files
        inFile.close();

      } // end try
      catch(IOException ioe)
      {
        System.out.println("IO Exception in copy:"+ioe);
        System.out.println("shutting down - output file incomplete");
        break;
      }
    } // end while(++fileNumber)<args.length)
    try
    {
      outFile.close();
    }
    catch(IOException ioe)
    {
      System.out.println("IOException closing output file - results questionable");
    }
    if(outputSegmentNumber > -1)
      System.out.print("segment " + segmentSuffixes[outputSegmentNumber]);
    else
      System.out.print("output file " + outputPath);
    System.out.println(" completed; bytes written:" + bytesOut +
      " records/messages written:" + outputRecordNumber);
    totalBytesOut += bytesOut;
    totalOutputRecordNumber += outputRecordNumber;
    System.out.println("RESEQ run complete");
    System.out.println("file bytes read:"+bytesIn);
    System.out.println("total bytes written:"+totalBytesOut);
    System.out.println("slide numbers generated/revised:"+slidesResequenced);
    System.out.println("total records/messages in input files:"+inputRecordNumber);
    System.out.println("total records/messages in output files:"+totalOutputRecordNumber);

  } // end RESEQ main method

 /**
  *
  */
  static byte[] makeSessionMessage(byte[] sourceAddress)
  {
    byte[] sessionMessage = // IP address goes at [0], [16], [28]
  {127,  0,  0,  1,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  1,127,  0,  0,  1,
     0,  0,  0,  1,  0,  0,  0,  1,127,  0,  0,  1,  0,  0,  0,  1,  0,  1,  0,  0,
   127,  0,  0,  1,  0,  0,  0,  7,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,  0};

    // copy source address to output
    System.arraycopy(sourceAddress,0,sessionMessage,0,4);
    System.arraycopy(sourceAddress,0,sessionMessage,16,4);
    System.arraycopy(sourceAddress,0,sessionMessage,28,4);

    return sessionMessage;
  }

 /** make a header for a session message
  */
  static byte[] makeSessionHeader(byte[] oldHeader, byte[] sessionMessage)
  {
    byte[] sessionHeader = new byte[9];
    System.arraycopy(oldHeader,0,sessionHeader,0,4);        // timestamp
    sessionHeader[4] = shortToByte1(sessionMessage.length); // length
    sessionHeader[5] = shortToByte2(sessionMessage.length);
    sessionHeader[6] = 0;                                   // not control
    sessionHeader[7] = whiteboard;                          // medium
    sessionHeader[8] = 0;                                   // restart

    return sessionHeader;
  }

  /** compare two byte[]; return true if equal
   *
   */
  static boolean bytesEqual(byte[]b1,byte[]b2,int length)
  {
    for(int scan=0; scan<length; ++scan)if(b1[scan]!=b2[scan])return false;
    return true;
  }

 /** add a new page number to conversion arrays
  *
  *  returns the new number
  */
  static int addPageNumber(byte[] oldPageNumberBytes,int offset)
  {
    long oldPageNumber = bytesToInt(oldPageNumberBytes,offset);
    oldSlideNumbers[slideNumberCount] = oldPageNumber;
    newSlideNumbers[slideNumberCount++] = ++slideNumber;
    ++slidesResequenced;
    return (int)slideNumber;
  }

 /** update a page number (4 bytes) to match the new number
  */
  static void updatePageNumber(byte[] udpMessage,int firstByteNumber)
  {
    byte[] convertNumberBytes = new byte[4];
    System.arraycopy(udpMessage,firstByteNumber,convertNumberBytes,0,4);
    convertNumberBytes[1] = (byte)inputSegmentNumber;
    long convertSlideNumber = bytesToInt(convertNumberBytes);
    int findSlideNumber = 0;
    for(; findSlideNumber<slideNumberCount; findSlideNumber++)
      if(oldSlideNumbers[findSlideNumber] == convertSlideNumber)break;
    if(findSlideNumber >= slideNumberCount)
      findSlideNumber = addPageNumber(convertNumberBytes,0);
    System.arraycopy(intToBytes(newSlideNumbers[findSlideNumber]),0,
      udpMessage,firstByteNumber,4);
  }

 /** utility method to convert four bytes to a long
  */
  public static long bytesToInt(byte[] bytes)
  {
    long long1 =
        ( bytes[0] >= 0 ? bytes[0] : 255 - ~bytes[0] ) * 16777216 +
        ( bytes[1] >= 0 ? bytes[1] : 255 - ~bytes[1] ) * 65536 +
        ( bytes[2] >= 0 ? bytes[2] : 255 - ~bytes[2] )  * 256 +
        ( bytes[3] >= 0 ? bytes[3] : 255 - ~bytes[3] );
    return long1;
  }

    public static long bytesToInt(byte[] bytes, int offset)
  {
    long long1 =
        ( bytes[offset] >= 0 ? bytes[offset] : 255 - ~bytes[offset] ) * 16777216 +
        ( bytes[offset+1] >= 0 ? bytes[offset+1] : 255 - ~bytes[offset+1] ) * 65536 +
        ( bytes[offset+2] >= 0 ? bytes[offset+2] : 255 - ~bytes[offset+2] )  * 256 +
        ( bytes[offset+3] >= 0 ? bytes[offset+3] : 255 - ~bytes[offset+3] );
    return long1;
  } // end bytesToInt method

 /** utility method to convert a long to four bytes
  */
  public static byte[] intToBytes(long long1)
  {
    byte[] result = new byte[4];
    result[0] = (byte ) (long1/16777216);
    long rem =  ( long1 - result[0] * 16777216);
    result[1] =  (byte) (rem  / 65536);
    rem =  ( rem - result[1] * 65536 );
    result[2] = (byte) (rem/256);
    rem =  ( rem - result[2] * 256) ;
    result[3] = (byte) rem;
    return result;
  } // end intToBytes lethod

 /** utility method to convert two bytes to an int
  */
  public static int bytesToShort(byte byte1, byte byte2)
  {
    int high_byte = byte1;
    if( high_byte < 0 ) high_byte = 127 - high_byte;
    int low_byte = byte2;
    if( low_byte < 0 ) low_byte = 127 - low_byte;
    return high_byte*256 + low_byte;
  } // end bytesToShort method

 /** utility method to convert to first byte of an int
  */
  public static byte shortToByte1( int input )
  {
    byte byte1;
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);
    return byte1;
  } // end shortToByte1 method

 /** utility method to convert to create the second byte of two from int
  */
  public static byte shortToByte2( int input )
  {
    byte byte1, byte2;

    // recalculate the high byte- this is quicker than invoking the method
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);

    // now form the low byte
    int lowByte = input - 256*highByte;
    if( lowByte <= 127 )byte2 = (byte)lowByte;
    else byte2 = (byte)(int)(127-lowByte);
    return byte2;
  } // end shortToByte2 method

 /** utility method to return the date and time
  */
  public static String dateTime()
  {
    GregorianCalendar gcalendar = new GregorianCalendar();
    int year = gcalendar.get(Calendar.YEAR);
    int month = gcalendar.get(Calendar.MONTH) + 1 ;
    int date = gcalendar.get(Calendar.DATE);
    int hour = gcalendar.get(Calendar.HOUR_OF_DAY);
    int minute = gcalendar.get(Calendar.MINUTE);
    int second = gcalendar.get(Calendar.SECOND);

    String strMonth = "" + ( month < 10 ? "0" + month : "" + month);
    String strDate = "" + ( date < 10 ? "0" + date : "" + date);
    String strHour = "" + ( hour < 10 ? "0" + hour : "" + hour );
    String strMinute = "" + ( minute < 10 ? "0" + minute : "" + minute );
    String strSecond = "" + ( second < 10 ? "0" + second : "" + second );
    return year + "." + strMonth + "." + strDate + "/" + strHour + ":"
      + strMinute + ":" + strSecond;
  } // end dateTime method

} // end RESEQ class
