/* 
 * TCPHandoff.java
 * Transport Layer Multicaster
 * Created on January 4, 2006
 * @author  mpullen & students
 * @version 5.2.4
 * @modified JMP 3/1/13
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
import java.io.*;
import java.net.*;
import java.nio.channels.SocketChannel;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;

public class TCPHandoff extends TLM implements Runnable
{
/*
  thread to read and apply setup messages from clients using TCP
  and locate them appropriately in the host table, given the complexity
  of possible second TCP connection for video and multiple connections
  from the same IP address under NAT

  this is done in a thread to avoid delay in accepting new connections
  while it is happening

  TCPHandoff class logic:
  -> look for an entry with thisAddress in host table that also has
     matching sessionVariable
    -> if found this will be the second (video) connection from some
       client; start its TCPVideo tunnel and return
  -> thisAddress is not in host table, so insert it; then start a
     new TCPTunnel and return

  Synchronization/authentication scheme scheme:
   1. In parallel:
   a. If using direct FC connection:
        FCConnect creates a new FCListen thread
        FCListen reads the sessionVariable from FC connection
        FCListen tries 20 times at 500ms intervals to find a
          matching TCP (or UDP?) connection from TLMC in host table;
          if none returns
      If FC conenction is through data tunnel (as of 4.5.1 always true)
	    Any FC message is passed from TCPTunnel to fcConnect for action
   b. TLM accepts a connection from TLMC, hands to a new TCPHandoff thread
      TCPHandoff reads the sessionVariable and other end port from TLMC connection
      TCPHandoff checks for existence of tunnel with matching
        IPAddress and otherEndPort; also sessionVariable unless it is all zeros
        indicating it is not used
      if match is found it must be video for existing tunnel. TCPHandoff
        starts a new TCPVideo thread and returns
      TCPHandoff looks for open hostTable entry; if none returns
      TCPHandoff enters this IPAddress and sessionVariable in hostTable
      if webAuthentication is in use, TCPHandoff checks for another connection
        with same username and IPaddress as the new one, and deletes the old
        (also does this if there is IPException establishing streams for
        the connection, but for any username at that address)
    2. In parallel:
    a. FCListen calls Auth.authenticate to validate the username and
         sessionVariable; tries 3 times, then calls closeTunnel and returns
       Upon authentication, FCListen tries 10 times at 1000ms intervals
         to confirm hostTable entry for this host has open tunnel resulting
         from successful TCPHandoff; on failure, calls closeTunnel and returns
       If open tunnel is confirmed, FCListen updates heartbeat times to
         current system clock and sets authenticatedHost, cleans up any loose-end
         connections from same hostm, then sleeps while network performance is
         checked (with timeout TUNNEL_ACTIVATION_TIMEOUT) and tunnel is activated
       FCListen then goes into normal operation mode, accepting FCClient input
         and taking appropriate action
       Upon disconnect of FC connection, FCListen calls closeTunnel and returns
    b. TCPTunnel thread sets TcpIsOpen for this host true
       TCPTunnel and TCPVideo threads wait indefinitely for authentication of
         host table entry from 2a; escape and return if TcpIsOpen is turned
         off for this host
       Upon authentication, TCPTunnel actiavtes tunnel; at this point it and
         TCPVideo go into normal operation  mode, accepting data inputs from
         TLMC and retransmitting them to other connected hosts via queueTunnel
         and TunnelSend
       Upon disconnect of TLMC connection, TCPTunnel and TCPVideo call
         closeTunnel and return
    3. In normal operation, FCClient sends a heartbeat message to FCListen
         every 10 seconds and TLMC sends a heartbeat message to TCPTunnel every
         1 second; this is not done when sending FC messages through TCPTunnel.
       Heartbeat checks the hostTable for all authenticated hosts once per second
         to ensure they are sending hearbeats; if allowable interval is expired
         (20000ms for FC, 10000ms for TLMC), calls closeTunnel for that host
    4. Whenever closeTunnel is called, it sends shutdown code, closes all
       connections for that host and resets hostTable position
	   
    Auto-reconnect added to the above 15 Dec 2009: if dataRateKbps in setup message
	is zero, this indicates a client has lost connection and is trying to reconnect
	without full setup, which would lose all state. TLM.closeTunnel preserves 
	enough state to permit this, using sessionVariable to identify the host. If the
	user logs in again the reconnect state is deleted. All state is deletd five
	minutes after last user logs out unless autoscheduling has set classInSession. 
*/	

  SocketChannel thisSocketChannel;
  InetAddress thisAddress;
  int newOtherEndPort;  // not the client's port, if NAT is involved
  String newSessionVariable, thisHostAddress;
  byte[] setupMessage = new byte[74];
  int firstByte = -1;

  public TCPHandoff(SocketChannel newSocketChannel)
  {
    thisSocketChannel = newSocketChannel;

  } // end constructor
    
    
  // reads SocketChannel until it gets numberOfBytes; puts them in recvBuffer
  // sleeps between reads to avoid thread lockup
  // quits after timeCount sleepTimes if timeCount > 0
  // returns true if disconnected, false otherwise
  // side-effect: sets firstByte to first byte read 
  // (or -1 if disconnected, -2 if timeout)
  boolean readBytesWithTimer(byte[] buffer, 
                             int numberOfBytes, 
                             int offset, 
                             int timerCount)
  {
    int countTries = timerCount;
    ByteBuffer bb = ByteBuffer.allocate(numberOfBytes);
    while(!stopTLM)
    {
      // read the channel and catch disconnect or exception
      try
      {
        if(thisSocketChannel.read(bb) < 0)
        {
          firstByte = -1;
          return true;
        }
      }
      catch(ClosedChannelException cce)
      {
        return true;
      }
      catch(IOException ioe)
      {
        System.out.println("*** IOException in TCPHandoff for:" +
                           thisAddress + "/" + newOtherEndPort + 
                           ":" + ioe.getMessage());
        ioe.printStackTrace(System.out);
        firstByte = -1;
        return true;
      }
      
      // when buffer is full copy ByteBuffer to recvBuffer and return,
      // making sure not to exceed length of recvBuffer
      if(!bb.hasRemaining())
      {
        int numberToTransfer = numberOfBytes;
        if((numberToTransfer + offset) > buffer.length)
          numberToTransfer = buffer.length - offset;
        bb.rewind();
        bb.get(buffer,offset,numberToTransfer);
        firstByte = buffer[0];
        return false;
      }
      
      // if timer exhausted, set firstByte to -2 and return
      if(timerCount > 0)
        if(countTries-- <= 0)
        {
          firstByte = -2;
          return false;
        }
      
      // sleep a while and try again
      try{Thread.sleep(sleepTime);}
      catch(InterruptedException ie){return false;}
      
    }// end while(!stopTLM)
      
    firstByte = -1;
    return false;
      
  }// end readBytes()
    
  
    
  // writes data to non-blocking channel
  // uses sleep to avoid lockup when receiver's buffer is full
  // returns true if disconnected
  boolean writeThisOutStream(byte[] data,int length)
  {
    // move data to a ByteBuffer to send
    ByteBuffer bb = ByteBuffer.wrap(data,0,length);
    
    // keep trying until the buffer gets sent
    while(!stopTLM)
    {
      // write the data to outstream
      try
      {
        thisSocketChannel.write(bb);
      }
      catch(ClosedChannelException cce)
      {
        return true;
      }
      catch(Exception e)
      {
        System.out.println("*** Exception in TunnelSend for " + 
                           thisAddress + "/" + newOtherEndPort + 
                           " :" + e);
        e.printStackTrace(System.out);
        return true;
      }
      
      // if data was not all written, wait sleepTime and try again
      if(bb.hasRemaining())
        try{Thread.sleep(sleepTime);}
        catch(InterruptedException ie)
        {
          return true;
        }
      else return false;
      
    }// end while(!stopTLM)
    
    return false;
    
  }// end writeThisOutStream()
    

  public void run()
  {
    long timeNow = System.currentTimeMillis();
	  long tunnelTimeoutMillisec = TUNNEL_ACTIVATION_TIMEOUT*1000;
	
    // extract the address and other end port for the connection
    thisAddress = thisSocketChannel.socket().getInetAddress();
    thisHostAddress = thisAddress.getHostAddress();
    if(thisHostAddress.equals(loopbackHostAddress))
    {
      thisAddress = localAddress;
      thisHostAddress = localHostAddress;
    }
    newOtherEndPort = thisSocketChannel.socket().getPort();
    System.out.println("\n" + dateTime() + 
	                   " new TCP connection request received from " +
                       thisAddress + "/" + newOtherEndPort + 
                       " (this may be external NAT address)");

    // open streams and read the setup message from the connection
    // message format client startup TLMC to TLM:
    // {
    //   transaction byte[1] value "64"
    //   messageLength unsigned short value 21 // 2 bytes
    //   clientVersionByte1 // byte
    //   clientVersionByte2 // byte
    //   clientVersionByte3 // byte
    //   sessionVariable[64]// bytes
    //   otherEndPort unsigned short // 2 bytes
    //   netCapacitykbps unsigned short // 2 bytes
    // }
    try
    {
      // read setupHeader
      if(readBytesWithTimer(setupMessage,3,0,(int)10000/sleepTime))
      {
        System.out.println("disconnected while receiving setup for " + 
                           thisAddress + "/" + newOtherEndPort);
        return;
      }
      if(firstByte == -2)
      {
        System.out.println("timed out waiting setupHeader for " +
                           thisAddress + "/" + newOtherEndPort);
        thisSocketChannel.close();
        return;
      }
           
      // verify this is a setup message
      int setupMessageLength = bytesToUshort(setupMessage[1], setupMessage[2]);
      if((setupMessage[0] != 64) || (setupMessageLength != 71))
      {
        System.out.println(dateTime() + " bad setup message received from " + 
                           thisAddress + "/" + newOtherEndPort + 
                           " TC:" + setupMessage[0] + 
                           " length:" + setupMessageLength);
        thisSocketChannel.close();
        return;
      }
      
      // read remainder of the full setup message
      if(readBytesWithTimer(setupMessage,71,3,0))
      {
        System.out.println("disconnected while receiving setup for " + 
                           thisAddress + "/" + newOtherEndPort);
        return;
      }
      newSessionVariable = new String(setupMessage,6,64);
    }
    catch (Exception e)
    {
      // this sometimes happens on reconnect attempt, if the
      // old connection has not been fully cleared out yet
      if(thisSocketChannel == null)return;
      System.out.println("Exception processing setup for " + thisAddress + " :" + e);
      e.printStackTrace(System.out);
      try{thisSocketChannel.close();}catch(IOException ioe2){}
      return;
    }
    System.out.println(thisAddress + "/" + newOtherEndPort +
                       " session variable:" + newSessionVariable);
    
    // extract the other end port number and appfile download data from setup
    int tlmcTcpPort = bytesToUshort(setupMessage[70],setupMessage[71]);// client's port
    int dataRateKbps = bytesToUshort(setupMessage[72],setupMessage[73]);
   
    // autoreconnect is indicated by zero dataRate in setup message
    // but is not possible when using file authentication  because
    // sessionVariable comes from authenticating via Moodle
    boolean thisIsReconnect = (dataRateKbps == 0) && (authFileName == null);
    if(thisIsReconnect)System.out.println("reconnect received from " + thisHostAddress);
          
    // look for an entry with thisAddress 
    // that also has matching sessionVariable
    // (could be video or autoreconnect)
    int hostIndex = videoAddressConfigured ? 1 : 0;
    for (; hostIndex < maxConnections; hostIndex++)
    {
      if(hostAddr[hostIndex] != null)
        if(!tunnelClosing[hostIndex])
          if(thisHostAddress.equals(hostAddr[hostIndex]) &&
             newSessionVariable.equals(sessionVariable[hostIndex]))
            break;
/*debugx			
	    // check for expired host table entry; clean it out so it can be used
	    if(hostAddr[hostIndex] != null)
	    {
	      long deltaT = timeNow - uRecvTime[hostIndex];
        
        // authenticatedHost is true when pending reconnect 
		    if((!authenticatedHost[hostIndex] && (deltaT > tunnelTimeoutMillisec)) ||
               (deltaT > reconnectTimeoutMillisec))
	      {
	        // set tunnelIsActive to allow closeTunnel to work
          tunnelIsActive[hostIndex] = true;
	        closeTunnel(hostIndex,false,"found timed out host table entry");
          initializeAutoReconnect(hostIndex);
	      }
	    }// end if((hostAddr[hostIndex] != null)
debugx*/
    }// end for(; hostIndex...
    
    // if host with same address and matching session variable is found, this could be
    // either a reconnection or the second (video) connection from this client 
    if(hostIndex < maxConnections)
    {
      // check for reconnect and matching IP address 
      if(thisIsReconnect)
      {
        // our end of connection may not have sensed disconnect
	      if(connectedByTCP[hostIndex])
		    {
	        closeTunnel(hostIndex,false,"tunnel status 'open' on reconnect");
			
          // sleep so tunnel can close
		      try{Thread.sleep(1);}catch(InterruptedException ie){}
        }
             
        // flag that this is a reconnect so TCPTunnel can treat accordingly
        reconnected[hostIndex] = true;
      }
        
      // if not reconnect it is a parallel video connection; 
      // start a video tunnel and return
      else 
      {   
        if(dataRateKbps == 0)
        {
          closeTunnel(hostIndex,true,
                      "attempted autoreconnect - not supported with file authentication");
          return;
        }
        if((videoPort != 0) || (multicastTunnelPort != 0))
        {
          if(!connectedByTCP[hostIndex])
          {
            closeTunnel(hostIndex,false,
                        "system error - attempted TCP video connection without audio connection");
            return;
          }
          tcpVideoSocketChannels[hostIndex] = thisSocketChannel;
          TCPVideo videoTunnel;
          try
          {
            // use a small buffer so excess video traffic is dumped  *** its not dumped - locks up! debugx
            // rather than causing long delay in delivery
            // quickfix patch 9-14-07 increased buffers from 12887 because
            // small size was truncating UDP video packets
            thisSocketChannel.socket().setReceiveBufferSize(videoSocketReceiveBuffersize);
            thisSocketChannel.socket().setSendBufferSize(videoSocketSendBuffersize);
            connectedVideoByTCP[hostIndex] = true;
            
            // instantiate a TCPVideoTunnel thread and pass the
            // SocketChannel to it for input
            videoTunnel = new TCPVideo(hostIndex);
          }
          catch (IOException ioe)
          {
            System.out.println(dateTime() + 
                              " unable to complete TCP video connection for host:" +
                               tcpRecvSocketChannel.socket().getInetAddress());
            if (thisSocketChannel != null)
              try{thisSocketChannel.close();}catch (IOException ioe2) {}
            return;
          }
          Thread videoThread = new Thread(videoTunnel);
          videoThread.start();
          System.out.println(dateTime() + " opened TCP video connection for host " + hostIndex +
                             " " + hostAddr[hostIndex] + "/" + newOtherEndPort);
          videoThreads[hostIndex] = videoThread;
          tcpTunnelInstance[hostIndex].setTCPVideo(videoTunnel);
          return;
        }// end if((videoPort != 0) || (multicastTunnelPort != 0))
      }// end else/if(thisIsReconnect...
    } // end if(hostIndex < maxConnections)
	
    // host is not in table; this can't be a reconnect;
    // reject it if it claims to be
    else 
    {
      // application level NACK - connection not acceptable
      if(thisIsReconnect)
      {
        System.out.println("\n*** " + dateTime() +
                           " rejected invalid reconnect attempt from "+
                           thisAddress + "/" + newOtherEndPort + "\n");
        byte[] reject = {0};
        writeThisOutStream(reject,1);
        try{thisSocketChannel.close();}catch (IOException ioe) {}
        return;
		
      }// end if(thisIsReconnect)
    }// end else/if(hostIndex < maxConnections)
     
    // on reconnect, clear out pending host delete timeout
    if(thisIsReconnect)
    {
		  if(tcpHandoffs[hostIndex] != null)
        if(tcpHandoffs[hostIndex].runningWait)
          tcpHandoffs[hostIndex].runAfterWait.interrupt();
            
      // we now have a new port at other end
      otherEndPort[hostIndex] = tlmcTcpPort;
    }
    else// (!thisIsReconnect)
    {
      // thisAddress/sessionVariable is not in host table, so insert it;
      // then start a new TCPTunnel
      hostIndex = addToHostTable(thisAddress,
                                 newSessionVariable,
                                 tlmcTcpPort,
                                 newOtherEndPort);
         
      // if the host table is full, reject the connection
      if (hostIndex < 0)
      {
        System.out.print("\nHost table full when connection attempted by " +
                         thisAddress + "/" + newOtherEndPort +
                         " - connection rejected\n");
        try
        {
          thisSocketChannel.socket().close();
        }
        catch (IOException ioe)
        {
          // other end closed the connection first
        }
        return;
		
      } // end if( hostIndex >= maxConnections )
	  
      // for initial connection, don't hold over previous bona fides
      clearUserDataFromHostTable(hostIndex);
	
    }// end else/if(thisIsReconnect)
    
    // flag this as TCP connection
    protocolIsTCP[hostIndex] = true;
    	  
    // save this TCPHandoff so we can invoke its ClearThisOneAfterWait
    tcpHandoffs[hostIndex] = this;
    
    // complete host table entries and connection parameters
    try 
    {   
      thisSocketChannel.socket().setSoLinger(false,0);
    }   
    catch(SocketException se)
    {   
      System.out.println("*** warning - unable to setSoLinger on Socket in TCPHandoff");
    }   
    tcpSocketChannels[hostIndex] = thisSocketChannel;
    uRecvTime[hostIndex] = System.currentTimeMillis();
    connectedByTCP[hostIndex] = true;
    if(!thisIsReconnect)authenticatedHost[hostIndex] = false;
    try
    {
      // set socket options for expanded buffers
      tcpSocketChannels[hostIndex].socket().setReceiveBufferSize(tcpSocketReceiveBuffersize);
      tcpSocketChannels[hostIndex].socket().setSendBufferSize(tcpSocketSendBuffersize);
    }
    catch (IOException ioe)
    {
      closeTunnel(hostIndex, false, "unable to complete TCP connection for host");
      return;
    }
       
    // spawn a TCPTunnel to take over the streams
    TCPTunnel tcpTunnel = new TCPTunnel(hostIndex,
                                        thisSocketChannel,
                                        setupMessage);
    Thread tcpTunnelThread = new Thread(tcpTunnel);
    loginTime[hostIndex] = System.currentTimeMillis();
    tcpTunnelInstance[hostIndex] = tcpTunnel;
    tcpTunnelThreads[hostIndex] = tcpTunnelThread;
    tcpTunnelThread.start();
	
    // if this is a reconnect, wait for tunnel 
    // to open, then reset floorHolder
    if(thisIsReconnect)
    {
      // update user and controller counts
      if(!tcpIsOpen[hostIndex])
        incrementUserCounts(hostIndex);
        
      // wait until tcpTunnel is running again  
      try
      {
        while(!tcpIsOpen[hostIndex])Thread.sleep(200);
      }
      catch(InterruptedException ie){}
              
      // application-level ACK reconnect
      byte[] connectAck = {1};
      if(writeThisOutStream(connectAck,1))return;
      System.out.println("\n" + dateTime() + " auto-reconnected host " + 
                         hostIndex + " " + thisAddress + "/" + 
                         newOtherEndPort + " user " + username[hostIndex]);
      fcConnect.reconnectFloorHolderStatus(hostIndex);
    }
    return;

  } // end run()
	

  // thread to reset autoschedule state five minutes after disconnect
  ClearThisOneAfterWait clearThisOneAfterWait;
  Thread runAfterWait;
  Boolean runningWait = new Boolean(false);
	
	
  // clear state for this connection five minutes after disconnect
  void clearThisConnectionAfterWait(int selectedHost)
  {
    if(runningWait)return;
	  clearThisOneAfterWait = new ClearThisOneAfterWait(selectedHost);
    runAfterWait = new Thread(clearThisOneAfterWait);
    runAfterWait.start();
    
  }// end clearThisConnectionAfterWait()

	
  // clears host table entry after a wait; length
  // of wait depends on whether that user is instructor
  class ClearThisOneAfterWait implements Runnable
  {
	long timeoutWait;
	int selectedHost;
	  
    public ClearThisOneAfterWait(int selectedHostRef)
    {
	    selectedHost = selectedHostRef;
      if(instructorPrivilege[selectedHost] == 0)
		    timeoutWait = normalReconnectTimeout;
      else 
		    timeoutWait = instructorReconnectTimeout;
	  }
	  
	  public void run()
	  {
      runningWait = true;
      try
      {
        Thread.sleep(timeoutWait);
      }
      catch (InterruptedException ie)
      {
        runningWait = false;
        return;
      }
      
      // if tunnel was reactivated during wait, no need to clear auto-reconnect
      if(tunnelIsActive[selectedHost])return;
      
      // otherwise, finish clearing host table state and print to log
      initializeAutoReconnect(selectedHost);
      clearHostDataFromHostTable(selectedHost);
      runningWait = false;
      System.out.println("\n" + dateTime() + 
						 " cleared auto-reconnect option for host " +
						 selectedHost);
      closeTunnel(selectedHost,true," disconnected host did not reconnect");
			
	  }// end run()
	
  }// end ClearThisOneAfterFiveMinutes class

}// end TCPHandoff class
