/*
 * TCPTunnel.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author  mpullen & students
 * @version 5.2.4
 * @modified JMP 3/1/2013
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

/* TCPTunnel.java class logic:

->Until the TCP connection is closed
-> read TCP message
-> strip port from first two byte of message and locate it in multicast port array
   or unicast port array
   if it is in the array, capture the corresponding port (unicast if it is in
   the multicast array and vice versa)  and flag that port as TCP for this host;
   else go back to reading TCP
-> if a loopback time probe message is received, send it back to the source
-> if not loopback or URL, it should be forwarded in the following ways
   -> update unicast last recv time array
   -> if the multicast group has been a source of messages (any port) within time
      window send unicast-udp message to multicast group with a modified TTL(15)
   -> send a copy of unicast UDP message to each unicast address that has been active
      within time window except the address it came from (source address) and queue
      it to be sent to each TCP and UDP tunnel address except the address it came
      from (tunnel gets the original message with port number prepended and revised
      transaction code)
-> if there is a record/play control connection open and the message bears rpCPort,
   send a copy through the tunnel connection to the control
-> When the connection is closed, set corresponding unicast TCP array to 0 and return
**************************************************************************** */
//package TLM;
import java.io.*;
import java.net.*;
import java.nio.channels.SocketChannel;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;

public class TCPTunnel extends TLM implements Runnable
{
    private MulticastSocket mcSocket;
    private DatagramSocket udpSocket;
    private SocketChannel thisSocketChannel;
    DatagramPacket sendMessage, sendPacket;
    TCPVideo thisHostVideoTunnel = null;
    private String remoteHost;
    byte[] recvBuffer, sendBuffer, setupMessage;
    private int portIndex, reportedMinRTT, reportedRTT;
    private int thisHostIndex;// position of this host in hostAddr[] from TLM
    private int thisUdpPort, udpLength, messageCount, loopbackCount, appNumber;
    private long startTime = 0; // time when a message is received
    private long checkTime = 0; // check for sending without being floorholder
    private boolean rpCIsThisHost, nonVideoPort, loginViaTCPTunnel = false;
    private boolean thisIsController, thisIsSupercontroller;
    private boolean signalSlideDone = false;
    int firstByte = -1;
    int audioControlPort, wbdControlPort, videoControlPort;
    int localFloorHolder; // floorholder for this breakout group
                          // (or whole group if no breakout)

//**********************************************************

    // constructor
    public TCPTunnel( int hostIndexParm,
                      SocketChannel tcpRecvSocketParm,
                      byte[] setupMessageParm )
    {
      thisHostIndex = (hostIndexParm);
      localFloorHolder = fcConnect.getFloorHolder(thisHostIndex);
      thisSocketChannel = tcpRecvSocketParm;
      setupMessage = setupMessageParm;
      reportedMinRTT = 0;
      reportedRTT = 0;
      messageCount = 0;
      loopbackCount = 0;
      audioControlPort = audioPort + 1;
      wbdControlPort = wbdPort + 1;
      videoControlPort = videoPort + 1;
      thisIsController = false;
      thisIsSupercontroller = false;
      
      // network structures
      recvBuffer = new byte[ MAX_MESSAGE+5 ];
      sendPacket = new DatagramPacket( recvBuffer, 0);
      sendBuffer = new byte[ MAX_MESSAGE ];
      
      // get local value for address of connecting host
      // and determine whether the record/playback control is this host
      remoteHost = hostAddr[ thisHostIndex ];
      if(rpAddr == null)rpCIsThisHost = false;
      else rpCIsThisHost = ( rpAddr.equals(hostInetAddr[thisHostIndex]) );
    }

    // sets local value of floorHolder when it changes
    // so we don't incur overhead of getFloorHolder every message 
    void setTCPTunnelFloorHolder(int hostIndex)
    {
      localFloorHolder = hostIndex;
      if(thisHostVideoTunnel != null)
        thisHostVideoTunnel.localFloorHolder = hostIndex;
    } 
    
    
    // sets local value of TCPVideo associated with this hsot
    void setTCPVideo(TCPVideo newTCPVideo)
    {
      thisHostVideoTunnel = newTCPVideo;
    }
    
    
    // reads thisSocketChannel until it gets numberOfBytes; puts them in recvBuffer
    // sleeps between reads to avoid thread lockup
    // returns true if disconnected, false otherwise
    // side-effect: sets firstByte to first byte read (or -1 if disconnected)
    boolean readBytes(int numberOfBytes, int offset)
    {      
      ByteBuffer bb = ByteBuffer.allocate(numberOfBytes);
      while(!stopTLM)
      {
        // read the channel and catch disconnect or exception
        try
        {
          if(thisSocketChannel.read(bb) < 0)
          {
            firstByte = -1; 
            return true;
          }
        }
        catch(ClosedChannelException cce)
        {
          return true;
        }
        catch(IOException ioe)
        {
          System.out.println("*** IOException in TCPTunnel for host:" +
                             thisHostIndex + ":" + ioe.getMessage());
          ioe.printStackTrace(System.out);
          firstByte = -1; 
          return true;
        }
        
        // when buffer is full copy ByteBuffer to recvBuffer and return,
        // making sure not to exceed length of recvBuffer
        if(!bb.hasRemaining())
        {
          int numberToTransfer = numberOfBytes;
          if((numberToTransfer + offset) > recvBuffer.length)
            numberToTransfer = recvBuffer.length - offset;
          bb.rewind();
          bb.get(recvBuffer,offset,numberToTransfer);
          firstByte = recvBuffer[0]; 
          return false;
        }
        
        // sleep a while and try again
        try{Thread.sleep(sleepTime);}
        catch(InterruptedException ie){return false;}
      
      }// end while(!stopTLM)
      
      firstByte = -1; 
      return false;
      
    }// end readBytes()
    
    
      
    // writes data to non-blocking channel
    // uses sleep to avoid lockup when receiver's buffer is full
    // returns true if disconnected
    boolean writeThisOutStream(byte[] data,int length)
    {
      // move data to a ByteBuffer to send
      ByteBuffer bb = ByteBuffer.wrap(data,0,length);
  
      // keep trying until the buffer gets sent
      while(!stopTLM)
      {
        // write the data to outstream
	      try
	      {
          thisSocketChannel.write(bb);
        }
        catch(ClosedChannelException cce)
        {
          return true;
        }
        catch(Exception e)
        {
	        closeTunnel(thisHostIndex,false,
            "*** Exception in TunnelSend for host:" + thisHostIndex + ":" + e);
          e.printStackTrace(System.out);  
	        return true;
        }
        
        // if data was not all written, wait sleepTime and try again
        if(bb.hasRemaining())
          try{Thread.sleep(sleepTime);}
          catch(InterruptedException ie)
          {
            return true;
          }
        else break;
        
      }// end while(!stopTLM)
            
      return false;
      
    }// end writeThisOutStream()
     
    
    public void run()
    {
      try
      {
        boolean congestionWarned = false;
        int badCode = 0;
        int testRTT = 0;
         
        // set tcpIsOpen array to indicate that this particular
        // unicast hostIndex is sending to tunnelTCPPort
        tcpIsOpen[thisHostIndex] = true;
System.out.println("TCPTunnel running host:"+thisHostIndex+" addr:"+remoteHost+" recon:"+reconnected[thisHostIndex]);//debugxx
        
        // test for error
        int udpLength = 0;
        byte[] sessionVariable = null;
        recvBuffer = new byte[ MAX_MESSAGE+5 ];
        
        // for reconnect, setup protocol is not used
        if(reconnected[thisHostIndex])
        {
          connectedByTCP[thisHostIndex] = true;//debugx
          activateTunnel(thisHostIndex);
          if(thisHostIndex == previousFloorHolder)
            localFloorHolder = thisHostIndex;
        }
        else
        {
          // first message through tunnel could be floor control connect;
          // read a message and check for encapsulated floor control
          while( !authenticatedHost[ thisHostIndex ] && !stopTLM )
          {
            // wait until 5-byte header is available
            if(readBytes(5,0))
            {
              // lost connection
              closeTunnel( thisHostIndex, true, "TCPTunnel disconnect" );
              return;
            }
            if(stopTLM)return;
            countTcpRecv++;
            firstByte = recvBuffer[0];
            
            // keep track of the time the message arrived
            startTime = System.currentTimeMillis();
            uRecvTime[ thisHostIndex ] = startTime;
            
            // if heartbeat message, go back to read the next message
            if( firstByte == 68 )
            {
              if(recvBuffer[3] != 1)continue;
              if((recvBuffer[1] != 0) || (recvBuffer[2] != 0) || (recvBuffer[4] != 0))continue; 
              closeTunnel(thisHostIndex,false,"client sent disconnect before setup");
              break;
            }
            
            // extract UDP message length and port from received header
            udpLength = bytesToUshort(recvBuffer[1],recvBuffer[2]);
            thisUdpPort = bytesToUshort(recvBuffer[3],recvBuffer[4]);
            
            // all FC messages have transactionCode 67
            if(firstByte != 67)break;
		    
            // if this is a floor control message, port will be zero
            if(thisUdpPort != 0)
            {
              closeTunnel(thisHostIndex, true, "invalid port code on FC login");
              return;
            }
            
            // read the UDP message from the TCP stream
            if( udpLength > 0 )
            {
              if(readBytes(udpLength,5))
              {
                closeTunnel(thisHostIndex, true,
                            "Error receiving UDP message from " + tunnelTcpPort);
                return;
              }
              int readLength = udpLength;

            }// end if( udpLength > 0 )
             
            // there are three possible types of FC message at this point
            if((recvBuffer[5] == 11) && (udpLength == 69))// sessionData message
            {
              sessionVariable = new byte[64];
              System.arraycopy(recvBuffer,8,sessionVariable,0,64);
              int tlmcTcpPort = bytesToUshort(recvBuffer[72],recvBuffer[73]);
              if(fcConnect.processSessionVariable(thisHostIndex,
                                                  sessionVariable,
                                                  hostAddr[thisHostIndex],
                                                  tlmcTcpPort))
                return;
              setupReceived[thisHostIndex] = true;
            
            }
            if((recvBuffer[5] == 0) && (udpLength == 33))// login data message
            {
              int fcLength = udpLength-3;
              byte[] loginMessage = new byte[fcLength];
              System.arraycopy(recvBuffer,8,loginMessage,0,fcLength);
              if(fcConnect.processLoginMessage(thisHostIndex,
                                               loginMessage,
                                               sessionVariable))
                return;
              loginViaTCPTunnel = true;
              break;
            }
            if((recvBuffer[5] == 72) && (udpLength == 5))// no tunnel floor control message
              break;
          
          }// end while( !authenticatedHost[ thisHostIndex ] && !stopTLM)
		  
          // check for network read error
          if(firstByte == -1)
          {
            // lost connection
            closeTunnel( thisHostIndex, true, "TCPTunnel closed while reading firstByte" );
            return;
          }
           
          // wait here until authentication is complete, timeout at maxLoginTime
          long endTime = System.currentTimeMillis() + maxLoginTime;
                 
          // read the tunnel and ignore input 
          // (except login timeout) until authenticated
          try
          {
            while(!authenticatedHost[ thisHostIndex ] && !stopTLM)
            {
              // show message received at this time to avoid connect race
              uRecvTime[ thisHostIndex ] = System.currentTimeMillis();
              
              // check for client disconnect
              if( !tcpIsOpen[ thisHostIndex ] )
              {
                closeTunnel( thisHostIndex, true, "TCPTunnel tcpIsOpen false" );
                return;
              }
              Thread.sleep(sleepTime);
              
              // check for login timeout and shut down if needed
              if( endTime < System.currentTimeMillis() )
              {
                byte[] timeoutBuffer = {69,0,0,3,0};
                writeThisOutStream(timeoutBuffer,5);
                Thread.sleep(1000);
                closeTunnel( thisHostIndex, false, "TCPTunnel login timeout" );
                return;
              } // end if( endTime...
            } // end while( !authenticatedHost[...
          } // end try
          catch( InterruptedException ie ){return;}
          
          // get floor controller status
          thisIsController =
            (controlPrivilege[ thisHostIndex ] > '0');
            
          // supercontroller can add to WBD even though not floorHolder
          thisIsSupercontroller =
            (controlPrivilege[ thisHostIndex ] > '0');
		  
          // unless this is a reconnection, send startup messsage
          if(!reconnected[ thisHostIndex ])
          {
            // message format client startup TLMC to TLM:
            // {
            //   transaction byte[1] value "64"
            //   messageLength unsigned short value 71 // 2 bytes
            //   clientVersionByte1 // byte
            //   clientVersionByte2 // byte
            //   clientVersionByte3 // byte
            //   sessionVariable[64]// bytes
            //   otherEndPort unsigned short // 2 bytes
            //   netCapacitykbps unsigned short // 2 bytes
            // }
            String clientVersion = new String(setupMessage,3,3);
            boolean clientIsOK = (setupMessage[0] == 64) &&
                                 (setupMessage[1] == 0) &&
                                 (setupMessage[2] == 71) &&
                                 (clientVersion.compareTo(minClientVersion) >= 0) &&
                                 (clientVersion.compareTo(maxClientVersion) <= 0);
             
            // extract networkCapacity specified in client's .nwm file and sent in setup message
            netCapacitykbps[thisHostIndex] = bytesToUshort(setupMessage[72],setupMessage[73]);
             
            // reply with server setup message
            //
            // message format client startup TLM to TLMC
            // {
            //  transaction byte value "64"
            //  messageLength u_short value 25    // 2 bytes
            //  clientVersionByte1                // byte
            //  clientVersionByte2                // byte
            //  clientVersionByte3                // byte
            //  connectionThreshold u_short       // 2 bytes
            //  congestionThreshold u_short       // 2 bytes
            //  recorderTCPPort u_short           // 2 bytes
            //  probeInterval unsigned int        // 2 bytes
            //  appCount byte
            //  whiteboardRingSize unsigned short // 2 bytes
            //  roomID byte[11]                   //MISTC-ID
            // }
            // send configuration to the client
            sendBuffer[0] = 64; // transaction code
            // message length does not include 5 byte header
            int messageLength = 25;
            sendBuffer[1] = intToByte1( messageLength );
            sendBuffer[2] = intToByte2( messageLength );
            sendBuffer[3] = 0;
            sendBuffer[4] = 0;
            sendBuffer[5] = 0;
            sendBuffer[6] = intToByte1( connectionThreshold );
            sendBuffer[7] = intToByte2( connectionThreshold );
            sendBuffer[8] = intToByte1( congestionThreshold );
            sendBuffer[9] = intToByte2( congestionThreshold );
            sendBuffer[10] = intToByte1( 0 ); // previously rpTcpPort
            sendBuffer[11] = intToByte2( 0 ); // previously rpTcpPort
            sendBuffer[12] = intToByte1( probeInterval );
            sendBuffer[13] = intToByte2( probeInterval );
            sendBuffer[14] = (byte)numberOfApplications;
            sendBuffer[15] = intToByte1(whiteboardRingSize);
            sendBuffer[16] = intToByte2(whiteboardRingSize);
            System.arraycopy(roomIDBytes,0,sendBuffer,17,11);
            
            // if the client is OK, send zeros in client field;
            // otherwise, send new version desired
            if(!clientIsOK)
            {
              sendBuffer[3] = maxClientVersionBytes[0];
              sendBuffer[4] = maxClientVersionBytes[1];
              sendBuffer[5] = maxClientVersionBytes[2];
            }
            
            // reply setup message is assembled; send and flush the TCP segment
            if(writeThisOutStream(sendBuffer,messageLength+5))
              closeTunnel ( thisHostIndex, false,
                            "error sending setup reply to host " + thisHostIndex);
                          
            // if the client is not OK, disconnect
            if(!clientIsOK)
            {
              try{Thread.sleep(1000);}catch(InterruptedException ie){return;}
              closeTunnel ( thisHostIndex, false,
                            "client version " + clientVersion + " incompatible" );
              return;
            }
          }// end if(!reconnected[ thisHostIndex])
        }// end else/if(reconnected...
		
        // start checking floor discipline after 20 seconds
        checkTime = System.currentTimeMillis()+20000;
		
        // read in data continuously, terminated by the remote host close
        while( !stopTLM )
        {
          // allocate a new read buffer so it can be saved until
          // written to all outstreams asynchronously
          recvBuffer = new byte[ MAX_MESSAGE+5 ];

          // read firstByte TCP non-blocking and test for error
          udpLength = 0;
          readBytes(1,0);
          if(firstByte == -1)
          {
            // lost connection
            closeTunnel( thisHostIndex, true, "TCPTunnel disconnected on read" );
            return;
          }
          
          // check whether protocol transaction code valid
          int badBytes = 0;
          if( firstByte != 67 )
          {
            if( (firstByte != 65) && (firstByte != 68) && (firstByte != 70) &&
                (firstByte != 71) && (firstByte != 75) && (firstByte != 76) &&
                (firstByte != 77) && (firstByte != 79) && (firstByte != 105) && 
                (firstByte != 103) && (firstByte != 106))
            {
              if(badBytes++ == 0)badCode = firstByte;
            }
          }
          // if not valid, read stream until a valid code is found
          // (ignore 77, it is setup only)
          if( badBytes > 0 )
          {
            while( (firstByte != 65) && (firstByte != 67) && (firstByte != 68) &&
                   (firstByte != 70) && (firstByte != 71) && (firstByte != 75) &&
                   (firstByte != 76) && (firstByte != 77) && (firstByte != 79) &&
                   (firstByte != 103) && (firstByte != 105) && (firstByte != 106))
            {
              if(readBytes(1,0))
              {
                // lost connection
                closeTunnel( thisHostIndex, true,
                             "TCPTunnel closed while reading firstByte" );
                return;
              }
              if(stopTLM)return;
            }
            System.out.println("*** received invalid protocol code:" + badCode
              + " from host #" + thisHostIndex + " - read " + badBytes
              + " input stream bytes to next valid code" );
              
          } // end if( badBytes > 0 )

          // read the rest of the five-byte header
          countTcpRecv++;
          if(readBytes(4,1))
          {
            // lost connection
            closeTunnel( thisHostIndex, true, "TCPTunnel error reading header" );
            return;
          }
          
          // keep track of the time the message arrived
          startTime = System.currentTimeMillis();
          uRecvTime[ thisHostIndex ] = startTime;
          if( thisHostIndex == localFloorHolder )checkTime = startTime;
          
          // extract UDP message length from received header
          // following code is inline version of bytesToUshort()
          // udpLength = bytesToUshort( recvBuffer[1], recvBuffer[2] );
          // done inline for efficiency
          int highByte = recvBuffer[1];
          if(highByte < 0)highByte = 127 - highByte;
          int lowByte = recvBuffer[2];
          if(lowByte < 0)lowByte = 127 - lowByte;
          udpLength = 256*highByte + lowByte;
          
          // extract UDP port number from received TCP segment
          // thisUdpPort = bytesToUshort( recvBuffer[3], recvBuffer[4] );
          // done inline for efficiency
          highByte = recvBuffer[3];
          if(highByte < 0)highByte = 127 - highByte;
          lowByte = recvBuffer[4];
          if(lowByte < 0)lowByte = 127 - lowByte;
          thisUdpPort = 256*highByte + lowByte;
          
          // check port number for validity
          boolean portIsGood = false;
          int controlPort = 0;
          for( portIndex = 0; portIndex < numberOfApplications; portIndex++ )
          {
            if( thisUdpPort == udpPort[ portIndex ] )
            {
              portIsGood = true;
              break;
            }
            if( thisUdpPort == ( udpPort[ portIndex ]+1) )
            {
              portIsGood = true;
              controlPort = 1;
              break;
              
            } // end if( thisUdpPort
          } // end for( portIndex
           
          // if an error message from TLMC, a URL, heartbeat, download
          // loopback request, or FC message, port need not be good
          if( !portIsGood )
            if( (firstByte != 65) && (firstByte != 68) && (firstByte != 70) &&
                (firstByte != 71) && (firstByte != 75) && (firstByte != 88) &&
                (firstByte != 76) && (firstByte != 77) && (firstByte != 79) &&
                (firstByte != 105) && (firstByte != 106) && (thisUdpPort != 0) )
          {
            System.out.println("Invalid UDP port " + thisUdpPort +
              " received through tunnel from " + remoteHost +
              " firstByte:" + firstByte + " - transmission ignored");
          } // end if(portIsGood...

          // read the UDP message from the TCP stream
          if(udpLength > 0)
          {
            if(readBytes(udpLength,5))
            {
              closeTunnel( thisHostIndex, true,
                "TCPTunnel found tcpIsOpen false while waiting on data" );
                return;
            }
            if(stopTLM)return;
            int readLength = udpLength;
            
            // check for message longer than our buffer
            if(udpLength > MAX_MESSAGE)
            {
              System.out.println( "*** Message of length " + udpLength
                + " received from " + remoteHost + " exceeds buffer size of "
                + MAX_MESSAGE + " and has been truncated" );
              readLength = MAX_MESSAGE;
            }
	        }// end if(udpLength > 0)

          // if heartbeat message, check for disconnect, if none go back to read the next message
          if( firstByte == 68 )
          {
            if(recvBuffer[3] != 1)continue;
            if((recvBuffer[1] != 0) || (recvBuffer[2] != 0) || (recvBuffer[4] != 0))continue; 
            closeTunnel(thisHostIndex,false,"client sent disconnect");
            break;
          }
          
          // ACK download message 
          if(firstByte == 103)
          {
            recordDownloadAcksReceived[thisHostIndex]++;
            continue;
          }
          
          // accept or reject download message
          if(firstByte == 105)
          {
            if(recordingDownloadFilename.length() == 0)
            {
              System.out.println("*** recording download approval out of sequence for host " +
                                 thisHostIndex);
              break;
            }
            if(thisUdpPort != 0)
            {
              // download approved - log it and turn on transfer
              System.out.println("\nstarted server recording file download " + 
                                 recordingDownloadFilename +
                                 " size " + recordingFileDownloadSize +
                                 " bytes mtime " + recordingFileDownloadMtime +
                                 " to host# " + thisHostIndex);
              tunnelSends[thisHostIndex].setRecordDownload(thisHostIndex,true);
            }
            else
            {
              System.out.println("host " + thisHostIndex +
                                 " user rejected server recording download " +
                                 recordingDownloadFilename);
              recordingDownloadFilename = "";
            }
            continue;
          }
          
          // cancel download message
          if(firstByte == 106)
          {
            serverRecordingDownload = false;
            recordingDownloadFilename = "";
            System.out.println("user stopped server recording download on host "+
                               thisHostIndex);
            continue;
          }
          
          // if unzipIsReady message, record status and
          // notify FC client to activate download panel
          if( firstByte == 75 )
          {
			      unzipIsReady[ thisHostIndex ] = downloadIsReady;
			      if(downloadIsReady)
			      {
              byte[] downloadDirectory = new byte[udpLength];
              System.arraycopy(recvBuffer,5,downloadDirectory,0,udpLength);
			        if(fcConnect.sendToFC(47,udpLength,downloadDirectory,thisHostIndex))
			          return;
		        }
			      continue;
	        }
          // if RTT probe message, capture RTT and send message back to source
          if(firstByte == 70)
          {
            if(writeThisOutStream(recvBuffer,udpLength+5))
            {
              // lost connection
              closeTunnel( thisHostIndex, true,
                "TCPTunnel disconnect writing RTT probe message" );
              return;
            }
                        
            // inline version of bytesToUshort()
            // reportedRTT = bytesToUshort( recvBuffer[3], recvBuffer[4] );
            highByte = recvBuffer[3];
            if(highByte < 0)highByte = 127 - highByte;
            lowByte = recvBuffer[4];
            if(lowByte < 0)lowByte = 127 - lowByte;
            reportedRTT = 256*highByte + lowByte;

            // when we receive the first non-zero RTT, that is minRTT
            if( reportedMinRTT == 0 )reportedMinRTT = reportedRTT;

            // capture the high-water RTT for this client
            if( reportedRTT > peakRTT[ thisHostIndex ] )
              peakRTT[ thisHostIndex ] = reportedRTT;

            // save difference between reported RTT and reportedMinRTT
            // for congestion management
            int rttIncrease = reportedRTT - reportedMinRTT;
            if( rttIncrease < 0 ) rttIncrease = 0;
            latestRTTIncrease[ thisHostIndex ] = rttIncrease;
            continue;

          } // end if(firstByte == 70)

          // transaction code 76 indicates successful network test
          // and provides appDataZipfile parameters
          if(firstByte == 76)
          {
            // extract the client's appDataZipfile parameters
            appFileMtime[ thisHostIndex ] = bytesToUint(recvBuffer,5);
            appFileSize[ thisHostIndex ] = bytesToUint(recvBuffer,9);
            appFileStatNew[ thisHostIndex ] = true;

            // start accepting tunnel messages the first time we come
            // here (message may be repeated to update file stats)
            connectionTestPassed[ thisHostIndex ] = true;
            if(tunnelIsActive[ thisHostIndex ])continue;
            activateTunnel( thisHostIndex );
            
            // hold here until the TunnelSend is running
            try
            {
              while(!tunnelIsActive[thisHostIndex])Thread.sleep(100);
            }
            catch(InterruptedException ie){return;}
            
            // initialize the Floor Control client if it's
            // passing messages through this tunnel
            if(loginViaTCPTunnel)
              fcConnect.sendFcInitMessage( thisHostIndex );
            continue;
            
          }// end if(firstByte == 76) 

          // transaction code 77 indicates failed network test
          if(firstByte == 77)
          {
            closeTunnel(thisHostIndex,false,"client reported failed network test");
            return;
            
          }// end if(firstByte == 77)
          
          // transaction code 79 is ACK for appDataZip download
          if(firstByte == 79)
          {
            // ACK comes in bytes normally used by UDP port
            tunnelSends[thisHostIndex].postACK(thisHostIndex,thisUdpPort);
            continue;
          }
          
          // port 0 indicates floor control message; invoke FC
          if(thisUdpPort == 0)
          {
            int transactionType = recvBuffer[5];
            byte[] fcMessage = null;
            if(udpLength > 3)
            {
              fcMessage = new byte[udpLength-3];
              System.arraycopy(recvBuffer,8,fcMessage,0,udpLength-3);
            }
            fcConnect.processFCMessage(transactionType,
                                         fcMessage,
                                         udpLength-3,
                                         thisHostIndex);
            continue;
          }

          // forwarding and floor control logic
          boolean okToForward = (firstByte != 67);
          if( thisHostIndex == localFloorHolder )okToForward = true;

          // floor control for WBD and video
          if( (thisUdpPort == wbdPort) || (thisUdpPort == wbdControlPort) )
            // allow WBD repair requests even under floor control
            // also allow Supercontroller to write to WBD at any time
            // unless in breakout mode
            if(inBreakoutMode)
            {
              if( ( !wbdFloorControl ||
                    (recvBuffer[15] == 2) ) &&
                  (thisHostIndex == localFloorHolder) )
                okToForward = true;
            }
            else
            {
              if( !wbdFloorControl ||
                  (recvBuffer[15] == 2) ||
                  thisIsSupercontroller )
                okToForward = true;
            }
          if( (thisUdpPort == videoPort) || (thisUdpPort == videoControlPort) )//debugx video here??
          {
            if( !videoFloorControl )okToForward = true;
          }
          else nonVideoPort = true;
          
          if( thisUdpPort == audioPort ) if( audioFloorControl )
          {
            if( thisHostIndex != localFloorHolder )
            {
              /* withdrawn 6-6-12 by JMP - just ignore the audio input
              // error message to TLMC if still sending 5 sec plus RTT after giving
              // up the floor; send every 20 seconds while this is true
              if( (startTime - checkTime) > (5000 + reportedRTT) )
              {
                if( lastWarningTime[thisHostIndex] < (startTime - 20000) )
                {
                  if(sendFloorWarning(thisHostIndex, thisUdpPort) )
                    return;
                  lastWarningTime[thisHostIndex] = startTime;
                } // end if( lastWarningTime
              } // end if( startTime
              end 6-6-12 withdrawn by JMP */
              continue;
            } // end else/if( thisHostIndex != localFloorHolder )
          } // end if( thisUdpPort == audioPort )

          if( thisUdpPort == wbdPort ) if( wbdFloorControl )
          {
            // maintain state of sending slide to support floor control
            signalSlideDone = false;
            byte testByte = recvBuffer[15]; // WBD pk_type
            if( thisHostIndex == localFloorHolder || thisIsSupercontroller )
            {
              // logic to avoid floorHolder change during slide send
              if (testByte == 0) // pk_type == pk_draw_msg
              {
                if (recvBuffer[40] == 17) // draw_cmd == DRW_NEW_LABEL
                {
                  sendingSlide = true;
                  slideTimestamp = bytesToUint(recvBuffer, 9);
                  slideIpAddress = bytesToUint(recvBuffer, 17);
                  slidePageNumber = bytesToUint(recvBuffer, 21);
                }
                else
                {
                  if(recvBuffer[40] == 12) // draw_cmd == DRW_GRP_PS
                    signalSlideDone =
                      ( (bytesToUint(recvBuffer, 9) == slideTimestamp) &&
                        (bytesToUint(recvBuffer, 17) == slideIpAddress) &&
                        (bytesToUint(recvBuffer, 21) == slidePageNumber) );
                }// end else/if(recvBuffer[40] == 11)
              }
            } // end if(thisHostIndex...
          } // end if( thisUdpPort == wbdPort )

          // check for error message from TLMC
          if( recvBuffer[0] == 65 )
          {
            receivedErrorMessage(recvBuffer,thisHostIndex);
            continue;
          }

          // could be a URL message from a playback server - do not forward
          if( firstByte == 71 )continue;
          
          // if this is not a forwardable message, don't
          if(!okToForward)continue;

          // for tunnel receivers, revise transaction code only
          recvBuffer[0] = 66;
          int tcpLength = udpLength + 5;
		  
		      // send to server-side recorder, if configured
		      if(recordOnServer && (serverRecord != null))
		      {
		        if( (thisUdpPort == audioPort) || (thisUdpPort == audioControlPort) )
			        serverRecord.queueForRecorder(recvBuffer,
                                            (thisUdpPort == audioControlPort) ? 1 : 0,
                                            0,
                                            udpLength);
			      else if((thisUdpPort == wbdPort) || (thisUdpPort == wbdControlPort) )
			        serverRecord.queueForRecorder(recvBuffer,
                                            (thisUdpPort == wbdControlPort) ? 1 : 0,
                                            1, 
                                            udpLength);
			      else if(recordVideo)
			        if( (thisUdpPort == videoPort) || (thisUdpPort == videoControlPort) )
			          serverRecord.queueForRecorder(recvBuffer,
                                              (thisUdpPort == videoControlPort) ? 1 : 0,
                                              2,
                                              udpLength);
            // should we also record floor control etc?
          }

          // send the received TCP message to every host with a tunnel open
          int hostIndex;
          for ( hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex )
          {
            if( !connectedByTCP[hostIndex] )continue;
            if( !connectionTestPassed[hostIndex] ) continue;
            	
            // in breakout mode, send only to our own group
            // (this should be implemented in all commutators) needs fixed debugx
            if(inBreakoutMode)
              if((thisHostIndex != breakoutController) || (breakoutGroup[thisHostIndex] > 0))
                if(breakoutGroup[hostIndex] != breakoutGroup[thisHostIndex])
                  continue;

            // place message in queue for tunnel send
            // audio traffic
            if( (thisUdpPort == audioPort) || (thisUdpPort == audioControlPort) )
            {
              if( queueTunnel( 1, hostIndex, recvBuffer, tcpLength ) )
              {
                if( ++audioDropped[ hostIndex ]%dropMessageInterval == 1 )
                {
                  System.out.println( dateTime() + " " +
                    audioDropped[ hostIndex ] + " audio messages for host #" +
                    hostIndex + " dropped due to congestion (delta RTT:" +
                    latestRTTIncrease[ hostIndex ] + ")" );
                  if( audioDropped[ hostIndex ] > audioCutoff )
                    closeTunnel(hostIndex,true,"TCP audio congestion");
                }
              } // end if( queueTunnel( 1...
              continue;
            } // end if( thisUdpPort == audioPort...

            // don't send message back to the source, except above audio case
            // (which is pulled off at TunnelSend after triggering download)
            if( hostIndex == thisHostIndex )continue;

            // whiteboard traffic
            if( (thisUdpPort == wbdPort) || (thisUdpPort == wbdControlPort) )
            {
              if( queueTunnel( 3, hostIndex, recvBuffer, tcpLength ) )
              {
                if( ++whiteboardDropped[ hostIndex ]%dropMessageInterval == 1 )
                {
                  System.out.println( dateTime() + " " +
                    whiteboardDropped[ hostIndex ] +
                    " whiteboard messages for host #" + hostIndex +
                    " dropped due to congestion (delta RTT:" +
                    latestRTTIncrease[ hostIndex ] + ")" );
                  if( whiteboardDropped[ hostIndex ] > whiteboardCutoff )
                        closeTunnel(hostIndex,true,"TCP whiteboard congestion");
                }
              } // end if( queueTunnel( 3...
              continue;
            } // end if( thisUdpPort == wbdPort...

            // video traffic
            if( (thisUdpPort == videoPort) || (thisUdpPort == videoControlPort) )
            {
              if( connectedVideoByTCP[ hostIndex ] || !connectedByTCP[ hostIndex ] )
              {
                if( queueTunnel( 4, hostIndex, recvBuffer, tcpLength ) )
                {
                  if( ++videoDropped[ hostIndex ]%dropMessageInterval == 1 )
                  {
                    System.out.println( dateTime() + " " +
                      videoDropped[ hostIndex ] + " video messages for host #" +
                      hostIndex + " dropped due to congestion (delta RTT:" +
                      latestRTTIncrease[ hostIndex ] + ")" );
                    if( videoDropped[ hostIndex ] > videoCutoff )
                        closeTunnel(hostIndex,true,"TCP video congestion");
                  }
                } // end if( queueTunnel( 4...
              } // end if( connectedVideoByTCP[ hostIndex ]...
            } // end thisUdpPort == videoPort

          } // end for( int sendIndex...

          // if this completed sending a slide, notify floor control hold thread
          if(signalSlideDone)
          {
            signalSlideDone = false;
            sendingSlide = false;
            synchronized(floorChangeHoldThread)
            {
              floorChangeHoldThread.notify();
            }
          }
          // **************** end TCP tunnel queueing ********************
          
          // ignore record/playback message in UDP and MC
          if(videoAddressConfigured)continue;//debugx

          // now send message received at TCP tunnel to non-tunnel UDP unicast hosts
          // send only if the packets have arrived within time window
          // and are not using tunnel for that application
          if( udpFlowsActive > 0 )
          {
            // set socket for UDP
            if( controlPort == 1 )udpSocket = udpSocket2Save[ portIndex ];
            else udpSocket = udpSocket1Save[ portIndex ];

            // in case videoAddressConfigured, some sockets may be null
            if( udpSocket != null )
            {
              sendPacket.setData( recvBuffer, 5, udpLength );
              sendPacket.setPort( udpPort[ portIndex ]+controlPort );
              for( hostIndex = 0; hostIndex < numberOfHosts; hostIndex++ )
              {
                // send only to active UDP tunnels
                if( !udpActive[ hostIndex ] )continue;

                // for UDP, forward only if the messages have arrived within time
                // window and the host is authenticated if that is required
                if( (startTime - uRecvTime[ hostIndex ]) < udpWindow )
                {
                  // do not send message back to source
                  if( hostIndex != thisHostIndex )
                  {
                    sendPacket.setAddress( hostInetAddr[ hostIndex ] );
                    udpSocket.send( sendPacket );
                    countUdpSend++;
                  } // end if(hostIndex...
                } // end if( (startTime
                else
                {
                  // address has stopped sending; delete from active set
                  deactivateFlow( hostIndex );
                } // end else/if( (startTime...
              } // end for( sendIndex...
            } // end if( udpFLowsActive > 0 )
          } // end if( udpFlowsActive > 0 )
          //************** end UDP unicast send ***********************

          // send audio/wbd/video packet received through TCP tunnel 
          // to local multicast
          if( multicastIsConfigured )
          {
            sendPacket.setData( recvBuffer, 5, udpLength );

            // set socket for multicast
            if( controlPort == 1 )mcSocket = mcSocket2Save[ portIndex ];
            else mcSocket = mcSocket1Save[ portIndex ];

            // send only if the packets have arrived within time window
            if( ( startTime - mRecvTime[portIndex] )  < udpWindow )
            {
              sendPacket.setAddress( mcGroup[ portIndex ] );
              sendPacket.setPort( mcPort[ portIndex ] );
              mcSocket.send( sendPacket );
              countMcSend++;
            } // end if( ( startTime...
          } // end if( multicastIsConfigured )
          
          // pause to let another listener thread run
          try
          {
            Thread.sleep( sleepTime );
          }
          catch( InterruptedException ie ){return;}

          // let other threads run
          Thread.yield();

        } // end while(!stopTLM...
      } // end main try
      catch( IOException ioe )
      {
        // lost connection - shut down tunnel
        closeTunnel( thisHostIndex, true, "TCPTunnel disconnect while queueing");
        return;
      }
      catch( NullPointerException npe )
      {
        // lost connection - shut down tunnel
        closeTunnel( thisHostIndex, true, "TCPTunnel NullPointerException");
		    npe.printStackTrace(System.out);
        return;
      }
      // shut down input stream
      try
      {
        if( thisSocketChannel != null )
          thisSocketChannel.close();
      }
      catch( IOException ioe ){ }
      if(hostAddr[thisHostIndex] != null)
        closeTunnel( thisHostIndex, true, "TCPTunnel disconnect while closing");
      return;
      
    } // end run

} // end TLM.TCPTunnel class
