/* modified JMP 3/1/13
 * TCPVideo.java
 * Transport Layer Multicaster
 * Created on August 8, 2002
 * @author  mpullen & students
 * @version 5.2.4 
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/
/*
  This separate thread is provided to allow a separate TCP video stream
  from a host to the TLM, to support users with adequate capacity for
  video while allowing the audio/whiteboard to be rate controlled at a
  lower 56 kbps modem rate. No recorder stream is supported in this thread,
  and it is assumed that all traffic received is video.

 TCPVideo.java class logic:

->Until the TCP connection is closed
-> read TCP message
-> strip port from first two byte of message and locate it in multicast port array
   or unicast port array
   if it is in the array, capture the corresponding port (unicast if it is in
   the multicast array and vice versa)  and flag that port as TCP for this host;
   else go back to reading TCP
-> if not loopback or URL, it should be forwarded in the following ways
   -> update unicast last recv time array
   -> if the multicast group has been a source of messages (any port) within time
      window send unicast-udp message to multicast group with a modified TTL(15)
   -> send a copy of unicast UDP message to each unicast address that has been active
      within time window except the address it came from (source address) and queue
      it to be sent to each TCP and UDP tunnel address except the address it came
      from (tunnel gets the original message with port number prepended and revised
      transaction code)
-> When the connection is closed, set corresponding unicast TCP array to 0 and return
******************************************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.nio.channels.SocketChannel;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;

public class TCPVideo extends TLM implements Runnable
{
    private MulticastSocket mcSocket;
    private DatagramSocket udpSocket;
    private SocketChannel thisSocketChannel;
    private String remoteHost;
    DatagramPacket sendMessage, sendPacket;
    byte[] recvBuffer, timeoutBuffer = {69,0,0,3,0};
    static long startTime = 0;
    public int portIndex;
    private int thisHostIndex;// position of this host in hostAddr[] from TLM
    private int firstByte = -1;
    int localFloorHolder; // floorholder for this breakout group
                          // (or whole group if no breakout)

//**********************************************************

    // constructor
    public TCPVideo( int hostIndexRef)
    {
      thisHostIndex = hostIndexRef;
      localFloorHolder = fcConnect.getFloorHolder(thisHostIndex);
      thisSocketChannel = tcpVideoSocketChannels[thisHostIndex];

      // network structures
      recvBuffer = new byte[ MAX_MESSAGE+5 ];
      sendPacket = new DatagramPacket( recvBuffer, 0);
      remoteHost = hostAddr[ thisHostIndex ];
    } 
    
    
    // reads thisSocketChannel until it gets numberOfBytes; puts them in recvBuffer
    // sleeps between reads to avoid thread lockup
    // returns true if disconnected, false otherwise
    // side-effect: sets firstByte to first byte read (or -1 if disconnected)
    boolean readBytes(int numberOfBytes, int offset)
    {      
      ByteBuffer bb = ByteBuffer.allocate(numberOfBytes);
      while(!stopTLM)
      {
        // read the channel and catch disconnect or exception
        try
        {
          if(thisSocketChannel.read(bb) < 0)
          {
            firstByte = -1; 
            return true;
          }
        }
        catch(ClosedChannelException cce)
        {
          return true;
        }
        catch(IOException ioe)
        {
          System.out.println("*** IOException in TCPTunnel for host:" +
                             thisHostIndex + ":" + ioe.getMessage());
          ioe.printStackTrace(System.out);
          firstByte = -1; 
          return true;
        }
        
        // when buffer is full copy ByteBuffer to recvBuffer and return,
        // making sure not to exceed length of recvBuffer
        if(!bb.hasRemaining())
        {
          int numberToTransfer = numberOfBytes;
          if((numberToTransfer + offset) > recvBuffer.length)
            numberToTransfer = recvBuffer.length - offset;
          bb.rewind();
          bb.get(recvBuffer,offset,numberToTransfer);
          firstByte = recvBuffer[0];
          return false;
        }
        
        // sleep a while and try again
        try{Thread.sleep(sleepTime);}
        catch(InterruptedException ie){return false;}
      
      }// end while(!stopTLM)
      
      firstByte = -1; 
      return false;
      
    }// end readBytes()

    public void run()
    {
      try
      {
        int bytesRead = 0;
        int badCode = 0;

        // wait here until authentication is complete, timeout at maxLoginTime
        long endTime = System.currentTimeMillis() + maxLoginTime;
        while( !authenticatedHost[ thisHostIndex ] )
        {
          // show message received at this time to avoid connect race
          uRecvTime[ thisHostIndex ] = System.currentTimeMillis();
          
          // check for client disconnect
          if( !tcpIsOpen[ thisHostIndex ] )
          {
            closeTunnel( thisHostIndex, true,
              "TCPVideo found tcpIsOpen false while waiting authentication" );
            return;
          }
          Thread.sleep(sleepTime);
          
          // check for login timeout and shut down if needed
          if( endTime < System.currentTimeMillis() )
          {
            Thread.sleep(1000);
            closeTunnel( thisHostIndex, false, "TCPVideo login timeout" );
            return;
          } // end if( endTime...
        } // end while( !authenticateHost[ ...

        // read in data continuously, terminated by the remote host close
        while( !stopTLM )
        {
	        // pause to let another listener thread run
          Thread.yield();

          // allocate a new read buffer so it can be saved until
          // written to all outstreams asynchronously
          recvBuffer = new byte[ MAX_MESSAGE+5 ];
          
          // read firstByte TCP non-blocking and test for error
          int thisUdpPort, udpLength = 0;
          readBytes(1,0);
          if(firstByte == -1)
          {
            // lost connection
            closeTunnel( thisHostIndex, true,
              "TCPVideo got lost connection reading first byte" );
            return;
          }
          
          // check if protocol transaction code valid
          if( firstByte != 67 )
          {
            System.out.println("*** bad protocol code:" + firstByte +
                               " found in TCPvideo for host " + thisHostIndex +
                               " - stopping video/multicast delivery from this host");
            return;
          }

          // read the rest of the five-byte header
          countTcpRecv++;
          if(readBytes(4,1))
          {
            // lost connection
            closeTunnel( thisHostIndex, true,
              "TCPVideo got lost connection reading header" );
            return;
          }

          // extract UDP message length from received header
          // following code is inline version of bytesToInt()
          // udpLength = bytesToInt( recvBuffer[1], recvBuffer[2] );
          // done inline for efficiency
          int highByte = recvBuffer[1];
          if(highByte < 0)highByte = 127 - highByte;
          int lowByte = recvBuffer[2];
          if(lowByte < 0)lowByte = 127 - lowByte;
          udpLength = 256*highByte + lowByte;

          // extract UDP port number from received TCP segment
          // thisUdpPort = bytesToInt( recvBuffer[3], recvBuffer[4] );
          // done inline for efficiency
          highByte = recvBuffer[3];
          if(highByte < 0)highByte = 127 - highByte;
          lowByte = recvBuffer[4];
          if(lowByte < 0)lowByte = 127 - lowByte;
          thisUdpPort = 256*highByte + lowByte;
                  
          // check whether it is a control port
          int controlPort = 0;
          if( thisUdpPort == videoPort+1 )controlPort = 1;

          // read the UDP message from the TCP stream
          if(udpLength > 0)
          {
            if(readBytes(udpLength,5))
            {
              closeTunnel( thisHostIndex, true,
                "TCPTunnel found tcpIsOpen false while waiting on data" );
                return;
            }
            if(stopTLM)return;
            int readLength = udpLength;
            
            // check for message longer than our buffer
            if(udpLength > MAX_MESSAGE)
            {
              System.out.println( "*** Message of length " + udpLength
                + " received from " + remoteHost + " exceeds buffer size of "
                + MAX_MESSAGE + " and has been truncated" );
              readLength = MAX_MESSAGE;
            }
	        }// end if(udpLength > 0)
//9-14-07 quickfix2: flush all input from non-controllers
//if(controlPrivilege[thisHostIndex] == 0)continue;//debugx
          // ignore the message if it is not for video port or multicast
          if( (thisUdpPort == videoPort) || 
              (thisUdpPort == videoPort+1) || 
              (thisUdpPort == multicastTunnelPort))
          {
            // for floor controlled applications, verify that sender has the floor
            // if not drop the message
            boolean okToForward = false;
            if( thisHostIndex == localFloorHolder )okToForward = true;
            if( (thisUdpPort == videoPort) &&
                (!videoFloorControl || (floorStatus == 1)) )okToForward = true;
            if( (thisUdpPort == multicastTunnelPort) &&
                (!multicastTunnelFloorControl || (floorStatus == 1)) )okToForward = true;
            // ready to forward the message...
            if( okToForward )
            {
              // for tunnel receivers, revise transaction code only
              recvBuffer[0] = 66;
              int tcpLength = udpLength + 5;

              // Send the received TCP message to every host with a tunnel open
              // except the one that sent it to us.
              for ( int hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex )
              {
                // don't send message back to the sender
                if( hostIndex == thisHostIndex )continue;

                // place message in queue for video tunnel send
                if( connectedVideoByTCP[ hostIndex ] )
                {
                  // video traffic
                  if( thisUdpPort == videoPort)
                  {
                    if( queueTunnel( 4, hostIndex, recvBuffer, tcpLength ) )
                    {
                      if( ++videoDropped[ hostIndex ]%dropMessageInterval == 1 )
                      {
                        System.out.println( dateTime() + " " +
                          videoDropped[ hostIndex ] +
                          " video messages for host #" + hostIndex +
                          " dropped due to congestion (delta RTT:" +
                          latestRTTIncrease[ hostIndex ] + ")" );
                        if( videoDropped[ hostIndex ] > videoCutoff )
                          closeTunnel(hostIndex,true,"TCP video congestion");
                      }
                    } // end if( queueTunnel( 4,...
                  }// end if( thisUdpPort == videoPort )
                  
                  // multicast tunnel traffic
                  if( thisUdpPort == multicastTunnelPort )
                  {
                    if( queueTunnel( 123, hostIndex, recvBuffer, tcpLength ) )
                    {
                      if( ++multicastDropped[ hostIndex ]%dropMessageInterval == 1 )
                      {
                        System.out.println( dateTime() + " " +
                          multicastDropped[ hostIndex ] + " multicast messages for host #" +
                          hostIndex + " dropped due to congestion (delta RTT:" +
                          latestRTTIncrease[ hostIndex ] + ")" );
                        if( multicastDropped[ hostIndex ] > multicastCutoff )
                          closeTunnel(hostIndex,true,"TCP multicast traffic congestion");
                      }// end if( ++multicastDropped...
                    } // end if( queueTunnel(123...
                  } // end thisUdpPort == multicastPort

                } // end if( connectedVideoByTCP...
              } // end for( int sendIndex...
              // **************** end TCP tunnel queueing ********************

              // send multicast packet from TCP tunnel to local multicast interface
              if( multicastTunnelServerSubnet )
                if( thisUdpPort == multicastTunnelPort )
                {
                  sendPacket.setData( recvBuffer, 5, udpLength );
                  
                  // send only if the packets have arrived within time window
                  if( ( startTime - mRecvTime[portIndex] )  < udpWindow )
                  {
                    sendPacket.setAddress( multicastTunnelGroup );
                    sendPacket.setPort( multicastTunnelPort );
                    mcTunnelSendSocket.send( sendPacket );
                    countMcSend++;
                  } // end if( ( startTime...
                } // end if( thisUdpPort...
              // ********************end tunnel multicast **********************

              // if no UDP to process, skip it
              if( (udpFlowsActive > 0)  && !videoAddressConfigured )
              {
                // set socket for UDP
                if( controlPort == 1 )udpSocket = udpSocket2Save[ portIndex ];
                else udpSocket = udpSocket1Save[ portIndex ];

                // now send packet received at TCP Port to other unicast hosts
                // send only if the packets have arrived within time window
                // and are not using TCP for that application
                sendPacket.setData( recvBuffer, 5, udpLength );
                sendPacket.setPort( udpPort[ portIndex ] + controlPort );
                for( int hostIndex = 0; hostIndex < udpFlowsActive; hostIndex++ )
                {
                  // send only to active UDP tunnels
                  if( !udpActive[ hostIndex ])continue;

                  // for UDP, forward only if the messages have arrived within time
                  // window and the host is authenticated if that is required
                  if( (startTime - uRecvTime[ hostIndex ]) < udpWindow )
                  {
                    // do not send message back to source
                    if( hostIndex != thisHostIndex )
                    {
                      sendPacket.setAddress( hostInetAddr[ hostIndex ] );
                      udpSocket.send( sendPacket );
                      countUdpSend++;
                    } // end if(hostIndex...
                  } // end if( (startTime
                  else
                  {
                    // address has stopped sending; delete from active set
                    deactivateFlow( hostIndex );
                  } // end else/if( (startTime...
                } // end for( sendIndex...
              } // end if( udpFlowsActive > 0 )
              //************** end UDP unicast send ***********************

              // now send packet received at TCP Port to multicast hosts
              if( multicastIsConfigured )
              {
                sendPacket.setData( recvBuffer, 5, udpLength );

                // set socket for multicast
                if( controlPort == 1 )mcSocket = mcSocket2Save[ portIndex ];
                else mcSocket = mcSocket1Save[ portIndex ];

                // send only if the packets have arrived within time window
                if( ( startTime - mRecvTime[portIndex] )  < udpWindow )
                {
                  sendPacket.setAddress( mcGroup[ portIndex ] );
                  sendPacket.setPort( mcPort[ portIndex ] );
                  mcSocket.send( sendPacket );
                  countMcSend++;
                } // end if( ( startTime...
              } // end if( multicastIsConfigured )
            }// end if( okToForward )
          } // end if( thisUdpPort...
          else
          {
            System.out.println("Invalid UDP port " + thisUdpPort +
              " received through TCP Video tunnel from " +
              hostAddr[thisHostIndex] + " - transmission ignored");
          } // end else/if(portIsGood...
        } // end while(!stopTLM...
      } // end main try
/*debugx
      catch( IOException ioe )
      {
        // lost connection - shut down tunnel
      }
*/
      catch( Exception e )
      {
        tcpTunnelInstance[thisHostIndex].thisHostVideoTunnel = null;
        System.out.println("Exception in TCPVideo for host:" + thisHostIndex
          + ":" + e + "\n" + e.getMessage());
      } // end main catch

      // shut down viedo tunnel
      closeTunnel( thisHostIndex, true, "TCPVideo lost connection during close" );
      return;
      
    } // end run
} // end TLM.TCPVideo class
