/* modified JMP 5/8/11
 * TunnelSend.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author  mpullen & students
 * @version 5.2.4
 * @modified 2/18/2013
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

/*********************SYSTEM LOGIC**********************************************
 * This may seem an odd way to send TCP: pass it through an internal UDP first!
 *
 * The problem was are solving here is that Java TCP does not have a non-blocking
 * send. As a result, any TCP flow that backs up due to congestion or just a
 * poor-quality modem link can block when TCP's window fills up. Our solution
 * to this is to use one thread per TCP output stream, so a blocked send
 * can't stop distribution to other receivers. The path between TCP receive and
 * TCP send is UDP sent on internal loopback. This means we automatically gain
 * a queuing mechanism (UDP buffer) for packets that will overflow if the stream
 * backs up.
 *
 * The appDataZip download process runs here, absorbing unused capacity. It
 * assumes that audio messages will be sent at the NEWSF standard rate of 5 per
 * second and uses them to trigger a new slot (this means if the floorholder is
 * not sending, we don't do any downloading).
 *
 * The serverRecording download also runs here, as last priority.
 *
 * TunnelSend.java class logic:
 * ->In an infinite loop
 * ->  dequeue a unicast message associated with this outstream
 *     on basis of priority
 * ->  send the message to the associated TCP outstream
 * ->  if we are doing download of application zipfiles,
 *       if there is unused capacity in this 1/5 second slot,
 *         queue enough download slices to use rated capacity
 *
 ******************************************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.nio.channels.ClosedChannelException;

public class TunnelSend extends TLM implements Runnable
{
  private DatagramSocket tunnelSendUDPSocket;
  private DatagramPacket sendMessage;     // UDP message
  private int thisHostIndex, udpLength;   // index in all [MAX_HOSTS] arrays
  private InetAddress thisInetAddress;    // destination of this tunnel
  SocketChannel thisSocketChannel;        // non-blocking Socket
  boolean thisTunnelSupportsDownload;
  boolean proceedWithDownload = false, startMessageSent;
  boolean serverRecordingDownload = false;
  FileInputStream recordingFile = null;
  byte[] recordingBuffer = new byte[MAX_MESSAGE+14];
  byte[] udpBuffer;
  int sendLength, bytesDownloaded = 0, bytesSentThisSlot = 0;
  int recordingBytesDownloaded = 0;
  byte[] downloadSegment = new byte[DOWNLOAD_MESSAGE_SIZE+6];
  int segmentSN = 0;
  int recordingSN = 0;
  public int lastSNAcked = -1;//debugx
  int lastAckRequestSent = -1;//debugx
  long timeLastAudioForwarded = 0; //debugx
  long startDownloadMillis;
  long waitTime = 0;
  int maxMessagesUnacked = 0x10000 / (MAX_MESSAGE+5);// TLMC receive buffer space

  // we rate limit download based on available bytes per slot
  // the slots are times by audio messages (5 per second)
  // so bytes per slot = 1000*(kbps/5)/8
  final int bytesPerSlot = (1000*netCapacitykbps[ thisHostIndex ])/40;
  final int bytesPerSlotMinus = bytesPerSlot - DOWNLOAD_MESSAGE_SIZE;
  final long oneSlotMillis = 200; // millseconds per slot debugx

  // constructor
  public TunnelSend ( int hostIndexParm )
  {
    // allocate buffer and initialize
    udpLength = 0;
    udpBuffer = new byte[MAX_MESSAGE+5];

    // copy parameters
    thisHostIndex = hostIndexParm;

    // we don't support download for connections less than 100kb/s
    thisTunnelSupportsDownload = (netCapacitykbps[ thisHostIndex ] >= 100)
                                 && (appFileMtime[thisHostIndex] != 0);//debugx

    // if this user tunnel is not TCP, open the return UDP tunnel
    if( tcpIsOpen[ thisHostIndex ] )
      thisSocketChannel = tcpSocketChannels[ thisHostIndex ];
    else
      try
      {
        tunnelSendUDPSocket = new DatagramSocket();
        udpSockets[ thisHostIndex ] = tunnelSendUDPSocket;
        thisInetAddress = hostInetAddr[ thisHostIndex ];
        sendMessage = new DatagramPacket( udpBuffer, udpLength,
                                          thisInetAddress, tunnelUdpPort+1 );
      }
      catch( SocketException se )
      {
        System.out.println("Socket exception in TunnelSend:" + se.getMessage());
      }

  }// end constructor

  public void run()
  {
    // set status so multicasters will use us
    tunnelIsActive[ thisHostIndex ] = true;
    	
    // clean out any previous host table positions occupied by this user
    checkForDeadHostTableEntry(username[thisHostIndex]);
    uRecvTime[ thisHostIndex ] = System.currentTimeMillis();
    	
    // wait for traffic to process
    try
    {    
      // perform this loop as long as messages arrive
      //
      // give priority to audio message, then whiteboard messages,
      // then video messages, finally app zipfile download
      int thisBufferIndex;
      while( !stopTLM && (uRecvTime[thisHostIndex] != 0) )
      {
        // if we've been told to shutdown, leave the loop
        if(!tunnelIsActive[thisHostIndex])break;
        bytesSentThisSlot = 0;
		
        // check the various traffic queues, in priority order
        synchronized(sendThreads[thisHostIndex])
        {
          synchronized(audioQueues[thisHostIndex])
          {
            // if buffer window is empty, go on to next priority
            thisBufferIndex = audioWindowStart[thisHostIndex];
            if (thisBufferIndex != audioWindowEnd[thisHostIndex])
            {
              // for audio (only), TCPTunnel does queueTunnel to host, to trigger
              // this code; but once the trigger is done we don't send it back to
              // the source host
              if(floorHolder != thisHostIndex)
              {
                bytesSentThisSlot = audioLengths[thisHostIndex][thisBufferIndex];
                sendBuffer(audioQueues[thisHostIndex][thisBufferIndex],
                           bytesSentThisSlot,
                           false);
              }
              else bytesSentThisSlot = 0;
              
              // save the time of audio forwarding; we need to
              // stimulate appDataDownload if no audio for 1 sec 
              timeLastAudioForwarded = uRecvTime[thisHostIndex];  //debugx

              // increment start index for window of buffers
              ++thisBufferIndex;
              if(thisBufferIndex >= audioRingSize)
                audioWindowStart[thisHostIndex] = 0;
              else
                audioWindowStart[thisHostIndex] = thisBufferIndex;
              continue;
            }
          }//end synchronized(audioQueues[thisHostIndex])

          synchronized (fcQueues[thisHostIndex])
          {
            // if buffer window is empty, go on to next priority
            thisBufferIndex = fcWindowStart[thisHostIndex];
            if(thisBufferIndex != fcWindowEnd[thisHostIndex])
            {
              sendLength = fcLengths[thisHostIndex][thisBufferIndex];
              bytesSentThisSlot += sendLength;
              sendBuffer(fcQueues[thisHostIndex][thisBufferIndex],
                         sendLength,
                         false);

              // increment start index for window of buffers
              ++thisBufferIndex;
              if(thisBufferIndex >= fcRingSize)
                fcWindowStart[thisHostIndex] = 0;
              else
                fcWindowStart[thisHostIndex] = thisBufferIndex;
              continue;
	          }
	        }//end synchronized (fcQueues[thisHostIndex])
          
          synchronized (whiteboardQueues[thisHostIndex])
          {
            // if buffer window is empty, go on to next priority
            thisBufferIndex = whiteboardWindowStart[thisHostIndex];
            if(thisBufferIndex != whiteboardWindowEnd[thisHostIndex])
            {
              sendLength = whiteboardLengths[thisHostIndex][thisBufferIndex];
              bytesSentThisSlot += sendLength;
              sendBuffer(whiteboardQueues[thisHostIndex][thisBufferIndex],
                         sendLength,
                         false);

              // increment start index for window of buffers
              ++thisBufferIndex;
              if(thisBufferIndex >= whiteboardRingSize)
                whiteboardWindowStart[thisHostIndex] = 0;
              else
                whiteboardWindowStart[thisHostIndex] = thisBufferIndex;
              continue;
	          }
	        }//end synchronized (whiteboardQueues[thisHostIndex])

          synchronized(videoQueues[thisHostIndex])
          {
            // if buffer is empty, nothing to send, so go to wait state
            thisBufferIndex = videoWindowStart[thisHostIndex];
            if(thisBufferIndex != videoWindowEnd[thisHostIndex])
            {
              sendLength = videoLengths[thisHostIndex][thisBufferIndex];
              bytesSentThisSlot += sendLength;
              sendBuffer(videoQueues[thisHostIndex][thisBufferIndex],
                         sendLength,
                         true);
              
              // increment start index for window of buffers
              ++thisBufferIndex;
              if(thisBufferIndex >= videoRingSize)
                 videoWindowStart[thisHostIndex] = 0;
              else
                 videoWindowStart[thisHostIndex] = thisBufferIndex;
              continue;
		        }
          }// end synchronized(videoQueues[thisHostIndex])

          synchronized(multicastQueues[thisHostIndex])
          {
            // if buffer is empty, nothing to send, so go to wait state
            thisBufferIndex = multicastWindowStart[thisHostIndex];
            if(thisBufferIndex != multicastWindowEnd[thisHostIndex])
            {
              sendLength = multicastLengths[thisHostIndex][thisBufferIndex];
              bytesSentThisSlot += sendLength;
              sendBuffer(multicastQueues[thisHostIndex][thisBufferIndex],
                         sendLength,
                         true);
             
              // increment start index for window of buffers
              ++thisBufferIndex;
              if(thisBufferIndex >= multicastRingSize)
                 multicastWindowStart[thisHostIndex] = 0;
              else
                 multicastWindowStart[thisHostIndex] = thisBufferIndex;
              continue;
            }
          }// end synchronized(multicastQueues[thisHostIndex])
          
          // there is no realtime data to send; check for download opportunity
          if(proceedWithDownload)
          {
            if(lastAckRequestSent == lastSNAcked)//debugx
            {
              while((bytesSentThisSlot) < bytesPerSlotMinus)
              {
                if(bytesDownloaded < appFileDownloadSize)
                {
                  sendDownload(false);
                  bytesSentThisSlot += DOWNLOAD_MESSAGE_SIZE;
                }
                else
                {
                  sendDownloadDone(false);
            //debugxx  proceedWithDownload = false;
                  break;
                }
                if(stopTLM)return;
              }// end while((bytesSentThisSlot...
            
              // add one more segment at end, with ACK request
              if(bytesDownloaded < appFileDownloadSize)//debugx
                sendDownload(true);
              
              // if we're not part of ongoing audio stream, limit wait()
              if((System.currentTimeMillis() - timeLastAudioForwarded) < 1000)//debugx
                waitTime = oneSlotMillis;
                
            }// end if(lastAckRequestSent 
          }// end if(proceedWithDownload)
          
          // if no appData to download, check to see if
          // user requested server recording download; if so
          // use maxMessageUnacked not slot limit as user will have been
          // informed this action takes over all available capacity
          else while(serverRecordingDownload)
            {
              if((recordDownloadMessagesSent[thisHostIndex] - recordDownloadAcksReceived[thisHostIndex])
                  > maxMessagesUnacked)break;
              sendRecordingDownload();
              if(stopTLM)return;
              recordDownloadMessagesSent[thisHostIndex]++;
              Thread.yield();
              continue;
            }
          
          // when there is nothing to send, wait one slot or until
          // notified that a buffer has been loaded
          try
          {
            sendThreads[thisHostIndex].wait(waitTime);//debugx
            continue;
          }
          catch (InterruptedException ie) {}

        } // end synchronized( sendThreads[ thisHostIndex ] )

      } // end while(!stopTLM && ...)
    }
    catch( IOException ioe )
    {
      // TCP error means lost connection- just close tunnel
      // UDP error means something basic is broken!
      if(!tcpIsOpen[thisHostIndex])
      {
        System.out.println("TunnelSend UDP exception on host #" +
                           thisHostIndex + " :" + ioe.getMessage() );
      }
    } // end catch
    closeTunnel( thisHostIndex, true, "TunnelSend found tunnel disconnected" );
    return;

  } // end run


  // this can be invoked from elsewhere in TLM after TunnelSend is running
  void startDownloading()
  {
    if(!fcIsReadyForAppData[thisHostIndex])return;
  
    // set startup state (startDownloading can happen more than once)
    startMessageSent = false;
    proceedWithDownload = false;
    segmentSN = 0;
    bytesDownloaded = 0;

    // create and start a thread to deal with exchange of filesize data
    CheckFileSpecs checkFileSpecs = new CheckFileSpecs();
    checkFileSpecs.start();

  }// end startDownloading()


  // send a floor control message to update the downloading status
  // codes: 0->new user  1->downloading  2-> can't download  3->done downloading
  void sendDownloadingStatus(int status)
  {
    fcConnect.sendUserStatus(thisHostIndex,(byte)status);
  }


  // send and flush an unzip-existing-appDataZipFile code
  void sendFilename(boolean unzipDownload)
  {
	  // build the message header
	  int appFileMessageLength = 27;
	  byte[] filenameMessage = new byte[appFileMessageLength+5];
	  filenameMessage[0] = 74;
	  filenameMessage[1] = intToByte1(appFileMessageLength);
	  filenameMessage[2] = intToByte2(appFileMessageLength);
	  if(unzipDownload)filenameMessage[3] = 1;
	  else filenameMessage[3] = 0;
	  filenameMessage[4] = 0;

	  // message contains file name, Mtime, length
          if(appFileDownloadName.length() > 19)
          appFileDownloadName = appFileDownloadName.substring(0,19);
	  while(appFileDownloadName.length() < 19)appFileDownloadName += " ";
	  System.arraycopy(appFileDownloadName.getBytes(),0,
	                   filenameMessage,5,
	                   19);
	  System.arraycopy(appFileDownloadMtimeBytes,0,
	                   filenameMessage,24,
	                   4);
    System.arraycopy(appFileDownloadLengthBytes,0,
                     filenameMessage,28,
                     4);

	  // send message to client
    if(writeThisOutStream(filenameMessage,appFileMessageLength+5))return;
    
    // log action
    if(unzipDownload)
      System.out.println(appFileDownloadName + " size " +
                         appFileSize[thisHostIndex] +
						             " mtime " + appFileMtime[thisHostIndex] +
                         " existing at host #" + thisHostIndex +
                         " OK to unzip");
                         
    // send status to system messages window on FC
    String systemMessage = (unzipDownload ? 
                           "unzipping appData " : "downloading appData ") 
                           + moodleFileName;
    byte[] unzipMessage = systemMessage.getBytes();
    fcConnect.sendToFC(49,
                       unzipMessage.length,
                       unzipMessage,
                       thisHostIndex);
                         
  }// end sendFilename()


  // send and flush a download segment
  // ack = 1 to request ACK, 0 not to request
  void sendDownload(boolean ack) throws IOException
  {
    // if no active TCP connection, don't do this
    if( !tcpIsOpen[ thisHostIndex ] )
    {
	    proceedWithDownload = false;
	    return;
    }

    // inform floor control display that download is starting
    if(!startMessageSent)
    {
      sendDownloadingStatus(1);
      startMessageSent = true;
      System.out.println("\nstarted app file download size " + appFileDownloadSize +
                         " bytes mtime " + appFileDownloadMtime +
						             " to host# " + thisHostIndex);
                         
      // send status to system messages window on FC
      byte[] downloadMessage = ("downloading appData " + moodleFileName).getBytes();
      fcConnect.sendToFC(49,
                         downloadMessage.length,
                         downloadMessage,
                         thisHostIndex);
    }
    
    // determine segment length; for final segment, update the header
    int downloadBytes = appFileDownloadSize - bytesDownloaded;
    boolean lastSegment = (downloadBytes <= DOWNLOAD_MESSAGE_SIZE);
    if(!lastSegment)downloadBytes = DOWNLOAD_MESSAGE_SIZE;
    int downloadSN = segmentSN;
    if(ack)//debugx
    {
      downloadSN += 10000;
      lastAckRequestSent = segmentSN;
    }
    downloadSegment[0] = 73;
    downloadSegment[1] = intToByte1(downloadBytes);
    downloadSegment[2] = intToByte2(downloadBytes);
    
    // insert segmentSN into downloadHeader
    downloadSegment[3] = intToByte1(downloadSN);
    downloadSegment[4] = intToByte2(downloadSN);
    segmentSN = (segmentSN+1)%10000;

    // send the segment
    System.arraycopy(appFileDownload,bytesDownloaded,
                     downloadSegment,5,
                     downloadBytes);
    if(writeThisOutStream(downloadSegment,downloadBytes+5))return;
    bytesSentThisSlot += downloadBytes;
    bytesDownloaded += downloadBytes;
    
    // after last segment, send unzip message
    if(lastSegment)sendFilename(true);
    
  } // end sendDownload


  // send a download segment of length zero, indicating download is done
  // breakoff is false for end of file; true to breakoff sending a file
  void sendDownloadDone(boolean breakoff) throws IOException
  {
    // we might get called when not downlaoding...
    if(!proceedWithDownload)return;
    proceedWithDownload = false;//debugx
    if(breakoff)segmentSN = 0;
    
     // send the segment
    downloadSegment[0] = 73;
    downloadSegment[1] = intToByte1(0);
    downloadSegment[2] = intToByte2(0);
    downloadSegment[3] = intToByte1(segmentSN);
    downloadSegment[4] = intToByte2(segmentSN);
    if(writeThisOutStream(downloadSegment,5))return;

    // inform floor control status panel that download is done
    sendDownloadingStatus(3);

    // log completion
    if(!breakoff)
      System.out.println("\ncompleted download size " + appFileDownloadName +
                         bytesDownloaded + " bytes mtime " + appFileDownloadMtime +
					               " to host #" + thisHostIndex + " in " + segmentSN + 
                         " segments " + (System.currentTimeMillis()-startDownloadMillis) +
                         " ms");
    else
      System.out.println("\nbroke off sending download for host:" + thisHostIndex);

  } // end sendDownloadDone
  

  // this can be invoked from elsewhere in TLM after TunnelSend is running
  void startRecordingDownload()
  {
    int recordingNameMessageLength = 38;
    recordDownloadMessagesSent[thisHostIndex] = 0;
    recordDownloadAcksReceived[thisHostIndex] = 0;
    waitTime = 1;
    
    // if no active TCP connection, don't do this
    if( !tcpIsOpen[ thisHostIndex ] )
    {
      serverRecordingDownload = false;
      recordingDownloadFilename = "";
      return;
    }
  
    // set startup state (startRecordingDownload can happen more than once)
    recordingBuffer[0] = 103;
    
    // confirm file exists
    String filePathname = 
      new String(recordingsSubdirectory + "/" + recordingDownloadFilename);
    File testFile = new File(filePathname);
    if(!testFile.exists()) 
    {
      System.out.println("*** can't find recording file to download:" +
                         filePathname + " for host " + thisHostIndex);
      recordingDownloadFilename = "";
      recordingFile = null;
      recordingNameMessageLength = 0;
    }
    
    // check for overlength filename
    if(recordingDownloadFilename.length() > 34)
    {
      System.out.println("*** error in recordingDownload: filename too long " +
                        recordingDownloadFilename);
      recordingDownloadFilename = "";
      recordingNameMessageLength = 0;
    }
    
    // check for zero length file
    if(testFile.length() == 0)
    {
      System.out.println("*** error in recordingDownload: file has zero length " +
                         recordingDownloadFilename);
      recordingDownloadFilename = "";
      recordingNameMessageLength = 0;
    }
    
    // confirm this TunnelSend is active
    if(!tunnelIsActive[thisHostIndex])
    {
      System.out.println(
        "*** error in recordingDownload: TunnelSend not active for host " +
        thisHostIndex);
      recordingDownloadFilename = "";
      recordingNameMessageLength = 0;
    }
    
    // open the file for stream read
    try 
    {
      recordingFile = new FileInputStream(filePathname);
    }
    catch(FileNotFoundException fnfe)
    {
      recordingDownloadFilename = "";
      recordingNameMessageLength = 0;
    }

    // prepare filename or error message to TLMC
    byte[] filenameMessage = new byte[recordingNameMessageLength+5];
    filenameMessage[0] = 104;
    filenameMessage[1] = intToByte1(recordingNameMessageLength);
    filenameMessage[2] = intToByte2(recordingNameMessageLength);
    filenameMessage[3] = 0;
    filenameMessage[4] = 0;

    if(recordingNameMessageLength == 38)
    {
      // create fixed length filename and move into message
      String filenameToSend = new String(recordingDownloadFilename);
	    while(filenameToSend.length() < 34)filenameToSend += " ";
	    System.arraycopy(filenameToSend.getBytes(),0,
	                     filenameMessage,5,
	                     34);
      
      // get the Mtime of the file and move into message
      recordingFileDownloadMtime = testFile.lastModified()/1000;
	    System.arraycopy(intToBytes((int)recordingFileDownloadMtime),0,
	                     filenameMessage,39,
	                     4);
                       
      // get the size to use in log output
      recordingFileDownloadSize = testFile.length();
      
    }
    
    // send message to client
    if(writeThisOutStream(filenameMessage,recordingNameMessageLength+5))return;
    
  }// end startRecordingDownload()
  

  // read and send a recording message download from file
  void sendRecordingDownload()
  {
    // if no active TCP connection, don't do this
    if( !tcpIsOpen[ thisHostIndex ] )
    {
      serverRecordingDownload = false;
      recordingDownloadFilename = "";
      return;
    }
    
    // download a recorded message; on error, stop
    int recordingMessageLength = readRecordingFile();
    if(recordingMessageLength == 0)
    {
      serverRecordingDownload = false;
      return;
    }
    
    // at end, send completion code to client
    if(recordingMessageLength < 0)
    {
      // restore wait time
      waitTime = 0;
    
      // send the segment
      recordingBuffer[1] = intToByte1(0);
      recordingBuffer[2] = intToByte2(0);
      recordingBuffer[3] = intToByte1(recordingSN);
      recordingBuffer[4] = intToByte2(recordingSN);
      if(writeThisOutStream(recordingBuffer,5))return;
      
      // log completion and return
      System.out.println("\ncompleted server recording download " + 
                         recordingDownloadFilename +
                         " size " + recordingFileDownloadSize +
                         " bytes mtime " + recordingFileDownloadMtime +
                         " to host #" + thisHostIndex);
      recordingDownloadFilename = "";
      serverRecordingDownload = false;
      return;
    }
    
    // insert length and segmentSN into downloadHeader
    recordingBuffer[1] = intToByte1(recordingMessageLength);
    recordingBuffer[2] = intToByte2(recordingMessageLength);
    recordingBuffer[3] = intToByte1(recordingSN);
    recordingBuffer[4] = intToByte2(recordingSN++);
    
    // send the message
    if(writeThisOutStream(recordingBuffer,recordingMessageLength+5))return;
    recordingBytesDownloaded += recordingMessageLength;

  } // end sendRecordingDownload
  
  
  // read and return the next MIST/C header+message from recordingFile
  // returns -1 if no more data, 0 for exception reading; messageLength otherwise
  int readRecordingFile()
  {
    int recordingMessageLength;
    try 
    {
      // read file; at endfile close file and return true
      if(recordingFile.read(recordingBuffer,5,9) < 0)
      {
        recordingFile.close();
        recordingFile = null;
        return -1;
      }
            
      // get length from header
      recordingMessageLength = bytesToUshort(recordingBuffer[9],recordingBuffer[10]);
      
      // confirm length is in right range
      if(recordingMessageLength > MAX_MESSAGE)
      {
        System.out.println("*** readRecordingFile found length out of range ( " +
                           recordingMessageLength + " > " +
                           MAX_MESSAGE + " ) in file " + 
                           recordingDownloadFilename + " - cancelled download");
        recordingFile.close();
        recordingFile = null;
        return 0;
      }
      
      // read the payload; if endfile close file and return error
      if(recordingFile.read(recordingBuffer,14,recordingMessageLength) < 0)
      {
        System.out.println("*** readRecordingFile found payload missing in file " +
                           recordingDownloadFilename);
        recordingFile.close();
        recordingFile = null;
        return 0;
      }
    }
    catch(Exception e)
    {
      System.out.println("*** Exception reading recording file " + 
                         recordingDownloadFilename + " " + e.getMessage());
      e.printStackTrace(System.out);
      try{recordingFile.close();}catch(Exception ee){}
      recordingFile = null;
      return 0;
    }
    return recordingMessageLength+9;
    
  }//end readRecordingFile


  // send and flush the message to the TCP or UDP stream associated with this port
  void sendBuffer(byte[] thisBuffer, 
                  int messageLength, 
                  boolean thisMessageIsVideo ) 
                  throws IOException
  {
    if( tcpIsOpen[thisHostIndex] )
    {
      // send by video or audio TCP tunnel
      if( thisMessageIsVideo )
      {
        if( connectedVideoByTCP[ thisHostIndex ] )
        {
          if(writeThisVideoOutStream(thisBuffer,messageLength))return;
          
        } // end if( connectedVideoByTCP[ thisHostIndex ] )
      } // end if( thisMessageIsVideo )
      else
      {
        if(writeThisOutStream(thisBuffer,messageLength))return;
      } // end else/if( tcpIsOpen[ thisHostIndex ] )
      countTcpSend++;
      
    } // end if( tcpIsOpen[ thisHostIndex ] )
    else if( (uRecvTime[thisHostIndex] != 0) && authenticatedHost[thisHostIndex])
    {
      // host using UDP tunnel- send back to it (unlike general UDP
      // in UCListen, tunnel is kept active by TCP floor control
      // connection, not by the interval between received messages)
      sendMessage.setData( thisBuffer, 0, messageLength );
      tunnelSendUDPSocket.send( sendMessage );
      countUdpSend++;
    } // end elseif( (uRecvTime...

  }// end sendBuffer()
  
  
  // post results of TC 79 download segment ACK
  void postACK(int ackHostIndex,int ackSN)
  {
    lastSNAcked = ackSN;
    synchronized(sendThreads[ackHostIndex])
    {
      sendThreads[ackHostIndex].notify();
    }
  
  }// end postACK()
  
  
  // writes the data to non-blocking channel
  // uses sleep to avoid lockup when receiver's buffer is full
  // returns true if disconnected or interrupted
  boolean writeThisOutStream(byte[] data,int length)
  {
    // move data to a ByteBuffer to send
    ByteBuffer bb = ByteBuffer.wrap(data,0,length);
  
    // keep trying until the buffer gets sent
    while(!stopTLM)
    {
      // write the data to outstream
	    try
	    {
        thisSocketChannel.write(bb);
      }
      catch(ClosedChannelException cce)
      {
        return true;
      }
      catch(IOException ioe)
      {
        // disconnected
        return true;
      }
      catch(Exception e)
      {
	      closeTunnel(thisHostIndex,false,
          "Exception in TunnelSend for host:" + thisHostIndex + ":" + e);
        e.printStackTrace(System.out);  
	      return true;
      }
      
      // if data was not all written, wait sleepTime and try again
      if(bb.hasRemaining())
        try{Thread.sleep(sleepTime);}
        catch(InterruptedException ie)
        {
          return true;
        }
      else break;
    
    }// end while(!stopTLM)
       
    return false;
      
  }// end writeThisOutStream()
  
  
  // writes the data to non-blocking video channel
  // uses sleep to avoid lockup when receiver's buffer is full
  // returns true if disconnected or interrupted
  boolean writeThisVideoOutStream(byte[] data,int length)
  {
    // move data to a ByteBuffer to send
    ByteBuffer bb = ByteBuffer.wrap(data,0,length);
  
    // keep trying until the buffer gets sent
    while(!stopTLM)
    {
      // write the data to outstream
	    try
	    {
        tcpVideoSocketChannels[thisHostIndex].write(bb);
      }
      catch(ClosedChannelException cce)
      {
        return true;
      }
      catch(IOException ioe)
      {
        // disconnected
        return true;
      }
      catch(Exception e)
      {
	      closeTunnel(thisHostIndex,false,
          "Exception in TunnelSend for host:" + thisHostIndex + ":" + e);
        e.printStackTrace(System.out);  
	      return true;
      }
      
      // if data was not all written, wait sleepTime and try again
      if(bb.hasRemaining())
        try{Thread.sleep(sleepTime);}
        catch(InterruptedException ie)
        {
          return true;
        }
      else break;
    
    }// end while(!stopTLM)
       
    return false;
      
  }// end writeThisVideoOutStream()
  
  
  // set status for recordings download
  void setRecordDownload(int recordHostIndex,boolean status)
  {
    serverRecordingDownload = true;
    synchronized(sendThreads[recordHostIndex])
    {
      sendThreads[thisHostIndex].notify();
    }
  }// end setRecordDownload()


  // thread class for asynchronous processing of download zipfile size data
  class CheckFileSpecs extends Thread
  {
    public void run()
    {
      // update download status with availability of buffered download file
      if(!thisTunnelSupportsDownload || (appFileMtime[thisHostIndex] == 0))
      {
        sendDownloadingStatus(2);
        return;
      }

      // to download, cached file must be ready
      if(!downloadIsReady)return;

      // send a request to the client for filesize of current
      // appDataZipFile, flagged to not unzip it
      sendFilename(false);

      // wait 30 seconds to receive a response, then go ahead with download
      appFileStatNew[ thisHostIndex ] = false;
      try
      {
        for(int timer=0; timer<120; ++timer)
        {
          if(appFileStatNew[ thisHostIndex ])break;
          Thread.sleep(250);
	    }
      }catch(InterruptedException ie){}
      System.out.println("host " + thisHostIndex + " reported appDataZipFile size " + 
                         appFileSize[thisHostIndex] + " bytes (cached download is " + 
                         appFileDownloadSize + ")" + " mtime " + 
                         appFileMtime[thisHostIndex] + " (cached " +
                         appFileDownloadMtime + ")");

      // check whether the zipfile size and length the client has is OK
      // if it is OK, invite unzip; if it is not, invite download;
      // allow client version to be off by exactly 3600 seconds to deal
      // with savings time bug
      if(((appFileMtime[thisHostIndex] == appFileDownloadMtime) ||
          (appFileMtime[thisHostIndex] == appFileDownloadMtime+3600) ||
          (appFileMtime[thisHostIndex] == appFileDownloadMtime-3600)) &&
          (appFileSize[thisHostIndex] == appFileDownloadSize))
      {
        sendFilename(true);
        System.out.println("sent appDataZipFile unzip to host:" + thisHostIndex);
      }
      else
      {
        sendFilename(false);
        System.out.println("started appDataZipFile download to host:" + thisHostIndex);
        startDownloadMillis = System.currentTimeMillis();
        proceedWithDownload = true;
        synchronized(sendThreads[thisHostIndex])
        {
          sendThreads[thisHostIndex].notify();
        }
      }
    }// end run()
  }// end class CheckFileSpecs
} //end class TunnelSend

