/* modified JMP 3/1/13
 * UCListen.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author  mpullen & students
 * @version 5.2.4
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/

/*********************SYSTEM LOGIC**************************
 *
 * UCListen.java class logic:
 * ->In an infinite loop
 * ->  read unicast UDP messages; forward only if they are in unicast array
 *     (insert in array if a wildcard slot is available)
 * ->  also, forward only if floorHolder or port is not floor controlled
 * ->  update unicast last recv time array
 * ->  if the multicast group has been a source of messages (any port) within time window
 *       send unicast-udp message to multicast group with a modified TTL(15)
 * ->  send a copy of unicast UDP message to each unicast addresses that has been active
 *       within time window except the address it came from (source address)
 *       unless the application is associated with TCP on that host
 * ->  and to each unicast TCP address/TCP tunnel port except the address it came from
 *       (after prepending unicast UDP port to message)
 ***********************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;

public class UCListen extends TLM implements Runnable
{
    MulticastSocket mcSocket;// multicast socket instance
    DatagramSocket udpSocket;// datagram socket instance
    DatagramPacket sendPacket;
    DatagramPacket recvMessage;
    static InetAddress sourceAddress,localAddress;
    // sourceAddress is inetaddress of the  packet received
    // localAddress is the inet address of the terminal on which TLM is running
    byte[] sendBuffer, recvBuffer;
    int udpPort, mcPort;
    int thisApplicationIndex = 0; // 0 = audio, 1 = WBD, 2 = video
    boolean nonVideoPort, forwardingAudio, forwardingWBD, forwardingVideo;
    private long startTime; // system time when a message is received
    private long checkTime = 0;

    // constructor
    public UCListen (  int applicationIndex, int udpPortParm, int mcPortParm,
      MulticastSocket mcSocketParm, DatagramSocket udpSocketParm )
    {
      // copy input parameters and set related values
      udpPort = udpPortParm;
      mcPort = mcPortParm;
      mcSocket = mcSocketParm;
      udpSocket = udpSocketParm;
      nonVideoPort = ( (udpPort != videoPort) && (udpPort != videoPort+1) );
      forwardingAudio = ( (udpPort == audioPort) || (udpPort == audioPort+1) );
      if( forwardingAudio )thisApplicationIndex = 0;
      forwardingWBD = ( (udpPort == wbdPort) || (udpPort == wbdPort+1) );
      if( forwardingWBD )thisApplicationIndex = 1;
      forwardingVideo = ( (udpPort == videoPort) || (udpPort == videoPort+1) );
      if( forwardingVideo )thisApplicationIndex = 2;

      // initialize and allocate
      recvBuffer = new byte[ MAX_MESSAGE ];
      sendPacket = new DatagramPacket( recvBuffer, 0, 0, localAddress, 0);

      try
      {
        localAddress = InetAddress.getLocalHost();
        udpSocket.setSoTimeout(1);
      }
      catch(IOException e)
      {
        System.out.println("UCListen: " + e.getMessage());
        // error message received when Unicast port is not open; execution halts
        stopTLM = true;
      }
    }

    public void run()
    {
      // allocate buffers and initialize
      int messageLength = 0;
      int hostIndex = 0;

      // loop as long as messages arrive
      try
      {
        while( !stopTLM )
        {
          // check for message, then give up control so other threads can run
          while( !stopTLM )
          {
            // receive UDP message, extract its length and source address
            try
            {
              recvMessage = new DatagramPacket( recvBuffer, MAX_MESSAGE );
              udpSocket.receive( recvMessage );
              messageLength = recvMessage.getLength();
              sourceAddress = recvMessage.getAddress();
              countUdpRecv++;
              break;
            }
            catch( InterruptedIOException e1 )
            {
              try
              {
                Thread.sleep(sleepTime);
              }
              catch( InterruptedException e2 )
              {
              } // end catch( InterruptedException
            } // end catch( InterruptedIOException
          } // end while(!stopTLM

          // allocate new send buffer to be queued for TunnelSend
          sendBuffer = new byte[ messageLength+5 ];

          // check for privileged video sender (permits no others)
          // in which case the only UCListen threads are those for video
          boolean okToForward = false;
          if( videoAddressConfigured )
          {
            if( sourceAddress.equals(videoInetAddress) )
            {
              okToForward = true;
            }
            else continue;
          }

          // look for sender in active flows
          int thisHostIndex;
          for( thisHostIndex = 0; thisHostIndex < numberOfHosts; ++thisHostIndex )
            if( udpActive[ thisHostIndex ] )
               if( sourceAddress.equals(hostAddr[ thisHostIndex ]) ) break;
          if( thisHostIndex >= maxConnections )
          {
            // if not in active list- locate sender in valid addresses
            // and activate (if not there, error; exit )
            for( thisHostIndex = 0; thisHostIndex < numberOfHosts; ++thisHostIndex )
              if( sourceAddress.equals(hostAddr[thisHostIndex]) )break;
            if( hostIndex < numberOfHosts )
              activateFlow( thisHostIndex, sourceAddress );
            else
            {
              // if it is not in the table, we may have acceptAnyHost; if so,
              // add it to host table and activeFlows[]; if not, ignore the message
              if( thisHostIndex >= numberOfHosts )
                if( !checkPassword || (thisHostIndex > MAX_HOSTS) )
                  continue; // while(!stopTLM)
              synchronized( hostInetAddr )
              {
                if(addToHostTable(sourceAddress,sixtyFourZeros,-1,-1) < 0)continue;
              }
              activateFlow( thisHostIndex, sourceAddress );
            } // end else/if( hostIndex < numberOfHosts )
          } // end if( flowIndex >= udpFlowsActive )
          // record time message received for source timeout
          startTime = System.currentTimeMillis();
          uRecvTime[ thisHostIndex ] = startTime;

          // floor control for WBD and video
          if( thisHostIndex == floorHolder )okToForward = true;
          if( forwardingWBD )
            // allow WBD repair requests even under floor control
            if( !wbdFloorControl ||
                (recvBuffer[15]==2) ||
                controlPrivilege[ thisHostIndex ] > '1' )
              okToForward = true;
          if( forwardingVideo )
            if( !videoFloorControl )okToForward = true;
          if( forwardingAudio )
          {
            if( (floorHolder >= 0) && (thisHostIndex == floorHolder) )
              checkTime = startTime;
            else
            {
              if( (startTime - checkTime) > 5000 )
              {
                if( lastWarningTime[thisHostIndex] < (startTime - 20000) )
                {
                  if( sendFloorWarning(thisHostIndex, udpPort) )
                    return;
                  lastWarningTime[thisHostIndex] = startTime;
                } // end if( lastWarningTime
              } // end if( startTime
            } // end if( (floorHolder >= 0)...

          } // end if( forwardingAudio )

          // if this is not a forwardable message, don't
          if(!okToForward)continue;

          // build outgoing TCP message
          // start message with protocol transaction code
          sendBuffer[0] = 66;

          // insert message length
          // inline version of intToByte()
          // sendBuffer[1] = intToByte1( messageLength );
          // sendBuffer[2] = intToByte2( messageLength );
          // coded inline for efficiency
          int highByte = messageLength >> 8;
          if( highByte <= 127 )
            sendBuffer[1] = (byte) highByte;
          else
            sendBuffer[1] = (byte)(int)(127-highByte);
          int lowByte = messageLength - 256*highByte;
          if( lowByte <= 127 )
            sendBuffer[2] = (byte)lowByte;
          else
            sendBuffer[2] = (byte)(int)(127-lowByte);

          // insert UDP port number
          // sendBuffer[3] = intToByte1( udpPort );
          // sendBuffer[4] = intToByte2( udpPort );
          // coded inline for efficiency
          highByte = udpPort >> 8;
          if( highByte <= 127 )
            sendBuffer[3] = (byte) highByte;
          else
            sendBuffer[3] = (byte)(int)(127-highByte);
          lowByte = udpPort - 256*highByte;
          if( lowByte <= 127 )
            sendBuffer[4] = (byte)lowByte;
          else
            sendBuffer[4] = (byte)(int)(127-lowByte);

          // copy UDP message to TCP segment
          System.arraycopy(recvBuffer,0,sendBuffer,5,messageLength);

          // send a copy of the received message to every host with tunnel open
          int tcpLength = messageLength + 5;
          for ( hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex )
          {
            if( !connectedByTCP[ hostIndex ])continue;

            // do not send message back to source
            if( hostIndex == thisHostIndex )continue;

            // place message in queue for TCP send
            if( forwardingAudio )
            {
              if( queueTunnel( 1, hostIndex, sendBuffer, tcpLength ) )
              {
                if( ++audioDropped[ hostIndex ]%dropMessageInterval == 1 )
                  System.out.println( dateTime() + " " +
                    audioDropped[ hostIndex ] + " audio messages for host #" +
                    hostIndex + " dropped due to congestion (delta RTT:" +
                    latestRTTIncrease[ hostIndex ] + ")" );
              } // end if( queueTunnel( 1...
            } // end if( forwardingAudio )
            if( forwardingWBD )
            {
              if( queueTunnel( 3, hostIndex, sendBuffer, tcpLength ) )
              {
                if( ++whiteboardDropped[ hostIndex ]%dropMessageInterval == 1 )
                  System.out.println( dateTime() + " " +
                    whiteboardDropped[ hostIndex ] +
                    " whiteboard messages for host #" + hostIndex +
                    " dropped due to congestion (delta RTT:" +
                    latestRTTIncrease[ hostIndex ] + ")" );
              } // end if( queueTunnel( 3...
            } // end if( forwardingWBD )
            if( forwardingVideo )
            {
              if( connectedVideoByTCP[ hostIndex ] || !connectedByTCP[ hostIndex ])
              {
                if( queueTunnel( 4, hostIndex, sendBuffer, tcpLength ) )
                {
                  if( ++videoDropped[ hostIndex ]%dropMessageInterval == 1 )
                    System.out.println( dateTime() + " " +
                      videoDropped[ hostIndex ] +
                      " video messages for host #" + hostIndex +
                      " dropped due to congestion (delta RTT:" +
                      latestRTTIncrease[ hostIndex ] + ")" );
                } // end if( queueTunnel( 4...
              } // end connectedVideoByTCP...
            }  // end if( forwardingVideo )
          } // end for(sendIndex...
          // **************** end queue send *******************************

          // if this is a heartbeat from local host, do not send to UDP
          if(sourceAddress.equals(localAddress))continue;
          sendPacket.setData(recvBuffer, 0, messageLength);
          for( hostIndex=0; hostIndex < numberOfHosts; hostIndex++ )
          {
            if( !udpActive[ hostIndex ] )continue;

            // for UDP, forward only if the messages have arrived within time window
            if( (startTime - uRecvTime[ hostIndex ]) < udpWindow )
            {
              // do not send message back to source
              if( hostAddr[ hostIndex ] != null)
              {
                if (hostIndex != thisHostIndex) {
                  // send UDP message
                  sendPacket.setAddress(hostInetAddr[hostIndex]);
                  sendPacket.setPort(udpPort);
                  udpSocket.send(sendPacket);
                  countUdpSend++;
                } // end if(hostIndex...
              }// end if hostAddr[ hostIndex ]...
            } //end if((startTime
            else
            {
              // address has stopped sending; delete from active set
              deactivateFlow( hostIndex );
            } // end else/if(startTime...
          } // end for(sendIndex...

          // send to multicast groups only if messages have arrived within window
          if( multicastIsConfigured )
          {
            sendPacket.setData(recvBuffer, 5, messageLength);
            if( ( startTime - mRecvTime[thisApplicationIndex] )  < udpWindow )
            {
              sendPacket.setAddress( mcGroup[ thisApplicationIndex ] );
              sendPacket.setPort( mcPort );
              mcSocket.send( sendPacket );
              countMcSend++;
            } // end if(startTime...
          } // end if( multicastIsConfigured )
        } // end while(!stopTLM)

        // let other threads run
        Thread.yield();

      } // end try receiving message
      catch( IOException ioe )
      {
        // lost connection - can't continue
      }
      catch( Exception e )
      {
        if( !stopTLM )
          System.out.println("Exception in UCListen:" + e + "\n" + e.getMessage());
      }
      if(udpSocket != null)udpSocket.close();
    } // end run

} //end TLM.UCListen class
