/*
 * UDPTunnel.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author  mpullen & students
 * @version 5.2.4
 * @modified JMP 3/1/13
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

/*********************SYSTEM LOGIC**************************
UDPTunnel.java class logic:
->As long as the UDP socket is open
-> read UDP tunnel message
-> strip port from first two byte of message and locate it in multicast port array
   or unicast port array
   if it is in the array, capture the corresponding port (unicast if it is in
   the multicast array and vice versa)  and flag that port as TCP for this host;
   else go back to reading TCP
-> if a loopback time probe message is received, send it back to the source
-> if not loopback or URL, it should be forwarded in the following ways
   -> update unicast last recv time array
   -> if the multicast group has been a source of messages (any port) within time
      window send unicast-udp message to multicast group with a modified TTL(15)
   -> send a copy of unicast UDP message to each unicast address that has been active
      within time window except the address it came from (source address) and queue
      it to be sent to each TCP and UDP tunnel address except the address it came
      from (tunnel gets the original message with port number prepended and revised
      transaction code)
-> if there is a record/play control connection open and the message bears rpCPort,
   send a copy through the tunnel connection to the control
-> When the connection is closed, set corresponding unicast TCP array to 0 and return
 ***********************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;

public class UDPTunnel extends TLM implements Runnable
{
    MulticastSocket mcSocket;// multicast send socket
    DatagramSocket udpSocket;// general UDP send socket
    DatagramSocket tunnelUDPRecvSocket, tunnelUDPSendSocket;
    DatagramPacket sendMessage, recvMessage, sendPacket;
    byte[] recvBuffer;
    byte[] setupBuffer;
    int setupLength, thisApplicationIndex;
    private long startTime; // system time when a message is received
    private long checkTime = 0;
    static InetAddress sourceAddress, localAddress;
    int portIndex, reportedMinRTT, reportedRTT;
    int thisUdpPort, sourceUdpPort, messageCount, loopbackCount, appNumber;
    InetAddress remoteHost = null;
    boolean rpCIsThisHost;
    int audioControlPort, wbdControlPort, videoControlPort;

    // constructor
    public UDPTunnel()
    {
      // initialize
      audioControlPort = audioPort + 1;
      wbdControlPort = wbdPort + 1;
      videoControlPort = videoPort + 1;

      // create port list message to respond to setup requests
      setupLength = 28;
      setupBuffer = new byte[ setupLength ];
      setupBuffer[0] = 64; // transaction code

      // initialize flags
      for(int hostNumber=0; hostNumber<maxConnections; ++hostNumber)
        udpSignalSlideDone[hostNumber] = false;

      // message length does not include 5 byte header
      int messageLength = setupLength - 5;
      // send setup information to the client
      setupBuffer[0] = 64; // transaction code
      setupBuffer[1] = intToByte1( messageLength );
      setupBuffer[2] = intToByte2( messageLength );
      setupBuffer[3] = 0;
      setupBuffer[4] = 0;
      setupBuffer[5] = 0;
      setupBuffer[6] = intToByte1( connectionThreshold );
      setupBuffer[7] = intToByte2( connectionThreshold );
      setupBuffer[8] = intToByte1( congestionThreshold );
      setupBuffer[9] = intToByte2( congestionThreshold );
      setupBuffer[10] = intToByte1( 0 ); // previously rpTcpPort
      setupBuffer[11] = intToByte2( 0 ); // previously rpTcpPort
      setupBuffer[12] = intToByte1( probeInterval );
      setupBuffer[13] = intToByte2( probeInterval );
      setupBuffer[14] = (byte)numberOfApplications;
      setupBuffer[15] = intToByte1(whiteboardRingSize);
      setupBuffer[16] = intToByte2(whiteboardRingSize);
      System.arraycopy(roomIDBytes,0,setupBuffer,17,11);
//debugx      System.arraycopy(thisSemesterBytes,0,setupBuffer,17,3);
//debugx      System.arraycopy(thisCourseNumBytes,0,setupBuffer,20,5);

      // create socket and packet; set their parameters
      recvBuffer = new byte[ MAX_MESSAGE+5 ];
      try
      {
        tunnelUDPRecvSocket = new DatagramSocket( tunnelUdpPort );
        tunnelUDPSendSocket = new DatagramSocket( );
        localAddress = InetAddress.getLocalHost();
        tunnelUDPRecvSocket.setSoTimeout(0);  // makes receive socket blocking
        sendPacket = new DatagramPacket( recvBuffer, 0, 0, remoteHost, 0);
      }
      catch(IOException e)
      {
        System.out.println("Panic error in UDPTunnel setup:" + e.getMessage());
        // error message received when Unicast port is not open; execution halts
        stopTLM = true;
        System.exit(-1);
      }
    }

    public void run()
    {
      // allocate buffers and initialize
      int messageLength = 0, imbeddedLength, highByte, lowByte;
      byte firstByte = 0;
      int audioThreshold = 2*congestionThreshold;
      int videoThreshold = congestionThreshold;

      // loop as long as messages arrive
      while( !stopTLM )
      {
        // allocate a new read buffer so it can be saved until
        // written to all outstreams asynchronously
        recvBuffer = new byte[ MAX_MESSAGE+5 ];

        // receive UDP message, extract its length and source address
        try
        {
          recvMessage = new DatagramPacket( recvBuffer, MAX_MESSAGE+5 );
          tunnelUDPRecvSocket.receive( recvMessage );
          messageLength = recvMessage.getLength();
          sourceAddress = recvMessage.getAddress();
          sourceUdpPort = recvMessage.getPort();
          countUdpTunnelRecv++;
          firstByte = recvBuffer[0];
        }
        catch( IOException ioe )
        {
          System.out.println("IOException in UDPTunnel receive:"
            + ioe.getMessage());
          continue; // while( !stopTLM )
        } // end catch( InterruptedIOException

        // look for sender in active tunnels
        int thisHostIndex;
        for(thisHostIndex = 0; thisHostIndex < numberOfHosts; ++thisHostIndex)
        {
          if(!tunnelClosing[ thisHostIndex ])
            if( udpActive[ thisHostIndex ] )
              if( sourceAddress.equals(hostAddr[ thisHostIndex ]) )
                if( sourceUdpPort == sourcePort[ thisHostIndex] )break;
        }
        // if not in active list- locate sender in valid addresses
        // if it is there, activate the tunnel
        boolean newParticipant = false;
        synchronized(hostAddr)
        {
          // look for the source address in address table - fixed addresses first
          // if there is a video source don't check its host
          if( videoAddressConfigured )thisHostIndex = 1;
          else thisHostIndex = 0;
          for( ; thisHostIndex < numberOfHosts; ++thisHostIndex)
          {
            if(hostAddr[thisHostIndex] != null)
              if(sourceAddress.equals(hostAddr[ thisHostIndex ]))
              {
                newParticipant = true;
                sourcePort[ thisHostIndex ] = sourceUdpPort;
                break;
              } // end if(sourceAddress...
          } // end for(thisHostIndex...

          // if it is not in the designated IP address part of the table,
          // we might be under acceptAnyHost
          // if not, ignore the message
          if( !newParticipant )
          {
            if( !checkPassword )
            {
              // look for a spot in host table to add this address;
              // if none, drop the message
              if((thisHostIndex =
                  addToHostTable(sourceAddress,sixtyFourZeros,-1,-1)) < 0)continue;
              newParticipant = true;
              sourcePort[ thisHostIndex] = sourceUdpPort;
            } // end if( !checkPassword )

            // if it still isn't in table, drop the message
            if( thisHostIndex >= numberOfHosts )continue; // to while(!stopTLM)

            // notify the sendTunnel so it can get past download check
            synchronized(sendThreads[thisHostIndex])
            {
              sendThreads[thisHostIndex].notify();
		    }

          } // end if( !newParticipant )

        } // end synchronized (hostAddr)

        // don't let the same host connect TCP and UDP
        if( connectedByTCP[ thisHostIndex ] )continue; // to while(!stopTLM)

        // record time message received
        startTime = System.currentTimeMillis();
        uRecvTime[ thisHostIndex ] = startTime;
        if( thisHostIndex == floorHolder )checkTime = startTime;
        
        // if this is a startup message, extract the sessionVariable
        // so FCListen can authenticate before the startup sequence ends
        if( firstByte == 64 )
        {
          if (messageLength < 24)continue;
          sessionVariable[ thisHostIndex ] = new String(recvBuffer, 6, 64);
          sourceVideoPort[ thisHostIndex ] =
            bytesToUshort(recvBuffer[70],recvBuffer[71]);
        }

        // has user authenticated with FCListen?
        if(!authenticatedHost[thisHostIndex])continue; // to while( !stopTLM )

        // client startup TLMC to TLM
        // {
        //  transaction byte[1] value 64
        //  messageLength unsigned short value 21 // 2 bytes
        //  clientVersionByte1 // byte
        //  clientVersionByte2 // byte
        //  clientVersionByte3 // byte
        //  sessionVariable byte[64]
        //  otherEndPort unsigned short    // 2 bytes
        //  netCapacitykbps unsigned short // 2 bytes
        // }
        if( firstByte == 64 ) // session initiation - respond with setup data
        {
          // check that first transmission has correct format
          // and client version is in range
          byte[] clientVersionBytes = new byte[3];
          System.arraycopy(recvBuffer,3,clientVersionBytes,0,3);
          String clientVersion = new String(clientVersionBytes);
          boolean clientIsOK = (recvBuffer[0] == 64) &&
                               (recvBuffer[1] == 0) &&
                               (recvBuffer[2] == 21) &&
                               (clientVersion.compareTo(minClientVersion) >= 0) &&
                               (clientVersion.compareTo(maxClientVersion) <= 0);

          // if the client is OK, send zeros in client field;
          // otherwise, send new version desired
          if(clientIsOK)
          {
            setupBuffer[3] = 0;
            setupBuffer[4] = 0;
            setupBuffer[5] = 0;
          }
          else
          {
            setupBuffer[3] = maxClientVersionBytes[0];
            setupBuffer[4] = maxClientVersionBytes[1];
            setupBuffer[5] = maxClientVersionBytes[2];
          }
          // respond to client setup message
          sendPacket.setAddress(sourceAddress);
          sendPacket.setPort(tunnelUdpPort+1);
          try
          {
            tunnelUDPSendSocket.send( sendPacket );
          }
          catch(IOException ioe)
          {
			closeTunnel ( thisHostIndex, false,
                          "IO Exception in UDP setup reply to "
                          + sourceAddress + ":" + ioe.getMessage());
            continue; // while( !stopTLM )
          }
          if(!clientIsOK)
          {
			closeTunnel (thisHostIndex, false,
                         "client version " + clientVersion + " incompatible" );
            continue;
	      }
          // could be a duplicate, because TLMC sends initiation request once
          // per second until a response is received; so only respond every
          // five seconds
          if( (startTime - udpSetupTime[ thisHostIndex ]) < 5000 ) continue;
          udpSetupTime[ thisHostIndex ] = startTime;
          DatagramPacket setupPacket =
            new DatagramPacket( setupBuffer,
                                setupLength,
                                sourceAddress,
                                tunnelUdpPort+1 );
            try
            {
              tunnelUDPSendSocket.send( setupPacket );
            }
            catch(IOException ioe)
            {
              System.out.println("IO Exception in UDP setup reply to "
                + sourceAddress + ":" + ioe.getMessage());
              continue; // while( !stopTLM )
            }
            countUdpSend++;
            System.out.println("Sent UDP tunnel setup message to host #" + thisHostIndex
             + " " + sourceAddress + " port " + (tunnelUdpPort+1) );
            continue; // while( !stopTLM )
        }
        
        // if heartbeat message, go back to read the next message
        if( firstByte == 68 )
          continue;// while( !stopTLM )

        // if unzipIsReady message, record status and
        // notify FC client to activate download panel
        if( firstByte == 75 )
        {
	      unzipIsReady[ thisHostIndex ] = downloadIsReady;
		  if(downloadIsReady)
		  {
            byte[] downloadDirectory = new byte[messageLength];
            System.arraycopy(recvBuffer,5,downloadDirectory,0,messageLength);
			if(fcConnect.sendToFC(47,messageLength,downloadDirectory,thisHostIndex))
			  return;
	      }
		  continue;
	    }

        // if RTT probe message, capture RTT and send it back to source
        if( firstByte == 70 )
        {
          try
          {
            sendMessage = new DatagramPacket( recvBuffer, messageLength,
              hostInetAddr[ thisHostIndex ], tunnelUdpPort+1 );
            tunnelUDPSendSocket.send( sendMessage );
            countUdpSend++;
          }
          catch( IOException ioe )
          {
            closeTunnel( thisHostIndex, true, "UDPTunnel IOException writing RTT probe" );
            return;
          }
          // capture loopback timing information
          //
          // inline version of bytesToUshort()
          // reportedRTT = bytesToUshort( recvBuffer[3], recvBuffer[4] );
          highByte = recvBuffer[3];
          if(highByte < 0)highByte = 127 - highByte;
          lowByte = recvBuffer[4];
          if(lowByte < 0)lowByte = 127 - lowByte;
          reportedRTT = 256*highByte + lowByte;

          // when we receive the third loopback message, that is minRTT
          if( ++udpLoopbackCount[ thisHostIndex ] == 3 )
            loopbackMinRTT[ thisHostIndex ] = reportedRTT;

          // capture the high-water RTT for this client
          if( reportedRTT > peakRTT[ thisHostIndex ] )
            peakRTT[ thisHostIndex ] = reportedRTT;

          // after 8 loopback messages, the UDP congestion probe is finished
          if( udpLoopbackCount[ thisHostIndex ] == 8 )
            activateTunnel( thisHostIndex );

          // save difference between reported RTT and reportedMinRTT for
          // congestion management; then quit processing this message
          int rttIncrease = reportedRTT - reportedMinRTT;
          if( rttIncrease < 0 ) rttIncrease = 0;
          latestRTTIncrease[ thisHostIndex ] = rttIncrease;
          continue;
        } // end if( firstByte == 70 )

        // transaction code 76 indicates successful network test
        // and provides appDataZipfile parameters
        if(firstByte == 76)
        {
	      // extract the client's appDataZipfile parameters
          appFileMtime[ thisHostIndex ] = bytesToUint(recvBuffer,5);
          appFileSize[ thisHostIndex ] = bytesToUint(recvBuffer,9);
          appFileStatNew[ thisHostIndex ] = true;

   		  // start accepting tunnel messages
		  activateTunnel( thisHostIndex );
		}

		// transaction code 77 indicates failed network test
		if(firstByte == 77)
		{
	      closeTunnel(thisHostIndex,false,"client reported failed network test");
		  return;
	    }

        // validate protocol code
        if( firstByte != 67 )
        {
          if( (firstByte != 65) && (firstByte != 71) && (firstByte != 88) )
          {
            System.out.println("Invalid protocol code " + firstByte
              + " received from host #" + thisHostIndex);
            continue; // while( !stopTLM )
          }
        }

        // extract encoded messageLength and compare to length received
        // inline version of bytesToUshort()
        // udpLength = bytesToUshort( recvBuffer[1], recvBuffer[2] );
        // done inline for efficiency
        highByte = recvBuffer[1];
        if(highByte < 0)highByte = 127 - highByte;
        lowByte = recvBuffer[2];
        if(lowByte < 0)lowByte = 127 - lowByte;
        imbeddedLength = 256*highByte + lowByte;
        if( imbeddedLength != (messageLength-5) )
        {
          System.out.println("Message length discrepancy in UDPTunnel for host #"
            + thisHostIndex + " received:"
            + messageLength + " imbedded:" + imbeddedLength);
          continue;
        }

        // extract imbedded UDP port number from received UDP tunnel message
        // thisUdpPort = bytesToUshort( recvBuffer[3], recvBuffer[4] );
        // done inline for efficiency
        highByte = recvBuffer[3];
        if(highByte < 0)highByte = 127 - highByte;
        lowByte = recvBuffer[4];
        if(lowByte < 0)lowByte = 127 - lowByte;
        thisUdpPort = 256*highByte + lowByte;

        // check port number for validity
        boolean portIsGood = false;
        int controlPort = 0;
        for( portIndex = 0; portIndex < numberOfApplications; portIndex++ )
        {
          if( thisUdpPort == udpPort[ portIndex ] )
          {
            portIsGood = true;
            break;
          }
          if( thisUdpPort == ( udpPort[ portIndex ]+1) )
          {
            portIsGood = true;
            controlPort = 1;
            break;
          } // end if( thisUdpPort...
        } // end for( portIndex...
        if( !portIsGood )
        {
          if( (firstByte != 65) && (firstByte != 71) && (firstByte != 88) )
          {
            System.out.println("Invalid port code " + thisUdpPort
              + " received from host #" + thisHostIndex);
            continue;
          }
        } // end if( !portIsGood )

        // forwarding and floor control logic
        boolean okToForward = (recvBuffer[0] != 67);
        boolean warnNotFloorHolder = false;
        if( thisHostIndex == floorHolder ) okToForward = true;

        // floor control for WBD and video
        if( (thisUdpPort == wbdPort) || (thisUdpPort == wbdPort+1) )
          // allow WBD repair requests even under floor control
          if( !wbdFloorControl ||
              (recvBuffer[10] ==2) ||
              (controlPrivilege[ thisHostIndex ] > '1') )
            okToForward = true;
        if( (thisUdpPort == videoPort) || (thisUdpPort == videoPort+1) )
        {
          if( !videoFloorControl )okToForward = true;
        }
        
        // check for multicastTunnel forwarding
        if((thisUdpPort == multicastTunnelPort) &&
           (!multicastTunnelFloorControl || (floorStatus == 1)))
          okToForward = true;
        
        if( thisUdpPort == audioPort ) if( audioFloorControl )
        {
          // error message to TLMC if still sending 5 secs plus RTT after giving
          // up the floor; send every 20 seconds while this is true
          if( thisHostIndex != floorHolder )
          {
            if( (startTime - checkTime) > (5000 + reportedRTT) )
            {
              if( lastWarningTime[thisHostIndex] < (startTime - 20000) )
              {
                if( sendFloorWarning(thisHostIndex, thisUdpPort) )
                  return;
                lastWarningTime[thisHostIndex] = startTime;
              } // end if( lastWarningTime
            } // end if( startTime
          }// end else/if(thisHostIndex == floorHolder)
        } // end if( thisUdpPort == audioPort )...

        if( thisUdpPort == wbdPort ) if( wbdFloorControl )
        {
          // maintain state of sending slide to support floor control
          udpSignalSlideDone[thisHostIndex] = false;
          byte testByte = recvBuffer[10]; // WBD pk_type
          if( (thisHostIndex == floorHolder) ||
              (controlPrivilege[ thisHostIndex ] > '1') )
          {
            // logic to avoid floorHolder change during slide send
            if (testByte == 0) // pk_type == pk_draw_msg
            {
              if (recvBuffer[35] == 17) // draw_cmd == DRW_NEW_LABEL
              {
                sendingSlide = true;
                slideTimestamp = bytesToUint(recvBuffer, 4);
                slideIpAddress = bytesToUint(recvBuffer, 12);
                slidePageNumber = bytesToUint(recvBuffer, 16);
              }
              else
              {
                if(recvBuffer[40] == 12) // draw_cmd == DRW_GRP_PS
                  udpSignalSlideDone[thisHostIndex] =
                    ( (bytesToUint(recvBuffer, 4) == slideTimestamp) &&
                      (bytesToUint(recvBuffer, 12) == slideIpAddress) &&
                      (bytesToUint(recvBuffer, 16) == slidePageNumber) );
              }
            }
          }
        } // end if( thisUdpPort == wbdPort )...

        // check for error message from TLMC
        if( recvBuffer[0] == 65 )
        {
          receivedErrorMessage(recvBuffer,thisHostIndex);
          continue; // while( !stopTLM )
        }

        // could be a URL message from a playback server - do not forward
        if( firstByte == 71 )continue;

        // if this is not a forwardable message, don't
        if(!okToForward)continue;

        // for tunnel receivers, revise transaction code only
        recvBuffer[0] = 66;
		  
        // send to server-side recorder, if configured
		if(recordOnServer && (serverRecord != null))
		{
         if( (thisUdpPort == audioPort) || (thisUdpPort == audioControlPort) )
           serverRecord.queueForRecorder(recvBuffer,
		                                 (thisUdpPort == audioControlPort) ? 1 : 0,
                                         0,
							             imbeddedLength);
		 else if((thisUdpPort == wbdPort) || (thisUdpPort == wbdControlPort) )
           serverRecord.queueForRecorder(recvBuffer,
							             (thisUdpPort == wbdControlPort) ? 1 : 0,
							             1,
							             imbeddedLength);
        else if(recordVideo)
		  if( (thisUdpPort == videoPort) || (thisUdpPort == videoControlPort) )
            serverRecord.queueForRecorder(recvBuffer,
			  			                  (thisUdpPort == videoControlPort) ? 1 : 0,
						                  2,
						                  imbeddedLength);
		// should we also record floor control etc?
	  }

        // send a copy of the received message to every host with TCP tunnel open
        int hostIndex;
        for ( hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex )
        {
          if( !connectedByTCP[ hostIndex ] )continue;

          // don't send message back to the sender
          if( hostIndex == thisHostIndex )continue;

          // place message in queue for tunnel send
          // audio traffic
          if( (thisUdpPort == audioPort) || (thisUdpPort == audioControlPort) )
          {
            if( queueTunnel( 1, hostIndex, recvBuffer, messageLength ) )
            {
              if( ++audioDropped[ hostIndex ]%dropMessageInterval == 1 )
                System.out.println( audioDropped[ hostIndex ]
                  + " audio messages for host #" + hostIndex
                  + " dropped due to congestion (delta RTT:"
                  + latestRTTIncrease[ hostIndex ] + ")" );
            } // end if( queueTunnel( 1...
            continue;
          } // end if( udpPort == audioPort...

          // whiteboard traffic
          if( (thisUdpPort == wbdPort) || (thisUdpPort == wbdControlPort) )
          {
            if( queueTunnel( 3, hostIndex, recvBuffer, messageLength ) )
            {
              if( ++whiteboardDropped[ hostIndex ]%dropMessageInterval == 1 )
                System.out.println( whiteboardDropped[ hostIndex ]
                  + " whiteboard messages for host #" + hostIndex
                  + " dropped due to congestion (delta RTT:"
                  + latestRTTIncrease[ hostIndex ] + ")" );
            } // end if( queueTunnel( 3...
            continue;
          } // end if( thisUdpPort == wbdPort...

          // video traffic
          if( (thisUdpPort == videoPort) || (thisUdpPort == videoControlPort) )
          { // if using TCP, send only to TCP video tunnel; if UDP, send anyway
            if( connectedVideoByTCP[ hostIndex ] || !connectedByTCP[ hostIndex ] )
            {
              if( queueTunnel( 4, hostIndex, recvBuffer, messageLength ) )
              {
                if( ++videoDropped[ hostIndex ]%dropMessageInterval == 1 )
                  System.out.println( videoDropped[ hostIndex ]
                    + " video messages for host #" + hostIndex
                    + " dropped due to congestion (delta RTT:"
                    + latestRTTIncrease[ hostIndex ] + ")" );
              } // end if( queueTunnel( 4...
            } // end if( connectedVideoByTCP[ hostIndex ]...
          } // end thisUdpPort == videoPort

          // multicast traffic
          if( thisUdpPort == multicastTunnelPort )
          { // if using TCP, send only to TCP video tunnel; if UDP, send anyway
            if( connectedVideoByTCP[ hostIndex ] || !connectedByTCP[ hostIndex ] )
            {
              if( queueTunnel( 123, hostIndex, recvBuffer, messageLength ) )
              {
                if( ++multicastDropped[ hostIndex ]%dropMessageInterval == 1 )
                  System.out.println( multicastDropped[ hostIndex ]
                    + " multicast messages for host #" + hostIndex
                    + " dropped due to congestion (delta RTT:"
                    + latestRTTIncrease[ hostIndex ] + ")" );
              } // end if( queueTunnel( 123...
            } // end if( connectedVideoByTCP[ hostIndex ]...
          } // end thisUdpPort == multicastPort

        } // end for( int sendIndex...

        // if this completed sending a slide, notify floor control hold thread
        if(udpSignalSlideDone[thisHostIndex])
        {
          udpSignalSlideDone[thisHostIndex] = false;
          sendingSlide = false;
          synchronized(floorChangeHoldThread)
          {
            floorChangeHoldThread.notify();
          }
        }
        // **************** end queueing from tunnel ********************

        // if this is a heartbeat from local host, do not send to UDP
        if(sourceAddress.equals(localAddress))continue;

        // now send message received at UDP tunnel to non-tunnel unicast hosts
        // send only if the packets have arrived within time window
        // and are not using tunnel for that application
        if(!videoAddressConfigured)
        if(udpFlowsActive > 0)
        {
          try
          {
            // select socket for UDP
            if( controlPort == 1 )udpSocket = udpSocket2Save[ portIndex ];
            else udpSocket = udpSocket1Save[ portIndex ];

            // in case videoAddressConfigured, some sockets may be null
            if( udpSocket != null )
            {
              sendPacket.setData(recvBuffer, 5, messageLength);
              synchronized( udpSocket )
              {
                for( hostIndex = 0; hostIndex < numberOfHosts; hostIndex++ )
                {
                  // send only to active UDP tunnels
                  if( !udpActive[ hostIndex ] )continue;

                  // non-tunnel UDP, forward only if the messages have arrived within
                  // time period w and the host is authenticated if that is required
                  if( (startTime - uRecvTime[ hostIndex ]) < udpWindow )
                  {
                    // do not send message back to source
                    if( hostIndex != thisHostIndex )
                    {
                      sendPacket.setAddress( hostInetAddr[ hostIndex ] );
                      sendPacket.setPort( udpPort[ portIndex ]+controlPort );
                      udpSocket.send( sendPacket );
                      countUdpSend++;
                    } // end if(hostIndex...
                  } // end if( (startTime
                  else
                  {
                    // address has stopped sending; delete from active set
                    deactivateFlow( hostIndex );
                  } // end else/if( (startTime...
                } // end for( sendIndex...
              } // end synchronized( udpSocket
            } // end if( udpSocket != null )
          } // end try
          catch(IOException ioe)
          {
            System.out.println("IOException in UDP send of message from "
              + sourceAddress + ":" + ioe.getMessage());
          }
        } // end if( udpFlowsActive > 0 )
        //************** end UDP unicast send ***********************

        // now send audio/WBD/video packet received through UDP tunnel to local multicast
        if( multicastIsConfigured )
        if( thisUdpPort != multicastTunnelPort )
        {
          sendPacket.setData(recvBuffer, 5, messageLength);

          // set socket for multicast
          if( controlPort == 1 )mcSocket = mcSocket2Save[ portIndex ];
          else mcSocket = mcSocket1Save[ portIndex ];

          // send only if the packets have arrived within time window
          if( ( startTime - mRecvTime[portIndex] )  < udpWindow )
          {
            sendPacket.setAddress( mcGroup[ portIndex ] );
            sendPacket.setPort( mcPort[ portIndex ] );
            try
            {
              synchronized( mcSocket )
              {
                // set ttl to 15 and send
                mcSocket.setTimeToLive( (int)15 );
                mcSocket.send( sendPacket );
                countMcSend++;
              } // end synchronized...
            } // end try
            catch(IOException ioe)
            {
              System.out.println("IOException in multicast send of message from "
                + sourceAddress + ":" + ioe.getMessage());
            } // end catch
          } // end if( ( startTime...
        } // end if( thisUdpPort...

        // send tunnel multicast packet received through TCP tunnel locally
        if( multicastTunnelServerSubnet )
        if(thisUdpPort == multicastTunnelPort)
        {
          sendPacket.setData( recvBuffer, 5, imbeddedLength );
          
          // send only if the packets have arrived within time window
          if( ( startTime - mRecvTime[portIndex] )  < udpWindow )
          {
            sendPacket.setAddress( multicastTunnelGroup );
            sendPacket.setPort( multicastTunnelPort );
            try
            {
              mcSocket.send( sendPacket );
            }
            catch(IOException ioe)
            {
              System.out.println("IOException sending in UPDTunnel:" + ioe);
              return;
            }
            countMcSend++;
          } // end if( ( startTime...
        } // end if( thisUdpPort...

        // let other threads run
        Thread.yield();

      } // end while(!stopTLM)
    } // end UDPTUnnel run()
} // end TLM.UDPTunnel class
